/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <memory.h>
#include "winvi.h"
#include "page.h"

BYTE BinaryPart[] = "WinVi binary:";

BOOL ClipboardCopy(void) {
	HGLOBAL	  hMem;
	char huge *pMem;

	/*prepare resources...*/
	if (!SelectCount) {
		Error(209);
		return (FALSE);
	}
	if (!OpenClipboard(hwndMain)) {
		ErrorBox(MB_ICONEXCLAMATION, 302);
		return (FALSE);
	}
	hMem = GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, 
					   (SelectCount+lstrlen(BinaryPart)+7));
	if (!hMem) {
		CloseClipboard();
		ErrorBox(MB_ICONEXCLAMATION, 304);
		return (FALSE);
	}
	pMem = GlobalLock(hMem);
	if (pMem == NULL) {
		CloseClipboard();
		GlobalFree(hMem);
		ErrorBox(MB_ICONEXCLAMATION, 305);
		return (FALSE);
	}

	/*copy into memory...*/
	{	POSITION Pos;
		ULONG    Bytes = SelectCount;
		UINT     i;
		BOOL	 NullbyteFound = FALSE;

		Pos = SelectStart;
		while (Bytes) {
			LPSTR p, pNull;
			CharAt(&Pos);	/*normalize position and enforce page load*/
			i = Pos.p->Fill - Pos.i;
			if (i > Bytes) i = (UINT)Bytes;
			p = Pos.p->PageBuf + Pos.i;
			if (!NullbyteFound && (pNull = _fmemchr(p, '\0', i)) != NULL) {
				if (ErrorBox(MB_ICONINFORMATION|MB_OKCANCEL, 328) == IDCANCEL) {
					CloseClipboard();
					GlobalUnlock(hMem);
					GlobalFree(hMem);
					return (FALSE);
				}
				NullbyteFound = TRUE;
				hmemcpy(pMem, p, i = pNull-p+1);
				pMem  += i;
				Bytes -= --i;
				Pos.i += i;
				hmemcpy(pMem, BinaryPart, i = lstrlen(BinaryPart));
				pMem  += i;
				hmemcpy(pMem, &Bytes, sizeof(Bytes));
				pMem  += sizeof(Bytes);
				continue;
			}
			hmemcpy(pMem, Pos.p->PageBuf + Pos.i, i);
			pMem  += i;
			Bytes -= i;
			Pos.i += i;
		}
	}

	/*unlock...*/
	GlobalUnlock(hMem);

	/*clear previous clipboard contents...*/
	if (!EmptyClipboard()) {
		CloseClipboard();
		GlobalFree(hMem);
		ErrorBox(MB_ICONEXCLAMATION, 303);
		return (FALSE);
	}

	/*finally, publish...*/
	SetClipboardData(CharSet ? CF_OEMTEXT : CF_TEXT, hMem);
	CloseClipboard();
	CheckClipboard();
	return (TRUE);
}

BOOL ClipboardCut(void) {
	if (!ClipboardCopy()) return (FALSE);
	DeleteSelected(17);
	GetXPos(&CurrCol);
	ShowEditCaret();
	return (TRUE);
}

BOOL ClipboardPaste(void) {
	HGLOBAL  hMem;
	char	 far *lpMem;
	UINT	 uFormat = CharSet ? CF_OEMTEXT : CF_TEXT;
	LONG	 lSize;
	MODEENUM SaveMode;

	if (IsViewOnly()) return(FALSE);

	/*prepare resources...*/
	if (!OpenClipboard(hwndMain)) {
		ErrorBox(MB_ICONEXCLAMATION, 302);
		return (FALSE);
	}
	hMem = GetClipboardData(uFormat);
	if (!hMem) {
		CloseClipboard();
		Error(306);
		return (FALSE);
	}
	lSize = GlobalSize(hMem);
	#if !defined(WIN32)
		if (lSize > 65535) {
			CloseClipboard();
			ErrorBox(MB_ICONSTOP, 307);
			return (FALSE);
		}
	#endif
	lpMem = GlobalLock(hMem);
	if (lpMem == NULL) {
		CloseClipboard();
		ErrorBox(MB_ICONEXCLAMATION, 305);
		return (FALSE);
	}
	SaveMode = Mode;
	if (Mode != InsertMode && Mode != ReplaceMode) {
		StartUndoSequence();
		Mode = InsertMode;
	}
	if (SelectCount) DeleteSelected(17);
	else EnterDeleteForUndo(&CurrPos, 0, 1);	/*force new undo element*/
	Mode = SaveMode;
	HideEditCaret();

	if (lSize) {
		LPSTR lp;
		LONG  Len = lstrlen(BinaryPart), lRemain;

		lp = _fmemchr(lpMem, '\0', (size_t)lSize);
		if (lp != NULL && (lp-lpMem+Len+7) <= lSize && lp[1]==BinaryPart[0]
					   && _fmemcmp(lp+1, BinaryPart, (size_t)Len) == 0) {
			/*insert first part of null containing binary data...*/
			if (*lpMem) InsertBuffer(lpMem, (UINT)(lp-lpMem), 1);
			Len += lp-lpMem+1;
			hmemcpy(&lRemain, lpMem+Len, sizeof(lRemain));
			if (lSize >= (lp-lpMem)+lRemain) lSize = lRemain;
			else lSize = 0;
			lpMem += Len + sizeof(lRemain);
		} else if (lp) lSize = lp-lpMem;
	}
	/*insert clipboard now...*/
	if (lSize) InsertBuffer(lpMem, (UINT)lSize, 1);
	GetXPos(&CurrCol);
	ShowEditCaret();

	/*clean up...*/
	GlobalUnlock(hMem);
	CloseClipboard();
	return (TRUE);
}
