/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <stdlib.h>
#include "resource.h"

#include "winvi.h"
#include "colors.h"
#include "paint.h"

UINT	 CurrColor;
COLORREF Colors[0x16] = {
	 RGB(000,000,000), RGB(128,000,000), RGB(000,128,000), RGB(128,128,000),
	 RGB(000,000,128), RGB(128,000,128), RGB(000,128,128), RGB(128,128,128),
	 RGB(192,192,192), RGB(255,000,000), RGB(000,255,000), RGB(255,255,000),
	 RGB(000,000,255), RGB(255,000,255), RGB(000,255,255), RGB(255,255,255),
};

static SysColors[4] = {COLOR_WINDOW,	COLOR_WINDOWTEXT,
					   COLOR_HIGHLIGHT, COLOR_HIGHLIGHTTEXT};
static HWND CurrFrame, CurrColorHwnd;
static int  CurrBmpInx;

LPARAM FAR PASCAL ColorProc(HWND hWnd, UINT uMsg, WPARAM wPar, LPARAM lPar)
	/*Color control window callback function*/
{	static char	Title[20];
	UINT		uColor;

	switch (uMsg) {
		case WM_KILLFOCUS:
			ShowWindow(hWnd, SW_HIDE);
			UpdateWindow(hWnd);
			ShowWindow(hWnd, SW_SHOW);
			break;

		case WM_SETFOCUS:
			InvalidateRect(hWnd, NULL, TRUE);
			if (!GetWindowText(hWnd, Title, sizeof(Title)))  break;
			if ((uColor = (int)strtol(Title, 0, 16)) > 0x15) break;
			{	PSTR p;

				switch (uColor) {
					case 0x10:	p = "COLOR_WINDOW";			break;
					case 0x11:	p = "COLOR_WINDOWTEXT";		break;
					case 0x12:	p = "COLOR_HIGHLIGHT";		break;
					case 0x13:	p = "COLOR_HIGHLIGHTTEXT";	break;
					case 0x14:	p = "";						break;
					case 0x15:	p = "";						break;
					default:
						CheckRadioButton(GetParent(hWnd),
										 IDC_COLORS, IDC_BITMAP, IDC_COLORS);
						p = NULL;
						break;
				}
				if (p) {
					BOOL EnableBitmap = uColor == 0x10;
					SetDlgItemText(GetParent(hWnd), IDC_SYSCOLOR, p);
					EnableWindow(GetDlgItem(GetParent(hWnd), IDC_SYSTEM),
								 *p != '\0');
					EnableWindow(GetDlgItem(GetParent(hWnd), IDC_BITMAP),
								 EnableBitmap);
					EnableWindow(GetDlgItem(GetParent(hWnd), IDC_BITMAPTITLE),
								 EnableBitmap);
					EnableWindow(GetDlgItem(GetParent(hWnd), IDC_BITMAPEDIT),
								 EnableBitmap);
					EnableWindow(GetDlgItem(GetParent(hWnd), IDC_BITMAPBROWSE),
								 EnableBitmap);
					if (CurrFrame) ShowWindow(CurrFrame, SW_HIDE);
					CurrFrame = GetDlgItem(GetParent(hWnd),
										   IDC_FRAME0-0x10+uColor);
					CurrColor = uColor;
					SetWindowLong(CurrColorHwnd, GWL_STYLE,
						GetWindowLong(CurrColorHwnd, GWL_STYLE) & ~WS_TABSTOP);
					CurrColorHwnd = hWnd;
					SetWindowLong(CurrColorHwnd, GWL_STYLE,
						GetWindowLong(CurrColorHwnd, GWL_STYLE) |  WS_TABSTOP);
					ShowWindow(CurrFrame, SW_SHOW);
					CheckRadioButton(GetParent(hWnd), IDC_COLORS, IDC_BITMAP,
						uColor==0x10 && *BitmapFilename ? IDC_BITMAP
							: wUseSysColors & (1 << uColor-0x10) ? IDC_SYSTEM
																 : IDC_COLORS);
				} else {
					if (CurrColor == 0x10) *BitmapFilename = '\0';
					wUseSysColors &= ~(1 << (CurrColor - 0x10));
					Colors[CurrColor] = Colors[uColor];
					InvalidateRect(CurrColorHwnd, 0, TRUE);
				}
			}
			break;

		case WM_LBUTTONDOWN:
			SetFocus(hWnd);
			break;

		case WM_GETDLGCODE:
			return (DLGC_WANTARROWS);

		case WM_KEYDOWN:
			if (!GetWindowText(hWnd, Title, sizeof(Title)))  break;
			if ((uColor = (int)strtol(Title, 0, 16)) > 0x15) break;
			{	UINT n;
				switch (wPar) {
					case VK_LEFT:	n = uColor>=0x10 ? -2 : -1;	break;
					case VK_RIGHT:	n = uColor>=0x10 ?  2 :  1;	break;
					case VK_UP:		n = uColor>=0x10 ? -1 : -8;	break;
					case VK_DOWN:	n = uColor>=0x10 ?  1 :  8;	break;
					default:		n = 0;
				}
				if (uColor <= 0x0fU) {
					   if (uColor+n > 0x0fU)					 n = 0;
				} else if (uColor+n > 0x15U || uColor+n < 0x10U) n = 0;
				if (n)
					SetFocus(GetDlgItem(GetParent(hWnd), IDC_COLOR0+uColor+n));
			}
			break;

		case WM_ERASEBKGND:
			{	HBRUSH		hbrBkgnd;
				RECT		Rect;

				GetClientRect(hWnd, &Rect);
				Rect.left+=1; Rect.top+=1; Rect.right-=1; Rect.bottom-=1;
				if (GetFocus() == hWnd) {
					FillRect((HDC)wPar, &Rect, SelBkgndBrush);
				}
				Rect.left+=2; Rect.top+=2; Rect.right-=2; Rect.bottom-=2;
				IntersectClipRect((HDC)wPar, Rect.left,  Rect.top,
											 Rect.right, Rect.bottom);
				if (!GetWindowText(hWnd, Title, sizeof(Title)))  break;
				if ((uColor = (int)strtol(Title, 0, 16)) > 0x15) break;
				if (uColor==0x10 && *BitmapFilename)
					PaintBmp((HDC)wPar, CurrBmpInx, &Rect, 0, 0);
				else {
					hbrBkgnd = CreateSolidBrush(Colors[uColor]);
					if (hbrBkgnd) {
						FillRect((HDC)wPar, &Rect, hbrBkgnd);
						DeleteObject(hbrBkgnd);
					}
				}
			}
			return (TRUE);

		default:
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));
	}
	return (0);
}

BOOL ColorsCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
	/*Color selection dialogbox callback function*/
{	static COLORREF	SaveFg, SaveBk, SaveSelFg, SaveSelBk, SaveShmFg, SaveShmBk;
	static char		SaveBm[260];
	static WORD		SaveSysColors;
	static BOOL		Applied, BitmapChanged;

	(VOID)lPar;		/*parameter not used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			lstrcpy(SaveBm, BitmapFilename);
			Colors[0x10]  = SaveBk	  = BackgroundColor;
			Colors[0x11]  = SaveFg	  = TextColor;
			Colors[0x12]  = SaveSelBk = SelBkgndColor;
			Colors[0x13]  = SaveSelFg = SelTextColor;
			Colors[0x14]  = SaveShmBk = ShmBkgndColor;
			Colors[0x15]  = SaveShmFg = ShmTextColor;
			SaveSysColors = wUseSysColors;
			CurrColor	  = 0x10;
			CurrColorHwnd = GetDlgItem(hDlg, IDC_EDIT_BG);
			CurrBmpInx	  = 0;
			Applied		  = FALSE;
			if (*BitmapFilename) {
				LowerCaseFname(BitmapFilename);
				SetDlgItemText(hDlg, IDC_BITMAPEDIT, BitmapFilename);
				CheckDlgButton(hDlg, IDC_BITMAP, TRUE);
			} else CheckDlgButton(hDlg,
								  wUseSysColors & 1 ? IDC_SYSTEM : IDC_COLORS,
								  TRUE);
			SetDlgItemText(hDlg, IDS_VERSION, GetViVersion());
			return (TRUE);

		case WM_COMMAND:
			if (NOTIFICATION != EN_KILLFOCUS) switch (COMMAND) {
				case IDC_COLORS:
					wUseSysColors &= ~(1 << (CurrColor - 0x10));
					if (CurrColor == 0x10) *BitmapFilename = '\0';
					InvalidateRect(CurrColorHwnd, NULL, TRUE);
					{	UINT	 i;
						COLORREF c = Colors[CurrColor];
						for (i=15; i; --i) if (Colors[i] == c) break;
						SetFocus(GetDlgItem(hDlg, IDC_COLOR0+i));
					}
					break;

				case IDC_SYSTEM:
					wUseSysColors |= 1 << (CurrColor - 0x10);
					if (CurrColor == 0x10) *BitmapFilename = '\0';
					Colors[CurrColor] = GetSysColor(SysColors[CurrColor-0x10]);
					InvalidateRect(CurrColorHwnd, NULL, TRUE);
					break;

				case IDC_BITMAP:
					GetDlgItemText(hDlg, IDC_BITMAPEDIT,
								   BitmapFilename, sizeof(BitmapFilename));
					CurrBmpInx = 1;
					ReleaseBmp(1);
					InvalidateRect(CurrColorHwnd, NULL, TRUE);
					break;

				case IDC_BITMAPBROWSE:
					{	HWND OldFocus = GetFocus();

						if (SelectBkBitmap(hDlg)) {
							LowerCaseFname(BitmapFilename);
							SetDlgItemText(hDlg, IDC_BITMAPEDIT,BitmapFilename);
							BitmapChanged = TRUE;
							CurrBmpInx = 1;
							ReleaseBmp(1);
							InvalidateRect(CurrColorHwnd, NULL, TRUE);
						}
						if (OldFocus) SetFocus(OldFocus);	/*for 16-bit*/
					}
					break;

				case IDC_BITMAPEDIT:
					GetDlgItemText(hDlg, IDC_BITMAPEDIT,
								   BitmapFilename, sizeof(BitmapFilename));
					if (*BitmapFilename) {
						OFSTRUCT Of;
						CheckRadioButton(hDlg,
										 IDC_COLORS, IDC_BITMAP, IDC_BITMAP);
						BitmapChanged = TRUE;
						if (OpenFile(BitmapFilename, &Of, OF_EXIST)
								!= HFILE_ERROR) {
							CurrBmpInx = 1;
							ReleaseBmp(1);
							InvalidateRect(CurrColorHwnd, NULL, TRUE);
						}
					}
					break;

				case ID_APPLY:
					Applied = TRUE;
					/*FALLTHROUGH*/
				case IDC_STORE:
				case IDOK:
					CurrBmpInx = 0;
					ReleaseBmp(1);
					if (BitmapChanged || *SaveBm != *BitmapFilename)
						ReleaseBmp(0);
					BackgroundColor	= SaveBk	= Colors[0x10];
					TextColor		= SaveFg	= Colors[0x11];
					SelBkgndColor	= SaveSelBk	= Colors[0x12];
					SelTextColor	= SaveSelFg	= Colors[0x13];
					ShmBkgndColor	= SaveShmBk	= Colors[0x14];
					ShmTextColor	= SaveShmFg	= Colors[0x15];
					SaveSysColors	= wUseSysColors;
					MakeBackgroundBrushes();
					if (hwndMain) InvalidateRect(hwndMain, NULL, TRUE);
					DlgResult = TRUE;
					BitmapChanged = FALSE;
					lstrcpy(SaveBm, BitmapFilename);
					switch (COMMAND) {
						case IDC_STORE:
							SaveColors();
							break;
						case IDOK:
							EndDialog(hDlg, TRUE);
							break;
					}
					break;

				case IDCANCEL:
					ReleaseBmp(1);
					if (BitmapChanged || *BitmapFilename != *SaveBm) {
						ReleaseBmp(0);
						BitmapChanged = FALSE;
						Applied = TRUE;
						lstrcpy(BitmapFilename, SaveBm);
					}
					if (Applied) {
						BackgroundColor	= SaveBk;
						TextColor		= SaveFg;
						SelBkgndColor	= SaveSelBk;
						SelTextColor	= SaveSelFg;
						ShmBkgndColor	= SaveShmBk;
						ShmTextColor	= SaveShmFg;
						wUseSysColors	= SaveSysColors;
						MakeBackgroundBrushes();
						if (hwndMain) InvalidateRect(hwndMain, NULL, TRUE);
					}
					EndDialog(hDlg, FALSE);
			}
			return (TRUE);
	}
	return (FALSE);
}
