/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <direct.h>
#include <malloc.h>
#include <assert.h>
#include "resource.h"
#include "winvi.h"
#include "page.h"
#include "status.h"
#include "toolbar.h"
#include "paint.h"
#include "exec.h"

#if !defined(_MSC_VER)
#define _chdir chdir	/*for Borland compiler*/
#endif

extern int nCurrFileListEntry, nNumFileListEntries;
extern int OemCodePage, AnsiCodePage;

PSTR	   Args;
int 	   CmdMask;
static int ExecError;
char	   FileNameBuf2[260];	/*used in [command.c]Compare:Substitute[#%],
								 *		  [command.c]CdCmd:_getcwd(),
								 *		  [winvi.c]MainProc:WM_DROPFILES and
								 *		  [file.c]Open/WriteCmd:OPENFILENAME
								 */

PSTR Compare(PSTR Cmd, PSTR CompareWith, int MinLength, int CheckMask)
	/* CheckMask: 1 = line allowed
	 *			  2 = line range allowed
	 *			  4 = arguments allowed
	 *			  8 = command may be forced with '!'
	 *		   0x10 = substitute '#'/'%' in arguments
	 */
{
	while (*Cmd == ' ') ++Cmd;
	IsAltFile = FALSE;
	for (;;) {
		int c = *Cmd++ ^ *CompareWith++;

		if (c && (c & ~('a'-'A') || CompareWith[-1] < 'a')) return (NULL);
		--MinLength;
		if (!isalpha(*Cmd) || !isalpha(CompareWith[-1])) {
			if (MinLength > 0) return (NULL);
			if (*Cmd == '!' && CheckMask & 8) {
				CommandWithExclamation = TRUE;
				++Cmd;
			}
			while (*Cmd == ' ') ++Cmd;
			{	int m = CmdMask;
				if (*Cmd) m |= 4;
				if (m &= ~CheckMask) {
					if		(m & 1) ExecError = 104;
					else if (m & 2) ExecError = 105;
					else if (m & 4) ExecError = 106;
					return (NULL);
				}
			}
			if (CheckMask & 0x10) {
				PSTR p, p1, p2;
				for (p = Cmd+lstrlen(Cmd)-1; p>=Cmd && *p==' '; --p)
					*p = '\0';
				p = Cmd;
				if (!lstrcmp(p, "#")) IsAltFile = TRUE;
				for (;;) {
					LPCSTR lp;
					int    len;

					p1 = strchr(p, '#');
					p2 = strchr(p, '%');
					if (p1!=NULL) p = p1;
					else if (p2==NULL) break;
					if (p2!=NULL && (p1==NULL || p2<p)) p = p2;
					if (Cmd != FileNameBuf2) {
						lstrcpy(FileNameBuf2, Cmd);
						p = FileNameBuf2 + (p-Cmd);
						Cmd = FileNameBuf2;
					}
					if (p[0] == p[1]) {
						/*special case:
						 *  "##" or "%%" treated as single character
						 *  which should not be converted (2.72 28-Feb-99)
						 */
						_fmemmove(p, p+1, lstrlen(p));
						++p;
						continue;
					}
					if	   (*p == '#')	 lp = AltFileName;
					else /*(*p == '%')*/ lp = CurrFile;
					if (!*lp) {
						ExecError = 113;
						return (NULL);
					}
					len = lstrlen(lp);
					_fmemmove(p+len, p+1, lstrlen(p));
					_fmemcpy(p, lp, len);
					p += len;
				}
			}
			return (Cmd);
		}
	}
}

#if defined(WIN32)
BOOL CALLBACK SetForegroundWndProc(HWND hWnd, LPARAM lParam)
{
	(VOID)lParam;	/*parameter not used*/
	SetForegroundWindow(hWnd);
	return (TRUE);
}
#endif

BOOL NewViEditCmd(PSTR Args)
{	PSTR p;
	UINT Err = 0;

	if ((p = malloc(lstrlen(Args) + 266)) != NULL) {
		*p = '"';
		if (GetModuleFileName(hInst, p+1, 260) > 0) {
			wsprintf(p+lstrlen(p), "\" \"%s\"", (LPSTR)Args);
			if (*Args) SetAltFileName(Args, NULL);
			#if defined(WIN32)
			{	static PROCESS_INFORMATION	pi;
				static STARTUPINFO			si;

				si.cb	   = sizeof(si);
				si.dwFlags = STARTF_FORCEONFEEDBACK;
				if (CreateProcess(NULL, p, NULL, NULL, FALSE, 
								  CREATE_DEFAULT_ERROR_MODE, NULL, NULL,
								  &si, &pi)) {
					WaitForInputIdle(pi.hProcess, 500);
					EnumThreadWindows(pi.dwThreadId, SetForegroundWndProc, 0);
				} else {
					FileError(p, 300, 332);
				}
			}
			#else
				if ((Err = WinExec(p, SW_NORMAL)) < 32)
					FileError(p, (int)Err+300, 332);
			#endif
		}
		free(p);
	} else Error(300);		/*fatal: no memory*/
	return (Err >= 32);
}

static BOOL RewindCmd(void)
{	LPSTR lp;

	if (nNumFileListEntries && (lp = GetFileListEntry(0), lp != NULL)) {
		if (!AskForSave(hwndMain, 1)) return (FALSE);
		Unsafe = FALSE;		/*already asked for save*/
		nCurrFileListEntry = 0;
		Open(lp, 0);
		return (TRUE);
	}
	Error(221);		/*no file list error*/
	return (FALSE);
}

BOOL MoreFilesWarned = FALSE, Exiting = FALSE;

BOOL QuitCmd(BOOL AskIfMoreFiles)
{
	if (nCurrFileListEntry < nNumFileListEntries-1 && !QuitIfMoreFlag) {
		if (!CommandWithExclamation && !MoreFilesWarned) {
			MoreFilesWarned = TRUE;
			if (!AskIfMoreFiles) {
				Error(232);
				return (FALSE);
			}
			ErrorBox(MB_ICONEXCLAMATION, 315);
			return (FALSE);
		}
	}
	SaveSettings();
	Exiting = TRUE;
	DestroyWindow(hwndMain);
	assert(hwndMain == 0);
	return (TRUE);
}

extern char Buffer[300];

static BOOL CdCmd(PSTR Dir)
{	PSTR p;

	if (*Dir) {
		extern BOOL NotEdited;
		BOOL		OtherLocation = FALSE;

		/*first check if current file has full path...*/
		do {	/*no loop, for break only*/
			if (!CurrFile || !*CurrFile) break;
			if (CharFlags[(BYTE)*CurrFile] & 64) {
				if (CharFlags[(BYTE)CurrFile[1]] & 64) break;
				if (CharFlags[(BYTE)Dir[0]] & 64) {
					if (!(CharFlags[(BYTE)Dir[1]] & 64)) break;
				} else {
					if (Dir[1] != ':') break;
					/*check if same drive...*/
					p = _getcwd(FileNameBuf2, sizeof(FileNameBuf2));
					if (p && p[1]==':' && *p==*Dir) break;
				}
			} else if (CurrFile[1] == ':' && CharFlags[(BYTE)CurrFile[2]] & 64)
				break;
			if (Unsafe && !AskForSave(hwndMain, 2)) return (FALSE);
			OtherLocation = TRUE;
		} while (FALSE);
		#if !defined(WIN32)
			if ((p = Dir+lstrlen(Dir)-1) >= Dir +
											(Dir[0] && Dir[1]==':' ? 3 : 1) &&
					 (*p=='\\' || *p=='/'))
				*p = '\0';
			AnsiToOem(Dir, Dir);
		#endif
		if (isalpha(Dir[0]) && Dir[1]==':'
				? Dir[2] && _chdir(Dir) || _chdrive(*Dir & 31) : _chdir(Dir)) {
			#if !defined(WIN32)
				OemToAnsi(Dir, Dir);
			#endif
			Error(227+TerseFlag, (LPSTR)Dir);
			return (FALSE);
		}
		NotEdited = OtherLocation;
	}
	p = _getcwd(FileNameBuf2, sizeof(FileNameBuf2));
	if (p == NULL) {
		Error(229);
		return (FALSE);
	}
	{	static char Fmt[30];
		#if !defined(WIN32)
			OemToAnsi(p, p);
		#endif
		LowerCaseFname(p);
		if (!*Fmt) LOADSTRING(hInst, 237, Fmt, sizeof(Fmt));
		wsprintf(Buffer, Fmt, (LPSTR)p);
	}
	NewStatus(0, Buffer, NS_NORMAL);
	return (TRUE);
}

int LinesVar, ColumnsVar;

void ResizeWithLinesAndColumns(void)
{	int  Width, Height;
	RECT r;

	GetWindowRect(hwndMain, &r);
	Width  = r.right  - r.left;
	Height = r.bottom - r.top;
	if (LinesVar) {
		Height -= ClientRect.bottom - ClientRect.top;	/*height of nc area*/
		Height += StatusHeight + TOOLS_HEIGHT + 1;		/*height of non-text*/
		Height += 2*YTEXT_OFFSET + LinesVar*TextHeight;	/*new height*/
	}
	if (ColumnsVar) {
		Width -= ClientRect.right - ClientRect.left;	/*width of nc area*/
		Width += GetSystemMetrics(SM_CXVSCROLL);		/*width of non-text*/
		Width += XTEXT_OFFSET-1 + StartX(FALSE) + ColumnsVar*AveCharWidth;
														/*new width*/
	}
	SetWindowPos(hwndMain, 0,0,0, Width, Height, SWP_NOMOVE | SWP_NOZORDER);
	if (ColumnsVar && BreakLineFlag) InvalidateText();
}

#define VF_BOOLEAN 1
#define VF_DECIMAL 2
#define VF_STRING  4

static BOOL SetCmd(PSTR Args)
{	static struct {
		WORD VarFlags;
		int  Min, Max;
		PSTR Name, Abbrev;
		int  *Variable;
	} VarDescr[] = {
		{0,			 0, 0,	 "all",			 "all",	  NULL},
		{VF_BOOLEAN, 0, 0,	 "ansimode",	 "am",	  (int*)&CharSet},
		{VF_DECIMAL, 0, 2000,"ansicodepage", "ansicp",(int*)&AnsiCodePage},
		{VF_BOOLEAN, 0, 0,	 "autoindent",	 "ai",	  (int*)&AutoIndentFlag},
		{VF_BOOLEAN, 0, 0,	 "autowrite",	 "aw",	  (int*)&AutoWriteFlag},
		{VF_DECIMAL, 8, 500, "columns",		 "co",			&ColumnsVar},
		{VF_BOOLEAN, 0, 0,	 "dosmode",		 "dm",	  (int*)&CharSet},
		{VF_BOOLEAN, 0, 0,	 "hexmode",		 "hm",	  (int*)&HexEditMode},
		{VF_BOOLEAN, 0, 0,	 "ignorecase",	 "ic",	  (int*)&IgnoreCaseFlag},
		{VF_DECIMAL, 1, 200, "lines",		 "li",			&LinesVar},
		{VF_BOOLEAN, 0, 0,	 "list",	   "list",	  (int*)&ListFlag},
		{VF_BOOLEAN, 0, 0,	 "magic",	  "magic",	  (int*)&MagicFlag},
		{VF_BOOLEAN, 0, 0,	 "number",		 "nu",	  (int*)&NumberFlag},
		{VF_DECIMAL, 0, 2000,"oemcodepage",	 "oemcp", (int*)&OemCodePage},
		{VF_BOOLEAN, 0, 0,	 "readonly",	 "ro",	  (int*)&ReadOnlyFlag},
		{VF_DECIMAL, 0, 80,	 "scroll",		 "sc",			&MaxScroll},
		{VF_DECIMAL, 1, 128, "shiftwidth",	 "sw",			&ShiftWidth},
		{VF_BOOLEAN, 0, 0,	 "showmatch",	 "sm",	  (int*)&ShowMatchFlag},
		{VF_DECIMAL, 1, 128, "tabstop",		 "ts",			&TabStop},
		{VF_STRING,  0, 0,	 "tags",	   "tags",	  (int*)&Tags},
		{VF_BOOLEAN, 0, 0,	 "terse",	  "terse",	  (int*)&TerseFlag},
		{VF_STRING,  0, 0,	 "tmpdirectory","tmp",	  (int*)&TmpDirectory},
		{VF_BOOLEAN, 0, 0,	 "viewonly",	 "vo",	  (int*)&ViewOnlyFlag},
		{VF_BOOLEAN, 0, 0,	 "wrapscan",	 "ws",	  (int*)&WrapScanFlag},
		{0,			 0, 0,	 NULL,			 NULL,	  NULL}
	};
	static char NameBuf[16];
	int			i;

	LinesVar = ColumnsVar = 0;
	if (!*Args) PostMessage(hwndMain, WM_COMMAND, ID_SETTINGS, 0);
	else do {
		BOOL Off = FALSE;
		if (LCASE(Args[0])=='n' && LCASE(Args[1])=='o') {
			Args += 2;
			Off = TRUE;
		}
		for (i=0; i<sizeof(NameBuf)-1 && isalpha(Args[i]); ++i) {
			NameBuf[i] = Args[i];
		}
		NameBuf[i] = '\0';
		Args += i;
		for (i=0; VarDescr[i].Name && lstrcmpi(NameBuf, VarDescr[i].Name)
								   && lstrcmpi(NameBuf, VarDescr[i].Abbrev);
				  ++i);
		if (VarDescr[i].Name) {
			if (VarDescr[i].VarFlags & VF_BOOLEAN) {
				if ((BOOL*)VarDescr[i].Variable == &HexEditMode) {
					if (*(BOOL*)VarDescr[i].Variable != !Off) ToggleHexMode();
				} else if ((UINT*)VarDescr[i].Variable == &CharSet) {
					if (*VarDescr[i].Name == 'a') Off = !Off;
					if (*(BOOL*)VarDescr[i].Variable != !Off)
						SetCharSet(!Off, 3);
				} else {
					*(BOOL*)VarDescr[i].Variable = !Off;
					if		((BOOL*)VarDescr[i].Variable == &ReadOnlyFlag)
						ReadOnlyFile = ReadOnlyFlag;
					else if ((BOOL*)VarDescr[i].Variable == &ViewOnlyFlag)
						CheckClipboard();
					else if ((BOOL*)VarDescr[i].Variable == &ListFlag ||
							 (BOOL*)VarDescr[i].Variable == &NumberFlag)
						if (!HexEditMode) InvalidateText();
				}
			} else if (VarDescr[i].VarFlags & VF_DECIMAL) {
				if (Off || *Args++!='=') Error(222);	/*not a boolean*/
				else {
					long Value;
					Value = strtol(Args, &Args, 10);
					if (Value < VarDescr[i].Min || Value > VarDescr[i].Max)
						Error(223, VarDescr[i].Min, VarDescr[i].Max);
					else if (*VarDescr[i].Variable != (int)Value) {
						*VarDescr[i].Variable = (int)Value;
						if (VarDescr[i].Variable == &TabStop) {
							HideEditCaret();
							InvalidateArea(&ScreenStart, 0, 7);
							NewPosition(&CurrPos);
							ShowEditCaret();
						} else if (VarDescr[i].Variable == &OemCodePage ||
								   VarDescr[i].Variable == &AnsiCodePage) {
							extern VOID SetNewCodePage(VOID);
							SetNewCodePage();
						}
					}
				}
			} else if (VarDescr[i].VarFlags & VF_STRING) {
				PSTR p;
				int  c;

				if (Off || *Args++!='=') Error(222);	/*not a boolean*/
				if (*Args=='"') {
					c = '"';
					++Args;
				} else c = '\0';
				for (p=Args; *Args && *Args!=c; ++Args);
				if (!*Args) c = '\0';
				else *Args = '\0';
				AllocString((LPSTR*)VarDescr[i].Variable, p);
				*Args = c;
				if (c) ++Args;
				DefaultTmpFlag = TmpDirectory==NULL || !*TmpDirectory;
			} else {
				/*set all*/
				if (Off || *Args=='=') {
					Error(226, (LPSTR)NameBuf);
					return (FALSE);
				}
				PostMessage(hwndMain, WM_COMMAND, ID_SETTINGS, 0);
			}
		} else {
			Error(226, (LPSTR)NameBuf);	 /*no such option*/
			return (FALSE);
		}
		while (*Args == ' ') ++Args;
	} while (*Args);
	if (*Args) {
		Error(106);
		return (FALSE);
	}
	ResizeWithLinesAndColumns();
	return (TRUE);
}

static void DeleteCmd(long StartLine, long EndLine, PSTR Args)
{
	(VOID)Args;		/*parameter not used*/
	SelectCount = 0;
	if (StartLine > 0 && EndLine >= StartLine) {
		Position(StartLine, 'G', 11);
		Position(1, '0', 11);
		Position(EndLine-StartLine+1, 'j', 31);
		DeleteSelected(1);
		GotChar('\033', 0);
		FindValidPosition(&CurrPos, 0);
		Position(1, '^', 0);
	}
}

PSTR LastCmd;

static BOOL ExecCmd(PSTR Args)
{	UINT Err;
	PSTR NewCmd;

	do {	/*no loop, for break in case of errors only*/
		if (*Args == '!') {
			if (!LastCmd) {
				Error(235);
				return (FALSE);
			}
			NewCmd = malloc(lstrlen(LastCmd) + lstrlen(Args));
			if (NewCmd == NULL) break;
			lstrcpy(NewCmd, LastCmd);
			lstrcat(NewCmd, Args+1);
		} else {
			NewCmd = malloc(lstrlen(Args) + 2);
			if (NewCmd == NULL) break;
			*NewCmd = '!';
			lstrcpy(NewCmd+1, Args);
		}
		if (LastCmd != NULL) free(LastCmd);
		LastCmd = NewCmd;
		NewStatus(0, LastCmd, NS_NORMAL);
		if ((Err = WinExec(LastCmd+1, SW_NORMAL)) < 32)
			FileError(LastCmd, (int)Err+300, 332);
		return (TRUE);
	} while (FALSE);
	Error(300);
	return (FALSE);
}

static void VersionCmd(void)
{
	wsprintf(Buffer, "WinVi - %s", (LPSTR)GetViVersion());
	NewStatus(0, Buffer, NS_NORMAL);
}

static PSTR GetAddr(PSTR Line, ULONG *Ptr)
{
	CmdMask = 0;
	switch (*Line) {
		case '.': *Ptr = CountLines(&CurrPos) + 1;
				  /*FALLTHROUGH*/
		case '$': if (*Line == '$') *Ptr = LinesInFile + IncompleteLastLine();
				  ++Line;
				  CmdMask |= 1;
				  break;

		case '\'':/*line of marker position...*/
				  {	  extern POSITION MarkPos[27];
					  int			  c;

					  if ((c = LCASE(Line[1])) < 'a' || c > 'z') {
						  Error(331);
						  return (NULL);
					  }
					  if (MarkPos[c &= 31].p == NULL) {
						  Error(332);
						  return (NULL);
					  }
					  Line += 2;
					  CmdMask |= 1;
					  *Ptr = CountLines(&MarkPos[c]) + 1;
				  }
				  break;
		case '/':
		case '?': {   POSITION    p;
					  extern BOOL MatchForward;
					  int		  Forward;

					  p = CurrPos;
					  Forward = *Line=='/' ? 1 : 0;
					  Line = BuildMatchList(Line, *Line, NULL);
					  if (!StartSearch()) return (NULL);
					  if (Forward)
					  	   do; while (!(CharFlags[CharAndAdvance(&p)] & 1));
					  else do; while (!(CharFlags[GoBackAndChar (&p)] & 1));
					  if (!SearchIt(&p, Forward | 2)) {
						  Error(Interrupted ? 218 : Forward ? 210 : 212);
						  if (Disabled) {
							  Enable();
							  ShowEditCaret();
						  }
						  return (NULL);
					  }
					  {	  BOOL SaveHexEdit = HexEditMode;
						  HexEditMode = FALSE;
						  *Ptr = CountLines(&p) + 1;
						  HexEditMode = SaveHexEdit;
					  }
					  if (Disabled) Enable();
				  }
				  CmdMask |= 1;
				  break;
	}
	while (*Line == ' ') ++Line;
	{	int  Sign;
		if (*Line == '+' || *Line == '-') {
			if (!(CmdMask & 1)) *Ptr = CountLines(&CurrPos) + 1;
			do {
				Sign = *Line=='-' ? -1 : 1;
				do; while (*++Line == ' ');
				if (*Line >= '0' && *Line <= '9') break;
				*Ptr += Sign;
			} while (*Line=='+' || *Line=='-');
			CmdMask |= 1;
		} else if (*Line >= '0' && *Line <= '9') {
			if (CmdMask & 1) return (Line); /*invalid*/
			CmdMask |= 1;
			*Ptr = 0;
			Sign = 1;
		} else return (Line);
		{	long n;
			n = 0;
			while (*Line >= '0' && *Line <= '9') n = 10 * n + *Line++ - '0';
			*Ptr += Sign * n;
		}
		while (*Line == ' ') ++Line;
	}
	return (Line);
}

BOOL CommandExec(PSTR Cmd)
{	BOOL  Ok = FALSE;
	ULONG StartLine, EndLine;
	/* StartLine:					EndLine:
	 *		-1: unspecified,			-1: unspecified,
	 *		 0: first line (0, %),		-2: last line ($, %)
	 *		>0:	specified line		  0,>0: specified line
	 */

	ExecError = 102;
	if (*Cmd != ':') {
		if (*Cmd=='/' || *Cmd=='?') {
			BuildMatchList(Cmd, *Cmd, NULL);
			return (TRUE);
		}
		if (*Cmd=='!') {
			CmdMask = 0;
			Args = Compare(Cmd, "!", 1, 0x14);	/*...replace '#' and '%'*/
			if (Args != NULL) Ok = ExecPiped(Args, 3);
			else assert(Args != NULL);
		} else Error(103);
		return (Ok);
	}
	do; while (*++Cmd==' ' || *Cmd==':');
	if (!*Cmd) return (TRUE);
	StartLine = EndLine = (ULONG)-1L;
	if (*Cmd == '%') {
		++Cmd;
		StartLine = 0;
		EndLine = (ULONG)-2L;
		CmdMask = 3;
	} else {
		Cmd = GetAddr(Cmd, &StartLine);
		if (Cmd != NULL && *Cmd == ',') {
			EndLine = CountLines(&CurrPos) + 1;
			if (!(CmdMask & 1)) StartLine = EndLine;
			Cmd = GetAddr(Cmd+1, &EndLine);
			if (EndLine >= LinesInFile + IncompleteLastLine() &&
				EndLine != (ULONG)-1L)
					EndLine = (ULONG)-2L;
			CmdMask = 3;
		}
		if (Cmd == NULL) return (FALSE);
	}
	if		  ((Args = Compare(Cmd, "write",   1, 0x1f)) != NULL) {
		Ok = WriteCmd(StartLine, EndLine, Args, (WORD)CommandWithExclamation);
	} else if ((Args = Compare(Cmd, "edit",	   1, 0x1c)) != NULL) {
		Ok = EditCmd(Args);
	} else if ((Args = Compare(Cmd, "vi",	   2, 0x1c)) != NULL) {
		Ok = NewViEditCmd(Args);
	} else if ((Args = Compare(Cmd, "rewind",  3,	 8)) != NULL) {
		Ok = RewindCmd();
	} else if ((Args = Compare(Cmd, "args",	   2,	 8)) != NULL) {
		ArgsCmd();
		Ok = TRUE;
	} else if ((Args = Compare(Cmd, "filename",1, 0x1c)) != NULL) {
		FilenameCmd(Args);
		Ok = TRUE;
	} else if ((Args = Compare(Cmd, "read",	   1, 0x15)) != NULL) {
		Ok = ReadCmd(StartLine, Args);
	} else if ((Args = Compare(Cmd, "next",	   1, 0x1c)) != NULL) {
		Ok = NextCmd(Args);
	} else if ((Args = Compare(Cmd, "wq",	   2, 0x1f)) != NULL) {
		if ((StartLine!=(ULONG)-1L && (StartLine>1U || EndLine!=(ULONG)-2L))
				&& !CommandWithExclamation)
			Error(224);
		else if (WriteCmd(StartLine, EndLine, Args, 
						  (WORD)CommandWithExclamation)) {
			Unsafe = FALSE;
			Ok = QuitCmd(FALSE);
		}
	} else if ((Args = Compare(Cmd, "xit",	   1, 0x1f)) != NULL) {
		if (StartLine!=(ULONG)-1L && (StartLine>1U || EndLine!=(ULONG)-2L))
			Error(225);
		else if (!Unsafe || WriteCmd(StartLine, EndLine, Args,
									 (WORD)CommandWithExclamation)) {
			Unsafe = FALSE;
			Ok = QuitCmd(FALSE);
		}
	} else if (((Args = Compare(Cmd, "cd",	   2, 0x1c)) != NULL) ||
			   ((Args = Compare(Cmd, "chdir",  3, 0x1c)) != NULL)) {
		Ok = CdCmd(Args);
	} else if ((Args = Compare(Cmd, "tag",	   2,	12)) != NULL) {
		Ok = Tag(Args, (WORD)(CommandWithExclamation | 2));
	} else if ((Args = Compare(Cmd, "quit",	   1,	 8)) != NULL) {
		if (AskForSave(hwndMain, 0)) {
			Ok = QuitCmd(FALSE);
			if (Ok) Unsafe = FALSE;
		}
	} else if ((Args = Compare(Cmd, "set",	   2,	 4)) != NULL) {
		Ok = SetCmd(Args);
	} else if ((Args = Compare(Cmd, "substitute", 1, 7)) != NULL) {
		if (!(CmdMask & 1)) StartLine = CountLines(&CurrPos) + 1;
		if (!(CmdMask & 2)) EndLine   = StartLine;
		Ok = SubstituteCmd(StartLine, EndLine, Args);
	} else if ((Args = Compare(Cmd, "delete",  1,	 3)) != NULL) {
		if (!(CmdMask & 1)) StartLine = CountLines(&CurrPos) + 1;
		else if (StartLine == 0) StartLine = 1;
		if (!(CmdMask & 2)) EndLine = StartLine;
		else if (EndLine == (ULONG)-2) EndLine = 2147000000;
		DeleteCmd(StartLine, EndLine, Args);
		Ok = TRUE;
	#if 0
	//	} else if ((Args = Compare(Cmd, "yank",	   1,	 3)) != NULL) {
	//	} else if ((Args = Compare(Cmd, "global",  1,	 3)) != NULL) {
	//	} else if ((Args = Compare(Cmd, "v",	   1,	 3)) != NULL) {
	#endif
	} else if ((Args = Compare(Cmd, "version", 2,	 0)) != NULL) {
		VersionCmd();
		Ok = TRUE;
	} else if ((Args = Compare(Cmd, "=",	   1,	 1)) != NULL) {
		wsprintf(Buffer, "%ld", CmdMask&1 ? StartLine : CountLines(&CurrPos)+1);
		NewStatus(0, Buffer, NS_NORMAL);
		Ok = TRUE;
	} else if ((Args = Compare(Cmd, "!",	   1, 0x14)) != NULL) {
		Ok = ExecCmd(Args);
	} else if (*Cmd) {
		for (Args=Cmd; *Args && *Args != ' '; ++Args);
		*Args = '\0';
		Error(ExecError, (LPSTR)Cmd);
		Ok = FALSE;
	} else {
		if (!StartLine) StartLine = 1;
		Ok = Position(StartLine, 'G', 0);
	}
	CommandWithExclamation = IsAltFile = FALSE;
	return (Ok);
}
