/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <assert.h>
#include <malloc.h>
#include "winvi.h"
#include "page.h"
#include "exec.h"

#if defined(WIN32)

extern PSTR	 LastCmd;

static LPSTR lpMemData;
static DWORD dwMemSize;

DWORD WINAPI InputThread(LPVOID lpParam)
	/*this thread handles command input*/
{	HANDLE Pipe = (HANDLE)lpParam;
	DWORD  Written;

	if (lpMemData) WriteFile(Pipe, lpMemData, dwMemSize, &Written, NULL);
	CloseHandle(Pipe);
	return (0);
}

static HANDLE			   hSemaphore;
static PROCESS_INFORMATION pi;
static volatile LPSTR	   pError = "No problem";
static volatile BOOL	   Done, GotFromPipe, WaitingForPaint;
static volatile HANDLE	   hThread1, hThread2, hThread3;

DWORD WINAPI OutputThread(LPVOID lpParam)
	/*this thread handles command output, i.e. WinVi input*/
{	HANDLE PipeVcv = (HANDLE)lpParam;
	BOOL   Ok	   = TRUE;

	for (;;) {
		static char	Buf[2048];
		DWORD		Read;

		WaitingForPaint = TRUE;
		WaitForSingleObject(hSemaphore, INFINITE);
		WaitingForPaint = FALSE;
		PeekNamedPipe(PipeVcv, NULL, 0, NULL, &Read, NULL);
		if (Read <= 0 && Ok && !Done) Read = 1;
		else if (Read > sizeof(Buf)) Read = sizeof(Buf);
		if (Read && !(Ok = ReadFile(PipeVcv, Buf, Read, &Read, NULL)))
			Read = 0;
		GotFromPipe = TRUE;
		if (Read <= 0) {
			CloseHandle(PipeVcv);
			PostMessage(hwndMain, WM_USER+1235, 0, 0);
			return (0);
		}
		PostMessage(hwndMain, WM_USER+1235, (WPARAM)Read, (LPARAM)Buf);
	}
}

DWORD WINAPI WatchThread(LPVOID lpParam)
{
	(VOID)lpParam;	/*parameter not used*/
	WaitForSingleObject((HANDLE)pi.hProcess, INFINITE);
	Done = TRUE;
	if (hThread1 && !WaitForSingleObject(hThread1, 0))
		TerminateThread(hThread1, 0);
	if (!WaitingForPaint) {
		GotFromPipe = FALSE;
		Sleep(1000);
		if (!GotFromPipe && !WaitingForPaint) {
			TerminateThread(hThread2, 0);
			PostMessage(hwndMain, WM_USER+1235, 0, 0);
		}
	}
	CloseHandle(pi.hThread);
	CloseHandle(pi.hProcess);
	pi.hProcess = NULL;
	return (0);
}

static POSITION InsertPos;

#endif

VOID ExecInput(WPARAM Read, LPARAM pBuf)
{
	#if defined(WIN32)
		if (Read) {
			LPSTR		Buf = (LPSTR)pBuf;
			int			nActPos;
			POSITION	OldPos;
			static BOOL	LastXOffState;

			nActPos = ComparePos(&CurrPos, &InsertPos);
			if (nActPos != 0) {
				OldPos  = CurrPos;
				CurrPos = InsertPos;
			}
			ViewOnlyFlag = FALSE;
			InsertBuffer(Buf, Read, 0);
			InsertPos = CurrPos;
			if (Buf[Read-1] == '\r') {
				GoBackAndChar(&InsertPos);
				Advance(&InsertPos, 1);
			}
			if (nActPos > 0) {
				while (OldPos.i > OldPos.p->Fill && OldPos.p->Next) {
					OldPos.i -= OldPos.p->Fill;
					OldPos.p = OldPos.p->Next;
				}
				Advance(&OldPos, Read);
				assert(ComparePos(&OldPos, &InsertPos) > 0);
			}
			NewPosition(nActPos==0 ? &InsertPos : &OldPos);
			ShowEditCaret();
			ViewOnlyFlag  = TRUE;
			if (XOffState != LastXOffState) {
				NewStatus(1, XOffState ? "Hold" : "", NS_NORMAL);
				LastXOffState = XOffState;
			}
			ReleaseSemaphore(hSemaphore, 1, NULL);
		} else if (ViewOnlyFlag) {
			GotChar('\033', 0);
			FindValidPosition(&CurrPos, 0);
			ShowEditCaret();
			CloseHandle(hSemaphore);
			ViewOnlyFlag = Interruptable = FALSE;
			CheckClipboard();
			if (pError) MessageBox(hwndMain, pError, "Test", MB_OK);
			if (hThread1) {CloseHandle(hThread1); hThread1 = 0;}
			if (hThread2) {CloseHandle(hThread2); hThread2 = 0;}
			if (hThread3) {CloseHandle(hThread3); hThread3 = 0;}
		}
	#endif
}

VOID ExecInterrupt(int Interrupted)
{
	#if defined(WIN32)
		if (pi.hProcess)
			GenerateConsoleCtrlEvent(Interrupted, pi.dwProcessId);
	#endif
}

BOOL ExecPiped(PSTR Cmd, WORD Flags)
	/*Flags: 1=use input  pipe (selected part piped from winvi)
	 *		 2=use output pipe (insert command output into current file)
	 */
{
	#if defined(WIN32)
		HANDLE					   PipeVvc, PipeCvc, PipeVcv, PipeCcv;
		/*PipeXyz =	pipe to be used in X for data flow from y to z,
		 *			V/v:	vi end of pipe,
		 *			C/c:	command end of pipe.
		 */
		static STARTUPINFO		   si;
		static SECURITY_ATTRIBUTES sa;
		DWORD					   dwFlags/*, Read;
		char					   Buf[2048]*/;
		PSTR					   Cmd2 = NULL, NewCmd;

		if (*Cmd == '!') {
			if (!LastCmd) {
				Error(235);
				return (FALSE);
			}
			NewCmd = malloc(lstrlen(LastCmd) + lstrlen(Cmd));
			if (NewCmd == NULL) {
				Error(300);
				return (FALSE);
			}
			lstrcpy(NewCmd, LastCmd);
			lstrcat(NewCmd, Cmd+1);
		} else {
			NewCmd = malloc(lstrlen(Cmd) + 2);
			if (NewCmd == NULL) {
				Error(300);
				return (FALSE);
			}
			*NewCmd = '!';
			lstrcpy(NewCmd+1, Cmd);
		}
		if (LastCmd) free(LastCmd);
		LastCmd = NewCmd;
		NewStatus(0, LastCmd, NS_NORMAL);
		Cmd = LastCmd + 1;

		sa.nLength = sizeof(sa);
		sa.bInheritHandle = TRUE;
		if (!CreatePipe(&PipeCvc, &PipeVvc, &sa, 2048)) return (FALSE);
		if (!CreatePipe(&PipeVcv, &PipeCcv, &sa, 2048)) {
			CloseHandle(PipeCvc);
			CloseHandle(PipeVvc);
			return (FALSE);
		}
		Interrupted = FALSE;
		#if 0
		//	Interruptable = TRUE;
		#endif
		dwFlags = CREATE_NEW_PROCESS_GROUP /*| CREATE_DEFAULT_ERROR_MODE*/;
		si.cb		  = sizeof(si);
		si.dwFlags	  = STARTF_USESTDHANDLES;
		si.hStdInput				 = PipeCvc;
		si.hStdOutput = si.hStdError = PipeCcv;
		if (Flags & 2) {
			Cmd2 = malloc(lstrlen(Cmd) + 19);
			if (Cmd2 != NULL) {
				static PSTR	 CmdName = "c:\\command.com";
				static HFILE hf		 = (HFILE)INVALID_HANDLE_VALUE;

				if (hf == (HFILE)INVALID_HANDLE_VALUE) {
					if (GetVersion() & 0x80000000U) {
						OFSTRUCT TmpOf;

						hf = OpenFile(CmdName, &TmpOf, OF_READ);
						if (hf != (HFILE)INVALID_HANDLE_VALUE) _lclose(hf);
						else CmdName += 3;	/*can hopefully be found in path*/
					} else	 CmdName = "cmd.exe";
					hf = 0;
				}
				wsprintf(Cmd2, "%s /c %s", CmdName, Cmd);
				Cmd = Cmd2;
			}
			#if 1
				si.wShowWindow = SW_SHOWMINNOACTIVE;
			#else
			//	si.wShowWindow = SW_HIDE;
			#endif
			si.dwFlags	  |= STARTF_USESHOWWINDOW;
		}
		if (CreateProcess(NULL,	   Cmd,  NULL, NULL, TRUE,
						  dwFlags, NULL, NULL, &si,  &pi)) {
			extern HGLOBAL YankHandles[27];
			extern ULONG   YankLength[27];
			extern int	   YankIndex;

			CloseHandle(PipeCvc);
			CloseHandle(PipeCcv);
			if (!SelectCount) Flags &= ~1;
			if (SelectCount) DeleteSelected(25);
			if (Flags & 1) {
				DWORD TId;
				lpMemData = (LPSTR)GlobalLock(YankHandles[YankIndex]);
				dwMemSize = YankLength[YankIndex];
				hThread1 = CreateThread(NULL, 0, InputThread, PipeVvc, 0, &TId);
			} else {
				CloseHandle(PipeVvc);
				hThread1 = NULL;
			}
			InsertPos = CurrPos;
			Interruptable = TRUE;
			if (Flags & 2) {
				DWORD /*ExitTime = 0,*/ TId2, TId3;
				Mode = InsertMode;
				NewPosition(&CurrPos);
				SwitchCursor(SWC_TEXTCURSOR);
				ShowEditCaret();
				XOffState = Done = WaitingForPaint = FALSE;
				pError = NULL;
				if (Cmd2) free(Cmd2);
				hSemaphore = CreateSemaphore(NULL, 1, 1, NULL);
				ViewOnlyFlag = TRUE;
				CheckClipboard();
				hThread3 = NULL;
				hThread2 =
					CreateThread(NULL, 0, OutputThread, PipeVcv, 0, &TId2);
				hThread3 = CreateThread(NULL, 0, WatchThread, NULL, 0, &TId3);
				return (TRUE);

#if 0
				for (;;) {
					extern BOOL XOffState;
					static BOOL LastXOffState;

					ViewOnlyFlag = TRUE;
					Interruptable = TRUE;
					MessageLoop(FALSE);
					if (Interrupted) {
						GenerateConsoleCtrlEvent
											(Interrupted-1, pi.dwProcessId);
						Interrupted = 0;
					}
					Sleep(0);
					if (XOffState != LastXOffState) {
						NewStatus(1, XOffState ? "Hold" : "", NS_NORMAL);
						LastXOffState = XOffState;
					}
					if (XOffState) continue;
					ViewOnlyFlag = FALSE;
					PeekNamedPipe(PipeVcv, NULL, 0, NULL, &Read, NULL);
					if (Read) {
						if (Read > sizeof(Buf)) Read = sizeof(Buf);
						if (ReadFile(PipeVcv, Buf, Read, &Read, NULL)) {
							if (Read > 0) {
								BOOL	 fActPos;
								POSITION OldPos;

								fActPos = ComparePos(&CurrPos, &InsertPos) == 0;
								if (!fActPos) {
									OldPos  = CurrPos;
									CurrPos = InsertPos;
								}
								InsertBuffer(Buf, Read, 0);
								InsertPos = CurrPos;
								if (Buf[Read-1] == '\r') {
									GoBackAndChar(&InsertPos);
									Advance(&InsertPos, 1);
								}
								NewPosition(fActPos ? &InsertPos : &OldPos);
								ShowEditCaret();
								ExitTime = 0;
								continue;
							}
						}
					}
					if (!WaitForSingleObject(pi.hProcess, 0)) {
						if (!ExitTime) ExitTime = GetCurrentTime();
						else if (GetCurrentTime()-ExitTime > 1000 /*1s*/) break;
					}
				}
				GotChar('\033', 0);
				FindValidPosition(&CurrPos, 0);
#endif
			} else hThread2 = hThread3 = NULL;
			ShowEditCaret();
			if (Flags & 1) {
				if (hThread1) {
					WaitForSingleObject(hThread1, INFINITE);
					CloseHandle(hThread1);
				}
				if (lpMemData) {
					GlobalUnlock(YankHandles[YankIndex]);
					lpMemData = NULL;
				}
			}
			CloseHandle(PipeVcv);
			CloseHandle(pi.hThread);
			CloseHandle(pi.hProcess);
			Interruptable = FALSE;
			return (TRUE);
		}
		if (Cmd2) free(Cmd2);
		CloseHandle(PipeVcv);
		CloseHandle(PipeCcv);
		CloseHandle(PipeVvc);
		CloseHandle(PipeCvc);
		Interruptable = FALSE;
	#else
		Error(125);
	#endif
	return (FALSE);
}
