/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include "winvi.h"
#include "page.h"

int  IndentChars;
LONG InsertCount;
BOOL SuppressMapOem = FALSE;

void InsertChar(int Char) {
	BYTE *pc;

	if (IsViewOnly()) return;
	SetUnsafe();
	if (Mode != ReplaceMode || (HexEditMode ? CharAt(&CurrPos) == C_EOF
											: CharFlags[CharAt(&CurrPos)] & 1))
		Reserve(&CurrPos, 1, 0);
	else {
		EnterDeleteForUndo(&CurrPos, 1, 0);
		EnterInsertForUndo(&CurrPos, 1);
		CharAt(&CurrPos);
	}
	++InsertCount;
	if (CharSet && Char >= 0200 && !SuppressMapOem &&
			(pc = memchr(MapOem, Char, sizeof(MapOem))) != NULL)
		Char = pc - MapOem;
	assert(CurrPos.p->PageBuf);
	CurrPos.p->PageBuf[CurrPos.i] = Char;
	#if 0
	//	CurrPos.p->Flags &= ~ISSAFE;
	#endif
	UnsafePage(&CurrPos);
	{	POSITION InvalidatePos;
		InvalidatePos = CurrPos;
		CharAndAdvance(&CurrPos);
		if (ComparePos(&LineInfo[CurrLine].Pos, &CurrPos) == 0) {
			/*insert at start of line*/
			LineInfo[CurrLine].Pos = InvalidatePos;
		}
		InvalidateArea(&InvalidatePos, 1, HexEditMode ? 7 : 9);
	}
	if (!ReplaceSingle) {
		HideEditCaret();
		UpdateWindow(hwndMain);
		NewPosition(&CurrPos);
		if (HexEditMode) NewScrollPos();
		ShowEditCaret();
		GetXPos(&CurrCol);
		if (Indenting)
			if (CharFlags[Char] & 16) ++IndentChars;
			else Indenting = FALSE;
	}
}

BOOL DeleteChar(void) {
	int c;

	if (IsViewOnly()) return (FALSE);
	if (HexEditMode) {
		if (!GoBack(&CurrPos, 1)) return (FALSE);
		if ((c = CharAt(&CurrPos)) == C_CRLF) c = ' ';
	} else if (CharFlags[c = GoBackAndChar(&CurrPos)] & 1) {
		if (c != C_EOF) {
			if (WrapPosFlag) {
				GotKeydown(VK_DELETE, 0);
				return (TRUE);
			}
			CharAndAdvance(&CurrPos);
		}
		return (FALSE);
	}
	if (InsertCount) --InsertCount;
	#if 0
	//	else return (FALSE);
	#endif
	SetUnsafe();
	HideEditCaret();
	NewPosition(&CurrPos);
	Reserve(&CurrPos, -1, (WORD)(CharFlags[c] & 1));
	InvalidateArea(&CurrPos, 0, HexEditMode ? 7 : 9);
	UpdateWindow(hwndMain);
	if (CharAt(&CurrPos)==C_EOF && !HexEditMode &&
								   !(CharFlags[GoBackAndChar(&CurrPos)] & 1))
		Advance(&CurrPos, 1);
	NewPosition(&CurrPos);
	if (HexEditMode) NewScrollPos();
	ShowEditCaret();
	GetXPos(&CurrCol);
	if (Indenting && IndentChars) --IndentChars;
	return (TRUE);
}

void SetCurrPos(VOID) {
	LineInfo[CurrLine+1].Pos		  = CurrPos;
	LineInfo[CurrLine+1].ByteCount	  = CountBytes(&CurrPos);
	LineInfo[CurrLine+1].nthContdLine = 0;
	LineInfo[CurrLine+1].Flags		  = LIF_VALID;
}

void ScrollDisplay(int Lines, WORD Flags) {
	/*Flags: 1=redraw current line
	 */
	int From = Lines>0 ? CurrLine+1 : CurrLine+2;
	int To   = Lines>0 ? CurrLine+2 : CurrLine+1;
	int ScrolledLines = VisiLines - CurrLine;
	int UpdateLino = 0;
	HDC hDC;
	extern BOOL Exiting;

	if (Exiting) {
		assert(!Exiting);
		return;
	}
	if (NumberFlag && LineInfo[VisiLines].Flags & LIF_EOF) {
		for (UpdateLino=VisiLines;
			 UpdateLino && LineInfo[--UpdateLino].Flags & LIF_EOF;
			);
		if (Lines > 0) ++UpdateLino;
	}
	memmove(LineInfo+To, LineInfo+From, ScrolledLines * sizeof(LINEINFO));
	if (Lines > 0) SetCurrPos();
	HideEditCaret();
	if (hwndMain && (hDC = GetDC(hwndMain)) != 0) {
		extern LONG	xOffset;
		RECT		r1, r2;
		int			x1;

		r1.left   = r1.top = 0;
		r1.right  = ClientRect.right;
		r1.bottom = ClientRect.bottom;
		TextAreaOnly(&r1, TRUE);
		x1 = r1.left;
		if (NumberFlag)
			r1.left = (int)(StartX(FALSE) - xOffset - (AveCharWidth>>1));
		r1.top += GetPixelLine(CurrLine+1, TRUE);
		ScrollDC(hDC, 0, Lines * TextHeight, &r1, &r1, NULL, &r2);
		ReleaseDC(hwndMain, hDC);
		if (Flags & 1) r2.top -= TextHeight;
		InvalidateRect(hwndMain, &r2, TRUE);
		if (*BitmapFilename) {
			/*refresh scrolled background bitmap (correct moved origin)*/
			while (To < VisiLines) LineInfo[To++].Flags |= LIF_REFRESH_BKGND;
			RefreshBitmap = TRUE;
		}
		UpdateWindow(hwndMain);
		if (UpdateLino) {
			r1.right = r1.left;
			r1.left  = x1;
			r1.bottom = (r1.top = GetPixelLine(UpdateLino, FALSE)) + TextHeight;
			InvalidateRect(hwndMain, &r1, TRUE);
			UpdateWindow(hwndMain);
		}
	}
}

void InsertEol(UINT Flags, int Char) {
	/*Flags: 1=redraw current line
	 *		 2=indent from next line (count autoindent in following line)
	 *		 4=newline while in insert mode, remove empty autoindent
	 *		 8=insert without autoindent
	 */
	int OldCurrLine = CurrLine;

	if (IsViewOnly()) return;
	SetUnsafe();
	++InsertCount;
	if (Char == C_CRLF) {
		++InsertCount;
		Reserve(&CurrPos, 2, 1);
		assert(CurrPos.p->PageBuf);
		CurrPos.p->PageBuf[CurrPos.i]   = '\r';
		CurrPos.p->PageBuf[CurrPos.i+1] = '\n';
	} else {
		if (Char=='\r' && CharAt(&CurrPos)=='\n') {
			/*glue CR and LF (11-May-98)...*/
			InvalidateArea(&CurrPos, 2, 9);
			Reserve(&CurrPos, -1, 1);
			Reserve(&CurrPos, 2, 1);
			assert(CurrPos.p->PageBuf && CurrPos.i < CurrPos.p->Fill);
			CurrPos.p->PageBuf[CurrPos.i]   = Char;
			CurrPos.p->PageBuf[CurrPos.i+1] = '\n';
			Advance(&CurrPos, 1);
			UpdateWindow(hwndMain);
			ShowEditCaret();
			return;
		}
		Reserve(&CurrPos, 1, 1);
		assert(CurrPos.p->PageBuf);
		CurrPos.p->PageBuf[CurrPos.i] = Char;
		if (Char == '\n') {
			if (GoBack(&CurrPos, 1))
				switch (CharAt(&CurrPos)) {
					case '\r':		/*may be split across page boundary*/
					case C_CRLF:
						if (CharAndAdvance(&CurrPos)!=C_CRLF || CurrPos.i==1) {
							Reserve(&CurrPos, 2, 1);
							CurrPos.p->PageBuf[CurrPos.i+0] = '\r';
							CurrPos.p->PageBuf[CurrPos.i+1] = '\n';
							if (GoBackAndChar(&CurrPos) == '\n')
								GoBackAndChar(&CurrPos);
							Reserve(&CurrPos, -2, 1);
							CharAndAdvance(&CurrPos);
						}
						/*Invalidating the previous line should be enough,*/
						/*but this is a very rare case, so I save code... */
						InvalidateText();
						return;
					default:
						CharAndAdvance(&CurrPos);
						/*FALLTHROUGH*/
					case C_EOF:
						break;
				}
		}
	}
	if (HexEditMode) InvalidateArea(&CurrPos, 1, 7);
	CharAndAdvance(&CurrPos);
	if (!HexEditMode) ScrollDisplay(1, (WORD)(Flags & 1));
	else SetCurrPos();
	if (!(Flags & 8) && AutoIndentFlag) {
		int		 Indent = 0, c;
		POSITION Pos;

		/*find start of autoindent reference line...*/
		Pos = CurrPos;
		if (Flags & 2) GoBackAndChar(&CurrPos);
		else {
			GoBackAndChar(&Pos);
			do; while (!(CharFlags[c = GoBackAndChar(&Pos)] & 1));
			if (c != C_EOF) CharAndAdvance(&Pos);
		}

		/*count indent...*/
		IndentChars = -1;
		for (;;) {
			++IndentChars;
			switch (c = CharAndAdvance(&Pos)) {
				case ' ':	++Indent;							  continue;
				case '\t':	Indent += TabStop - Indent % TabStop; continue;
			}
			break;
		}
		if (Flags & 4 && CharFlags[c] & 1 && Indent) {
			/*remove empty indent...*/
			GoBackAndChar(&Pos);
			GoBack(&CurrPos, IndentChars);
			GoBack(&Pos,     IndentChars);
			Reserve(&Pos,   -IndentChars, 0);
			if ((InsertCount -= IndentChars) < 0) InsertCount = 0;
		}

		/*insert indent now...*/
		IndentChars = 0;
		Indenting = TRUE;
		{	MODEENUM SaveMode = Mode;

			Mode = InsertMode;	/*insert indent even in replace mode*/
			if (TabStop > 1) {
				while (Indent >= TabStop) {
					InsertChar('\t');
					Indent -= TabStop;
				}
			}
			while (Indent--) InsertChar(' ');
			Mode = SaveMode;
		}
		if (Flags & 2) CharAndAdvance(&CurrPos);
	}
	NewPosition(&CurrPos);
	if (hwndMain) {
		if (OldCurrLine >= CrsrLines-1) {
			/*screen has been scrolled up by NewPosition(),*/
			/*invalid area has also moved, invalidate again...*/
			RECT r;
			r = ClientRect;
			TextAreaOnly(&r, TRUE);
			r.top += GetPixelLine(CurrLine-1, TRUE);
			InvalidateRect(hwndMain, &r, TRUE);
		}
		UpdateWindow(hwndMain);
	}
	NewScrollPos();
	ShowEditCaret();
	GetXPos(&CurrCol);
}

void DeleteEol(BOOL RedrawCurrline) {
	int c = CharAt(&CurrPos);

	if (IsViewOnly()) return;
	switch (c) {
		case C_EOF:		return;
		case C_CRLF:	Reserve(&CurrPos, -2, 1); --InsertCount; break;
		default:		Reserve(&CurrPos, -1, 1); break;
	}
	--InsertCount;
	ScrollDisplay(-1, 0);
	if (hwndMain && RedrawCurrline) {
		RECT r;
		r = ClientRect;
		TextAreaOnly(&r, TRUE);
		r.top += GetPixelLine(CurrLine, TRUE);
		r.bottom = r.top + TextHeight;
		InvalidateRect(hwndMain, &r, TRUE);
		UpdateWindow(hwndMain);
	}
	NewScrollPos();
}

void DeleteForward(UINT Number) {
	int      i = 1;
	POSITION Pos;
	Pos = CurrPos;
	if (!Number) ++Number;
	do {
		if (!HexEditMode && CharFlags[CharAndAdvance(&Pos)] & 1) break;
		--i;
	} while (Number==(UINT)-1 || --Number);
	if (i <= 0) {
		if (i < 0) Reserve(&CurrPos, i, 0);
		CharAndAdvance(&CurrPos);
		++InsertCount;
		DeleteChar();
	}
	if (Mode == CommandMode && (HexEditMode || CharFlags[CharAt(&CurrPos)] & 1))
		Position(1, 'h', 0);
}

void Join(void) {
	int  c, LastC = C_UNDEF;
	BOOL InsertSpace = TRUE;

	if (IsViewOnly()) return;
	while (!(CharFlags[c = CharAt(&CurrPos)] & 1))
		LastC = CharAndAdvance(&CurrPos);
	if (c == C_EOF) {
		if (CharFlags[LastC] & 1) GoBackAndChar(&CurrPos);
	} else {
		SetUnsafe();
		DeleteEol(TRUE);
		while ((c = CharAt(&CurrPos)) == '\t' || c == ' ')
			Reserve(&CurrPos, -1, 0);
		InvalidateArea(&CurrPos, 1, 9);
		switch (LastC) {
			case '(': case '{': case '[':
			case ' ': case '\t':
			case C_UNDEF:
				InsertSpace = FALSE;
		}
		switch (c) {
			case ')': case '}': case ']':
			case '.': case ',': case ';': case ':':
				InsertSpace = FALSE;
		}
		if (!(CharFlags[c] & 1)) {
			if (InsertSpace) {
				InsertChar(' ');
				GoBackAndChar(&CurrPos);
			}
		} else if (LastC != C_UNDEF && !(CharFlags[LastC] & 1))
			if (c != C_EOF)
				if (CharFlags[c = GoBackAndChar(&CurrPos)] & 1 && c != C_EOF)
					CharAndAdvance(&CurrPos);
	}
	#if 0
	//	NewPosition(&CurrPos);
	//	ShowEditCaret();
	#endif
}

void DeleteSelected(int Flags) {
	/*Flags: 1=Show updated window
	 *		 2=No yank (called from undo or substitute)
	 *		 4=Suppress SetUnsafe()
	 *		 8=Suppress cursor setting to displayable position
	 *		16=Area was really selected (reselect when undoing)
	 */
	ULONG	 Newlines = SelectStart.p->Prev ? SelectStart.p->Prev->Newlines : 0;
	ULONG	 SelectBytePos;
	int		 n, InvalidateFlags = 3;
	POSITION EndSel;

	if (!SelectCount) return;
	if (IsViewOnly()) return;
	SelectBytePos = CountBytes(&SelectStart);
	EnterDeleteForUndo(&SelectStart, SelectCount, (WORD)((Flags & 16) != 0));
	EndSel = SelectStart;
	Advance(&EndSel, SelectCount);
	if (!(Flags & 2)) Yank(&SelectStart, SelectCount,
						   (WORD)(CharFlags[GoBackAndChar(&EndSel)] & 1));
	if (ComparePos(&SelectStart, &ScreenStart) < 0) {
		/*move start of (currently invisible) SelectStart line into view...*/
		int c;
		ScreenStart = SelectStart;
		if (HexEditMode) {
			GoBack(&ScreenStart, ((int)CountBytes(&ScreenStart) & 15) + 32);
		} else {
			for (n=3; n; --n)
				do; while (!(CharFlags[c = GoBackAndChar(&ScreenStart)] & 1));
			if (c != C_EOF) CharAndAdvance(&ScreenStart);
			FirstVisible = CountLines(&ScreenStart);
		}
		InvalidateFlags |= 4;
	} else if (HexEditMode) InvalidateFlags |= 4;
	if (Flags & 1) InvalidateArea(&SelectStart, SelectCount, InvalidateFlags);
	while (SelectStart.i + SelectCount > SelectStart.p->Fill) {
		n = SelectStart.p->Fill - SelectStart.i;
		if (n) {
			CorrectLineInfo(0, 0, 0, 0, SelectBytePos, -n);
			SelectCount -= n;
			if (SelectStart.i &&
					(SelectStart.p->PageBuf || ReloadPage(SelectStart.p))) {
				Newlines = SelectStart.p->Newlines - CountNewlinesInBuf
									(SelectStart.p->PageBuf + SelectStart.i,
									 SelectStart.p->Fill    - SelectStart.i,
									 NULL);
			}
			#if 0
			//	SelectStart.p->Flags &= ~ISSAFE;
			#endif
			UnsafePage(&SelectStart);
		}
		SelectStart.p->Newlines = Newlines;
		SelectStart.p->Fill = SelectStart.i;
		SelectStart.p = SelectStart.p->Next;
		if (SelectStart.p == NULL) {
			SelectStart.p = LastPage;
			SelectCount = 0;
			break;
		}
		SelectStart.i = 0;
	}
	n = SelectStart.p->Fill - SelectStart.i - (UINT)SelectCount;
	if (n) {
		CharAt(&SelectStart);	/*ensure page to be loaded*/
		assert(SelectStart.p->PageBuf);
		_fmemmove(SelectStart.p->PageBuf + SelectStart.i,
				  SelectStart.p->PageBuf + SelectStart.i + (UINT)SelectCount,
				  n);
		#if 0
		//	SelectStart.p->Flags &= ~ISSAFE;
		#endif
		UnsafePage(&SelectStart);
	}
	CorrectLineInfo(SelectStart.p, SelectStart.i + (UINT)SelectCount,
					SelectStart.p, SelectStart.i,
					SelectBytePos, -(int)SelectCount);
	SelectStart.p->Fill	-= (UINT)SelectCount;
	if (SelectStart.p->PageBuf == NULL) ReloadPage(SelectStart.p);
	assert(SelectStart.p->PageBuf);
	Newlines = SelectStart.p->Newlines -
		(CountNewlinesInBuf(SelectStart.p->PageBuf, SelectStart.p->Fill, NULL)
		 + Newlines);
	SelectCount = 0;
	if (Newlines) {
		LPPAGE p = SelectStart.p;
		while (p) {
			p->Newlines -= Newlines;
			p = p->Next;
		}
		LinesInFile -= Newlines;
	}
	if (Flags & 1) {
		UpdateWindow(hwndMain);
		if (ComparePos(&CurrPos, &SelectStart) || Newlines)
			NewPosition(&SelectStart);
		if (!(Flags & 8))
			FindValidPosition(&CurrPos, (WORD)(Mode==InsertMode ? 3 : 0));
		NewScrollPos();
	} else CurrPos = SelectStart;
	if (!(Flags & 4)) SetUnsafe();
	CheckClipboard();
}

BOOL InsertBuffer(LPSTR Buffer, ULONG Length, WORD Flags) {
	/*Flags: 1=call to FindValidPosition()
	 */
	ULONG    i = Length;
	POSITION InsertStart;
	UINT     wFlags = 3;
	BOOL     DoAdvance = FALSE, Disabled = FALSE;

	if (!Length || !hwndMain) return (FALSE);
	if (IsViewOnly()) return (FALSE);
	InsertStart = CurrPos;
	if (GoBackAndChar(&InsertStart) != C_EOF) ++DoAdvance;
	Interrupted = FALSE;
	while (i--) {
		int c;
		c = *Buffer++;
		if (CharFlags[c] & 1) {
			if (i && c == '\r' && *Buffer == '\n') {
				++Buffer;
				--i;
				Reserve(&CurrPos, 2, 1);
				++InsertCount;
				assert(CurrPos.p->PageBuf);
				CurrPos.p->PageBuf[CurrPos.i++] = '\r';
				c = '\n';
			} else if (c == '\n') {
				if (CharAt(&CurrPos) == '\n') {
					Reserve(&CurrPos, 1, 1);
					/*TODO: assert(PrevChar!='\r' || CurrPos.i>0);*/
				} else {
					ULONG n = GoBack(&CurrPos, 1);
					if (n && CharAt(&CurrPos) == '\r') {
						/*cr/lf combination...*/
						#if 0	/*did not work this way (29-Jul-98)*/
						//	Reserve(&CurrPos, -1, 1);
						//	Reserve(&CurrPos,  2, 1);
						//	CurrPos.p->PageBuf[CurrPos.i++] = '\r';
						#else
							Advance(&CurrPos, 1);
							/*TODO: assert(CurrPos.i > 0);*/
							Reserve(&CurrPos, 1, 0);
							if (!GoBack(&InsertStart, 1)) DoAdvance = FALSE;
						#endif
					} else {
						Advance(&CurrPos, n);
						Reserve(&CurrPos, 1, 1);
					}
				}
			} else if (c=='\r' && CharAt(&CurrPos)=='\n') {
				Reserve(&CurrPos, 1, 0);
				/*TODO: assert(CurrPos.i+1 < CurrPos.Len);*/
			} else Reserve(&CurrPos, 1, 1);
			wFlags |= 4;
		} else Reserve(&CurrPos, 1, 0);
		++InsertCount;
		assert(CurrPos.p->PageBuf);
		CurrPos.p->PageBuf[CurrPos.i] = c;
		CharAndAdvance(&CurrPos);
		if (!((Length-i) & 8191)) {
			if (!Disabled) {
				Disable(Disabled = TRUE);
				SetUnsafe();
			}
			MessageLoop(FALSE);
			if (Interrupted) {
				static CHAR Buffer[80];

				LOADSTRING(hInst, 218, Buffer, sizeof(Buffer));
				NewStatus(0, Buffer, NS_ERROR);
				break;
			}
		}
	}
	if (Disabled) Enable();
	SetUnsafe();
	if (DoAdvance) CharAndAdvance(&InsertStart);
	else {
		InsertStart.p = FirstPage;
		assert(InsertStart.i == 0);
		CharAt(&InsertStart);	/*...normalize position pointer*/
		ScreenStart = InsertStart;
	}
	if (HexEditMode) wFlags |= 4;
	InvalidateArea(&InsertStart, Length, wFlags);
	UpdateWindow(hwndMain);
	if (Flags & 1)
		FindValidPosition(&CurrPos,
						  (WORD)((Mode==InsertMode) | (HexEditMode << 1)));
	NewScrollPos();
	Indenting = FALSE;
	return (TRUE);
}

extern int DblQuoteChar;
HGLOBAL    YankHandles[36];
ULONG      YankLength[36];
int		   YankSubIndex[10];
int        YankIndex = 0;
/*Meaning of elements in the arrays above:
 *	0:	   yank/delete buffer used if none of the following apply,
 *	1-9:   automatically assigned for delete row operations
 *		   (rotated, 1=most recent, 9=least recent, buffer names "1 to "9)
 *		   may be a copy of one of the following entries,
 *		   then YankSubIndex[] contains the corresponding index,
 *	10-35: buffers used for named buffers "a through "z.
 *		   If using "A to "Z, the text will be appended.
 */

BOOL Yank(PPOSITION Pos, ULONG Length, WORD Flags) {
	/*Flags: 1: Enter in YankHandles[1...9] (delete row(s) operation)
	 */
	POSITION  YankPos;
	HGLOBAL   YankHandle;
	char huge *YankMem;
	ULONG	  Expand = 0;
	int		  NumericIndex;

	YankPos = *Pos;
	if (DblQuoteChar) {
		if (isdigit(DblQuoteChar)) YankIndex = DblQuoteChar & 15;
		else if (!isalpha(DblQuoteChar)
				 || (YankIndex = (DblQuoteChar & 31) + 9) >= 36)
		   YankIndex = 0;
	} else YankIndex = 0;
	if (Flags & 1) {

		if (isdigit(DblQuoteChar)) NumericIndex = YankIndex;
		else {
			/*exchange YankHandles 1 and 9 and shift...*/
			HGLOBAL TempHandle;
			ULONG	TempLength;
			int		TempSubIndex;

			TempHandle	 = YankHandles[9];
			TempLength	 = YankLength[9];
			TempSubIndex = YankSubIndex[9];
			memmove((VOID*)	/*cast should not be necessery,
							 *but VC 1.5 does not understand it without*/
					(YankHandles+2), YankHandles+1,  8*sizeof(YankHandles[1]));
			memmove(YankLength+2,	 YankLength+1,   8*sizeof(YankLength[1]));
			memmove(YankSubIndex+2,	 YankSubIndex+1, 8*sizeof(YankSubIndex[1]));
			YankHandles[1]	= TempHandle;
			YankLength[1]	= TempLength;
			YankSubIndex[1]	= TempSubIndex;
			NumericIndex	= 1;
		}
		if (!YankSubIndex[NumericIndex] && YankHandles[NumericIndex] != NULL)
			GlobalFree(YankHandles[NumericIndex]);
		YankHandles [NumericIndex] = NULL;
		YankSubIndex[NumericIndex] = 0;
		if (!YankIndex) YankIndex = NumericIndex;
	} else NumericIndex = 0;
	if (YankIndex > 9) {
		int i;

		for (i=1; i<=9; ++i)
			if (YankSubIndex[i] == YankIndex) {
				/*don't free, still in use...*/
				YankHandles[YankIndex] = NULL;
				YankSubIndex[i] = 0;
			}
	}
	if ((YankHandle = YankHandles[YankIndex]) != NULL) {
		if (DblQuoteChar<'A' || DblQuoteChar>'Z') {
			GlobalFree(YankHandle);
			YankHandle = NULL;
		} else Expand = YankLength[YankIndex];
	}
	if (Length) {
		char huge *p;

		if (YankHandle != NULL) {
			HGLOBAL h = GlobalReAlloc(YankHandle, Expand+Length, GMEM_MOVEABLE);
			if (!h) {
				ErrorBox(MB_ICONSTOP, 308);
				ShowEditCaret();
				return (FALSE);
			}
			YankHandle = h;
		} else YankHandle = GlobalAlloc(GMEM_MOVEABLE, Length);
		if (YankHandle == NULL || (YankMem = GlobalLock(YankHandle)) == NULL) {
			if (YankHandle != NULL) GlobalFree(YankHandle);
			YankHandles[YankIndex] = NULL;
			ErrorBox(MB_ICONSTOP, 308);
			ShowEditCaret();
			return (FALSE);
		}
		YankLength[YankIndex] = Expand + Length;
		p = YankMem + Expand;
		while (Length) {
			UINT n;

			CharAt(&YankPos);	/*...normalize pos and enforce page load*/
			n = YankPos.p->Fill - YankPos.i;
			if ((ULONG)n > Length) n = (UINT)Length;
			assert(n > 0 && n <= PAGESIZE);
			assert(YankPos.p->PageBuf != NULL);
			hmemcpy(p, YankPos.p->PageBuf + YankPos.i, n);
			Length	  -= n;
			YankPos.i += n;
			p		  += n;
		}
		GlobalUnlock(YankHandle);
	}
	YankHandles[YankIndex] = YankHandle;
	if (NumericIndex) {
		YankHandles [NumericIndex] = YankHandle;
		YankLength  [NumericIndex] = YankLength[YankIndex];
		YankSubIndex[NumericIndex] = YankIndex > 9 ? YankIndex : 0;
	}
	return (TRUE);
}

BOOL Put(long Number, BOOL Behind) {
	int       c, Index;
	BOOL      LineFlag;
	POSITION  SavePos;
	char huge *YankMem;

	if (isdigit(DblQuoteChar)) Index = DblQuoteChar & 15;
	else if (!isalpha(DblQuoteChar) || (Index = (DblQuoteChar & 31) + 9) >= 36)
		Index = YankIndex;
	if (!YankHandles[Index] || IsViewOnly()) return (FALSE);
	HideEditCaret();
	YankMem = GlobalLock(YankHandles[Index]);
	if (!YankMem) return (FALSE);
	LineFlag = CharFlags[YankMem[YankLength[Index]-1]] & 1;
	if (LineFlag) {
		/*insert as lines...*/
		if (Behind) {
			do; while (!(CharFlags[CharAndAdvance(&CurrPos)] & 1));
		} else {
			do; while (!(CharFlags[c = GoBackAndChar(&CurrPos)] & 1));
			if (c != C_EOF) CharAndAdvance(&CurrPos);
		}
	} else if (Behind && !(CharFlags[CharAt(&CurrPos)] & 1))
		CharAndAdvance(&CurrPos);
	SavePos = CurrPos;
	c = GoBackAndChar(&SavePos);
	do InsertBuffer(YankMem, YankLength[Index], 0);
	while (--Number > 0);
	GlobalUnlock(YankHandles[Index]);
	CurrPos = SavePos;
	if (c != C_EOF || CharFlags[CharAt(&CurrPos)] & 16)
		do; while (CharFlags[c = AdvanceAndChar(&CurrPos)] & 16 && LineFlag);
	FindValidPosition(&CurrPos, 0);
	ShowEditCaret();
	return (TRUE);
}

BOOL Shift(long Amount, WORD Flags) {
	/*Flags: 1=cursor may be positioned behind last character of line
	 *		 2=retain spacing if line is empty
	 *		 4=position at previous location
	 *		 8=shift in (else shift out)
	 *		16=remove all indenting chars
	 */
	int		 (*Advance)(PPOSITION);
	int		 c, Indent, nOldChars, LineCount = 0;
	BOOL	 More = TRUE;
	POSITION EolPos, LineStartPos, p;
	int		 SaveCurrLine = CurrLine;

	/*to be implemented: Disable(), <Ctrl/Break>*/

	if (IsViewOnly()) return (FALSE);
	EolPos = CurrPos;
	while (!(CharFlags[c = GoBackAndChar(&EolPos)] & 1)) ++Amount;
	if (Amount < 0) {
		++Amount;
		if (c == C_CRLF) ++Amount;
		Amount  = -Amount;
		Advance = GoBackAndChar;
	} else {
		if (c == C_EOF) --Amount;
		Advance = AdvanceAndChar;
	}
	SetUnsafe();
	do {
		LineStartPos = EolPos;
		if (CharFlags[c] & 0x20) {
			c = CharAt(&LineStartPos);
			if (Advance != AdvanceAndChar) More = FALSE;
		} else {
			c = AdvanceAndChar(&LineStartPos);
			if (Amount < 0) More = FALSE;
		}
		p = LineStartPos;
		Indent = nOldChars = 0;
		for (;;) {
			switch (c) {
				case '\t': Indent += TabStop - Indent % TabStop;
						   ++nOldChars;
						   c = AdvanceAndChar(&p);
						   continue;
				case ' ':  ++Indent;
						   ++nOldChars;
						   c = AdvanceAndChar(&p);
						   continue;
				default:   break;
			}
			break;
		}
		if (CharFlags[c] & 1 && !(Flags & 2)) Indent = 0;
		else if (Flags & 8) Indent += ShiftWidth;
		else if (Flags & 16 || (Indent -= ShiftWidth) < 0) Indent = 0;
		if (Advance == AdvanceAndChar)
			Amount += Indent/TabStop + Indent%TabStop - nOldChars;
		Reserve(&LineStartPos, -nOldChars, 0);
		p = LineStartPos;
		if (TabStop > 1) {
			while (Indent >= TabStop) {
				Reserve(&p, 1, 0);
				assert(p.p->PageBuf);
				p.p->PageBuf[p.i] = '\t';
				Indent -= TabStop;
				AdvanceAndChar(&p);
				if (!p.i)
					for (c=0; c<VisiLines; ++c)
						if (!LineInfo[c].Pos.i && LineInfo[c].Pos.p==p.p) {
							do LineInfo[c].Pos.p = LineInfo[c].Pos.p->Prev;
							while (!LineInfo[c].Pos.p->Fill);
							LineInfo[c].Pos.i = LineInfo[c].Pos.p->Fill-1;
						}
			}
		}
		while (Indent--) {
			Reserve(&p, 1, 0);
			assert(p.p->PageBuf);
			p.p->PageBuf[p.i] = ' ';
			AdvanceAndChar(&p);
		}
		++LineCount;
		if (ComparePos(&LineStartPos, &ScreenStart) >= 0)
			InvalidateArea(&LineStartPos, 1, 9);
		UpdateWindow(hwndMain);
		c = 0;
		while (More) {
			if (Amount-- < 0 && Advance == AdvanceAndChar) More = FALSE;
			else if (CharFlags[c = (*Advance)(&EolPos)] & 1) {
				if (Advance == AdvanceAndChar) {
					if (Amount < 0) More = FALSE;
					++CurrLine;
					/*...CurrLine update for Reserve()->CorrectLineInfo()...*/
				} else if (CurrLine > 0) --CurrLine;
				break;
			}
		}
		if (c == C_CRLF) --Amount;
	} while (More);
	CurrLine = SaveCurrLine;
	if (Flags & 4) {
		while (ComparePos(&EolPos, &LineStartPos) < 0) CharAndAdvance(&EolPos);
		NewPosition(&EolPos);
	} else {
		CurrPos = LineInfo[CurrLine].Pos;
		Position(1, '^', (Flags & 1) | 0x400);
	}
	if (LineCount > 1) {
		extern char Buffer[300];
		LOADSTRING(hInst, 334 + !(Flags & 8), Buffer+100, sizeof(Buffer)-100);
		if (Buffer[100]) {
			wsprintf(Buffer, Buffer+100, LineCount);
			NewStatus(0, Buffer, NS_NORMAL);
		}
	}
	return (TRUE);
}
