/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <string.h>
#include "winvi.h"
#include "intercom.h"

#define MAXDISPLAY 20

extern PCSTR ClassName;

static UINT	 IcMessage, IcCount, IcUnassigned;
static BOOL	 IcLowest;
static UINT	 IcIndex = MAXDISPLAY;
static HMENU IcMenu;
static HWND	 IcWindows[MAXDISPLAY];
static DWORD IcStartTime, IcAwait, IcReceiveTime;

BOOL CALLBACK EnumWindowsProc(HWND hWnd, LPARAM lPar)
{	CHAR Buf[20];

	(VOID)lPar;		/*parameter never used*/
	if (GetClassName(hWnd, Buf, sizeof(Buf)) &&	!lstrcmp(Buf, ClassName)) {
		DWORD i;

		#if defined(WIN32)
			SendMessageTimeout(hWnd, IcMessage, (WPARAM)hWnd, 0,
							   SMTO_NORMAL, 2000, &i);
		#else
			i = SendMessage(hWnd, IcMessage, (WPARAM)hWnd, 0);
		#endif
		if (i < MAXDISPLAY) {
			if (!IcWindows[(int)i]) IcWindows[(int)i] = hWnd;
			#if 0
			//	else OutputDebugString("Entry already in use!\n");
			#endif
		} else ++IcUnassigned;
	}
	return (TRUE);
}

VOID InterCommInit(HWND hWnd)
{	INT i;

	(VOID)hWnd;		/*parameter never used*/
	IcMessage = RegisterWindowMessage("WinVi - Message");
	for (;;) {
		IcLowest = TRUE;
		IcCount = IcUnassigned = 0;
		EnumWindows(EnumWindowsProc, 0);
		if (IcCount <= 1 /*own window (might already be closed)*/) break;
		if (!IcLowest && IcUnassigned > 1) {
			IcCount = 0;
			for (i=1000; i>0; --i) {
				MessageLoop(FALSE);
				if (IcCount && i > MAXDISPLAY) i = MAXDISPLAY;
			}
		}
	}
	/*search for an empty entry index to use for this WinVi instance...*/
	for (i=0; i<MAXDISPLAY && IcWindows[i]; ++i) {}
	IcIndex=i;
}

BOOL InterCommCommand(WORD uCmd)
{	HWND hwndThis;

	if ((uCmd -= 2000) < MAXDISPLAY && IsWindow(hwndThis = IcWindows[uCmd])) {
		if (IsIconic(hwndThis)) ShowWindow(hwndThis, SW_RESTORE);
		else {
			#if defined(WIN32)
				BringWindowToTop(hwndThis);
			#else
				SetActiveWindow(hwndThis);
			#endif
		}
		return (TRUE);
	}
	return (FALSE);
}

static VOID RebuildWindowList(VOID)
{	static WNDENUMPROC EnumProc;

	/*request window titles from all running WinVis*/
	if (EnumProc == NULL)
		EnumProc=(WNDENUMPROC)MakeProcInstance((FARPROC)EnumWindowsProc, hInst);
	memset((VOID*)IcWindows, 0, sizeof(IcWindows));
	EnumWindows(EnumProc, 0);
}

VOID InterCommMenu(HWND hWnd, HMENU hMenu)
{	INT  i;
	UINT LastItem = 0;

	if (!IcMessage) return;

	/*clear old list of menu entries*/
	do; while (DeleteMenu(hMenu, 0, MF_BYPOSITION));

	RebuildWindowList();
	for (i=0; i<MAXDISPLAY; ++i) {
		static CHAR Buf[266];
		PSTR		p, p2;

		if (!IcWindows[i]) continue;
		GetWindowText(IcWindows[i], Buf+4, sizeof(Buf)-4);
		if (!memcmp(Buf+4, "WinVi - ", 8)) p = Buf + 8;
		else {
			Buf[lstrlen(Buf+4)-4] = '\0';
			p = Buf;
		}
		p2 = p+2;
		while ((p2 = strchr(p2+2, '&')) != NULL)
			memmove(p2+1, p2, lstrlen(p2)+1);
		p[0] = '&'; p[1] = i<9 ? '1'+i : 'a'-9+i; p[2] = ' '; p[3] = '\t';
		#if defined(WIN32)
		{	static MENUITEMINFO mii = {sizeof(MENUITEMINFO),
									   MIIM_ID | MIIM_TYPE | MIIM_STATE,
									   MFT_RADIOCHECK | MFT_STRING};

			mii.dwTypeData = p;
			mii.wID		   = 2000 + i;
			mii.fState	   = IcWindows[i]==hWnd ? MFS_CHECKED : 0;
			if (InsertMenuItem(hMenu, LastItem, TRUE, &mii)) {
				++LastItem;
				continue;
			}
		}
		#endif
		InsertMenu(hMenu, LastItem++,
				   IcWindows[i]!=hWnd ? MF_BYPOSITION | MF_STRING
									  : MF_CHECKED | MF_BYPOSITION | MF_STRING,
				   2000+i, p);
	}
}

BOOL InterCommActivate(BOOL Backwards)
{	INT  i, Start;
	HWND Activate = 0;

	if (!IcMessage) return (FALSE);
	RebuildWindowList();
	for (Start=0; IcWindows[Start]!=hwndMain && Start<MAXDISPLAY; ++Start);
	if (Start >= MAXDISPLAY) return (FALSE);
	i = Start;
	while (++i != Start) {
		if (i >= MAXDISPLAY) {
			i = 0;
			if (!Start) break;
		}
		if (IcWindows[i]) {
			Activate = IcWindows[i];
			if (!Backwards) break;
		}
	}
	if (!Activate) return (FALSE);
	#if defined(WIN32)
		BringWindowToTop(Activate);
	#else
		SetActiveWindow(Activate);
	#endif
	return (TRUE);
}

LRESULT InterCommWindowProc(HWND hWnd, UINT uMsg, WPARAM wPar, LPARAM lPar)
{
	(VOID)lPar;		/*parameter never used*/
	if (!uMsg || uMsg!=IcMessage) return ((LRESULT)-1);
	++IcCount;
	if ((HWND)wPar < hWnd) IcLowest = FALSE;
	return (IcIndex);
}
