/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <string.h>
#include <assert.h>
#include "winvi.h"
#include "resource.h"
#include "page.h"
#include "paint.h"
#include "status.h"
#include "toolbar.h"
#include "ctl3d.h"

HPEN   BlackPen;
HBRUSH SelBkgndBrush, BkgndBrush;
HPEN   HilitePen, ShadowPen;	/*3d paint tools*/
HBRUSH FaceBrush;				/*3d paint tools*/
HBRUSH DlgBkBrush;
INT    CaretX, CaretY, CaretWidth, CaretHeight;
INT    VisiLines;	/*number of visible screen lines*/
INT    CrsrLines;	/*number of screen lines with positionable cursor*/
HFONT  SpecialFont = 0;
HWND   hwndScroll;
INT    TextAreaHeight, InsertCaretWidth = 2;
LONG   xOffset;

void SystemColors(BOOL New)
{
	if (HilitePen)	{DeleteObject(HilitePen);  HilitePen  = NULL;}
	if (ShadowPen)	{DeleteObject(ShadowPen);  ShadowPen  = NULL;}
	if (FaceBrush)	{DeleteObject(FaceBrush);  FaceBrush  = NULL;}
	if (DlgBkBrush)	{DeleteObject(DlgBkBrush); DlgBkBrush = NULL;}
	if (New) {
		HDC hDC;
		HilitePen = CreatePen(PS_SOLID, 1,	  GetSysColor(COLOR_BTNHIGHLIGHT));
		ShadowPen = CreatePen(PS_SOLID, 1,	  GetSysColor(COLOR_BTNSHADOW));
		FaceBrush =			 CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
		#if defined(WIN32)
				DlgBkBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
		#else
		   if (Ctl3dEnabled())
				DlgBkBrush = CreateSolidBrush(GetSysColor(COLOR_BTNFACE));
		   else DlgBkBrush = CreateSolidBrush(GetSysColor(COLOR_WINDOW));
		#endif
		StatusStyle[NS_NORMAL].Color =
		StatusStyle[NS_TOOLTIP].Color =		  GetSysColor(COLOR_BTNTEXT);
		if (hwndMain && IsWindowVisible(hwndMain)) {
			RECT r;
			r = ClientRect;
			r.top = r.bottom - StatusHeight;
			InvalidateRect(hwndMain, &r, FALSE);
		}
		hDC = GetDC(NULL);
		if (hDC) {
			PrepareBitmap(hDC);
			ReleaseDC(NULL, hDC);
		}
		MakeBackgroundBrushes();
	}
}

INT ScrollX;

void InitPaint(void)
{
	if (!TextHeight) TextHeight = 13;
	if (!TextColor)  TextColor	= RGB(0,0,0);
	LOADSTRING(hInst, 902, NumLock,  sizeof(NumLock));
	LOADSTRING(hInst, 903, CapsLock, sizeof(CapsLock));
	BlackPen   = GetStockObject(BLACK_PEN);
	SystemColors(TRUE);
	{	static LOGFONT LogStatusFont = {
			16, 0, 0, 0, 400, 0, 0, 0,
			ANSI_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
			DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, "Helv"
		};
		HDC hDC = GetDC(NULL);

		if (hDC) {
			LogStatusFont.lfHeight =
				-MulDiv(10/*pt*/, GetDeviceCaps(hDC, LOGPIXELSY), 72);
			ReleaseDC(NULL, hDC);
		}
		StatusFont  = CreateFontIndirect(&LogStatusFont);
	}
	RecalcStatusRow();
	TextAreaHeight = ClientRect.bottom - StatusHeight - nToolbarHeight;
	ScrollX = ClientRect.right - GetSystemMetrics(SM_CXVSCROLL);
	if (WinVersion >= MAKEWORD(95,3)) ScrollX -= 2;
	hwndScroll = CreateWindow("ScrollBar", "",
				WS_CHILDWINDOW | WS_VISIBLE | SBS_VERT, ScrollX, nToolbarHeight,
				GetSystemMetrics(SM_CXVSCROLL), TextAreaHeight,
				hwndMain, NULL, hInst, 0L);
	VisiLines = (TextAreaHeight + TextHeight - 1 - 2*YTEXT_OFFSET) / TextHeight;
	CrsrLines = (TextAreaHeight              + 2 - 2*YTEXT_OFFSET) / TextHeight;
	LineInfo = (PLINEINFO)LocalAlloc(LPTR, (VisiLines+2) * sizeof(LINEINFO));
	if (!LineInfo) ErrorBox(MB_ICONSTOP, 300);
	CheckClipboard();
	hcurArrow	  = LoadCursor(NULL, IDC_ARROW);
	hcurHourGlass = LoadCursor(NULL, IDC_WAIT);
	hcurIBeam	  = LoadCursor(NULL, IDC_IBEAM);
	hcurSizeNwSe  = LoadCursor(NULL, IDC_SIZENWSE);
	hcurCurrent	  = hcurHourGlass;
	#if 0
	//	if (!SystemParametersInfo(SPI_GETCARETWIDTH, 0, &InsertCaretWidth, 0))
	//		/*...available in Windows 2000 only, cannot test*/
	#endif
	InsertCaretWidth = 2;
}

void ReleasePaintResources(void)
{
	SystemColors(FALSE);
	if (TextFont)		{DeleteObject(TextFont);	  TextFont		= 0;}
	if (SpecialFont)	{DeleteObject(SpecialFont);	  SpecialFont	= 0;}
	if (StatusFont)		{DeleteObject(StatusFont);	  StatusFont	= 0;}
	if (TooltipFont)	{DeleteObject(TooltipFont);	  TooltipFont	= 0;}
	if (Buttons)		{DeleteObject(Buttons);		  Buttons		= 0;}
	if (SelBkgndBrush)	{DeleteObject(SelBkgndBrush); SelBkgndBrush	= 0;}
	if (BkgndBrush)		{DeleteObject(BkgndBrush);	  BkgndBrush	= 0;}
}

void AdjustWindowParts(INT OldBottom, INT NewBottom)
	/*called if main window changes height or width
	 *or when language has changed
	 */
{	INT OldVisiLines = VisiLines;
	INT StartLine = 0;

	TextAreaHeight = NewBottom - StatusHeight - nToolbarHeight;
	if (TextHeight) {
		VisiLines = (TextAreaHeight+TextHeight-1-2*YTEXT_OFFSET) / TextHeight;
		CrsrLines = (TextAreaHeight           +2-2*YTEXT_OFFSET) / TextHeight;
	}
	if (LineInfo && CrsrLines>0) {
		PLINEINFO	NewLineInfo = (PLINEINFO)LocalAlloc
								  (LPTR, (VisiLines+2) * sizeof(LINEINFO));
		INT			NewScrollX;
		static RECT r;

		if (NewLineInfo) {
			if (OldVisiLines) {
				INT LinesToCopy = OldVisiLines;
				if (VisiLines <= OldVisiLines) {
					LinesToCopy = VisiLines;
					while (CurrLine && CurrLine >= CrsrLines) {
						++StartLine;
						--CurrLine;
						CaretY -= TextHeight;
					}
				}
				memcpy(NewLineInfo, LineInfo + StartLine,
					   (LinesToCopy + 2) * sizeof(LINEINFO));
				if (StartLine) {
					ScreenStart = NewLineInfo[0].Pos;
					FirstVisible += StartLine;
				}
			}
			LocalFree((LOCALHANDLE)LineInfo);
			LineInfo = NewLineInfo;
		} else ErrorBox(MB_ICONSTOP, 300);
		NewScrollX = ClientRect.right - GetSystemMetrics(SM_CXVSCROLL);
		if (WinVersion >= MAKEWORD(95,3)) NewScrollX -= 2;
		SetWindowPos(hwndScroll, 0, NewScrollX, nToolbarHeight,
					 GetSystemMetrics(SM_CXVSCROLL), TextAreaHeight,
					 SWP_NOZORDER);
		if (hwndMain) {
			r.top	 = nToolbarHeight;
			r.bottom = ClientRect.bottom - StatusHeight;
			if (StartLine) {
				r.left	= 0;
				r.right = ClientRect.right - GetSystemMetrics(SM_CXVSCROLL);
				InvalidateRect(hwndMain, &r, TRUE);
			}
			if (WinVersion >= MAKEWORD(95,3)) {
				--r.top;
				++r.bottom;
				if (NewScrollX > ScrollX) {
					r.left = ScrollX;
					r.right = ClientRect.right;
					InvalidateRect(hwndMain, &r, TRUE);
				} else {
					r.left = (r.right = ClientRect.right) - 2;
					InvalidateRect(hwndMain, &r, FALSE);
				}
			}
			r.left   = 0;
			r.top	 = (NewBottom < OldBottom ? NewBottom : OldBottom)
					   - StatusHeight - YTEXT_OFFSET;
			r.bottom = NewBottom;
			r.right  = ClientRect.right;
			InvalidateRect(hwndMain, &r, TRUE);
		}
		ScrollX = NewScrollX;
	}
	if (hwndCmd) SetWindowPos(hwndCmd, 0, StatusFields[0].x+1,
							  ClientRect.bottom-StatusHeight+5,
							  StatusFields[0].Width-1, StatusTextHeight,
							  SWP_NOZORDER);
	if (LineInfo) {
		UpdateWindow(hwndMain);
		if (StartLine) ShowEditCaret();
	}
}

void Jump(PPOSITION NewPos)
{	INT i, c;

	ScreenStart = *NewPos;
	if (HexEditMode) {
		GoBack(&ScreenStart,
			   (CountBytes(&ScreenStart) & 15) + (((VisiLines - 1) >> 1) << 4));
	} else {
		for (i = (VisiLines + 1) >> 1; i; --i)
			do; while (!(CharFlags[c = GoBackAndChar(&ScreenStart)] & 1));
		if (c != C_EOF) CharAndAdvance(&ScreenStart);
	}
	HideEditCaret();
	InvalidateText();
}

INT ScrollShift, ScrollRange, ScrollPos;

void NewScrollPos(void)
{	LONG		Range;
	static BOOL	Disabled;

	if (LineInfo == NULL /*not yet initialized*/) return;
	if (HexEditMode) {
		LPPAGE p;

		Range = 15;
		for (p=FirstPage; p; p=p->Next) Range += p->Fill;
		Range >>= 4;
	} else Range = LinesInFile + IncompleteLastLine();
	Range = Range < (LONG)CrsrLines ? 0 : Range - CrsrLines;
	FirstVisible = FirstPage ? CountLines(&ScreenStart) : 0;
	ScrollShift	 = 0;
	while (Range > 32000) {
		++ScrollShift;
		Range >>= 1;
	}
	ScrollPos	= (INT)(FirstVisible>>ScrollShift);
	ScrollRange	= (INT)Range;
	if ((ScrollRange || ScrollPos) == Disabled) {
		EnableWindow(hwndScroll, Disabled);
		Disabled ^= TRUE^FALSE;
	}
	if (Disabled) return;
	#if defined(WIN32)
		if (WinVersion >= MAKEWORD(95,3)) {
			#if !defined(SBM_SETSCROLLINFO)
				typedef struct tagSCROLLINFO {
					ULONG cbSize; 
					ULONG fMask; 
					long  nMin; 
					long  nMax; 
					ULONG nPage; 
					long  nPos; 
					long  nTrackPos; 
				} SCROLLINFO, FAR *LPSCROLLINFO;
				#define SBM_SETSCROLLINFO	0x00e9
				#define SIF_RANGE			1
				#define SIF_PAGE			2
				#define SIF_POS				4
				#define SIF_DISABLENOSCROLL	8
			#endif
			static SCROLLINFO ScrollInfo;
			INT				  ThumbSize;

			if (!(ThumbSize = CrsrLines >> ScrollShift)) ++ThumbSize;
			ScrollInfo.cbSize	 = sizeof(ScrollInfo);
			ScrollInfo.fMask	 = SIF_PAGE | SIF_POS | SIF_RANGE |
								   SIF_DISABLENOSCROLL;
			ScrollInfo.nMin		 = 0;
			ScrollInfo.nMax		 = ScrollRange + ThumbSize - 1;
			ScrollInfo.nPage	 = ThumbSize;
			ScrollInfo.nPos		 = ScrollPos;
			ScrollInfo.nTrackPos = ScrollPos;
			#if !defined(DIRECT_CALL_SETSCROLLINFO)
			{	extern char	 QueryString[150];
				extern DWORD QueryTime;
				DWORD		 i;

				i = SendMessage(hwndScroll, SBM_SETSCROLLINFO, (WPARAM)TRUE,
								(LPARAM)(LPSCROLLINFO)&ScrollInfo);
				if (i==(DWORD)ScrollPos && sizeof(INT)==4 || i) return;
			}
			#else
				do {	/*no loop, for break only*/
					static DWORD p32SetScrollInfo;
					if (!p32SetScrollInfo) {
						DWORD		h32Dll;
						static BOOL	fAlreadyTried;

						if (fAlreadyTried) break;
						fAlreadyTried = TRUE;
						h32Dll = LoadLibraryEx32W("USER32", NULL, 0);
						if (!h32Dll) break;
						p32SetScrollInfo =
							GetProcAddress32W(h32Dll, "SetScrollInfo");
						if (!p32SetScrollInfo) {
							FreeLibrary32W(h32Dll);
							break;
						}
					}
					CallProcEx32W(CPEX_DEST_STDCALL | 4, 4, p32SetScrollInfo,
						/*params...*/ hwndScroll,
									  (DWORD)SB_CTL,
									  (LPSCROLLINFO)&ScrollPos,
									  (DWORD)TRUE);
					return;
				} while (FALSE);
			#endif
		}
	#endif
	SetScrollRange(hwndScroll, SB_CTL, 0, ScrollRange, FALSE);
	SetScrollPos  (hwndScroll, SB_CTL,    ScrollPos,   TRUE);
}

extern POSITION	SelectAnchor;
extern ULONG	SelectAnchorCount;

void ExtendSelection(void)
{	ULONG BytePos = CountBytes(&CurrPos);

	if ((BytePos<SelectAnchorCount) != (SelectBytePos<SelectAnchorCount)) {
		/*reverting select direction from anchor...*/
		if (SelectCount > 0) {
			/*remove old selection totally...*/
			InvalidateArea(&SelectStart, SelectCount, 1);
			SelectCount = 0;
			UpdateWindow(hwndMain);
		}
		if (BytePos < SelectAnchorCount) {
			SelectStart = CurrPos;
			SelectCount = SelectAnchorCount-BytePos;
			SelectBytePos = BytePos;
		} else {
			SelectStart = SelectAnchor;
			SelectCount = BytePos-SelectAnchorCount;
			SelectBytePos = SelectAnchorCount;
		}
		if (SelectCount) InvalidateArea(&SelectStart, SelectCount, 0);
	} else if (BytePos>=SelectBytePos && BytePos<=SelectBytePos+SelectCount) {
		/*reducing previous selection...*/
		if (BytePos < SelectAnchorCount) {
			if (BytePos != SelectBytePos)
				InvalidateArea(&SelectStart, BytePos-SelectBytePos, 1);
			SelectStart = CurrPos;
			SelectCount = SelectAnchorCount-BytePos;
			SelectBytePos = BytePos;
		} else {
			if (BytePos != SelectBytePos+SelectCount)
				InvalidateArea(&CurrPos, SelectBytePos+SelectCount-BytePos, 1);
			SelectCount = BytePos-SelectBytePos;
		}
	} else {
		/*extending selection...*/
		if (BytePos < SelectBytePos) {
			InvalidateArea(&CurrPos, SelectBytePos-BytePos, 0);
			SelectStart = CurrPos;
			SelectCount += SelectBytePos-BytePos;
			SelectBytePos = BytePos;
		} else {
			POSITION Inv = SelectStart;

			Advance(&Inv, SelectCount);
			InvalidateArea(&Inv, BytePos-SelectBytePos-SelectCount, 0);
			SelectCount = BytePos-SelectBytePos;
		}
	}
	UpdateWindow(hwndMain);
}

extern BOOL Exiting;

void ScrollJump(INT Where)
{	ULONG	 Line;
	POSITION Pos;
	INT		 SaveLine, Tmp;
	BOOL	 Selecting;

	if (Exiting) {
		assert(!Exiting);
		return;
	}
	if (Where == ScrollPos) return;
	Selecting = GetKeyState(VK_SHIFT) < 0;
	if (Selecting) {
		if (!SelectCount) {
			SelectStart	  = SelectAnchor	  = CurrPos;
			SelectBytePos = SelectAnchorCount = CountBytes(&CurrPos);
		}
	} else {
		if (SelectCount) {
			InvalidateArea(&SelectStart, SelectCount, 1);
			SelectCount = 0;
			UpdateWindow(hwndMain);
		}
	}
	if (!ScrollRange) Line = 0;
	else Line = ((ULONG)Where << ScrollShift) +
				((1<<ScrollShift) * Where /*- 1*/) / ScrollRange;
	Pos.p = FirstPage;
	if (HexEditMode) {
		Pos.i = 0;
		Advance(&Pos, (Line + CrsrLines /*- 1*/) << 4);
		GoBack(&Pos, ((CrsrLines - 1) << 4) + 1);
		Tmp = (INT)((Line = CountBytes(&Pos)) & 15);
		if (Tmp) GoBack(&Pos, Tmp);
		Line >>= 4;
	} else {
		ULONG Max;

		if (Line > (Max = LinesInFile + IncompleteLastLine() - CrsrLines))
			Line = Max;
		while (Pos.p && Line > Pos.p->Newlines) Pos.p = Pos.p->Next;
		if (!Pos.p) {
			/*should not occur, but nobody knows...*/
			assert(0);
			Pos.p = LastPage;
			Pos.i = LastPage->Fill;
		} else {
			Pos.i = 0;
			if (Pos.p->Prev) Tmp = (INT)(Line - Pos.p->Prev->Newlines);
			else Tmp = (INT)Line;
			while (Tmp) {
				INT c;

				if ((c = CharFlags[CharAndAdvance(&Pos)]) & 1) --Tmp;
				if (c & 0x20) break;
			}
		}
	}
	HideEditCaret();

	/*scroll if ScreenStart and Pos are less than CrsrLines apart...*/
	Line -= CountLines(&ScreenStart);
	if ((LONG)Line < CrsrLines && -(LONG)Line < CrsrLines) {
		/*scroll (using single ScrollDC for more speed)...*/
		RECT r1, r2;
		HDC  hDC;

		if (hwndMain && (hDC = GetDC(hwndMain)) != 0) {
			r1 = ClientRect;
			TextAreaOnly(&r1, TRUE);
			Tmp = (INT)Line;
			ScrollDC(hDC, 0, Tmp * -TextHeight, &r1, &r1, NULL, &r2);
			ReleaseDC(hwndMain, hDC);

			if (Tmp > 0) {
				memmove(LineInfo, LineInfo+Tmp,
				 		(VisiLines-Tmp+1) * sizeof(LINEINFO));
				memset(LineInfo+VisiLines-Tmp+1, 0, Tmp * sizeof(LINEINFO));
				r2.bottom += YTEXT_OFFSET;	/*invalidate bitmap border area*/
			} else {
				memmove(LineInfo-Tmp, LineInfo,
						(VisiLines+Tmp+1) * sizeof(LINEINFO));
				r2.top -= YTEXT_OFFSET;		/*invalidate bitmap border area*/
			}
			ScreenStart = Pos;
			SaveLine = CurrLine;
			FirstVisible += Tmp;
			InvalidateRect(hwndMain, &r2, TRUE);
			UpdateWindow(hwndMain);
			if (*BitmapFilename) {
				/*repaint other end of scrolled area...*/
				if (Tmp > 0) {
					r1.top	  = nToolbarHeight;
					r1.bottom = r1.top + YTEXT_OFFSET;
				} else {
					r1.bottom = ClientRect.bottom - StatusHeight;
					r1.top	  = r1.bottom - YTEXT_OFFSET;
				}
				InvalidateRect(hwndMain, &r1, TRUE);
			}
			Tmp = 0;
			while ((CurrPos = LineInfo[SaveLine].Pos).p == NULL) {
				--SaveLine;
				++Tmp;
			}
			NewScrollPos();
			#if 0		/*position to start of line*/
				if (Tmp) Position(Tmp, '\r', 0);
				else Position(1, '^', 0);
			#else		/*position to original column (new 7-Oct-2000)*/
				if (Tmp) Position(Tmp, 'j', 0);
				else {
					AdvanceToCurr(&CurrPos, /*SaveLine,*/
								  Mode==InsertMode || Mode==ReplaceMode);
					NewPosition(&CurrPos);
					if (Selecting) ExtendSelection();
				}
			#endif
			ShowEditCaret();
			Indenting = FALSE;
			return;
		}
	}

	ScreenStart = Pos;
	SaveLine = CurrLine;
	InvalidateText();
	CurrPos = LineInfo[SaveLine].Pos;
	NewScrollPos();
	#if 0
		Position(1, '^', 0);
	#else
		AdvanceToCurr(&CurrPos, /*SaveLine,*/
					  Mode==InsertMode||Mode==ReplaceMode);
		NewPosition(&CurrPos);
		if (Selecting) ExtendSelection();
	#endif
	ShowEditCaret();
	Indenting = FALSE;
}

BOOL ScrollLine(INT Dir)
{	HDC  hDC;
	RECT r1, r2;

	if (Exiting) {
		assert(!Exiting);
		return (FALSE);
	}
	if (Dir > 0) memmove(LineInfo, LineInfo+1, VisiLines * sizeof(LINEINFO));
	else		 memmove(LineInfo+1, LineInfo, VisiLines * sizeof(LINEINFO));
	hDC = GetDC(hwndMain);
	if (hDC) {
		r1 = ClientRect;
		TextAreaOnly(&r1, TRUE);
		ScrollDC(hDC, 0, Dir>0 ? -TextHeight : TextHeight, &r1, &r1, NULL, &r2);
		ReleaseDC(hwndMain, hDC);
	} else {
		r2 = ClientRect;
		TextAreaOnly(&r2, TRUE);
	}
	if (Dir > 0) {
		ScreenStart = LineInfo[0].Pos;
		LineInfo[VisiLines].Flags = 0;
		++FirstVisible;
		r2.bottom += YTEXT_OFFSET;
	} else {
		if (HexEditMode) GoBack(&ScreenStart, 16);
		else {
			INT c;
			GoBackAndChar(&ScreenStart);
			do c = GoBackAndChar(&ScreenStart);
			while (!(CharFlags[c] & 1));
			if (c != C_EOF) CharAndAdvance(&ScreenStart);
		}
		LineInfo[0].Flags = 0;
		LineInfo[1].Flags &= ~LIF_VALID;
		--FirstVisible;
		r2.top -= YTEXT_OFFSET;
	}
	if (hwndMain) {
		InvalidateRect(hwndMain, &r2, TRUE);
		UpdateWindow(hwndMain);
		if (*BitmapFilename) {
			/*repaint other end of scrolled area...*/
			if (Dir > 0) {
				r1.top	  = nToolbarHeight;
				r1.bottom = r1.top + YTEXT_OFFSET;
			} else {
				r1.bottom = ClientRect.bottom - StatusHeight;
				r1.top	  = r1.bottom - YTEXT_OFFSET;
			}
			InvalidateRect(hwndMain, &r1, TRUE);
		}
	}
	if (Dir > 0) {
		if (!LineInfo[VisiLines].Flags)
			/*not painted probably because invisible, fake painting...*/
			TextLine(NULL, VisiLines-1,
					 nToolbarHeight+YTEXT_OFFSET+(VisiLines-1)*TextHeight);
	}
	return (hDC != 0);
}

INT  AveCharWidth;
CHAR StatusLineFormat[8]="%05lu", StatusColFormat[8]="%03lu";

long NewPosition(PPOSITION NewPos)
{	static char  LineBuf[12], ColumnBuf[12], PercentBuf[10];
	static ULONG OldLine, OldColumn, OldLines;
	static INT	 OldPercent, ScrollMax;
	POSITION	 ScreenEnd;
	ULONG		 Bytes, Lines, NoOfLines;

	CurrPos = *NewPos;
	Bytes	= CountBytes(NewPos);
	Lines	= CountLines(NewPos);
	if (!(ScrollMax=MaxScroll) && (ScrollMax=CrsrLines-2)<0) ScrollMax = 0;

	if (Exiting) {
		assert(!Exiting);
		return (0);
	}
	/*first, move current position into view...*/
	if (ScreenStart.p && (NewPos->p->PageNo <  ScreenStart.p->PageNo
					   || NewPos->p->PageNo == ScreenStart.p->PageNo
					   && NewPos->i			<  ScreenStart.i)) {
		/*go up in file...*/
		if (FirstVisible - (LONG)Lines > (LONG)ScrollMax)
			Jump(NewPos); /*hard jump*/
		else {
			/*scroll back (rolling down)*/
			HideEditCaret();
			do ScrollLine(-1);
			while (Bytes < LineInfo[0].ByteCount);
		}
		NewScrollPos();
	} else if ((ScreenEnd = LineInfo[CrsrLines].Pos).p
			&& (NewPos->p->PageNo >  ScreenEnd.p->PageNo
			 || NewPos->p->PageNo == ScreenEnd.p->PageNo
			 && NewPos->i		  >= ScreenEnd.i)
			&& !(LineInfo[CrsrLines].Flags & LIF_EOF)) {
		/*go down...*/
		if ((long)(Lines - CountLines(&ScreenEnd)) >= (long)ScrollMax)
			Jump(NewPos); /*hard jump*/
		else {
			/*scroll forward (rolling up)*/
			HideEditCaret();
			do ScrollLine(1);
			while (Bytes >= LineInfo[CrsrLines].ByteCount &&
				   !(LineInfo[CrsrLines].Flags & LIF_EOF));
		}
		NewScrollPos();
	} else {
		/*calculate new caret position within screen...*/
		INT i;
		for (i = CrsrLines; i && (!(LineInfo[i].Flags & LIF_VALID) ||
									LineInfo[i].ByteCount > Bytes  ||
									LineInfo[i].Flags & LIF_EOF); --i);
		if (LineInfo[i].ByteCount <= Bytes) {
			INT InitCaretX, CaretUpperY;

			/*set caret pos and dimension just for the case sth goes wrong...*/
			CaretX = InitCaretX = StartX(FALSE) - (Mode == InsertMode);
			CaretY = CaretUpperY = nToolbarHeight + YTEXT_OFFSET + i*TextHeight;
			if (Mode == ReplaceMode) {
				CaretHeight = 3;
				CaretY += TextHeight - 3;
			} else CaretHeight = TextHeight;
			CaretWidth = Mode==InsertMode ? InsertCaretWidth : AveCharWidth;
			TextLine(0, i, CaretUpperY);
			while (!(LineInfo[i+1].Flags & LIF_EOF) &&
					 LineInfo[i+1].Flags & LIF_VALID &&
					 LineInfo[i+1].ByteCount <= Bytes) {
				if (i+1 >= CrsrLines) {
					ScrollLine(1);
					UpdateWindow(hwndMain);
				} else {
					++i;
					CaretY		+= TextHeight;
					CaretUpperY	+= TextHeight;
					TextLine(0, i, CaretUpperY);
				}
			}
			if (CharAt(&CurrPos)==C_EOF && CaretX==InitCaretX &&
					!(LineInfo[i].Flags & LIF_EOF))
				CaretY += TextHeight;
		}
	}

	/*finally, update status row...*/
	if (HexEditMode) {
		ScreenEnd.p = LastPage;
		ScreenEnd.i = LastPage->Fill;
		NoOfLines = (CountBytes(&ScreenEnd) + 15) >> 4;
		CurrColumn = Bytes & 15;
	} else NoOfLines = LinesInFile + IncompleteLastLine();
	++Lines;
	{	static ULONG StatusLineLimit=100000, StatusColLimit=1000;
		static INT	 StatusLineDigits=5,	 StatusColDigits=3;
		PSTR		 p;

		if (Lines >= StatusLineLimit || CurrColumn >= StatusColLimit) {
			/*enlarge line number and/or column field in status line...*/
			RECT r;

			while (CurrColumn >= StatusColLimit) {
				StatusColLimit *= 10;
				++StatusColDigits;
				OldColumn = 0;
				if (StatusLineLimit < StatusColLimit*100
					&& StatusLineLimit < 1000000000L) {
						/*The field for the current line number should be wider
						 *than the field for the current column if possible.
						 */
						StatusLineLimit *= 10;
						++StatusLineDigits;
						OldLine = 0;
				}
			}
			while (Lines >= StatusLineLimit && StatusLineLimit < 1000000000L) {
				StatusLineLimit *= 10;
				++StatusLineDigits;
				OldLine = 0;
			}
			wsprintf(StatusLineFormat, p="%%0%dlu", StatusLineDigits);
			wsprintf(StatusColFormat,  p, StatusColDigits);
			RecalcStatusRow();
			r = ClientRect;
			r.top = r.bottom - StatusHeight + 4;
			InvalidateRect(hwndMain, &r, TRUE);
		}
		if (Lines != OldLine || NoOfLines != OldLines) {
			INT Percent;

			wsprintf(LineBuf, StatusLineFormat, OldLine = Lines);
			Percent = NoOfLines
					  ? Lines<=21474836L ? (INT)((100 * Lines) / NoOfLines)
										 : (INT)(Lines / (NoOfLines / 100))
					  : 0;
			if (Percent != OldPercent || NoOfLines != OldLines) {
				wsprintf(PercentBuf, "--%d%%--", OldPercent = Percent);
				NewStatus(4, PercentBuf, NS_NORMAL);
				OldLines = NoOfLines;
			}
			NewStatus(5, LineBuf, NS_NORMAL);
		}
		if (CurrColumn != OldColumn) {
			wsprintf(ColumnBuf, StatusColFormat, OldColumn = CurrColumn);
			NewStatus(6, ColumnBuf, NS_NORMAL);
		}
	}
	return (0);
}

BOOL PaintAll;
LONG ExpandRedraw, ShrinkRedraw;

void InvalidateText(void)
{	RECT r;

	if (!hwndMain) return;
	StartX(TRUE);
	r = ClientRect;
	TextAreaOnly(&r, FALSE);
	FirstVisible = CountLines(&ScreenStart);
	ValidateRect(hwndMain, &r);
	UpdateWindow(hwndMain);
	memset(LineInfo, 0, (VisiLines+2) * sizeof(LINEINFO));
	PaintAll = TRUE;
	ExpandRedraw = 2000000000L;
	ShrinkRedraw = 0;
	InvalidateRect(hwndMain, &r, TRUE);
	UpdateWindow(hwndMain);
	if (!LineInfo[0].Flags || !LineInfo[CrsrLines].Flags) {
		/*not painted, probably hidden: fake painting...*/
		HDC hDC = GetDC(hwndMain);

		if (hDC) {
			Paint(hDC, &r);
			ReleaseDC(hwndMain, hDC);
		}
	}
}

INT ComparePos(PPOSITION First, PPOSITION Second)
{
	if (First->p != Second->p)
		return (First->p->PageNo < Second->p->PageNo ? -1 : 1);
	return ((INT)First->i - (INT)Second->i);
}

#if !defined(WIN32)
BOOL GetTextExtentExPoint(HDC hDC, LPSTR lpStr, INT nLen, INT nMaxExtent,
						  LPINT lpnFit, LPINT nMustBeNULL, LPSIZE lpSize) {
	INT nTryLen = nLen;

	do {
		if (!GetTextExtentPoint(hDC, lpStr, nTryLen, lpSize)) return (FALSE);
		if (lpSize->cx <= nMaxExtent) {
			*lpnFit = nTryLen;
			return (TRUE);
		}
	} while (--nTryLen >= 0);
	return (FALSE);
}
#endif

BOOL IsFixedFont = FALSE;

INT GetTextWidth(HDC hDC, PSTR String, INT Len)
{
	/*----------------------------------------------------------------------*/
	/* called from:	TextLine, GetTextPos									*/
	/*----------------------------------------------------------------------*/
	if (IsFixedFont) return (Len * AveCharWidth);
	if (!hDC) {
		/*No output, calculating cursor position only*/
		HDC   hDC;
		INT   Ret;
		HFONT Font = HexEditMode ? HexFont : TextFont, OldFont;

		hDC = GetDC(NULL);
		if (!hDC) return (0);
		if (Font) OldFont = SelectObject(hDC, Font);
		{	SIZE Size;

			Ret = GetTextExtentPoint(hDC, String, Len, &Size) ? Size.cx : 0;
		}
		if (Font) SelectObject(hDC, OldFont);
		ReleaseDC(NULL, hDC);
		return (Ret);
	}
	{	SIZE Size;
		return (GetTextExtentPoint(hDC, String, Len, &Size) ? Size.cx : 0);
	}
}

INT GetTextPos(PPOSITION StartParam, PPOSITION End, INT ColParam)
{	LONG        Len = 0, Col = ColParam;
	static char Buf[40];
	PSTR        p = Buf;
	POSITION    Start;

	if (HexEditMode) return (0);
	Start = *StartParam;
	while (Start.i >= Start.p->Fill && Start.p->Next) {
		Start.i -= Start.p->Fill;
		Start.p  = Start.p->Next;
	}
	while (ComparePos(&Start, End) < 0) {
		INT c = CharAndAdvance(&Start);
		if (CharFlags[c] & 1) {
			Col = Len = 0;
			p = Buf;
		} else switch (c) {
			case '\t':	Col += p - Buf;
						Col = Col + TabStop - Col % TabStop;
						Len = Col * AveCharWidth;
						p = Buf;
						break;

			default:	if (CharSet) c = MapOem[c];
						if (CharFlags[c] & 2) {
							if (c < ' ') {
								*p++ = '^';
								*p++ = c + ('A'-1);
							} else p += wsprintf(p, "\\%02x", c);
						} else *p++ = c;
		}
		if (p >= Buf+sizeof(Buf)-3) {
			Len += GetTextWidth(NULL, Buf, p-Buf);
			Col += p - Buf;
			p = Buf;
		}
	}
	if (p > Buf) Len += GetTextWidth(NULL, Buf, p-Buf);
	return (Len-xOffset <= ClientRect.right
			? (INT)(Len-xOffset) : ClientRect.right);
}

void InvalidateArea(PPOSITION Start, long Length, UINT Flags)
	/*Flags: 1 = erase flag for invalidation
	 *		 2 = invalidate to end of screen if area contains newlines
	 *		 4 = invalidate to end of screen even if doesn't seem necessary
	 *		 8 = invalidate to end of last line
	 */
{	INT		 i = 0, Cmp;
	RECT	 r;
	POSITION End;
	BOOL     NewLineFound = Flags & 4;

	if (!hwndMain) return;
	r = ClientRect;
	TextAreaOnly(&r, TRUE);
	/*find first involved line...*/
	while (i < VisiLines && ComparePos(&LineInfo[i+1].Pos, Start) <= 0
						 && !(LineInfo[i].Flags & LIF_EOF)) {
		++i;
		r.top += TextHeight;
	}
	if (LineInfo[i].Flags & LIF_EOF)
		if (i && ComparePos(&LineInfo[i].Pos, Start) >= 0) {
			--i;
			r.top -= TextHeight;
			Flags |= 8;
		}
	/*advance to first involved character in line...*/
	if (ComparePos(&LineInfo[i].Pos, Start) >= 0) r.left = 0;
	else r.left = StartX(FALSE) + GetTextPos(&LineInfo[i].Pos, Start, 0);
	/*find end position...*/
	End = *Start;
	while ((long)End.i + Length >= (long)End.p->Fill && End.p->Next) {
		Length -= End.p->Fill - End.i;
		End.p = End.p->Next;
		End.i = 0;
	}
	if ((long)End.i + Length > (long)End.p->Fill)
		Length = End.p->Fill - End.i;
	End.i += (INT)Length;
	/*finally, invalidate all following lines...*/
	while (i < VisiLines) {
		r.bottom = r.top + TextHeight;
		if ((Cmp = ComparePos(&LineInfo[++i].Pos, &End)) >= 0) {
			if (!NewLineFound && Cmp > 0 || !(Flags & 2)) {
				if (Cmp > 0 && !(Flags & 10) && !HexEditMode)
					r.right = StartX(FALSE) + 1 +
							  GetTextPos(&LineInfo[i-1].Pos, &End, 0);
				InvalidateRect(hwndMain, &r, Flags & 1);
				break;
			}
		}
		InvalidateRect(hwndMain, &r, Flags & 1);
		NewLineFound = TRUE;
		r.left = 0;
		r.top += TextHeight;
	}
}

void TextAreaOnly(RECT *r, BOOL WithinOffsets)
{
	if (WithinOffsets) {
		if (r->top	  < nToolbarHeight + YTEXT_OFFSET)
			r->top	  = nToolbarHeight + YTEXT_OFFSET;
		if (r->bottom > ClientRect.bottom-StatusHeight - YTEXT_OFFSET)
			r->bottom = ClientRect.bottom-StatusHeight - YTEXT_OFFSET;
	} else {
		if (r->top	  < nToolbarHeight)
			r->top	  = nToolbarHeight;
		if (r->bottom > ClientRect.bottom-StatusHeight)
			r->bottom = ClientRect.bottom-StatusHeight;
	}
	if (WinVersion >= MAKEWORD(95,3)) {
		if (r->left		< 2) r->left = 2;
		if (r->right	> ClientRect.right-GetSystemMetrics(SM_CXVSCROLL)-2)
			r->right	= ClientRect.right-GetSystemMetrics(SM_CXVSCROLL)-2;
	} else if (r->right > ClientRect.right-GetSystemMetrics(SM_CXVSCROLL))
			   r->right = ClientRect.right-GetSystemMetrics(SM_CXVSCROLL);
}

static BYTE	Buf[40];
static BOOL	MeasureCaret, NewCaret;
INT 		DispLine = -1, TabStop = 8, RightOffset = 0;
extern BOOL	ShowParenMatch;

static void Flush(HDC hDC, LONG x, INT y, PSTR Buf, INT Len, BOOL Selected)
{
	/*----------------------------------------------------------------------*/
	/* called from:	TextLine												*/
	/*----------------------------------------------------------------------*/
	if (hDC && Len && (x-=xOffset) > -1000 && x < ClientRect.right) {
		if (Selected) {
			SetBkMode(hDC, OPAQUE);
			if (ShowParenMatch) {
				SetBkColor	(hDC, ShmBkgndColor);
				SetTextColor(hDC, ShmTextColor);
			} else {
				SetBkColor	(hDC, SelBkgndColor);
				SetTextColor(hDC, SelTextColor);
			}
		} else	SetTextColor(hDC, TextColor);
		TextOut(hDC, (INT)x, y, Buf, Len);
		if (Selected) SetBkMode(hDC, TRANSPARENT);
	}
}

static BOOL EnforcePainting;
BOOL SuppressPaint;

static void PrepareNextLine(PPOSITION p, ULONG ByteCount, BOOL Continued)
{
	if (Continued)
		 LineInfo[DispLine].Flags |=  LIF_CONTD_NEXTLINE;
	else LineInfo[DispLine].Flags &= ~LIF_CONTD_NEXTLINE;
	++DispLine;
	if (LineInfo[DispLine].Pos.p != p->p		  ||
		LineInfo[DispLine].Pos.i != p->i		  ||
		LineInfo[DispLine].ByteCount != ByteCount ||
		LineInfo[DispLine].nthContdLine !=
			 (Continued ? LineInfo[DispLine-1].nthContdLine+1 : 0) ||
		!(LineInfo[DispLine].Flags & LIF_VALID))
	{
		LineInfo[DispLine].Pos = *p;
		LineInfo[DispLine].Flags |= LIF_VALID;
		LineInfo[DispLine].nthContdLine =
							Continued ? LineInfo[DispLine-1].nthContdLine+1 : 0;
		LineInfo[DispLine].ByteCount = ByteCount;
		EnforcePainting = TRUE;
	}
	if (CharAt(p) == C_EOF) LineInfo[DispLine].Flags |= LIF_EOF;
	else LineInfo[DispLine].Flags &= ~LIF_EOF;
}

INT DefaultNewLine = C_CRLF;

INT StartX(BOOL OkToRecalc)
{	static INT NumberDigits, NumberPower10;

	if (!NumberFlag || HexEditMode) return (XTEXT_OFFSET);
	if (!NumberDigits || OkToRecalc) {
		NumberDigits = 4;
		NumberPower10 = 1000;
		while (LinesInFile / NumberPower10 >= 9) {
			++NumberDigits;
			NumberPower10 *= 10;
		}
	}
	return (XTEXT_OFFSET + (NumberDigits + 1) * AveCharWidth);
}

void TextLine(HDC hDC, INT Line, INT YPos)
	/*----------------------------------------------------------------------*/
	/* calls:		CharAndAdvance	  CountBytes	Flush	 GetTextWidth	*/
	/*	  WinApi:	wsprintf												*/
	/*																		*/
	/* called from:	NewPosition		  Paint									*/
	/*----------------------------------------------------------------------*/
{	static POSITION p;
	INT 			i, c, Pos, i2, x2;
	LONG			x, LineStart;
	ULONG			ByteCount;
	BOOL			WithinSelect = FALSE;

	if (SuppressPaint) return;
	EnforcePainting = FALSE;
	if (!Line || LineInfo[Line].Flags & LIF_VALID) {
		DispLine = Line;
		p = Line ? LineInfo[Line].Pos : ScreenStart;
	} else {
		p = ScreenStart;
		if (!p.p) return;
		DispLine = 0;
		while (DispLine != Line) {
			if (HexEditMode) Advance(&p, 16);
			else if (!BreakLineFlag) {
				for (;;) {
					c = CharAndAdvance(&p);
					if (c < 0 || (CharFlags[c] & 1)) break;
				}
				if (c == '\r' && CharAt(&p) == '\n') CharAndAdvance(&p);
			} else {
				/*line too long, must be split*/
			}
			++DispLine;
		}
	}
	i = i2 = x2 = Pos = 0;
	MeasureCaret = FALSE;
	x = StartX(Line==0);
	if (NumberFlag && !HexEditMode) {
		/*display line number...*/

		if (CharAt(&p) != C_EOF) {
			char NumberBuf[12];
			INT	 Width, Len;
			Len = wsprintf (NumberBuf, "%lu", CountLines(&p) + 1);
			Width = GetTextWidth(hDC, NumberBuf, Len);
			Flush(hDC, x - AveCharWidth - Width, YPos, NumberBuf, Len, 0);
		}
	}
	LineStart = x;
	LineInfo[DispLine].Pos = p;
	LineInfo[DispLine].Flags = LIF_VALID;
	LineInfo[DispLine].ByteCount = ByteCount = CountBytes(&p);
	if (SelectCount && ByteCount > SelectBytePos
					&& ByteCount < SelectBytePos+SelectCount) {
		WithinSelect = TRUE;
	}
	for (;;) {
		extern BOOL HexEditTextSide;
		static char Buf2[18];	/*right column in hex edit mode*/

		if (!p.p->Fill && p.p->Next) p.p = p.p->Next;
		/*...I do not understand this currently.
		 *Without this line, the input of
		 *   "i<blank><ctrl+bksp><esc><pgdown><crsr-right><ctrl+home>"
		 *does not change the cursor position (after <ctrl+home>), though
		 *in fact, the cursor is at the home position.	8-Aug-97 Ramo
		 */

		if (p.i==CurrPos.i && p.p==CurrPos.p) {
			if (!Line || Line <= CrsrLines &&
					(p.i < p.p->Fill || i || x != LineStart ||
					 !HexEditMode &&
					 !(LineInfo[Line-1].Flags & LIF_EOF) &&
					 !(LineInfo[Line-1].Flags & LIF_CONTD_NEXTLINE))) {
				if (i) {
					INT Width = GetTextWidth(hDC, Buf, i);

					Pos += i;
					if (WithinSelect && HexEditMode &&
							ByteCount >= SelectBytePos+SelectCount && i)
						--i;
					if (x-xOffset < ClientRect.right)
						Flush(hDC, x, YPos, Buf, i, WithinSelect);
					x += Width;
					i = 0;
				}
				CaretX = (INT)x - (Mode == InsertMode);
				if (HexEditMode && HexEditTextSide) {
					if (i2) {
						Flush(hDC, x2, YPos, Buf2, i2, WithinSelect);
						x2 += GetTextWidth(hDC, Buf2, i2);
						i2 = 0;
					}
					CaretX += (INT)(x2 - x);
				}
				if (Mode == ReplaceMode) {
					CaretHeight = 3;
					CaretY = YPos + TextHeight - 3;
				} else {
					CaretY = YPos;
					CaretHeight = TextHeight;
				}
				NewCaret = MeasureCaret = TRUE;
				CurrLine = Line;
				CurrColumn = Pos + 1;
			}
		}
		if (SelectCount &&
				(WithinSelect ? ByteCount >= SelectBytePos+SelectCount
							  : ByteCount == SelectBytePos)) {
			if (HexEditMode) {
				if (i2) {
					Flush(hDC, x2, YPos, Buf2, i2, WithinSelect);
					x2 += GetTextWidth(hDC, Buf2, i2);
					i2 = 0;
				}
				if (WithinSelect && i) --i;
			}
			if (i) {
				Pos += i;
				if (x-xOffset < ClientRect.right)
					Flush(hDC, x, YPos, Buf, i, WithinSelect);
				x += GetTextWidth(hDC, Buf, i);
				i = 0;
				if (WithinSelect && HexEditMode) Buf[i++] = ' ';
			}
			WithinSelect ^= TRUE;
		}
		++ByteCount;
		if (HexEditMode) {
			INT			c2;

			switch (c = CharAndAdvance(&p)) {
				case C_EOF:
					if (!i && x==LineStart) {
						LineInfo[DispLine].Flags |= LIF_EOF;
						if (hDC) {
							SetTextColor(hDC, RGB(128,0,0));
							TextOut(hDC, (INT)x, YPos+2, MapOem+c, 1);
						}
					}
					#if 0
					//	else LineInfo[DispLine].Flags |= LIF_CONTD_NEXTLINE;
					#endif
					--ByteCount;
					if (MeasureCaret) {
						if (Mode == InsertMode) CaretWidth = InsertCaretWidth;
						else {
							CaretWidth = AveCharWidth;
							if (!HexEditTextSide) CaretWidth <<= 1;
							if (Mode != ReplaceMode) {
								CaretY += TextHeight >> 1;
								CaretHeight = 2;
							}
						}
						MeasureCaret = FALSE;
						if (!hDC) return;
					}
					break;
				case C_CRLF:
					GoBack(&p, 1);
					c = '\r';
					/*FALLTHROUGH*/
				default:
					if ((ByteCount & 15) == 1 /*already incremented*/) {
						INT Width;

						if (HexNumberFlag)
							 i = wsprintf(Buf, ByteCount>0x1000000L ?  "%07lx:"
																	: " %06lx:",
										  ByteCount-1);
						else i = wsprintf(Buf, ByteCount>  1000000L	?  "%07ld:"
																	: " %06ld:",
										  ByteCount-1);
						Flush(hDC, x, YPos, Buf, i, FALSE);
						i = 0;
						x += Width = 10 * AveCharWidth;
						Buf2[0]    = '|';
						Buf2[i2=1] = '\0';
						x2 = 62 * AveCharWidth;
						if ((WithinSelect || MeasureCaret) && i2) {
							Flush(hDC, x2, YPos, Buf2, i2, FALSE);
							x2 += GetTextWidth(hDC, Buf2, i2);
							i2 = 0;
						}
						if (MeasureCaret)
							if (HexEditTextSide) CaretX = x2;
							else CaretX += Width;
					}
					c2 = CharSet ? MapOem[c] : c;
					if (CharFlags[c] & 1 || CharFlags[c2] & 2) c2 = '\267';
					Buf2[i2++] = c2;
					i += wsprintf(Buf+i, "%02x", c);
					if (ByteCount & 3) Buf[i++] = ' ';
					if (MeasureCaret) {
						extern INT HexEditFirstNibble;

						if (Mode == InsertMode) CaretWidth = InsertCaretWidth;
						else if (HexEditTextSide)
							 CaretWidth = GetTextWidth(hDC, Buf2, 1);
						else CaretWidth = GetTextWidth(hDC, Buf,  2);
						if (HexEditFirstNibble != -1) {
							CaretX += AveCharWidth;
							if (Mode != InsertMode)
								CaretWidth = GetTextWidth(hDC, Buf+1, 1);
						}
						MeasureCaret = FALSE;
						if (!hDC) return;
					}
			}
			if (!(ByteCount & 3) || c == C_EOF) {
				Flush(hDC, x, YPos, Buf, i, WithinSelect);
				x += GetTextWidth(hDC, Buf, i);
				i = 0;
				if (!(ByteCount & 15) || c == C_EOF) {
					if (c != C_EOF && p.i >= p.p->Fill) continue;
					if (i2) {
						Flush(hDC, x2, YPos, Buf2, i2, WithinSelect);
						x2 += GetTextWidth(hDC, Buf2, i2);
					}
					if (x2) Flush(hDC, x2, YPos, "|", 1, FALSE);
					PrepareNextLine(&p, ByteCount, FALSE);
					return;
				}
				x = (INT)(10 * AveCharWidth + LineStart +
						((51 * AveCharWidth * (ByteCount & 15)) >> 4));
			}
		} else switch (c = CharAndAdvance(&p)) {
			case C_EOF:    /*end of file*/
			case C_ERROR:  /*error*/
				if (!i && x==LineStart) {
					LineInfo[DispLine].Flags |= LIF_EOF;
					if (hDC && xOffset < 32000) {
						SetTextColor(hDC, RGB(128,0,0));
						TextOut(hDC, (INT)(XTEXT_OFFSET-xOffset), YPos+2,
								MapOem+c, 1);
					}
				} else {
					LineInfo[DispLine].Flags |= LIF_CONTD_NEXTLINE;
					if (c == C_EOF) c = C_UNDEF;
				}
				--ByteCount;
				/*FALLTHROUGH*/
			case C_CRLF:
			case '\r':
			case '\n':
			case '\36':	/*Record Separator*/
			case '\0':	/*real Null byte*/
				{	INT Width = GetTextWidth(hDC, Buf, i);

					if (i && x-xOffset<ClientRect.right)
						Flush(hDC, x, YPos, Buf, i, WithinSelect);
					x += Width;
					if (x > ShrinkRedraw) ShrinkRedraw = x;
				}
				if (WithinSelect && hDC) {
					RECT r;

					r.left   = (INT)(x - xOffset);
					r.top    = YPos;
					r.right  = (INT)(ClientRect.right - XTEXT_OFFSET);
					r.bottom = YPos + TextHeight;
					if (r.left < r.right) FillRect(hDC, &r, SelBkgndBrush);
				}
				if (c == C_CRLF) ++ByteCount;
				if (c != C_EOF && (c != DefaultNewLine || ListFlag) && hDC &&
						x-xOffset > -32000 && x-xOffset < ClientRect.right) {
					HFONT OldFont = SelectObject(hDC, SpecialFont);
					INT	  y = YPos;

					SetTextColor(hDC, RGB(128,0,0));
					if (x != LineStart) x += 2;
					TextOut(hDC, (INT)(x-xOffset), y, MapOem+c, 1);
					if (OldFont) SelectObject(hDC, OldFont);
				}
				PrepareNextLine(&p, ByteCount, FALSE);
				if (MeasureCaret) {
					if (Mode != ReplaceMode) {
						if (Mode != InsertMode) {
							CaretWidth = AveCharWidth;
							/*other caret shape for visualizing empty lines...*/
							CaretY += CaretHeight >> 1;
							CaretHeight = 2;
						} else CaretWidth = InsertCaretWidth;
					}
					MeasureCaret = FALSE;
				}
				return;

			case '\t':
				if (!ListFlag) {
					if (i) {
						Pos += i;
						if (x-xOffset < ClientRect.right)
							Flush(hDC, x, YPos, Buf, i, WithinSelect);
						x += GetTextWidth(hDC, Buf, i);
						if (x > ShrinkRedraw) ShrinkRedraw = x;
						i = 0;
					}
					Pos += TabStop - Pos % TabStop;
					{	INT NewX;

						NewX = (INT)(LineStart + Pos * AveCharWidth);
						if (NewX > x) {
							if (MeasureCaret) {
								CaretWidth = Mode != InsertMode
												  || NewX-x  < InsertCaretWidth
											 ? (INT)(NewX-x) : InsertCaretWidth;
								MeasureCaret = FALSE;
								if (!hDC) return;
							}
							if (WithinSelect && hDC) {
								RECT r;

								r.left   = (INT)(x - xOffset);
								r.top    = YPos;
								r.right  = (INT)(NewX - xOffset);
								r.bottom = YPos + TextHeight;
								if (r.left < r.right)
									FillRect(hDC, &r, SelBkgndBrush);
							}
							x = NewX;
						} else if (MeasureCaret) {
							CaretWidth = Mode==InsertMode ? InsertCaretWidth
														  : AveCharWidth;
							MeasureCaret = FALSE;
							if (!hDC) return;
						}
					}
					if (BreakLineFlag &&
							x-xOffset > ClientRect.right-RightOffset) {
						PrepareNextLine(&p, ByteCount, TRUE);
						if (x < ExpandRedraw) ExpandRedraw = x;
						return;
					}
					break;
				}
				/*FALLTHROUGH*/
			default:
				{	INT OldI = i, c2;
					INT Width;

					c2 = CharSet ? MapOem[c] : c;
					if (CharFlags[c2] & 2) {
						if (c < ' ') i += wsprintf(Buf+i, "^%c", c + 'A'-1);
						else		 i += wsprintf(Buf+i, "\\%02X", c);
					} else Buf[i++] = c2;
					Width = GetTextWidth(hDC, Buf, i);
					if (MeasureCaret) {
						CaretWidth = Mode==InsertMode ? InsertCaretWidth
													  : Width;
						MeasureCaret = FALSE;
						if (!hDC) return;
					}
					if (BreakLineFlag
							&& x-xOffset+Width > ClientRect.right-RightOffset) {
						Flush(hDC, x, YPos, Buf, OldI, WithinSelect);
						if (x+Width < ExpandRedraw) ExpandRedraw = x+Width;
						Width = GetTextWidth(hDC, Buf, OldI);
						if (x+Width > ShrinkRedraw) ShrinkRedraw = x+Width;
						GoBack(&p, 1);
						PrepareNextLine(&p, ByteCount-1, TRUE);
						return;
					}
					if (i >= sizeof(Buf)-3) {
						Pos += i;
						if (x-xOffset < ClientRect.right)
							Flush(hDC, x, YPos, Buf, i, WithinSelect);
						x += Width;
						i = 0;
					}
				}
				break;
		}
	}
}

void Paint(HDC hDC, PRECT pRect)
	/*----------------------------------------------------------------------*/
	/* calls:		PaintStatus			PaintTools			TextLine		*/
	/*	  WinApi:	GetTextMetrics		GetSystemMetrics	GetClientRect	*/
	/*				IntersectClipRect	SelectObject		SetBkMode		*/
	/*																		*/
	/* called from:	(winvi.c)MainProc(WM_PAINT)								*/
	/*----------------------------------------------------------------------*/
{	extern BOOL SessionEnding;
	INT			y = nToolbarHeight + YTEXT_OFFSET, xStart = 0;
	HFONT		Font = HexEditMode ? HexFont : TextFont, OldFont;

	if (SessionEnding) return;
	if (Font) OldFont = SelectObject(hDC, Font);
	if (!AveCharWidth) {
		static TEXTMETRIC tm;
		GetTextMetrics(hDC, &tm);
		AveCharWidth = tm.tmAveCharWidth + tm.tmOverhang;
		IsFixedFont  = GetTextWidth(hDC, "W", 1) == GetTextWidth(hDC, "i", 1);
		/*there are flags for fixed pitch, but it's not working correctly*/
	}
	if (!RightOffset)
		 RightOffset = XTEXT_OFFSET + GetSystemMetrics(SM_CXVSCROLL) - 2;
	if (!ClientRect.right) GetClientRect(hwndMain, &ClientRect);
	if (pRect->top < nToolbarHeight) PaintTools(hDC, pRect);
	if (pRect->bottom > ClientRect.bottom-StatusHeight)
		PaintStatus(hDC, pRect, &ClientRect);
	if (WinVersion >= MAKEWORD(95,3)) {
		POINT Line[2];
		HPEN  OldPen;

		if (pRect->left < 2) {
			/*paint shadow lines on left side...*/
			Line[0].x = Line[1].x = 0;
			Line[0].y = nToolbarHeight - 1;
			Line[1].y = ClientRect.bottom - StatusHeight + 2;
			OldPen = SelectObject(hDC, ShadowPen);
			Polyline(hDC, Line, 2);
			Line[0].x = Line[1].x = 1;
			++Line[0].y; --Line[1].y;
			SelectObject(hDC, BlackPen);
			Polyline(hDC, Line, 2);
			SelectObject(hDC, OldPen);
			xStart = 2;
		}
		if (pRect->right > ClientRect.right-2) {
			/*paint shadow lines on right of scrollbar*/
			Line[0].x = Line[1].x = ClientRect.right - 2;
			Line[0].y = nToolbarHeight - 1;
			Line[1].y = ClientRect.bottom - StatusHeight + 1;
			OldPen = SelectObject(hDC, ShadowPen);
			Polyline(hDC, Line, 2);
			Line[0].x = ++Line[1].x;
			SelectObject(hDC, HilitePen);
			Polyline(hDC, Line, 2);
			SelectObject(hDC, OldPen);
		}
	}
	if (pRect->bottom > ClientRect.bottom-StatusHeight-YTEXT_OFFSET ||
			xStart || pRect->right > ClientRect.right-2)
		IntersectClipRect(hDC, xStart, 0,
			ClientRect.right  - (RightOffset  - XTEXT_OFFSET + 2),
			ClientRect.bottom - (StatusHeight + YTEXT_OFFSET));
	SetBkMode(hDC, TRANSPARENT);
	NewCaret = FALSE;
	if (LineInfo) {
		INT line = 0, Lower;

		Lower = ClientRect.bottom - StatusHeight - YTEXT_OFFSET;
		if (Lower > pRect->bottom && !PaintAll && LineInfo &&
							 (LineInfo[VisiLines].Flags & LIF_EOF ||
							  LineInfo[VisiLines].Pos.p != NULL))
			Lower = pRect->bottom;
		PaintAll = FALSE;
		while (y < Lower) {
			if (EnforcePainting || PaintAll || pRect->top <= y+TextHeight)
				TextLine(hDC, line, y);
			++line;
			y += TextHeight;
		}
	} /*else fatal error*/
	if (Font && OldFont) SelectObject(hDC, OldFont);
}

INT GetPixelLine(INT Line, BOOL WithinTextArea)
{	INT i = Line * TextHeight;

	if (!WithinTextArea) i += nToolbarHeight + YTEXT_OFFSET;
	return (i);
}

void GetXPos(PCOLUMN pPos)
{	INT Line = (CaretY - nToolbarHeight - YTEXT_OFFSET) / TextHeight;

	pPos->ScreenLine = 0;
	while (Line > 0 && LineInfo[--Line].Flags & LIF_CONTD_NEXTLINE)
		++pPos->ScreenLine;
	/*TODO: first screen line may be a continuation line*/
	pPos->PixelOffset = CaretX - StartX(FALSE) + (Mode==InsertMode);
}

char CommandBuf[280];

void CommandEdit(UINT StartChar)
{
	if (Disabled || hwndCmd) return;
	for (;;) {
		static char Buf[2];
		DWORD		LastError;

		Buf[0] = StartChar;
		NewStatus(0, "", NS_NORMAL);
		hwndCmd = CreateWindow("Edit", Buf,
							   WS_CHILD | ES_AUTOHSCROLL | ES_LEFT,
							   StatusFields[0].x+1,
							   ClientRect.bottom-StatusHeight+5,
							   StatusFields[0].Width-1, StatusTextHeight,
							   hwndMain, (HMENU)ID_COMMAND, hInst, 0L);
		if (hwndCmd) break;
		#if defined(WIN32)
		{	extern char  QueryString[150];
			extern DWORD QueryTime;
			char		 LastErrString[32];

			if ((LastError = GetLastError()) == ERROR_NOT_ENOUGH_MEMORY)
				 lstrcpy (LastErrString, "ERROR_NOT_ENOUGH_MEMORY");
			else wsprintf(LastErrString, "%ld", LastError);
			wsprintf(QueryString, "CreateWindow failed, LastError=%s",
								  LastErrString);
			QueryTime = GetTickCount();
		}
		#else
			#if !defined(ERROR_NOT_ENOUGH_MEMORY)
				#define  ERROR_NOT_ENOUGH_MEMORY 8
			#endif
			LastError =  ERROR_NOT_ENOUGH_MEMORY;
		#endif
		if (LastError==ERROR_NOT_ENOUGH_MEMORY && FreeSpareMemory()) continue;
		ErrorBox(MB_ICONSTOP, 301);
		return;
	}
	SendMessage(hwndCmd, EM_LIMITTEXT, sizeof(CommandBuf)-1, 0L);
	SendMessage(hwndCmd, WM_SETFONT,    (WPARAM)StatusFont,	 0L);
	#if defined(WIN32)
		SendMessage(hwndCmd, EM_SETSEL, 1, 1);
	#else
		SendMessage(hwndCmd, EM_SETSEL, (WPARAM)FALSE, MAKELPARAM(1,1));
	#endif
	ShowWindow(hwndCmd, SW_SHOW);
	SetFocus(hwndCmd);
}

void CommandDone(BOOL Escape)
{
	if (hwndCmd) {
		HWND ToDestroy = hwndCmd;

		GetWindowText(hwndCmd, CommandBuf, sizeof(CommandBuf));
		hwndCmd = 0;  /*must be cleared when window disappears*/
		DestroyWindow(ToDestroy);
		if (!Escape) {
			if (*CommandBuf) CommandExec(CommandBuf);
		} else if (SelectCount) {
			InvalidateArea(&SelectStart, SelectCount, 1);
			SelectCount = 0;
		}
		CommandLineDone(Escape, CommandBuf);
		if (hwndMain) {
			ShowEditCaret();
			SetFocus(hwndMain);
		}
	}
}

static char ErrBuf1[200], ErrBuf2[280];
extern char Buffer[300];

void Error(INT StringNo, ...)
{
	LOADSTRING(hInst, StringNo, ErrBuf1, sizeof(ErrBuf1));
	if (!*ErrBuf1) return;
	wvsprintf(Buffer, ErrBuf1, (char*)(&StringNo+1));
	NewStatus(0, Buffer, NS_ERROR);
}

char ErrorTitle[16] = {'\0'};

INT ErrorBox(INT Flags, INT StringNo, ...)
{
	LOADSTRING(hInst, StringNo, ErrBuf1, sizeof(ErrBuf1));
	if (!*ErrBuf1 /*error string not defined in resource string table*/)
		return(IDCANCEL);
	if (!*ErrorTitle) LOADSTRING(hInst, 314, ErrorTitle, sizeof(ErrorTitle));
	wvsprintf(ErrBuf2, ErrBuf1, (char*)(&StringNo+1));
	return (MessageBox(hwndErrorParent, ErrBuf2,
			Flags & MB_ICONINFORMATION ? ApplName : ErrorTitle, MB_OK | Flags));
}
