/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <commdlg.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <direct.h>
#include "winvi.h"
#include "paint.h"
#include "colors.h"

extern PCSTR   ApplName;
extern RECT    WholeWindow;
extern int	   x, y, Width, Height;
extern BOOL    Maximized, RefreshFlag;
extern int	   OemCodePage, AnsiCodePage;

#if defined(WIN32)
	   BOOL	   LcaseFlag		 = FALSE;
#else
	   BOOL	   LcaseFlag		 = TRUE;
#endif

static LPCSTR  ProfileSection	 = "Ramo's WinVi";
static LPCSTR  pfLanguage		 = "Language";
static LPCSTR  pfFontFace	 	 = "FontName";
static LPCSTR  pfFontSize	 	 = "FontSize";
static LPCSTR  pfFontAttr	 	 = "FontAttributes";
static LPCSTR  pfFontCharSet	 = "FontCharSet";
static LPCSTR  pfHexFontFace 	 = "HexFontName";
static LPCSTR  pfHexFontSize 	 = "HexFontSize";
static LPCSTR  pfHexFontAttr	 = "HexFontAttributes";
static LPCSTR  pfHexFontCharSet	 = "HexFontCharSet";
static LPCSTR  pfXPos 			 = "X-Position";
static LPCSTR  pfYPos 			 = "Y-Position";
static LPCSTR  pfWidth			 = "Width";
static LPCSTR  pfHeight			 = "Height";
static LPCSTR  pfMaximized		 = "Maximized";
static LPCSTR  pfFgColor		 = "TextColor";
static LPCSTR  pfBkColor		 = "BackgroundColor";
static LPCSTR  pfSelFgColor		 = "SelectedText";
static LPCSTR  pfSelBkColor		 = "SelectedBackground";
static LPCSTR  pfShmFgColor		 = "ShowMatchText";
static LPCSTR  pfShmBkColor		 = "ShowMatchBackground";
static LPCSTR  pfTerse			 = "Terse";
static LPCSTR  pfLcaseFnames	 = "LowerCaseFnames";
static LPCSTR  pfReadOnly		 = "ReadOnly";
static LPCSTR  pfAutoRestore	 = "AutoRestore";
static LPCSTR  pfAutoIndent		 = "AutoIndent";
static LPCSTR  pfAutoWrite		 = "AutoWrite";
static LPCSTR  pfShowMatch		 = "ShowMatch";
static LPCSTR  pfWrapScan		 = "WrapScan";
static LPCSTR  pfWrapPos		 = "WrapPos";
static LPCSTR  pfIgnoreCase		 = "IgnoreCase";
static LPCSTR  pfMagic			 = "Magic";
static LPCSTR  pfQuitIfMore		 = "QuitIfMoreFiles";
static LPCSTR  pfTabStop		 = "TabStop";
static LPCSTR  pfUndoLimit		 = "UndoLimit";
static LPCSTR  pfShiftWidth		 = "ShiftWidth";
static LPCSTR  pfDefaultInsert	 = "DefaultInsert";
static LPCSTR  pfLineNumber		 = "LineNumber";
static LPCSTR  pfHexLineNumber	 = "HexModeLineNumber";
static LPCSTR  pfTags			 = "Tags";
static LPCSTR  pfTmpDirectory	 = "TmpDirectory";
static LPCSTR  pfPageThreshold	 = "PageThreshold";
static LPCSTR  pfBitmap			 = "Bitmap";
static LPCSTR  pfRefreshBkgnd	 = "RefreshBackground";
static LPCSTR  pfFullRowExtend	 = "FullRowExtend";
static LPCSTR  pfBreakLine		 = "BreakLine";
static LPCSTR  pfSearchBoxX		 = "SearchBox-X";
static LPCSTR  pfSearchBoxY		 = "SearchBox-Y";
static LPCSTR  pfOemCodePage	 = "OemCodePage";
static LPCSTR  pfAnsiCodePage	 = "AnsiCodePage";
static LPCSTR  pfPrintBorderLeft = "PrintBorderLeft";
static LPCSTR  pfPrintBorderTop	 = "PrintBorderTop";
static LPCSTR  pfPrintBorderRight= "PrintBorderRight";
static LPCSTR  pfPrintBorderBottom= "PrintBorderBottom";
static LPCSTR  pfPrintNumbers	 = "PrintNumbers";
static LPCSTR  pfPrintFontFace	 = "PrintFontFace";
static LPCSTR  pfPrintFontAttr	 = "PrintFontAttributes";
static LPCSTR  pfPrintFontCharSet= "PrintFontCharSet";
static LPCSTR  pfPrintFontSize1	 = "PrintFontSize1";
static LPCSTR  pfPrintFontSize2	 = "PrintFontSize2";
static LPCSTR  pfPrintFrame		 = "PrintFrame";
static LPCSTR  pfPrintFrame2	 = "PrintFrame2";
static LPCSTR  pfPrintFrame3	 = "PrintFrame3";
static LPCSTR  pfPrintHeader	 = "PrintHeader";
static LPCSTR  pfPrintFooter	 = "PrintFooter";
static LPCSTR  pfPrintOrientation= "PrintOrientation";
static LPCSTR  pfRecentFiles	 = "RecentFiles";
static LPCSTR  pfRecentFile		 = "RecentFile%d";
static LPCSTR  pfRecentDir		 = "RecentDir%d";
static LPCSTR  pfMaxScroll		 = "MaxScroll";
static LPCSTR  pfoPortrait		 = "Portrait";
static LPCSTR  pfoLandscape		 = "Landscape";
static LPCSTR  pfoDoubleColumn	 = "DoubleColumn";

static DWORD MyGetProfileLong(LPCSTR Key, DWORD Default, WORD SysDefaultMask)
{	char Buf[16];

	GetProfileString(ProfileSection, Key, "", Buf, sizeof(Buf));
	if (!*Buf) {
		wUseSysColors |= SysDefaultMask;
		return (Default);
	}
	return (strtol(Buf, NULL, 0));
}

static VOID MyGetProfileString(LPCSTR Key, LPCSTR Default, LPSTR *pResult)
{	char Buf[260];

	GetProfileString(ProfileSection, Key, Default, Buf, sizeof(Buf));
	AllocString(pResult, Buf);
}

static BOOL MyGetProfileBool(LPCSTR Key, BOOL Default)
{	char Buf[4];

	GetProfileString(ProfileSection, Key, "", Buf, sizeof(Buf));
	if (!*Buf) return (Default);
	return (*Buf == '1' || (*Buf = LCASE(*Buf)) == 'y' || *Buf == 't');
}

static void MyWriteProfileInt(LPCSTR Key, PSTR Format, long Value)
{	char Buf[16];

	wsprintf(Buf, Format, Value);
	WriteProfileString(ProfileSection, Key, Buf);
}

static void MyWriteProfileString(LPCSTR Key, LPCSTR Value)
{
	WriteProfileString(ProfileSection, Key, Value);
}

static void MyWriteProfileBool(LPCSTR Key, BOOL Value)
{
	WriteProfileString(ProfileSection, Key, Value ? "Yes" : "No");
}

static VOID GetFontAttributes(LPCSTR ProfileKey, PLOGFONT LogFont)
{	char Buf[20], *p;

	if (GetProfileString(ProfileSection, ProfileKey, "", Buf,sizeof(Buf))) {
		LogFont->lfWeight = (int)strtol(Buf, &p, 10);
		--p;
		while (*++p == ',') {
			switch (*++p) {
				case 'i':	LogFont->lfItalic	 = TRUE;	continue;
				case 'u':	LogFont->lfUnderline = TRUE;	continue;
				case 's':	LogFont->lfStrikeOut = TRUE;	continue;
			}
			break;
		}
	}
}

static VOID WriteFontAttributes(LPCSTR ProfileKey, PLOGFONT LogFont)
{	char Buf[20];

	wsprintf(Buf, "%d", LogFont->lfWeight);
	if (LogFont->lfItalic)	  lstrcat(Buf, ",i");
	if (LogFont->lfUnderline) lstrcat(Buf, ",u");
	if (LogFont->lfStrikeOut) lstrcat(Buf, ",s");
	MyWriteProfileString(ProfileKey, Buf);
}

char *LdFormat  = "%ld";
char *L6xFormat = "0x%06lx";
BOOL SearchBoxPosition;
int  SearchBoxX, SearchBoxY;

void SaveColors(void)
{
	if (wUseSysColors & 2)
		MyWriteProfileString(pfFgColor,	   NULL);
	else MyWriteProfileInt(pfFgColor,	   L6xFormat,	 TextColor);
	if (wUseSysColors & 1)
		MyWriteProfileString(pfBkColor,	   NULL);
	else MyWriteProfileInt(pfBkColor,	   L6xFormat,	 BackgroundColor);
	if (wUseSysColors & 8)
		MyWriteProfileString(pfSelFgColor, NULL);
	else MyWriteProfileInt(pfSelFgColor,   L6xFormat,	 SelTextColor);
	if (wUseSysColors & 4)
		MyWriteProfileString(pfSelBkColor, NULL);
	else MyWriteProfileInt(pfSelBkColor,   L6xFormat,	 SelBkgndColor);
	MyWriteProfileInt	  (pfShmFgColor,   L6xFormat,	 ShmTextColor);
	MyWriteProfileInt	  (pfShmBkColor,   L6xFormat,	 ShmBkgndColor);
	MyWriteProfileString(pfBitmap,
						 *BitmapFilename ? (LPSTR)BitmapFilename : (LPSTR)NULL);
}

static void SaveFont(PLOGFONT LogFont, LPCSTR FaceKey, LPCSTR HeightKey,
									   LPCSTR AttrKey, LPCSTR CharsetKey)
{
	if (*LogFont->lfFaceName) {
		MyWriteProfileString(FaceKey,				 LogFont->lfFaceName);
		MyWriteProfileInt	(HeightKey,	 LdFormat,   LogFont->lfHeight);
		MyWriteProfileInt	(CharsetKey, LdFormat,   LogFont->lfCharSet);
		WriteFontAttributes(AttrKey, LogFont);
	}
}

extern LOGFONT TextLogFont, HexLogFont;

void SaveSettings(void)
{
	if (!UnsafeSettings) return;
	if (  (UINT)((WholeWindow.left + WholeWindow.right) / 2)
				< (UINT)GetSystemMetrics(SM_CXSCREEN) &&
		  (UINT)((WholeWindow.top + WholeWindow.bottom) / 2)
				< (UINT)GetSystemMetrics(SM_CYSCREEN)) {
		MyWriteProfileInt(pfXPos,	LdFormat, WholeWindow.left);
		MyWriteProfileInt(pfYPos,	LdFormat, WholeWindow.top);
	}
	MyWriteProfileInt(pfWidth,	LdFormat, WholeWindow.right -WholeWindow.left);
	MyWriteProfileInt(pfHeight, LdFormat, WholeWindow.bottom-WholeWindow.top);
	MyWriteProfileBool(pfMaximized, IsZoomed(hwndMain));
	if (SearchBoxPosition) {
		MyWriteProfileInt(pfSearchBoxX, LdFormat, SearchBoxX);
		MyWriteProfileInt(pfSearchBoxY, LdFormat, SearchBoxY);
	}
}

void SaveFontSettings(void)
{
	if (wUseSysColors & 2)
		MyWriteProfileString(pfFgColor,	   NULL);
	else MyWriteProfileInt(pfFgColor,	   L6xFormat,	 TextColor);
	SaveFont(&TextLogFont, pfFontFace, pfFontSize, pfFontAttr, pfFontCharSet);
	SaveFont(&HexLogFont,  pfHexFontFace, pfHexFontSize, pfHexFontAttr,
						   pfHexFontCharSet);
}

void SaveEditSettings(void)
{
	MyWriteProfileBool(pfReadOnly,				  ReadOnlyFlag);
	MyWriteProfileBool(pfAutoIndent,			  AutoIndentFlag);
	MyWriteProfileBool(pfAutoWrite,				  AutoWriteFlag);
	MyWriteProfileBool(pfWrapScan,				  WrapScanFlag);
	MyWriteProfileBool(pfWrapPos,				  WrapPosFlag);
	MyWriteProfileBool(pfIgnoreCase,			  IgnoreCaseFlag);
	MyWriteProfileBool(pfMagic,					  MagicFlag);
	MyWriteProfileBool(pfQuitIfMore,			  QuitIfMoreFlag);
	MyWriteProfileInt (pfUndoLimit,		LdFormat, UndoLimit);
	MyWriteProfileInt (pfShiftWidth,	LdFormat, ShiftWidth);
	MyWriteProfileBool(pfDefaultInsert,			  DefaultInsFlag);
	MyWriteProfileBool(pfFullRowExtend,			  FullRowFlag);
	MyWriteProfileString(pfTmpDirectory, DefaultTmpFlag ? NULL : TmpDirectory);
	MyWriteProfileString(pfTags,
						 Tags!=NULL && lstrcmp(Tags, "tags") ? Tags : NULL);
	MyWriteProfileInt (pfPageThreshold, LdFormat, PageThreshold);
	#if defined(WIN32)
		if (OemCodePage==0)
			 MyWriteProfileString(pfOemCodePage,  NULL);
		else MyWriteProfileInt	 (pfOemCodePage,  LdFormat, OemCodePage);
		if (AnsiCodePage==0)
			 MyWriteProfileString(pfAnsiCodePage, NULL);
		else MyWriteProfileInt	 (pfAnsiCodePage, LdFormat, AnsiCodePage);
	#endif
}

void SaveDisplaySettings(void)
{
	MyWriteProfileBool(pfTerse,					  TerseFlag);
	MyWriteProfileBool(pfAutoRestore,			  AutoRestoreFlag);
	MyWriteProfileBool(pfShowMatch,				  ShowMatchFlag);
	MyWriteProfileInt (pfTabStop,		LdFormat, TabStop);
	MyWriteProfileInt (pfMaxScroll,		LdFormat, MaxScroll);
	MyWriteProfileBool(pfLineNumber,			  NumberFlag);
	MyWriteProfileString(pfHexLineNumber, HexNumberFlag ? "hexadecimal" : NULL);
	MyWriteProfileBool(pfRefreshBkgnd,			  RefreshFlag);
	#if defined(WIN32)
		if (OemCodePage==0)
			 MyWriteProfileString(pfOemCodePage,  NULL);
		else MyWriteProfileInt	 (pfOemCodePage,  LdFormat, OemCodePage);
		if (AnsiCodePage==0)
			 MyWriteProfileString(pfAnsiCodePage, NULL);
		else MyWriteProfileInt	 (pfAnsiCodePage, LdFormat, AnsiCodePage);
	#endif
}

void SaveFileSettings(void)
{
	MyWriteProfileBool(pfLcaseFnames,			  LcaseFlag);
	MyWriteProfileBool(pfQuitIfMore,			  QuitIfMoreFlag);
	MyWriteProfileString(pfTmpDirectory, DefaultTmpFlag ? NULL : TmpDirectory);
	MyWriteProfileString(pfTags,
						 Tags!=NULL && lstrcmp(Tags, "tags") ? Tags : NULL);
	MyWriteProfileInt (pfPageThreshold, LdFormat, PageThreshold);
}

void LoadFontInfo(LPCSTR FaceKey, LPCSTR HeightKey, LPCSTR AttrKey,
				  LPCSTR CharsetKey, PLOGFONT LogFont)
{	extern int	HexTextHeight, TextTextHeight;

	if (GetProfileString(ProfileSection, FaceKey, "Courier",
						 LogFont->lfFaceName, sizeof(LogFont->lfFaceName))) {
		LogFont->lfHeight  = GetProfileInt(ProfileSection, HeightKey, 0);
		if (!LogFont->lfHeight) {
			HDC hDC = GetDC(NULL);

			if (hDC) {
				LogFont->lfHeight =
					-MulDiv(10/*pt*/, GetDeviceCaps(hDC, LOGPIXELSY), 72);
				ReleaseDC(NULL, hDC);
			} else LogFont->lfHeight = 10;
		}
		GetFontAttributes(AttrKey, LogFont);
		LogFont->lfCharSet = GetProfileInt(ProfileSection, CharsetKey,
										  DEFAULT_CHARSET);
										  /*don't care ansi, etc.*/
		if (LogFont == &TextLogFont) {
			NewFont(&TextFont, LogFont, &TextTextHeight);
			if (!HexEditMode) TextHeight = TextTextHeight;
		} else {
			NewFont(&HexFont,  LogFont, &HexTextHeight);
			if (HexEditMode) TextHeight = HexTextHeight;
		}
	}
}

COLORREF ShmTextColor  = RGB(255,255,255);
COLORREF ShmBkgndColor = RGB(255,000,000);
BOOL	 SuppressAsterisk;

void LoadSettings(void)
{
	LoadFontInfo(pfFontFace, pfFontSize, pfFontAttr, pfFontCharSet,
				 &TextLogFont);
	LoadFontInfo(pfHexFontFace, pfHexFontSize, pfHexFontAttr, pfHexFontCharSet,
				 &HexLogFont);
	x		   = GetProfileInt(ProfileSection, pfXPos,		 x);
	y		   = GetProfileInt(ProfileSection, pfYPos,		 y);
	Width	   = GetProfileInt(ProfileSection, pfWidth,		 Width);
	Height	   = GetProfileInt(ProfileSection, pfHeight,	 Height);
	TabStop	   = GetProfileInt(ProfileSection, pfTabStop,	 TabStop);
	ShiftWidth = GetProfileInt(ProfileSection, pfShiftWidth, ShiftWidth);
	MaxScroll  = GetProfileInt(ProfileSection, pfMaxScroll,	 MaxScroll);
	WholeWindow.left   = x;
	WholeWindow.top    = y;
	WholeWindow.right  = x + Width;
	WholeWindow.bottom = y + Height;
	SelTextColor  = GetSysColor(COLOR_HIGHLIGHTTEXT);
	SelBkgndColor = GetSysColor(COLOR_HIGHLIGHT);
	Maximized		= MyGetProfileBool(pfMaximized,		FALSE);
	TextColor		= MyGetProfileLong(pfFgColor,		TextColor,		 2);
	BackgroundColor	= MyGetProfileLong(pfBkColor,		BackgroundColor, 1);
	SelTextColor	= MyGetProfileLong(pfSelFgColor,	SelTextColor,	 8);
	SelBkgndColor	= MyGetProfileLong(pfSelBkColor,	SelBkgndColor,	 4);
	ShmTextColor	= MyGetProfileLong(pfShmFgColor,	ShmTextColor,	 0);
	ShmBkgndColor	= MyGetProfileLong(pfShmBkColor,	ShmBkgndColor,	 0);
	UndoLimit		= MyGetProfileLong(pfUndoLimit,		UndoLimit,		 0);
	GetProfileString(ProfileSection, pfBitmap, "",
					 BitmapFilename, sizeof(BitmapFilename));
	TerseFlag		= MyGetProfileBool(pfTerse, 		FALSE);
	LcaseFlag		= MyGetProfileBool(pfLcaseFnames,	LcaseFlag);
	if (MyGetProfileBool(pfReadOnly, FALSE)) ReadOnlyFlag = TRUE;
	AutoRestoreFlag	= MyGetProfileBool(pfAutoRestore,	TRUE);
	AutoIndentFlag	= MyGetProfileBool(pfAutoIndent,	FALSE);
	AutoWriteFlag	= MyGetProfileBool(pfAutoWrite,		FALSE);
	ShowMatchFlag	= MyGetProfileBool(pfShowMatch,		FALSE);
	WrapScanFlag	= MyGetProfileBool(pfWrapScan,		TRUE);
	WrapPosFlag		= MyGetProfileBool(pfWrapPos,		FALSE);
	IgnoreCaseFlag	= MyGetProfileBool(pfIgnoreCase,	TRUE);
	MagicFlag		= MyGetProfileBool(pfMagic,			TRUE);
	QuitIfMoreFlag	= MyGetProfileBool(pfQuitIfMore,	FALSE);
	DefaultInsFlag	= MyGetProfileBool(pfDefaultInsert,	TRUE);
	NumberFlag		= MyGetProfileBool(pfLineNumber,	FALSE);
	RefreshFlag		= MyGetProfileBool(pfRefreshBkgnd,	FALSE);
	FullRowFlag		= MyGetProfileBool(pfFullRowExtend,	TRUE);
	BreakLineFlag	= MyGetProfileBool(pfBreakLine,		FALSE);
	SuppressAsterisk= MyGetProfileBool("SuppressCaptionAsterisk", FALSE);
	MyGetProfileString(pfHexLineNumber, "", &TmpDirectory);
	HexNumberFlag	= LCASE(*TmpDirectory) == 'h';
	MyGetProfileString(pfTmpDirectory,  "", &TmpDirectory);
	MyGetProfileString(pfTags, "tags", &Tags);
	DefaultTmpFlag	= !*TmpDirectory;
	PageThreshold	= MyGetProfileLong(pfPageThreshold,	PageThreshold,	 0);
	SearchBoxX		= GetProfileInt(ProfileSection, pfSearchBoxX, -500);
	if (SearchBoxX != -500) {
		SearchBoxY	= GetProfileInt(ProfileSection, pfSearchBoxY, 50);
		SearchBoxPosition = TRUE;
	}
	MakeBackgroundBrushes();
	OemCodePage		= GetProfileInt(ProfileSection, pfOemCodePage,  0);
	AnsiCodePage	= GetProfileInt(ProfileSection, pfAnsiCodePage, 0);
	{	extern VOID SetNewCodePage(VOID);

		SetNewCodePage();
	}
}

extern int	 nPrtLeft, nPrtTop, nPrtRight, nPrtBottom;
extern BOOL  LineNumbering, Frame, Frame2, Frame3;
extern BOOL	 OverideOrientation, DoubleColumn;
extern int	 nOrientation;
extern LPSTR Header, Footer;
extern LOGFONT    lf;
extern CHOOSEFONT cf;

static char PrintBuf[82];

VOID StorePrintSettings(VOID)
{	static PSTR Fmt = "%d";

	MyWriteProfileInt	(pfPrintBorderLeft,   Fmt, nPrtLeft);
	MyWriteProfileInt	(pfPrintBorderTop,    Fmt, nPrtTop);
	MyWriteProfileInt	(pfPrintBorderRight,  Fmt, nPrtRight);
	MyWriteProfileInt	(pfPrintBorderBottom, Fmt, nPrtBottom);
	MyWriteProfileString(pfPrintOrientation,
		OverideOrientation ? (DoubleColumn ? pfoDoubleColumn :
							  (nOrientation ? pfoLandscape : pfoPortrait)) :
							 NULL);
	MyWriteProfileBool(pfPrintNumbers,		LineNumbering);
	if (lf.lfHeight) {
		WriteFontAttributes(pfPrintFontAttr,   &lf);
		MyWriteProfileString(pfPrintFontFace,	    lf.lfFaceName);
		MyWriteProfileInt (pfPrintFontCharSet, Fmt, lf.lfCharSet);
		MyWriteProfileInt (pfPrintFontSize1,   Fmt, lf.lfHeight);
		MyWriteProfileInt (pfPrintFontSize2,   Fmt, cf.iPointSize);
	}
	MyWriteProfileBool(pfPrintFrame,  Frame);
	MyWriteProfileBool(pfPrintFrame2, Frame2);
	MyWriteProfileBool(pfPrintFrame3, Frame3);
	MyWriteProfileString(pfPrintHeader, Header ? Header : "");
	MyWriteProfileString(pfPrintFooter, Footer ? Footer : "");
}

VOID LoadPrintSettings(VOID)
{
	{	static BOOL SettingsLoaded;
		if (SettingsLoaded) return;
		SettingsLoaded = TRUE;
	}
	nPrtLeft   = (int)MyGetProfileLong(pfPrintBorderLeft,   nPrtLeft,   0);
	nPrtTop    = (int)MyGetProfileLong(pfPrintBorderTop,    nPrtTop,    0);
	nPrtRight  = (int)MyGetProfileLong(pfPrintBorderRight,  nPrtRight,  0);
	nPrtBottom = (int)MyGetProfileLong(pfPrintBorderBottom, nPrtBottom, 0);
	GetProfileString(ProfileSection, pfPrintOrientation, "",
					 PrintBuf, sizeof(PrintBuf));
	switch (*PrintBuf) {
		case 'd': case 'D': OverideOrientation = nOrientation = DoubleColumn =
								TRUE;
							break;
		case 'l': case 'L': OverideOrientation = nOrientation = TRUE;
							break;
		case 'p': case 'P': OverideOrientation = TRUE;
							break;
	}
	LineNumbering =		 MyGetProfileBool(pfPrintNumbers,	   LineNumbering);
	GetProfileString(ProfileSection, pfPrintFontFace, "",
					 lf.lfFaceName, sizeof(lf.lfFaceName));
	GetFontAttributes(pfPrintFontAttr, &lf);
	lf.lfCharSet  = (int)MyGetProfileLong(pfPrintFontCharSet, lf.lfCharSet,	 0);
	lf.lfHeight   = (int)MyGetProfileLong(pfPrintFontSize1,	  lf.lfHeight,	 0);
	cf.iPointSize = (int)MyGetProfileLong(pfPrintFontSize2,	  cf.iPointSize, 0);
	Frame		  =		 MyGetProfileBool(pfPrintFrame,		  Frame);
	Frame2		  =		 MyGetProfileBool(pfPrintFrame2,	  Frame2);
	Frame3		  =		 MyGetProfileBool(pfPrintFrame3,	  Frame3);
	GetProfileString(ProfileSection, pfPrintHeader, "&n",
					 PrintBuf, sizeof(PrintBuf));
	if (Header) _ffree(Header);
	if (!*PrintBuf) Header = NULL;
	else if ((Header = _fcalloc(lstrlen(PrintBuf) + 1, 1)) != NULL)
		lstrcpy(Header, PrintBuf);
	{	char Def[10];
		LOADSTRING(hInst, 247, Def, sizeof(Def));
		GetProfileString(ProfileSection, pfPrintFooter, Def,
						 PrintBuf, sizeof(PrintBuf));
	}
	if (Footer) _ffree(Footer);
	if (!*PrintBuf) Footer = NULL;
	else if ((Footer = _fcalloc(lstrlen(PrintBuf) + 1, 1)) != NULL)
		lstrcpy(Footer, PrintBuf);
}

/******** list of recently used files... ********/

#define NUMFILES 6

static int  Index[NUMFILES] = {-1,-1,-1,-1,-1,-1};
static char Buff[264];
static int  AlreadyInsertedLruFiles;

static void GetIndexList(void)
{	int  i;
	char *p;
	BOOL Free[NUMFILES];

	GetProfileString(ProfileSection, pfRecentFiles, "", Buff, sizeof(Buff));
	for (i=0; i<NUMFILES; ++i) Free[i] = TRUE;
	for (i=0, p=Buff; *p; ++p) {
		if (*p>='0' && *p<'0'+NUMFILES && Free[*p-'0']) {
			Index[i++] = *p - '0';
			Free[*p-'0'] = FALSE;
		}
	}
}

static void WriteIndexList(void)
{	char *p = Buff;
	int  i;

	for (i=0; i<NUMFILES; ++i) if (Index[i] != -1) {
		*p++ = ',';
		*p++ = Index[i] + '0';
	}
	*p = '\0';
	WriteProfileString(ProfileSection, pfRecentFiles, Buff+1);
}

void InitFileList(void)
{
	AlreadyInsertedLruFiles = 0;
}

void InsertFileList(HMENU hMenu)
{	int         i;
	char	    Key[16];

	GetIndexList();
	for (i=0; i<NUMFILES; ++i) {
		if (Index[i] == -1) break;
		wsprintf(Key, pfRecentFile, Index[i]);
		wsprintf(Buff, "&%c   ", i+'1');
		if (!GetProfileString(ProfileSection, Key, "", Buff+5, sizeof(Buff)-5))
			continue;
		if (lstrlen(Buff) > 32) {
			char *p = Buff + lstrlen(Buff) - 32;

			while (p>Buff && *p!='\\' && *p!='/') --p;
			if (p > Buff+14) {
				memmove(Buff+8, p, lstrlen(p)+1);
				Buff[5] = Buff[6] = Buff[7] = '.';
			}
		}
		{	PSTR p = strchr(Buff+5, '&');
			while (p!=NULL && lstrlen(Buff)<sizeof(Buff)-1) {
				memmove(p+1, p, lstrlen(p)+1);
				p = strchr(p+2, '&');
			}
		}
		if (i >= AlreadyInsertedLruFiles) {
			if (!AlreadyInsertedLruFiles) {
				/*insert another separator...*/
				InsertMenu(hMenu, 13, MF_BYPOSITION | MF_SEPARATOR, 0, NULL);
			}
			InsertMenu(hMenu, 13+i, MF_BYPOSITION | MF_STRING, 3000+i, Buff);
			++AlreadyInsertedLruFiles;
		} else {
			ModifyMenu(hMenu, 13+i, MF_BYPOSITION | MF_STRING, 3000+i, Buff);
		}
	}
}

void  AddToFileList(LPCSTR FileName)
{	int  i, j;
	char Key[16];

	if (!*FileName) return;
	GetIndexList();
	for (i=0; i<NUMFILES; ++i) {
		if (Index[i] != -1) {
			wsprintf(Key, pfRecentFile, Index[i]);
			GetProfileString(ProfileSection, Key, "", Buff, sizeof(Buff));
			if (lstrcmpi(Buff, FileName) == 0) {
				int  j, k = Index[i];
				BOOL Differs = lstrcmp(Buff, FileName) != 0;

				wsprintf(Key, pfRecentDir, Index[i]);
				if (GetProfileString(ProfileSection,Key,"",Buff,sizeof(Buff))) {
					char Dir[260];
					_getcwd(Dir, sizeof(Dir));
					if (lstrcmpi(Dir, Buff) != 0) {
						if ((FileName[1]!=':' ||
							 FileName[2]!='\\' && FileName[2]!='/') &&
							(FileName[0]!='\\' && FileName[0]!='/' ||
							 FileName[1]!=FileName[0]))
							continue;
					}
					if (lstrcmp(Dir, Buff) != 0)
						WriteProfileString(ProfileSection, Key, Dir);
				}
				if (Differs) {
					wsprintf(Key, pfRecentFile, Index[i]);
					GetProfileString(ProfileSection, Key,"", Buff,sizeof(Buff));
					WriteProfileString(ProfileSection, Key, FileName);
				}
				for (j=i; j>0; --j) Index[j] = Index[j-1];
				Index[0] = k;
				WriteIndexList();
				return;
			}
		}
	}
	{	BOOL Occupied[NUMFILES];

		for (i=0; i<NUMFILES; ++i) Occupied[i] = FALSE;
		for (i=0; i<NUMFILES; ++i) {
			if (Index[i] == -1) break;
			Occupied[Index[i]] = TRUE;
		}
		if (i == NUMFILES) j = Index[--i];
		else for (j=0; j<NUMFILES && Occupied[j]; ++j);
	}
	while (i > 0) {
		Index[i] = Index[i-1];
		--i;
	}
	Index[0] = j;
	WriteIndexList();
	wsprintf(Key, pfRecentFile, j);
	WriteProfileString(ProfileSection, Key, FileName);
	wsprintf(Key, pfRecentDir, j);
	_getcwd(Buff, sizeof(Buff));
	WriteProfileString(ProfileSection, Key, Buff);
}

void FileListCommand(int Command)
{
	if (Command>=3000 && Command<3000+NUMFILES && Index[Command-=3000]!=-1) {
		char Key[16];

		if (!AskForSave(hwndMain, 1)) return;
		wsprintf(Key, pfRecentDir, Index[Command]);
		lstrcpy(Buff, ":cd ");
		if (GetProfileString(ProfileSection, Key, "", Buff+4, sizeof(Buff)-4)) {
			NewStatus(0, Buff, NS_READING_WRITING);
			#if defined(WIN32)
				SetCurrentDirectory(Buff+4);
			#else
				if (Buff[4] && Buff[5]==':') _chdrive(Buff[4] & 31);
				_chdir(Buff+4);
			#endif
		}
		wsprintf(Key, pfRecentFile, Index[Command]);
		if (GetProfileString(ProfileSection, Key, "", Buff, sizeof(Buff))) {
			extern char CommandBuf[260];

			lstrcpyn(CommandBuf, Buff, sizeof(CommandBuf));
			NewStatus(0, "", NS_NORMAL);
			CommandWithExclamation = TRUE;	/*prevent another AskForSave()*/
			Open(CommandBuf, 0);
			CommandWithExclamation = FALSE;
		} else Error(412);
	}
}

/************* Language handling... *************/

char SysLang[6], ViLang[4];

int GetLanguage(BOOL SysDefault)
{	int  Language;
	char *p = SysLang;

	GetProfileString("intl",		 "sLanguage", "", SysLang, sizeof(SysLang));
	GetProfileString(ProfileSection, pfLanguage,  "", ViLang,  sizeof(ViLang));
	if (!SysDefault && *ViLang) p = ViLang;
	p[0] = LCASE(p[0]);
	p[1] = LCASE(p[1]);
	if		(p[0]=='d' && p[1]=='e') Language =  7000;
	else if (p[0]=='e' && p[1]=='s') Language = 10000;
	else if (p[0]=='f' && p[1]=='r') Language = 12000;
	else							 Language =  9000;
	return (Language);
}

VOID SaveLanguage(LPSTR pLang)
{
	WriteProfileString(ProfileSection, pfLanguage, pLang);
}

BOOL DebugMessages(VOID)
{
	return (MyGetProfileBool("Debug", FALSE));
}
