/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "winvi.h"
#include "page.h"
#include "resource.h"

#if !defined(DM_REPOSITION)
# define DM_REPOSITION (DM_SETDEFID+1)
#endif

#define ISHEX(x) (x>='0'&&x<='9' || (x|('a'-'A'))>='a' && (x|('a'-'A'))<='f')

#define W 255
#define INCLUDE(m, c, i) (m)[c][(i)>>3] |=  (1 << ((i) & 7))
#define EXCLUDE(m, c, i) (m)[c][(i)>>3] &= ~(1 << ((i) & 7))
#define TEST(m, c, i)   ((m)[c][(i)>>3] &   (1 << ((i) & 7)))

typedef BYTE MASK[2][33];

MASK NoneMask = {0}, AnyMask = {
{W-1,W-4-32,W,W-64,W,W,W,W,W,W,W,W,W,W,W,W, W,W,W,W,W,W,W,W,W,W,W,W,W,W,W,W, 0},
{W-1,W-4-32,W,W-64,W,W,W,W,W,W,W,W,W,W,W,W, W,W,W,W,W,W,W,W,W,W,W,W,W,W,W,W, 0},
};

typedef struct tagSEARCHPART {
	struct tagSEARCHPART *Next, *Prev;
	enum {SrchChoice, SrchString, SrchSpecial} Type;
	union {
		struct {
			MASK Mask;				/*Mask[0] for Ansi, Mask[1] for Oem*/
			BYTE Inverted;			/*choice built with [^...]*/
			BYTE MatchMulti;		/*appended '*' in pattern*/
		} ChoiceMatch;
		struct {
			WORD AllowConvert[4];	/*(sizeof(String)-1)/(8*sizeof(WORD))+1*/
			BYTE FirstUcase, FirstLcase;
			BYTE Length;
			BYTE String[56];
		} StringMatch;
		struct {
			char     Char;			/*one of: '(', '|', ')', '<', '>'*/
			POSITION Pos;			/*matched position of "\(" / "\)"*/
			struct tagSEARCHPART *NextBranch;
			   /*circular forward pointing list of substrings divided by "\|",*/
			   /*\) entry points back to \( entry, NULL if only one choice*/
		} SpecialMatch;
	} u;
} SEARCHPART, *PSEARCHPART;
PSEARCHPART FirstMatch, FirstString, LastMatch;
BOOL		MatchAtStartOfLine, MatchAtEndOfLine, MatchForward, MatchValid;
int			RecursiveBranch;

BOOL CheckFwd (PPOSITION, PSEARCHPART);
BOOL CheckBkwd(PPOSITION, PSEARCHPART);

int SingleCharAndAdvance(PPOSITION Pos) {
	int c;
	if ((c = CharAndAdvance(Pos)) == C_CRLF) {
		GoBack(Pos, 1);
		c = '\r';
	}
	return (c);
}

int GoBackAndSingleChar(PPOSITION Pos) {
	int c;
	if (!GoBack(Pos, 1)) c = C_EOF;
	else if ((c = CharAt(Pos)) == C_CRLF) c = '\r';
	return (c);
}

BOOL MatchSpecial(PSEARCHPART *pSrch, PPOSITION Pos, BOOL Backward) {
	int c;
	PSEARCHPART Srch = *pSrch;

	switch (Srch->u.SpecialMatch.Char) {
		case '<':
			if (!(CharFlags[c = CharAt(Pos)]		& 8)) return (FALSE);
			if (  CharFlags[c = GoBackAndChar(Pos)]	& 8 ) return (FALSE);
			if (c != C_EOF) CharAndAdvance(Pos);
			break;

		case '>':
			if (  CharFlags[c = CharAt(Pos)]		& 8 ) return (FALSE);
			if (!(CharFlags[c = GoBackAndChar(Pos)]	& 8)) return (FALSE);
			CharAndAdvance(Pos);
			break;

		case '|':
			while (Srch->u.SpecialMatch.NextBranch != NULL &&
				   Srch->u.SpecialMatch.Char != (Backward ? '(' : ')'))
				Srch = Srch->u.SpecialMatch.NextBranch;
			*pSrch = Srch;
			/*FALLTHROUGH*/
		case '(':
		case ')':
			Srch->u.SpecialMatch.Pos = *Pos;
			if (Srch->u.SpecialMatch.NextBranch != NULL) {
				POSITION PosSav;

				PosSav = *Pos;
				if (Backward) {
					if (Srch->u.SpecialMatch.Char == '(') break;
					Srch = Srch->u.SpecialMatch.NextBranch;
					assert (Srch != NULL);
					assert (Srch->u.SpecialMatch.Char == '(');
					Srch = Srch->u.SpecialMatch.NextBranch;
					assert (Srch != NULL);
					assert (Srch->u.SpecialMatch.Char == '|');
					++RecursiveBranch;
					for (;;) {
						if (CheckBkwd(Pos, Srch)) break;
						Srch = Srch->u.SpecialMatch.NextBranch;
						assert (Srch != NULL);
						if (Srch->u.SpecialMatch.Char == '(') {
							--RecursiveBranch;
							return (FALSE);
						}
						*Pos = PosSav;
					}
				} else {
					if (Srch->u.SpecialMatch.Char == ')') break;
					++RecursiveBranch;
					for (;;) {
						if (CheckFwd(Pos, Srch->Next)) break;
						Srch = Srch->u.SpecialMatch.NextBranch;
						assert (Srch != NULL);
						if (Srch->u.SpecialMatch.Char == ')') {
							--RecursiveBranch;
							return (FALSE);
						}
						*Pos = PosSav;
					}
				}
				--RecursiveBranch;
				while (Srch->u.SpecialMatch.Char != ')')
					Srch = Srch->u.SpecialMatch.NextBranch;
				if (Backward)
					Srch = Srch->u.SpecialMatch.NextBranch;
				*pSrch = Srch;
			}
			break;
	}
	return (TRUE);
}

BYTE TestBit(PSEARCHPART pS, int CharSet, int Char) {
	BYTE		Result;
	static BYTE c[2];

	if ((Result = TEST(pS->u.ChoiceMatch.Mask, CharSet, Char) != 0)
			  != pS->u.ChoiceMatch.Inverted || !IgnoreCaseFlag || Char > 255)
		return (Result);
	c[0] = Char;
	if (CharSet) OemToAnsiBuff(c, c, 1);
	if (IsCharLower(c[0])) AnsiUpperBuff(c, 1);
	else				   AnsiLowerBuff(c, 1);
	return (TEST(pS->u.ChoiceMatch.Mask, 0, c[0]));
}

BOOL CheckFwd(PPOSITION Pos, PSEARCHPART Srch) {
	while (Srch) {
		switch (Srch->Type) {
			case SrchString:
				{	PBYTE pCmp = Srch->u.StringMatch.String;
					int   Len  = Srch->u.StringMatch.Length;
					int   i	   = 0;
					BYTE  ch;

					while (Len--) {
						int c = SingleCharAndAdvance(Pos);
						if (c > 0377) return (FALSE);
						ch = (BYTE)c;
						if (CharSet && Srch->u.StringMatch.AllowConvert[i>>4] &
									   (1 << (i & 15)))
							OemToAnsiBuff(&ch, &ch, 1);
						if (ch != *pCmp) {
							if (!(Srch->u.StringMatch.AllowConvert[i>>4] &
									   (1 << (i & 15))) ||
									!IgnoreCaseFlag ||
									#if defined(WIN32)
									   AnsiLower((LPTSTR)(DWORD)ch) !=
									   AnsiLower((LPTSTR)(DWORD)*pCmp))
									#else
									   (char)(DWORD)AnsiLower(MAKELP(0, ch)) !=
									   (char)(DWORD)AnsiLower(MAKELP(0, *pCmp)))
									#endif
								return (FALSE);
						}
						++i;
						++pCmp;
					}
				}
				break;

			case SrchChoice:
				if (Srch->u.ChoiceMatch.MatchMulti) {
					int i = 0, c;

					if ((c = CharAt(Pos)) == C_CRLF) c = '\r';
					while (TestBit(Srch, CharSet, c)) {
						++i;
						Advance(Pos, 1);
						if ((c = CharAt(Pos)) == C_CRLF) c = '\r';
					}
					do {
						POSITION SavePos;
						SavePos = *Pos;
						if (!Srch->Next || CheckFwd(Pos, Srch->Next))
							return (!MatchAtEndOfLine ||
									CharFlags[CharAt(Pos)] & 0x21);
						*Pos = SavePos;
						GoBack(Pos, 1);
					} while (i--);
					return (FALSE);
				} else {
					int c = SingleCharAndAdvance(Pos);

					if (!TestBit(Srch, CharSet, c))
						return (FALSE);
				}
				break;

			case SrchSpecial:
				{	BOOL RecursivelyCalled = Srch->u.SpecialMatch.Char!='(';
					if (!MatchSpecial(&Srch, Pos, FALSE)) return (FALSE);
					if (RecursiveBranch && Srch->u.SpecialMatch.NextBranch)
						if (RecursivelyCalled) return (TRUE);
				}
				break;
		}
		Srch = Srch->Next;
	}
	return (!MatchAtEndOfLine || (CharFlags[CharAt(Pos)] & 0x21) == 1);
}

BOOL MinimumMatchBkwd;

BOOL CheckBkwd(PPOSITION Pos, PSEARCHPART Srch) {
	/*Current Srch need not be checked, check preceeding chain elements only.
	 */
	if (Srch) while ((Srch = Srch->Prev) != NULL) {
		int c;
		switch (Srch->Type) {
			case SrchString:
				{	PBYTE pCmp = Srch->u.StringMatch.String;
					int   Len  = Srch->u.StringMatch.Length;
					BYTE  ch;

					pCmp += Len-1;
					while (Len--) {
						int c = GoBackAndSingleChar(Pos);

						if (c > 0377) return (FALSE);
						ch = (BYTE)c;
						if (CharSet && Srch->u.StringMatch.AllowConvert[Len>>4]&
									   (1 << (Len & 15)))
							OemToAnsiBuff(&ch, &ch, 1);
						if (ch != *pCmp) {
							if (!(Srch->u.StringMatch.AllowConvert[Len>>4] &
									   (1 << (Len & 15))) ||
									!IgnoreCaseFlag ||
									#if defined(WIN32)
									   AnsiLower((LPTSTR)(DWORD)ch) !=
									   AnsiLower((LPTSTR)(DWORD)*pCmp))
									#else
									   (char)(DWORD)AnsiLower(MAKELP(0, ch)) !=
									   (char)(DWORD)AnsiLower(MAKELP(0, *pCmp)))
									#endif
								return (FALSE);
						}
						--pCmp;
					}
				}
				break;

			case SrchChoice:
				if (Srch->u.ChoiceMatch.MatchMulti) {
					int		 i = 0;
					POSITION SavePos;

					if (MinimumMatchBkwd) {
						for (;;) {
							SavePos = *Pos;
							if (CheckBkwd(Pos, Srch)) return (TRUE);
							*Pos = SavePos;
							c = GoBackAndSingleChar(Pos);
							if (!TestBit(Srch, CharSet, c))
								return (FALSE);
						}
					}
					for (;;) {
						c = GoBackAndSingleChar(Pos);
						if (!TestBit(Srch, CharSet, c))
							break;
						++i;
					}
					do {
						if (c == C_EOF) c = 0;
						else Advance(Pos, 1);
						SavePos = *Pos;
						if (CheckBkwd(Pos, Srch)) return (TRUE);
						*Pos = SavePos;
					} while (i--);
					return (FALSE);
				} else {
					c = GoBackAndSingleChar(Pos);
					if (!TestBit(Srch, CharSet, c))
						return (FALSE);
				}
				break;

			case SrchSpecial:
				{	BOOL RecursivelyCalled = Srch->u.SpecialMatch.Char!=')';
					if (!MatchSpecial(&Srch, Pos, TRUE)) return (FALSE);
					if (RecursiveBranch && Srch->u.SpecialMatch.NextBranch)
						if (RecursivelyCalled) return (TRUE);
				}
				break;
		}
	}
	if (MatchAtStartOfLine) {
		int c;
		if (Pos->i && Pos->p->PageBuf) {
			c = (BYTE)Pos->p->PageBuf[Pos->i-1];
			if (c=='\r' && CharAt(Pos)=='\n' ||
				Pos->i==Pos->p->Fill && CharAt(Pos)==C_EOF) return (FALSE);
		} else {
			if (GoBack(Pos, 1) != 1) return (TRUE);
			c = CharAt(Pos);
			Advance(Pos, 1);
			if (c==C_CRLF /*position inbetween*/ || CharAt(Pos)==C_EOF)
				return (FALSE);
		}
		if (!(CharFlags[c] & 0x21)) return (FALSE);
	}
	return (TRUE);
}

LONG StartTime;

BOOL CheckInterrupt(void) {
	LONG Time;
	Time = GetCurrentTime();
	if (Disabled) MessageLoop(FALSE);
	else if (Time-StartTime > 200) Disable(TRUE);
	return (Interrupted);
}

VOID StartInterruptCheck(void) {
	Interrupted = FALSE;
	StartTime = GetCurrentTime();
}

int		 OffsetTo1stString;
BOOL	 RescanMaximumMatch, Wrapped;
POSITION EndMatch;

BOOL SearchIt(PPOSITION pPos, int Flags) {
	/*pPos:  on input, it specifies the start position for search
	 *		 on return, it marks the start of the area found
	 *Flags: 1=forward
	 *		 2=don't advance position before search
	 *		 4=don't pre-advance 2 positions at end of line (insert mode)
	 */
	POSITION Pos;

	Wrapped = FALSE;
	if (CheckInterrupt()) return (FALSE);
	Pos = *pPos;
	MinimumMatchBkwd = FALSE;
	{	int c;
		BOOL Done = FALSE;
		int (*Next)(PPOSITION) =
			Flags & 1 ? SingleCharAndAdvance : GoBackAndSingleChar;
		if (!(Flags & 2)) {
			if (CharFlags[(c = (*Next)(&Pos))] & 0x21) {
				if (c > C_CRLF || FirstString) {
					if (!(Flags & 1) && c != C_EOF) Advance(&Pos, 1);
					Done = TRUE;
				}
				if (c=='\r' && CharAt(&Pos)=='\n') CharAndAdvance(&Pos);
			} else if ((Flags & 5) == 1 && CharFlags[CharAt(&Pos)] & 1)
				CharAndAdvance(&Pos);
		}
		if (!Done) for (;;) {
			EndMatch = Pos;
			if (CheckFwd(&EndMatch,FirstMatch) && CheckBkwd(&Pos,FirstMatch)) {
				*pPos = Pos;
				return (TRUE);
			}
			if (Done) break;
			if (CharFlags[c = (*Next)(&Pos)] & 0x21) {
				if (CheckInterrupt()) return (FALSE);
				if (c > C_CRLF || FirstString) Done = TRUE;
			}
		}
		if (c > C_CRLF) {
			if (c != C_EOF || !WrapScanFlag) return (FALSE);
			Wrapped = TRUE;
			if (Flags & 1) {
				Pos.p = FirstPage;
				Pos.i = 0;
			} else {
				Pos.p = LastPage;
				Pos.i = LastPage->Fill;
			}
			CharAt(&Pos);
			if (!FirstString) for (;;) {
				if (Pos.p == CurrPos.p && Pos.i == CurrPos.i) return (FALSE);
				EndMatch = Pos;
				if (CheckFwd(&EndMatch, FirstMatch) &&
						CheckBkwd(&Pos, FirstMatch)) {
					*pPos = Pos;
					return (TRUE);
				}
				if (CharFlags[c = (*Next)(&Pos)] & 0x21)
					if (CheckInterrupt()) return (FALSE);
			}
		}
	}
	{	BYTE Ucase = FirstString->u.StringMatch.FirstUcase;
		BYTE Lcase = FirstString->u.StringMatch.FirstLcase;
		CharAt(&Pos);	/*page must be loaded*/
		if (Flags & 1) {
			/*search forward*/
			LPSTR lpUc, lpLc;
			for (;;) {
				if (Pos.i >= Pos.p->Fill) {
					Pos.i = 0;
					if ((Pos.p = Pos.p->Next) == NULL) {
						if (!WrapScanFlag) return (FALSE);
						Pos.p = FirstPage;
						Wrapped = TRUE;
					}
					if (CheckInterrupt()) return (FALSE);
					if (!Pos.p->PageBuf && !ReloadPage(Pos.p)) return (FALSE);
				}
				lpUc = _fmemchr(Pos.p->PageBuf + Pos.i,
								Ucase, Pos.p->Fill - Pos.i);
				if (Lcase == Ucase) lpLc = NULL;
				else lpLc = _fmemchr(Pos.p->PageBuf + Pos.i,
									 Lcase, Pos.p->Fill - Pos.i);
				if (lpUc || lpLc) {
					do {
						LPSTR lp;
						if (lpUc && lpLc) lp = lpUc < lpLc ? lpUc : lpLc;
						else lp = lpUc ? lpUc : lpLc;
						Pos.i = lp - Pos.p->PageBuf;
						if (Wrapped && Pos.i >= CurrPos.i && Pos.p == CurrPos.p)
							return (FALSE);
						EndMatch = Pos;
						if (CheckFwd(&EndMatch, FirstString)) {
							POSITION ChkBack;
							ChkBack = Pos;
							if (CheckBkwd(&ChkBack, FirstString)) {
								if (Wrapped==(ComparePos(&EndMatch, pPos)<=0)) {
									*pPos = ChkBack;
									if (!RescanMaximumMatch) return (TRUE);
									EndMatch = ChkBack;
									return (CheckFwd(&EndMatch, FirstMatch));
								}
							}
						}
						if (++Pos.i >= Pos.p->Fill)
							if (lp == lpUc) lpUc = NULL;
							else lpLc = NULL;
						else if (lp == lpUc)
							lpUc = _fmemchr(Pos.p->PageBuf + Pos.i,
											Ucase,
											Pos.p->Fill - Pos.i);
						else lpLc = _fmemchr(Pos.p->PageBuf + Pos.i,
											 Lcase,
											 Pos.p->Fill - Pos.i);
					} while (lpUc || lpLc);
					++Pos.i;
				} else {
					if (Pos.p == CurrPos.p && Wrapped) return (FALSE);
					Pos.i = Pos.p->Fill;
				}
			}
		} else {
			/*search backward*/
			LPSTR lp;
			MinimumMatchBkwd = TRUE;
			for (;;) {
				BYTE c;
				while (!Pos.i) {
					if ((Pos.p = Pos.p->Prev) == NULL) {
						if (!WrapScanFlag) return (FALSE);
						Pos.p = LastPage;
						Wrapped = TRUE;
					}
					Pos.i = Pos.p->Fill;
					if (CheckInterrupt()) return (FALSE);
				}
				if (!Pos.p->PageBuf && !ReloadPage(Pos.p)) return (FALSE);
				lp = NULL;
				do {
					c = Pos.p->PageBuf[--Pos.i];
					if (c == Ucase || c == Lcase) {
						lp = Pos.p->PageBuf + Pos.i;
						break;
					}
				} while (Pos.i);
				if (lp) {
					Pos.i = lp - Pos.p->PageBuf;
					if (Wrapped && Pos.i <= CurrPos.i && Pos.p == CurrPos.p)
						return (FALSE);
					EndMatch = Pos;
					if (CheckFwd(&EndMatch, FirstString)) {
						POSITION ChkBack;
						ChkBack = Pos;
						if (CheckBkwd(&ChkBack, FirstString)) {
							if (Wrapped == (ComparePos(&ChkBack, pPos)>=0)) {
								*pPos = ChkBack;
								if (!RescanMaximumMatch) return (TRUE);
								EndMatch = ChkBack;
								return (CheckFwd(&EndMatch, FirstMatch));
							}
						}
					}
				} else if (Pos.p == CurrPos.p && Wrapped) return (FALSE);
			}
		}
	}
}

int SearchErrNo = 216;

BOOL StartSearch(void) {
	if (!MatchValid) {
		if (SearchErrNo == (TerseFlag ? 216 : 217)) SearchErrNo ^= 216^217;
		Error(SearchErrNo);
		return (FALSE);
	}
	StartInterruptCheck();
	return (TRUE);
}

static char		SrchDispBuf[100];
static POSITION	StartPos;

BOOL SearchAndStatus(PPOSITION Pos, WORD Flags) {
	/*Flags: 1=Reverse search
	 *		 2=don't advance position before search
	 */
	BOOL Return, Forward = MatchForward ^ (Flags & 1);

	if (!StartSearch()) return (FALSE);
	if (*SrchDispBuf != (Forward ? '/' : '?')) {
		int  c, len;
		PSTR p = SrchDispBuf;

		c = *SrchDispBuf ^= '/' ^ '?';
		for (;;) {
			do if (*p++=='\\' && *p) ++p;
			while (*p && *p!=c);
			if (!*p) break;
			len = lstrlen(p) + 1;
			if (len >= (int)sizeof(SrchDispBuf)-(p-SrchDispBuf)) {
				--len;
				lstrcpy(SrchDispBuf + sizeof(SrchDispBuf) - 5, "...");
			}
			_fmemmove(p+1, p, len);
			*p = '\\';
		}
	}
	NewStatus(0, SrchDispBuf, NS_NORMAL);
	Flags &= 2;
	if (Forward) Flags |= Mode==InsertMode ? 5 : 1;
	if (!SearchIt(Pos, Flags)) {
		Error(Interrupted  ? 218 :
			  WrapScanFlag ? 214 :
			  Forward	   ? 210 :
							 212);
		if (Disabled) Enable();
		ShowEditCaret();
		Return = FALSE;
	} else {
		extern BOOL SuppressStatusBeep;
		BOOL OldSupp = SuppressStatusBeep;

		SuppressStatusBeep = TRUE;
		if (Wrapped) Error(Forward ? 258 : 259);
		SuppressStatusBeep = OldSupp;
		if (Disabled) Enable();
		StartPos = *Pos;
		Return = TRUE;
	}
	EnableToolButton(IDB_SRCHAGN, TRUE);
	FindValidPosition(Pos, (WORD)((Mode==InsertMode) | 4));
	return (Return);
}

PSEARCHPART NewSearchPart(void)
{	PSEARCHPART p;

	if ((p = malloc(sizeof(SEARCHPART))) == NULL) {
		MatchValid = FALSE;
		return (NULL);
	}
	if (LastMatch) {
		LastMatch->Next = p;
		p->Prev = LastMatch;
		LastMatch = p;
	} else {
		p->Prev = NULL;
		FirstMatch = LastMatch = p;
	}
	p->Next = NULL;
	return (p);
}

void NewChoice(MASK Mask)
{	PSEARCHPART p;

	if (!MatchValid) return;
	if ((p = NewSearchPart()) == NULL) return;
	p->Type = SrchChoice;
	memcpy(p->u.ChoiceMatch.Mask, Mask, sizeof(p->u.ChoiceMatch.Mask));
	p->u.ChoiceMatch.MatchMulti = p->u.ChoiceMatch.Inverted = 0;
}

BOOL MagicFlag = TRUE;
int  Level;

void EnterMatch(BYTE Char, BOOL MayConvert, PSTR Asterisk)
{
	if (!MatchValid) return;
	if (MagicFlag ? Asterisk[1]=='*'
				  : Asterisk[1]=='\\' && Asterisk[2]=='*') {
		NewChoice(NoneMask);
		INCLUDE(LastMatch->u.ChoiceMatch.Mask, 0, Char);
		if (MayConvert) AnsiToOemBuff(&Char, &Char, 1);
		INCLUDE(LastMatch->u.ChoiceMatch.Mask, 1, Char);
		LastMatch->u.ChoiceMatch.MatchMulti = 1;
		return;
	}
	if (!LastMatch || LastMatch->Type != SrchString ||
					  LastMatch->u.StringMatch.Length >=
					  sizeof(LastMatch->u.StringMatch.String)-1) {
		PSEARCHPART p;

		if ((p = NewSearchPart()) == NULL) return;
		p->Type = SrchString;
		p->u.StringMatch.Length = 0;
		{	WORD *pc = p->u.StringMatch.AllowConvert;
			*pc = 0; *++pc = 0; *++pc = 0; *++pc = 0;
		}
		{	BYTE ch[2];
			ch[0] = ch[1] = Char;
			if (MayConvert) {
				AnsiUpperBuff(ch+0, 1);
				AnsiLowerBuff(ch+1, 1);
				if (CharSet) AnsiToOemBuff(ch, ch, 2);
			}
			p->u.StringMatch.FirstUcase = ch[0];
			p->u.StringMatch.FirstLcase = ch[1];
		}
		if (!FirstString && !Level) {
			FirstString = p;
			while ((p = p->Prev) != NULL)
				if (!p->u.ChoiceMatch.MatchMulti) ++OffsetTo1stString;
				else RescanMaximumMatch = TRUE;
		}
	}
	if (MayConvert)
		LastMatch->u.StringMatch.AllowConvert
				 [LastMatch->u.StringMatch.Length >> 4] |=
			1 << (LastMatch->u.StringMatch.Length & 15);
	LastMatch->u.StringMatch.String[LastMatch->u.StringMatch.Length++] = Char;
	LastMatch->u.StringMatch.String[LastMatch->u.StringMatch.Length] = '\0';
}

static PSTR pDisp;

void SetSrchDisp(int c)
{
	if (pDisp < SrchDispBuf + sizeof(SrchDispBuf) - 4) *pDisp++ = c;
	else if (pDisp == SrchDispBuf + sizeof(SrchDispBuf) - 4) {
		lstrcpy(pDisp, "...");
		pDisp += 3;
	}
}

PSTR EnterChoice(PSTR Cmd)
{	BOOL Inverse;

	Inverse = *++Cmd=='^';
	if (Inverse) {
		NewChoice(AnyMask);
		SetSrchDisp(*Cmd);
		++Cmd;
		LastMatch->u.ChoiceMatch.Inverted = 1;
	} else NewChoice(NoneMask);
	for (;;) {
		BYTE cw, cd;
		MASK *m;

		SetSrchDisp(cw = cd = (BYTE)*Cmd);
		AnsiToOemBuff(&cd, &cd, 1);
		switch (cw) {
			case '\0':
				return (Cmd);

			case '\\':
				SetSrchDisp(cw = (BYTE)*++Cmd);
				if (!cw) return (Cmd);
				if (cw>='0' && cw<'8') {
					cw -= '0';
					if (Cmd[1]>='0' && Cmd[1]<'8') {
						cw = (cw << 3) + *++Cmd - '0';
						SetSrchDisp(*Cmd);
					}
					if (cw<040 && Cmd[1]>='0' && Cmd[1]<'8') {
						cw = (cw << 3) + *++Cmd - '0';
						SetSrchDisp(*Cmd);
					}
				} else if (cw == '%') {
					cw = Cmd[1];
					if (ISHEX(cw)) {
						++Cmd;
						SetSrchDisp(cw);
						if (cw >= 'A') cw += (BYTE)(10-'A');
						cw &= 15;
						if (ISHEX(Cmd[1])) {
							cw <<= 4;
							cw  += *++Cmd & 15;
							if (*Cmd >= 'A') cw += (10-'A') & 15;
							SetSrchDisp(*Cmd);
						}
					} else cw = *Cmd;
				} else if (!MagicFlag && cw==']') return (Cmd);
				cd = cw;
				/*FALLTHROUGH*/
			case ']':
				if (MagicFlag && cw==']') return (Cmd);
			default:
				m = &LastMatch->u.ChoiceMatch.Mask;
				if (Inverse) {
					EXCLUDE(*m, 0, cw);
					EXCLUDE(*m, 1, cd);
				} else {
					INCLUDE(*m, 0, cw);
					INCLUDE(*m, 1, cd);
				}
				{	static BOOL WithinRange;
					static BYTE WinRngStart, DosRngStart;

					if (!WithinRange && Cmd[1]=='-' && Cmd[2] && Cmd[2]!=']') {
						WithinRange = TRUE;
						WinRngStart = cw;
						DosRngStart = cd;
						SetSrchDisp(*++Cmd);
					} else if (WithinRange) {
						WithinRange = FALSE;
						if (Inverse) {
							while (cw > WinRngStart) {--cw; EXCLUDE(*m, 0, cw);}
							while (cd > DosRngStart) {--cd; EXCLUDE(*m, 1, cd);}
						} else {
							while (cw > WinRngStart) {--cw; INCLUDE(*m, 0, cw);}
							while (cd > DosRngStart) {--cd; INCLUDE(*m, 1, cd);}
						}
					}
				}
		}
		++Cmd;
	}
}

PSTR BuildMatchList(PSTR Cmd, int Delimiter, int *ErrNo)
{	PSTR		Start = Cmd;
	BOOL		ForcedMagic = FALSE;
	PSEARCHPART	FirstBranch[10], LastBranch[10];

	Level = 0;
	FirstBranch[0] = LastBranch[0] = NULL;
	pDisp = SrchDispBuf;
	MatchForward = (*pDisp++ = *Cmd++) == '/';
	if (*Cmd && *Cmd != Delimiter) {
		BOOL Done = FALSE;
		while (FirstMatch) {
			LastMatch = FirstMatch->Next;
			free(FirstMatch);
			FirstMatch = LastMatch;
		}
		FirstString = NULL;
		OffsetTo1stString = 0;
		RescanMaximumMatch = MatchAtStartOfLine = MatchAtEndOfLine = FALSE;
		if (*Cmd == '^') {
			MatchAtStartOfLine = TRUE;
			SetSrchDisp(*Cmd++);
		}
		MatchValid = TRUE;
		for (;;) {
			if (*Cmd == Delimiter) {
				++Cmd;
				break;
			}
			if (!ForcedMagic) SetSrchDisp(*Cmd);
			switch (*Cmd) {
				case '\0':
					Done = TRUE;
					break;
				case '.':
					if (MagicFlag || ForcedMagic) {
						ForcedMagic = FALSE;
						NewChoice(AnyMask);
						if (MagicFlag ? Cmd[1]=='*' :
										Cmd[1]=='\\' && Cmd[2]=='*')
							LastMatch->u.ChoiceMatch.MatchMulti = 1;
					} else EnterMatch(*Cmd, FALSE, Cmd);
					break;
				case '[':
					if (MagicFlag || ForcedMagic) {
						ForcedMagic = FALSE;
						if (!*(Cmd = EnterChoice(Cmd))) Done = TRUE;
						else if (MagicFlag ? Cmd[1]=='*' :
											 Cmd[1]=='\\' && Cmd[2]=='*')
							LastMatch->u.ChoiceMatch.MatchMulti = 1;
					} else EnterMatch(*Cmd, FALSE, Cmd);
					break;
				case '/':
				case '?':
					if (*Cmd==*Start && pDisp[-1]==*Cmd) {
						pDisp[-1] = '\\';
						SetSrchDisp(*Cmd);
					}
					EnterMatch(*Cmd, FALSE, Cmd);
					break;
				case '$':
					if (!Cmd[1] || Cmd[1]==Delimiter) {
						MatchAtEndOfLine = TRUE;
						break;
					}
					EnterMatch(*Cmd, FALSE, Cmd);
					break;
				case '\\':
					SetSrchDisp(*++Cmd);
					switch (*Cmd) {
						BYTE c;

						case '(':
							if (Level >= 9) {
								if (ErrNo) *ErrNo = 251;
								SearchErrNo = 251;
								MatchValid = FALSE;
								Done = TRUE;
								break;
							}
							++Level;
							FirstBranch[Level] = LastBranch[Level] = NULL;
							/*FALLTHROUGH*/
						case ')': case '|':
						case '<': case '>':
							if (!NewSearchPart()) break;
							LastMatch->Type = SrchSpecial;
							LastMatch->u.SpecialMatch.Char  = *Cmd;
							LastMatch->u.SpecialMatch.Pos.p = NULL;
							LastMatch->u.SpecialMatch.NextBranch = NULL;
							if (*Cmd=='|') {
								if (!Level) {
									if (ErrNo) *ErrNo = 252;
									SearchErrNo = 252;
									MatchValid = FALSE;
									Done = TRUE;
									break;
								}
								if (LastBranch[Level]!=NULL) {
									LastBranch[Level]->u.SpecialMatch.NextBranch
										= LastMatch;
									LastBranch[Level] = LastMatch;
								}
							} else if (*Cmd == '(') {
								FirstBranch[Level] = LastBranch[Level] =
									LastMatch;
							} else if (*Cmd == ')') {
								if (LastBranch[Level] != NULL &&
									LastBranch[Level] != FirstBranch[Level]) {
										LastBranch[Level]->u.SpecialMatch.
														 NextBranch = LastMatch;
										LastMatch->u.SpecialMatch.NextBranch =
											FirstBranch[Level];
								}
								if (!Level) {
									if (ErrNo) *ErrNo = 250;
									SearchErrNo = 250;
									MatchValid = FALSE;
									Done = TRUE;
									break;
								}
								if (Level) --Level;
							}
							break;

						case '0': case '4':
						case '1': case '5':
						case '2': case '6':
						case '3': case '7':
							c = *Cmd - '0';
							if (Cmd[1] >= '0' && Cmd[1] < '8') {
								c = (c << 3) + *++Cmd - '0';
								SetSrchDisp(*Cmd);
							}
							if (c < 040 && Cmd[1] >= '0' && Cmd[1] < '8') {
								c = (c << 3) + *++Cmd - '0';
								SetSrchDisp(*Cmd);
							}
							EnterMatch(c, FALSE, Cmd);
							break;

						case '%':	/*WinVi extension: hex*/
							c = Cmd[1];
							if (ISHEX(c)) {
								++Cmd;
								SetSrchDisp(c);
								if (c >= 'A') c += (BYTE)(10-'A');
								c &= 15;
								if (ISHEX(Cmd[1])) {
									c <<= 4;
									c  += *++Cmd & 15;
									if (*Cmd >= 'A') c += (10-'A') & 15;
									SetSrchDisp(*Cmd);
								}
							} else c = *Cmd;
							EnterMatch(c, FALSE, Cmd);
							break;

						case '\0':
							EnterMatch(*--Cmd, FALSE, "-");
							Done = TRUE;
							break;

						case '.': case '[':
							if (!MagicFlag) {
								ForcedMagic = TRUE;
								continue;
							}
							/*FALLTHROUGH*/
						default:
							EnterMatch(*Cmd, FALSE, Cmd);
					}
					break;
				default:
					EnterMatch(*Cmd, TRUE, Cmd);
			}
			if (Done) break;
			if (MagicFlag) {
				if (*++Cmd == '*') SetSrchDisp(*Cmd++);
			} else if (*++Cmd == '\\' && Cmd[1] == '*') {
				SetSrchDisp(*Cmd++);
				SetSrchDisp(*Cmd++);
			}
		}
		*pDisp = '\0';
	} else {
		if (*Cmd) ++Cmd;
		if (!MatchValid) {
			SearchAndStatus(&CurrPos, 0);	/*error display only*/
			return (Cmd);
		}
	}
	if (MatchValid && Level) {
		if (ErrNo) *ErrNo = 249;
		SearchErrNo = 249;
		MatchValid = FALSE;
	}
	if (!MatchValid) MessageBeep(MB_ICONEXCLAMATION);
	return (Cmd);
}

typedef struct tagSEARCH_DESCR {
	BOOL MatchValid, MatchForward;
	BOOL RescanMaximumMatch, MatchAtStartOfLine, MatchAtEndOfLine;
	PSEARCHPART FirstMatch, FirstString, LastMatch;
	char DisplayBuffer[1];
} SEARCH_DESCR;
SEARCH_DESCR FAR *SavedSearch;

BOOL SaveMatchList(VOID)
{	int Size = sizeof(SEARCH_DESCR) + lstrlen(SrchDispBuf);

	if ((SavedSearch = _fcalloc(Size, 1)) == NULL) return (FALSE);
	lstrcpy(SavedSearch->DisplayBuffer, SrchDispBuf);
	SavedSearch->FirstMatch			= FirstMatch;
	SavedSearch->FirstString		= FirstString;
	SavedSearch->LastMatch			= LastMatch;
	SavedSearch->RescanMaximumMatch	= RescanMaximumMatch;
	SavedSearch->MatchAtStartOfLine	= MatchAtStartOfLine;
	SavedSearch->MatchAtEndOfLine	= MatchAtEndOfLine;
	SavedSearch->MatchForward		= MatchForward;
	SavedSearch->MatchValid			= MatchValid;
	FirstMatch = FirstString = LastMatch = NULL;
	*SrchDispBuf = '\0';
	return (TRUE);
}

BOOL RestoreMatchList(VOID)
{
	if (SavedSearch == NULL) return (FALSE);
	while (FirstMatch) {
		LastMatch = FirstMatch->Next;
		free(FirstMatch);
		FirstMatch = LastMatch;
	}
	lstrcpy(SrchDispBuf,  SavedSearch->DisplayBuffer);
	FirstMatch			= SavedSearch->FirstMatch;
	FirstString			= SavedSearch->FirstString;
	LastMatch			= SavedSearch->LastMatch;
	RescanMaximumMatch	= SavedSearch->RescanMaximumMatch;
	MatchAtStartOfLine	= SavedSearch->MatchAtStartOfLine;
	MatchAtEndOfLine	= SavedSearch->MatchAtEndOfLine;
	MatchForward		= SavedSearch->MatchForward;
	MatchValid			= SavedSearch->MatchValid;
	_ffree(SavedSearch);
	SavedSearch = NULL;
	return (TRUE);
}

typedef struct tagREPLIST {
	struct tagREPLIST FAR *Next;
	BYTE Buf[80];
} REPLIST, FAR *LPREPLIST;
REPLIST   RepList;
LPREPLIST lpLastRep = &RepList;
int       nRepListCount = 0;

VOID FreeRepList(VOID)
{	LPREPLIST ToDel;

	while ((ToDel = RepList.Next) != NULL) {
		RepList.Next = ToDel->Next;
		_ffree(ToDel);
	}
	lpLastRep = &RepList;
	nRepListCount = 0;
}

LPBYTE ExpandRepList(LPBYTE Prev)
{
	if (Prev==NULL || ++Prev < lpLastRep->Buf + sizeof(lpLastRep->Buf))
		return (Prev);
	if ((lpLastRep->Next = _fcalloc(sizeof(REPLIST), 1)) == NULL)
		return (NULL);
	++nRepListCount;
	lpLastRep = lpLastRep->Next;
	return (lpLastRep->Buf);
}

PSTR pLastReplace;
int	 nParenPairs;

BOOL CheckParenPairs(VOID)
{	PSEARCHPART	pSp;
	int			Level;

	Level = nParenPairs = 0;
	for (pSp=FirstMatch; pSp; pSp=pSp->Next) {
		if (pSp->Type == SrchSpecial) {
			switch (pSp->u.SpecialMatch.Char) {
				case '(': ++nParenPairs;
						  ++Level;
						  continue;
				case ')': if (--Level < 0) break;
						  /*FALLTHROUGH*/
				default:  continue;
			}
			break;
		}
	}
	if (Level) {
		Error(231);
		return (FALSE);
	}
	return (TRUE);
}

BOOL BuildReplaceString
	 (PBYTE Source, LONG *DestLen, PPOSITION Pos, LONG Len, WORD Flags)
	/*Flags: 1=hex input flag checked in dialog box
	 *		 2=display error in box
	 *		 4=double any \'s (nomagic)
	 */
{	LPBYTE		pRep = RepList.Buf-1;
	PSEARCHPART	pSp;
	int			c;

	while (*Source) {
		switch (*Source) {
			case '\\':
				if (Flags & 4) {
					if ((pRep = ExpandRepList(pRep)) != NULL) *pRep = '\\';
					break;
				}
				switch (c = *++Source) {
					case '\0':
						if ((pRep = ExpandRepList(pRep)) != NULL)
							*pRep = *--Source;
						break;
					case '1': case '2': case '3': case '4': case '5':
					case '6': case '7': case '8': case '9':
						if (c-'0' <= nParenPairs) {
							/*extract \(...\)*/
							POSITION	p;
							int			i  = c-'0';

							pSp = FirstMatch;
							for (;;) {
								while (pSp && (pSp->Type != SrchSpecial ||
									   pSp->u.SpecialMatch.Char != '('))
									pSp = pSp->Next;
								if (!pSp || --i <= 0) break;
								pSp = pSp->Next;
							}
							assert(pSp);
							if (pSp) {
								p = pSp->u.SpecialMatch.Pos;
								while ((pSp = pSp->Next) != NULL) {
									if (pSp->Type == SrchSpecial) {
										switch (pSp->u.SpecialMatch.Char) {
											case '(': ++i;
													  continue;
											case ')': if (--i < 0) break;
													  /*FALLTHROUGH*/
											default:  continue;
										}
										break;
									}
								}
							}
							assert(pSp);
							if (pSp) {
								while (ComparePos(&p,
											 &pSp->u.SpecialMatch.Pos) < 0) {
									if ((c = CharAt(&p)) == C_CRLF) c = '\r';
									if ((pRep = ExpandRepList(pRep)) != NULL)
										*pRep = c;
									Advance(&p, 1);
								}
							}
							break;
						}
						if (*Source >= '8') {
							Error(230);
							if (Flags & 2) ErrorBox(MB_ICONSTOP, 230);
							if (Disabled) Enable();
							FreeRepList();
							return (FALSE);
						}
						/*FALLTHROUGH*/
					case '0':
						if ((pRep = ExpandRepList(pRep)) != NULL) {
							*pRep = *Source - '0';
							if (Source[1] >= '0' && Source[1] < '8') {
								*pRep = (*pRep<<3) + *++Source - '0';
								if (*pRep < '\040')
									if (Source[1] >= '0' && Source[1] < '8')
										*pRep = (*pRep<<3) + *++Source - '0';
							}
						}
						break;
					case '%':
						if ((pRep = ExpandRepList(pRep)) != NULL) {
							if (ISHEX(Source[1])) {
								*pRep = *++Source;
								if (*Source >= 'A')
									*pRep += (BYTE)(10-'A');
								*pRep &= 15;
								if (ISHEX(Source[1])) {
									*pRep <<= 4;
									*pRep += *++Source & 15;
									if (*Source >= 'A')
										*pRep += (BYTE)((10-'A') & 15);
								}
							} else *pRep = *Source;
						}
						break;
					case '&':
						if (!MagicFlag) goto HandleAmpersand;
						/*FALLTHROUGH*/
					default:
						if ((pRep = ExpandRepList(pRep)) != NULL)
							*pRep = *Source;
				}
				break;
			case '&':
				if (MagicFlag && !(Flags & 4)) {
					POSITION p;
					LONG	 n;
				HandleAmpersand:
					p = *Pos;
					n = Len;
					while (n--) {
						if ((c = CharAndAdvance(&p)) == C_CRLF) {
							if (n) {
								if ((pRep = ExpandRepList(pRep)) != NULL)
									*pRep = '\r';
								c = '\n';
								--n;
							} else c = '\r';
						}
						if ((pRep = ExpandRepList(pRep)) != NULL) *pRep = c;
					}
					break;
				}
				/*FALLTHROUGH*/
			default:
				if (Flags & 1) {
					if (ISHEX(*Source)) {
						if ((pRep = ExpandRepList(pRep)) != NULL) {
							if ((*pRep = *Source) >= 'A')
								*pRep += (BYTE)(10-'A');
							*pRep &= 15;
							if (ISHEX(Source[1])) {
								*pRep <<= 4;
								*pRep  += *++Source & 15;
								if (*Source >= 'A') *pRep += (10-'A') & 15;
							}
						}
					} else if (*Source != ' ') {
						Error(239);
						if (Flags & 2) ErrorBox(MB_ICONSTOP, 239);
						FreeRepList();
						return (FALSE);
					}
				} else if ((pRep = ExpandRepList(pRep)) != NULL) {
					PSTR pc;
					if (CharSet && *Source >= 0200 &&
						   (pc=memchr(MapOem, *Source, sizeof(MapOem))) != NULL)
						 *pRep = pc - MapOem;
					else *pRep = *Source;
				}
				break;
		}
		++Source;
	}
	if (pRep == NULL) {
		Error(121);
		if (Flags & 2) ErrorBox(MB_ICONSTOP, 121);
		FreeRepList();
		return (FALSE);
	}
	*DestLen = nRepListCount * sizeof(lpLastRep->Buf) + (pRep+1-lpLastRep->Buf);
	return (TRUE);
}

BOOL SubstituteCmd(LONG StartLine, LONG EndLine, PSTR Args)
{	PBYTE		pSearch = NULL, pReplace = NULL;
	int			c;
	BOOL		SaveWrapFlag, ReportLineCount, GlobalFlag = FALSE;
	BOOL		DoSetUnsafe = FALSE, AtByte0 = FALSE;
	LONG		lMatched;
	POSITION	Pos, Start, SkipNullMatch;
	LONG		LineCount = 0, LastSubstLine = -1, DiffLines = 0;

	if (IsViewOnly()) return (FALSE);

	/*unassemble arguments...*/
	if ((c = *Args) > ' ' && !(CharFlags[c] & 8)) {
		pSearch = ++Args;
		while (*Args && *Args != c)
			if (*Args++ == '\\' && *Args) ++Args;
		if (*Args) {
			*Args++ = '\0';
			pReplace = Args;
			while (*Args && *Args != c)
				if (*Args++ == '\\' && *Args) ++Args;
			if (*Args) *Args++ = '\0';
		}
	} else if (!pLastReplace) {
		Error(234);
		return (FALSE);
	}
	for (;;) {
		switch (*Args++) {
			case 'g':	GlobalFlag = TRUE;
						/*FALLTHROUGH*/
			case ' ':	continue;
			case '\0':	break;
			default:	Error(106);
						return (FALSE);
		}
		break;
	}

	/*set up search structure...*/
	if (pSearch) {
		if (*pSearch) {
			*--pSearch = '/';
			pSearch = BuildMatchList(pSearch, c, NULL);
			assert(*pSearch == '\0');
		}
		if (!pReplace) pReplace = "";
		if (pLastReplace) free(pLastReplace);
		if ((pLastReplace = malloc(lstrlen(pReplace)+1)) == NULL) {
			Error(312);
			return (FALSE);
		}
		lstrcpy(pLastReplace, pReplace);
	}
	if (!StartSearch()) return (FALSE);
	Start.p = NULL;

	if (!CheckParenPairs()) return (FALSE);

	/*jump to start line...*/
	if (StartLine != EndLine) {
		SwitchCursor(SWC_HOURGLASS_ON);
		ReportLineCount = GlobalFlag;
	} else ReportLineCount = FALSE;
	Pos.p = FirstPage; Pos.i = 0;
	while (--StartLine > 0) {
		do; while (!(CharFlags[CharAndAdvance(&Pos)] & 1));
	}

	/*set up some variables...*/
	SaveWrapFlag = WrapScanFlag;
	lMatched = 0;
	WrapScanFlag = Interrupted = FALSE;
	StartUndoSequence();
	SkipNullMatch.p = NULL;

	/*and finally, do the substitutions...*/
	while (!(CharFlags[CharAt(&Pos)] & 0x20)) {
		LONG Length, CurrLine, ReplaceLength, Lines;
		BOOL ContinueOnSameLine = GlobalFlag;

		/*search...*/
		if (!SearchIt(&Pos, 3)) break;
		if (SkipNullMatch.p != NULL) {
			if (ComparePos(&EndMatch, &SkipNullMatch) == 0) {
				if (!Advance(&Pos, 1)) break;
				SkipNullMatch.p = NULL;
				continue;
			}
			SkipNullMatch.p = NULL;
		}
		CurrLine = CountLines(&Pos);
		if (EndLine >= 0 && CurrLine >= EndLine+DiffLines) break;
		Length = CountBytes(&EndMatch) - CountBytes(&Pos);
		if (CurrLine != LastSubstLine) {
			LastSubstLine = CurrLine;
			++LineCount;
		}

		/*build replace buffer...*/
		if (!BuildReplaceString(pLastReplace, &ReplaceLength, &Pos,Length, 0)) {
			WrapScanFlag = SaveWrapFlag;
			ShowEditCaret();
			SwitchCursor(SWC_HOURGLASS_OFF);
			return (FALSE);
		}

		if (!lMatched++) {
			Start = Pos;
			if (CountBytes(&Start) == 0) AtByte0 = TRUE;
		}
		Lines = CountLines(&EndMatch) - CurrLine;
		if (Lines) {
			if (CharAt(&EndMatch)=='\n' && EndMatch.i && EndMatch.p->PageBuf &&
					EndMatch.p->PageBuf[EndMatch.i-1] == '\r') --Lines;
			if (Lines) {
				DiffLines -= Lines;
				ContinueOnSameLine = TRUE;
			}
		}

		/*replace...*/
		SelectCount = Length;
		if (SelectCount) {
			SelectStart = Pos;
			SelectBytePos = CountBytes(&SelectStart);
			DeleteSelected(6);
			DoSetUnsafe = TRUE;
		}
		CurrPos = Pos;
		if (ReplaceLength) {
			LPREPLIST CurrRep = &RepList;
			LPBYTE	  p = RepList.Buf;
			BOOL	  AddCr = FALSE;

			DoSetUnsafe = TRUE;
			while (ReplaceLength-- > 0) {
				if (AddCr && *p=='\n') {
					Reserve(&Pos, 2, 1);
					assert(Pos.p->PageBuf);
					Pos.p->PageBuf[Pos.i++] = '\r';
					Pos.p->PageBuf[Pos.i]   = *p;
					AddCr = FALSE;
					++DiffLines;
				} else {
					if (AddCr) {
						Reserve(&Pos, 1, 1);
						assert(Pos.p->PageBuf);
						Pos.p->PageBuf[Pos.i] = '\r';
						AddCr = FALSE;
						++DiffLines;
					}
					if (CharFlags[*p] & 1) {
						#if 0
						//if (EndLine >= 0) ++EndLine;
						#endif
						if (*p=='\r' && ReplaceLength>0) AddCr = TRUE;
						else {
							if (*p=='\r' && CharAt(&Pos)=='\n') {
								Reserve(&Pos, -1, 1);
								Reserve(&Pos,  2, 1);
								assert(Pos.p->PageBuf);
								Pos.p->PageBuf[Pos.i]   = '\r';
								Pos.p->PageBuf[Pos.i+1] = '\n';
							} else {
								Reserve(&Pos, 1, 1);
								assert(Pos.p->PageBuf);
								Pos.p->PageBuf[Pos.i] = *p;
								++DiffLines;
							}
						}
					} else {
						Reserve(&Pos, 1, 0);
						assert(Pos.p->PageBuf);
						Pos.p->PageBuf[Pos.i] = *p;
					}
				}
				if (!AddCr) Advance(&Pos, 1);
				if (++p >= CurrRep->Buf + sizeof(CurrRep->Buf)) {
					if ((CurrRep = CurrRep->Next) == NULL) break;
					p = CurrRep->Buf;
				}
			}
		}
		FreeRepList();

		if (!ContinueOnSameLine)
			do; while (!(CharFlags[CharAndAdvance(&Pos)] & 1));
		else SkipNullMatch = Pos;
	}
	WrapScanFlag = SaveWrapFlag;
	if (DoSetUnsafe) SetUnsafe();
	if (Interrupted) Error(218);
	else if (!lMatched) Error(214);
	else {
		/*display number of matches...*/
		extern char	Buffer[300];
		static char	Fmt[36];
		if (ReportLineCount && lMatched>1)
			 LOADSTRING(hInst, 242 + (LineCount!=1), Fmt, sizeof(Fmt));
		else LOADSTRING(hInst, 240 + (lMatched!=1),  Fmt, sizeof(Fmt));
		wsprintf(Buffer, Fmt, lMatched, LineCount);
		NewStatus(0, Buffer, NS_NORMAL);
	}
	if (lMatched) {
		if (AtByte0 && LineInfo[0].Pos.p==Start.p && LineInfo[0].Pos.i==0) {
			ScreenStart.p = FirstPage;
			ScreenStart.i = 0;
			InvalidateText();
		} else {
			GoBackAndChar(&Start);
			InvalidateArea(&Start, CountBytes(&Pos) - CountBytes(&Start), 15);
		}
		UpdateWindow(hwndMain);
		NewPosition(&CurrPos);
		FindValidPosition(&CurrPos, (WORD)(Mode==InsertMode));
	}
	if (Disabled) Enable();
	ShowEditCaret();
	SwitchCursor(SWC_HOURGLASS_OFF);
	return (TRUE);
}

void SearchNewCset(void)
{
	if (MatchValid) {
		PSEARCHPART p;
		for (p=FirstMatch; p; p=p->Next) {
			if (p->Type == SrchString) {
				BYTE ch[2];
				ch[0] = ch[1] = *p->u.StringMatch.String;
				if (p->u.StringMatch.AllowConvert[0] & 1) {
					AnsiUpperBuff(ch+0, 1);
					AnsiLowerBuff(ch+1, 1);
					if (CharSet) AnsiToOemBuff(ch, ch, 2);
				}
				p->u.StringMatch.FirstUcase = ch[0];
				p->u.StringMatch.FirstLcase = ch[1];
			}
		}
	}
}

BOOL SearchAgain(WORD Flags)
	/*Flags: 1=Reverse (search in opposite direction)
	 *		 2=pop up dialog if not searched before
	 */
{	BOOL  Return;
	ULONG OldLength;

	HideEditCaret();
	if (Flags & 2 && !MatchValid) {
		SearchDialog();
		return (TRUE);
	}
	OldLength = SelectCount;
	if (OldLength) {
		InvalidateArea(&SelectStart, SelectCount, 1);
		SelectCount = 0;
		UpdateWindow(hwndMain);
	}
	SelectStart = CurrPos;
	Return = SearchAndStatus(&SelectStart, (WORD)(Flags & 1 | 2));
	if (Return) {
		if (OldLength==0 && Mode!=InsertMode ||
				ComparePos(&SelectStart, &CurrPos) == 0) {
			if (CountBytes(&EndMatch) - CountBytes(&StartPos) == OldLength) {
				SelectStart = CurrPos;
				Return = SearchAndStatus(&SelectStart, (WORD)(Flags & 1));
			}
		}
		if (Return) {
			NewPosition(&SelectStart);
			UpdateWindow(hwndMain);
			SelectStart   = StartPos;
			SelectBytePos = CountBytes(&StartPos);
			SelectCount   = CountBytes(&EndMatch) - SelectBytePos;
			InvalidateArea(&SelectStart, SelectCount, 0);
			UpdateWindow(hwndMain);
		}
	}
	CheckClipboard();
	ShowEditCaret();
	GetXPos(&CurrCol);
	return (Return);
}

extern char CommandBuf[280];
BOOL Magic, HexInput;

void ReplaceSearched(HWND hDlg)
{	MODEENUM SaveMode = Mode;
	LONG	 Length;

	Magic	 = IsDlgButtonChecked(hDlg, IDC_MAGIC);
	HexInput = IsDlgButtonChecked(hDlg, IDC_HEXSEARCH);
	GetDlgItemText(hDlg, IDC_REPLACESTRING, CommandBuf, sizeof(CommandBuf));
	hwndErrorParent = hDlg;			/*change parent of message box*/
	if (Magic && !CheckParenPairs() ||
		!BuildReplaceString(CommandBuf, &Length,
							&CurrPos, SelectCount,
							(WORD)(HexInput | 2 | (Magic ? 0 : 4)))) {
			hwndErrorParent = hwndMain;			/*change parent of message box*/
			return;
	}
	hwndErrorParent = hwndMain;			/*change parent of message box*/
	Mode = InsertMode;
	StartUndoSequence();
	if (SelectCount) DeleteSelected(19);
	HideEditCaret();
	{	LPREPLIST lpRep = &RepList;

		while (Length > (LONG)sizeof(lpRep->Buf)) {
			InsertBuffer(lpRep->Buf, sizeof(lpRep->Buf), 0);
			Length -= sizeof(lpRep->Buf);
			if ((lpRep = lpRep->Next) == NULL) break;
		}
		if (Length && lpRep != NULL) InsertBuffer(lpRep->Buf, Length, 0);
		FreeRepList();
	}
	GoBackAndChar(&CurrPos);
	Mode = SaveMode;
	FindValidPosition(&CurrPos, (WORD)(Mode==InsertMode));
	SendMessage(hDlg, DM_SETDEFID, IDOK, 0);
	SetFocus(GetDlgItem(hDlg, IDOK));
	EnableWindow(GetDlgItem(hDlg, IDC_REPLACE), FALSE);
}

char SearchBuf[256];
BOOL WholeWord, ReplaceOpen;

PSTR SetupSearchString(HWND hDlg, int *Error)
{	PSTR pS = SearchBuf, pD = CommandBuf;
	BOOL Ok = TRUE, OpenBracket = FALSE;

	Magic		   = IsDlgButtonChecked(hDlg, IDC_MAGIC);
	HexInput	   = IsDlgButtonChecked(hDlg, IDC_HEXSEARCH);
	WholeWord	   = IsDlgButtonChecked(hDlg, IDC_WHOLEWORD);
	IgnoreCaseFlag = IsDlgButtonChecked(hDlg, IDC_MATCHCASE) == FALSE;
	WrapScanFlag   = IsDlgButtonChecked(hDlg, IDC_WRAPSCAN);
	*CommandBuf	   = IsDlgButtonChecked(hDlg, IDC_FORWARD) ? '/' : '?';
	GetDlgItemText(hDlg, IDC_SEARCHSTRING, SearchBuf, sizeof(SearchBuf));
	if (WholeWord) {
		*++pD = '\\';
		*++pD = '<';
	}
	for (;;) {
		switch (*++pD = *pS++) {
			case '\\':
				if (Magic) {
					switch (*pS) {
						case '\0':
							break;
						case '%':
							*++pD = *pS++;
							if (!ISHEX(*pS)) break;
							*++pD = *pS++;
							if (ISHEX(*pS)) *++pD = *pS++;
							continue;
						case '0': case '1': case '2': case '3':
							*++pD = *pS++;
							if (*pS >= '0' && *pS <= '7') *++pD = *pS++;
							if (*pS >= '0' && *pS <= '7') *++pD = *pS++;
							continue;
						default:
							*++pD = *pS++;
							continue;
					}
				}
				if (HexInput) Ok = FALSE;
				break;
			case '/': case '?':
				if (HexInput) {
					Ok = FALSE;
					break;
				}
				if (*pD != *CommandBuf) continue;
				/*FALLTHROUGH*/
			case '\0':
				break;
			case '[':
				OpenBracket = TRUE;
				/*FALLTHROUGH*/
			case '*': case '.':
				if (HexInput && !Magic) {
					Ok = FALSE;
					break;
				}
				if (!MagicFlag != !Magic) break;
				continue;
			case '^': case '$':
				if (!Magic) {
					if (HexInput) Ok = FALSE;
					break;
				}
				continue;
			case '-': case ']':
				if (Magic && OpenBracket) {
					if (!(OpenBracket = *pD != ']') && !MagicFlag) break;
					continue;
				}
				/*FALLTHROUGH*/
			default:
				if (HexInput) {
					if (ISHEX(*pD)) {
						pD[2] = *pD;
						*pD++ = '\\'; *pD++ = '%';
						if (ISHEX(*pS)) *++pD = *pS++;
					} else if (*pD-- != ' ') {
						Ok = FALSE;
						break;
					}
				}
				continue;
		}
		if (!Ok) {
			if (Error != NULL) *Error = 239;
			return (NULL);
		}
		if (*pD == '\0') break;
		pD[1] = *pD;
		*pD++ = '\\';
	}
	if (WholeWord) lstrcpy(pD, "\\>");
	return (BuildMatchList(CommandBuf, *CommandBuf, Error));
}

void SearchOk(HWND hDlg)
{	int  ErrNo = 0;
	HWND hSearchButton, hFocus;

	hFocus = GetFocus();
	EnableWindow(hSearchButton = GetDlgItem(hDlg, IDOK), FALSE);
	if (SetupSearchString(hDlg, &ErrNo))
		if (!MatchValid) ErrNo = SearchErrNo;
		else if (!SearchAgain(0)) ErrNo = 238;
	EnableWindow(hSearchButton, TRUE);
	if (hFocus == hSearchButton) SetFocus(hSearchButton);
	if (!ErrNo) {
		extern void HorizontalScrollIntoView(void);

		if (!ReplaceOpen) {
			if (!ViewOnlyFlag)
				EnableWindow(GetDlgItem(hDlg, IDC_SHOWREPLACE), TRUE);
		} else {
			EnableWindow(GetDlgItem(hDlg, IDC_REPLACE), TRUE);
			SendMessage(hDlg, DM_SETDEFID, IDC_REPLACE, 0);
			SendDlgItemMessage(hDlg, IDOK, BM_SETSTYLE,
							   (WPARAM)BS_PUSHBUTTON, TRUE);
		}
		/*main window does not have the focus...*/
		HorizontalScrollIntoView();
	} else {
		if (ReplaceOpen)
			 EnableWindow(GetDlgItem(hDlg, IDC_REPLACE),	 FALSE);
		else EnableWindow(GetDlgItem(hDlg, IDC_SHOWREPLACE), FALSE);
		hwndErrorParent = hDlg;			/*change parent of message box*/
		ErrorBox(ErrNo==216 || ErrNo==217 ? MB_ICONINFORMATION : MB_ICONSTOP,
				 ErrNo);
		hwndErrorParent = hwndMain;		/*reset to normal value*/
	}
}

void EnableControls(HWND hDlg, BOOL Enable)
{
	EnableWindow(GetDlgItem(hDlg, IDC_SEARCHSTRING),  Enable);
	EnableWindow(GetDlgItem(hDlg, IDC_REPLACESTRING), Enable);
	EnableWindow(GetDlgItem(hDlg, IDOK),			  Enable);
	EnableWindow(GetDlgItem(hDlg, IDC_REPLACE),		  Enable);
	EnableWindow(GetDlgItem(hDlg, IDC_REPLACEALL),	  Enable);
}

BOOL ReplacingAll = FALSE;

void GlobalSubst(HWND hDlg)
{	PSTR p = CommandBuf;
	int	 n;

	Magic	 = IsDlgButtonChecked(hDlg, IDC_MAGIC);
	HexInput = IsDlgButtonChecked(hDlg, IDC_HEXSEARCH);
	ReplacingAll = TRUE;
	p += n = wsprintf(p, ":%%s//");
	n += GetDlgItemText(hDlg, IDC_REPLACESTRING, p, sizeof(CommandBuf)-n);
	while (*p) {
		switch (*p) {
			case '\\':
				if (Magic) {
					if	 (*++p >= '0' && *p <= '7') {
						if (*p >= '0' && *p <= '3') ++p;
						if (*p >= '0' && *p <= '7') ++p;
						if (*p >= '0' && *p <= '7') ++p;
						continue;
					}
					if (*p == '%') {
						if (ISHEX(p[1])) ++p;
						if (ISHEX(p[1])) ++p;
					}
					break;
				}
				/*FALLTHROUGH*/
			case '&':
				if (*p=='&' && !Magic!=MagicFlag) break;
				/*FALLTHROUGH*/
			case '/':
				if (n+1 < sizeof(CommandBuf)) {
					memmove(p+1, p, lstrlen(p)+1);
					*p++ = '\\';
					++n;
				}
				break;
			default:
				if (HexInput) {
					if (!ISHEX(*p)) {
						if (*p != ' ') {
							hwndErrorParent = hDlg;
							ErrorBox(MB_ICONSTOP, 239);
							hwndErrorParent = hwndMain;
							return;
						}
						memmove(p, p+1, lstrlen(p));
						--n;
						continue;
					}
					if (n+2 < sizeof(CommandBuf)) {
						memmove(p+2, p, lstrlen(p)+1);
						*p++ = '\\';
						*p++ = '%';
						if (ISHEX(p[1])) ++p;
						n += 2;
					}
				}
		}
		++p;
	}
	if (n+2 < sizeof(CommandBuf)) lstrcpy(p, "/g");
	SendMessage(hDlg, DM_SETDEFID, IDCANCEL, 0);
	SetFocus(GetDlgItem(hDlg, IDCANCEL));
	EnableControls(hDlg, FALSE);
	CommandExec(CommandBuf);
	EnableControls(hDlg, TRUE);
	ReplacingAll = FALSE;
}

HWND hwndSearch;

BOOL FAR PASCAL SearchCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	extern BOOL SearchBoxPosition, HexEditTextSide;
	extern int  SearchBoxX, SearchBoxY;
	static RECT DlgRect;
	static BOOL Enabled;
	static char CloseString[10];

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		RECT r;

		case WM_INITDIALOG:
			GetWindowRect(hDlg, &DlgRect);
			GetWindowRect(GetDlgItem(hDlg, IDC_REPLACE), &r);
			SetWindowPos(hDlg, 0, SearchBoxX, SearchBoxY,
								  DlgRect.right - DlgRect.left,
								  r.top - DlgRect.top,
						 SearchBoxPosition ? SWP_NOZORDER
										   : SWP_NOZORDER | SWP_NOMOVE);
			SendMessage(hDlg, DM_REPOSITION, 0, 0);
			EnableWindow(GetDlgItem(hDlg, IDC_SHOWREPLACE), FALSE);
			if (HexInput != (HexEditMode && !HexEditTextSide)) {
				HexInput ^= TRUE;
				*SearchBuf = '\0';
			}
			if (SelectCount &&
					(HexInput ? 3 : 1) * SelectCount < sizeof(SearchBuf)) {
				POSITION SelPos;
				int		 i, c;
				PSTR	 p = SearchBuf;

				SelPos = SelectStart;
				for (i=(int)SelectCount; i; --i) {
					if ((c = CharAt(&SelPos)) == C_CRLF) c = '\r';
					if (Advance(&SelPos, 1) != 1) break;
					if (HexInput) p += wsprintf(p, "%02x ", c);
					else *p++ = c;
				}
				if (HexInput) --p;
				*p = '\0';
				if (*SearchBuf && !ViewOnlyFlag)
					EnableWindow(GetDlgItem(hDlg, IDC_SHOWREPLACE), TRUE);
			}
			Enabled = *SearchBuf!='\0';
			if (Enabled) {
				EnableWindow(GetDlgItem(hDlg, IDOK), TRUE);
				SetDlgItemText(hDlg, IDC_SEARCHSTRING, SearchBuf);
				SendMessage(hDlg, DM_SETDEFID, IDOK, 0L);
			} else {
				SendMessage(hDlg, DM_SETDEFID, IDCANCEL, 0L);
				EnableWindow(GetDlgItem(hDlg, IDOK), FALSE);
			}
			EnableWindow(GetDlgItem(hDlg, IDC_REPLACESTRING), FALSE);
			EnableWindow(GetDlgItem(hDlg, IDC_REPLACE),       FALSE);
			EnableWindow(GetDlgItem(hDlg, IDC_REPLACEALL),    FALSE);
			CheckDlgButton(hDlg, *SrchDispBuf=='?' ? IDC_BACKWARD : IDC_FORWARD,
								 TRUE);
			CheckDlgButton(hDlg, IDC_MATCHCASE,	!IgnoreCaseFlag);
			CheckDlgButton(hDlg, IDC_MAGIC,		Magic);
			CheckDlgButton(hDlg, IDC_HEXSEARCH, HexInput);
			CheckDlgButton(hDlg, IDC_WHOLEWORD,	WholeWord);
			CheckDlgButton(hDlg, IDC_WRAPSCAN,	WrapScanFlag);
			LOADSTRING(hInst, 909, CloseString, sizeof(CloseString));
			ReplaceOpen = FALSE;
			PostMessage(hDlg, WM_COMMAND, 4569, 0);	/*for Wine*/
			return (TRUE);

		case WM_COMMAND:
			switch (COMMAND) {
				case 4569:
					/*Disable/enable again for Wine...*/
					EnableWindow(GetDlgItem(hDlg, IDOK), Enabled);
					break;
				case IDOK:
					SearchOk(hDlg);
					break;
				case IDCANCEL:
					if (ReplacingAll) Interrupted = TRUE;
					PostMessage(hDlg, WM_CLOSE, 0, 0);
					break;
				case IDC_SEARCHSTRING:
					GetDlgItemText(hDlg, IDC_SEARCHSTRING, CommandBuf, 2);
					if (Enabled != (*CommandBuf != '\0')) {
						if (Enabled ^= TRUE) {
							EnableWindow(GetDlgItem(hDlg, IDOK), TRUE);
							SendMessage(hDlg, DM_SETDEFID, IDOK, 0L);
						} else {
							SendMessage(hDlg, DM_SETDEFID, IDCANCEL, 0L);
							EnableWindow(GetDlgItem(hDlg, IDOK), FALSE);
						}
						if (ReplaceOpen) {
						  EnableWindow(GetDlgItem(hDlg,IDC_REPLACE),   Enabled);
						  EnableWindow(GetDlgItem(hDlg,IDC_REPLACEALL),Enabled);
						}
					}
					break;
				case IDC_SHOWREPLACE:
					EnableWindow(GetDlgItem(hDlg, IDC_REPLACESTRING), TRUE);
					EnableWindow(GetDlgItem(hDlg, IDC_REPLACE), SelectCount!=0);
					EnableWindow(GetDlgItem(hDlg, IDC_REPLACEALL),	  TRUE);
					SetWindowPos(hDlg, 0,0,0, DlgRect.right-DlgRect.left,
											  DlgRect.bottom-DlgRect.top,
								 SWP_NOZORDER | SWP_NOMOVE);
					SendMessage(hDlg, DM_REPOSITION, 0, 0);
					SetFocus(GetDlgItem(hDlg, IDC_REPLACESTRING));
					SendMessage(hDlg, DM_SETDEFID, IDC_REPLACE, 0L);
					EnableWindow(GetDlgItem(hDlg, IDC_SHOWREPLACE),  FALSE);
					ReplaceOpen = TRUE;
					break;
				case IDC_REPLACE:
					ReplaceSearched(hDlg);
					SetDlgItemText(hDlg, IDCANCEL, CloseString);
					break;
				case IDC_REPLACEALL:
					SetupSearchString(hDlg, NULL);
					GlobalSubst(hDlg);
					SetDlgItemText(hDlg, IDCANCEL, CloseString);
			}
			return (TRUE);

		case WM_MOVE:
		case WM_CLOSE:
			SearchBoxPosition = TRUE;
			GetWindowRect(hDlg, &r);
			SearchBoxX = r.left;
			SearchBoxY = r.top;
			if (uMsg == WM_CLOSE) {
				DestroyWindow(hDlg);
				hwndSearch = NULL;
			}
			return (TRUE);
	}
	return (FALSE);
}

void SearchDialog(void)
{	static DLGPROC DlgProc;

	if (!DlgProc)
		 DlgProc = (DLGPROC)MakeProcInstance((FARPROC)SearchCallback, hInst);
	if (hwndSearch) BringWindowToTop(hwndSearch);
	else hwndSearch =
		 CreateDialog(hInst, MAKEINTRES(IDD_SEARCH), hwndMain, DlgProc);
}
