/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <commdlg.h>
#include <string.h>
#include "WinVi.h"
#include "Colors.h"
#include "PrintH.h"
#include "TabCtrl.h"
#include "resource.h"

#define WITH_PROFILES 0

BOOL FontCallback(HWND, UINT, WPARAM, LPARAM);

HWND		   hwndSettings, hwndSubSet;
static DLGPROC SettingsDlgProc, SubSetDlgProc;
char		   QueryString[150];
DWORD		   QueryTime;
static BOOL	   SuppressFocusChange, FocusOnFirstControl;

struct tagTABSHEET {
	CHAR	*Label;
	BOOL	(*Callback)(HWND,UINT,WPARAM,LPARAM);
	HWND	hDlg;
	INT		Dialog, Flags, FirstControl, LastControl;
} TabSheets[] = {
#if WITH_PROFILES
	{/*Activation*/	NULL, NULL,						NULL, 320, 0,
					IDC_DRIVESHOSTSPATHS, IDC_ALLUSERS},
	{/*Modes*/		NULL, NULL,						NULL, 321, 0,
					IDC_AUTO, IDC_OEMCP},
#endif
	{/*Display*/	NULL, DisplaySettingsCallback,	NULL, 322, 0,
					1023, IDC_MAXSCROLL},
	{/*Edit*/		NULL, EditSettingsCallback,		NULL, 323, 0,
					IDC_DEFAULTINSERT, IDC_MAGIC},
	{/*Files*/		NULL, FileSettingsCallback,		NULL, 324, 0,
					IDC_PAGE_THRESHOLD, 1005},
#if 0
	{/*Macros*/		NULL, MapCallback,				NULL, 325, 0,
					IDC_MACROLIST, IDC_COMMANDLINE},
#endif
	{/*Colors*/		NULL, ColorsCallback,			NULL, 326, 0,
					1043, 1051 /*1062*/},
	{/*Font*/		NULL, FontCallback,				NULL, 327, 0,
					IDC_CHANGEFONT_TEXT, IDC_CHANGEFONT_HEX},
	{/*Printing*/	NULL, PrintCallback,			NULL, 328, 0,
					1066, 1084},
	{/*Language*/	NULL, LanguageCallback,			NULL, 329, 0,
					IDC_ENGLISH, IDC_AUTOLANG},
	{				NULL, NULL,						NULL,   0, 0}
};
INT CurrTab = 0;

LOGFONT			  TextLogFont, HexLogFont, NewTextLogFont, NewHexLogFont;
HFONT			  NewTextFont, NewHexFont;
INT				  NewTextHeight, NewHexHeight;
static CHOOSEFONT cfFont;
static INT		  Count;

static INT CALLBACK CountFonts
#if defined(WIN32)
	(PLOGFONT plf, PTEXTMETRIC ptm, DWORD dwType, LPARAM lp)
#else
	(LPLOGFONT plf, LPTEXTMETRIC ptm, INT dwType, LPARAM lp)
#endif
{
	(VOID)plf;		/*parameter never used*/
	(VOID)ptm;		/*parameter never used*/
	(VOID)dwType;	/*parameter never used*/
	(VOID)lp;		/*parameter never used*/
	++Count;
	return (0);
}

typedef struct {
	HINSTANCE hInst;
	INT		  FontTitle;
} CHOOSEFONTPARAM, FAR *LPCHOOSEFONTPARAM;
CHOOSEFONTPARAM ChooseFontParam;

UINT CALLBACK ChooseFontHook(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{
	(VOID)wPar;		/*parameter never used*/
	if (uMsg==WM_INITDIALOG) {
		LPCHOOSEFONT	  lpCf = (LPCHOOSEFONT)lPar;
		LPCHOOSEFONTPARAM cfp  = (LPCHOOSEFONTPARAM)lpCf->lCustData;

		if (cfp->FontTitle) {
			CHAR Title[48];

			LoadString(cfp->hInst, cfp->FontTitle, Title, sizeof(Title));
			SetWindowText(hDlg, Title);
		}
	}
	return (0);
}

static BOOL SelectFont(HWND hWnd, PLOGFONT LogFont)
{	HWND OldFocus = GetFocus();
	BOOL Result;

	ChooseFontParam.hInst = hInst;
	cfFont.lStructSize = sizeof(CHOOSEFONT);
	cfFont.hwndOwner   = hWnd;
	cfFont.rgbColors   = TextColor;
	cfFont.lCustData   = (LPARAM)(LPCHOOSEFONTPARAM)&ChooseFontParam;
	cfFont.lpLogFont   = LogFont;
	if (cfFont.lpfnHook == NULL)
		cfFont.lpfnHook = (UINT (CALLBACK*)(HWND, UINT, WPARAM, LPARAM))
						  MakeProcInstance((FARPROC)ChooseFontHook, hInst);
	cfFont.Flags	   = CF_SCREENFONTS | CF_EFFECTS | CF_INITTOLOGFONTSTRUCT
										| CF_ENABLEHOOK;
	cfFont.hInstance   = hInst;
	if (!*LogFont->lfFaceName) {
		lstrcpy(LogFont->lfFaceName, "Fixedsys");
		LogFont->lfHeight = 12;
	}
	#if 0
	//	{	HDC hDC = GetDC(hwndMain);
	//		/*special handling for Wine...*/
	//
	//		if (hDC) {
	//			static OLDFONTENUMPROC Proc;
	//
	//			Count = 0;
	//			if (!Proc) Proc = (OLDFONTENUMPROC)
	//							  MakeProcInstance((FARPROC)CountFonts,hInst);
	//			EnumFonts(hDC, LogFont->lfFaceName, Proc, 0);
	//			if (Count == 0) {
	//				lstrcpy(LogFont->lfFaceName, "Lucidatypewriter");
	//				LogFont->lfHeight = 10;
	//			}
	//			ReleaseDC(hwndMain, hDC);
	//		}
	//	}
	#endif
	Result = ChooseFont(&cfFont);
	if (OldFocus) SetFocus(OldFocus);	/*for 16-bit*/
	return (Result);
}

LPARAM ChooseFontCustData(INT FontTitle)
{
	ChooseFontParam.hInst	  = hInst;
	ChooseFontParam.FontTitle = FontTitle+Language;
	return ((LPARAM)(LPCHOOSEFONTPARAM)&ChooseFontParam);
}

static BOOL ShowFont(VOID)
{	extern BOOL IsFixedFont;
	BOOL        Repaint = FALSE;

	UpdateWindow(hwndMain);
	if (TextFont && NewTextFont) {
		DeleteObject(TextFont);
		TextFont = NewTextFont;
		NewTextFont = 0;
		if (!HexEditMode) {
			TextHeight = NewTextHeight;
			Repaint = TRUE;
		}
		TextLogFont = NewTextLogFont;
	}
	if (HexFont && NewHexFont) {
		DeleteObject(HexFont);
		HexFont = NewHexFont;
		NewHexFont = 0;
		if (HexEditMode) {
			TextHeight = NewHexHeight;
			Repaint = TRUE;
		}
		HexLogFont = NewHexLogFont;
	}
	if (TextColor != cfFont.rgbColors) {
		TextColor  = cfFont.rgbColors;
		wUseSysColors &= ~2;
		Repaint = TRUE;
	}
	if (Repaint) {
		AveCharWidth = 0;		/*will be set again in first paint operation*/
		IsFixedFont	 = FALSE;	/*...*/
		AdjustWindowParts(ClientRect.bottom, ClientRect.bottom);
		InvalidateText();
	}
	ShowEditCaret();
	return (TRUE);
}

static VOID DisplayFont(PLOGFONT pLogFont, HFONT Font, HWND hDlg,
						INT NameControl, INT SampleControl)
{	CHAR Buf[64];
	INT  PointSize;
	HDC  hDC = GetDC(hwndMain);

	if (!hDC) return;
	lstrcpyn(Buf, pLogFont->lfFaceName, 44);
	PointSize = -MulDiv(pLogFont->lfHeight, 72, GetDeviceCaps(hDC, LOGPIXELSY));
	ReleaseDC(hwndMain, hDC);
	wsprintf(Buf+lstrlen(Buf), ",  %d pt.", PointSize);
	SetDlgItemText(hDlg, NameControl, Buf);
	SendMessage(GetDlgItem(hDlg, SampleControl), WM_SETFONT, (WPARAM)Font, 0);
}

static INT  nDefId;
static HWND hDefParent;

VOID CheckDefButton(VOID)
{	INT			d1, d2;
	HWND		h, hParent;
	INT			i;
	static BOOL	IsActive = FALSE;
	static HWND PrevH = 0, LastDefId = 0;

	if (IsActive) return;
	IsActive = TRUE;
	h = GetFocus();
	if (h == PrevH) {
		IsActive = FALSE;
		return;
	}
	if (!h || (hParent=GetParent(h))!=hwndSubSet && hParent!=hwndSettings) {
		if (LastDefId)
			SendMessage(LastDefId, BM_SETSTYLE, (WPARAM)BS_PUSHBUTTON, 1);
		PrevH = 0;
		IsActive = FALSE;
		return;
	}
	PrevH = h;
	d1 = d2 = 102;
	switch (i = GetDlgCtrlID(h)) {
		case IDOK:
		case IDCANCEL:
		case ID_APPLY:
			nDefId = d1 = i;
			hDefParent = hwndSettings;
			break;

		case IDC_BITMAPBROWSE:
		case IDC_PRINTFONT:
		case IDC_CHANGEFONT_TEXT:
		case IDC_CHANGEFONT_HEX:
			nDefId = d2 = i;
			hDefParent = hwndSubSet;
			break;

		default:
			nDefId = d1 = IDOK;
			hDefParent = hwndSettings;
			h = GetDlgItem(hwndSettings, IDOK);
	}
	if (h != LastDefId /*&& IsWindowVisible(hwndSubSet)*/) {
		if (nDefId == d2) SendMessage(hwndSettings,	DM_SETDEFID, d1, 0);
		SendMessage					 (hwndSubSet,	DM_SETDEFID, d2, 0);
		if (nDefId == d1) SendMessage(hwndSettings,	DM_SETDEFID, d1, 0);
		if (LastDefId)
			SendMessage(LastDefId, BM_SETSTYLE, (WPARAM)BS_PUSHBUTTON,	  1);
		if (nDefId != 102)
			SendMessage(h,		   BM_SETSTYLE, (WPARAM)BS_DEFPUSHBUTTON, 1);
		LastDefId = nDefId==102 ? (HWND)0 : h;
	}
	IsActive = FALSE;
}

BOOL FontCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	BOOL		Result = TRUE;
	static CHAR Sample[16], Buffer[128];
	static CHAR	*TextSample = "%s  AaBbCcXxYyZz\r\n"
							  "The quick brown fox jumps over a lazy dog.",
				*HexSample  = "%s  AaBbCcXxYyZz\r\n"
							  "14 38 7c ff  |.8|.|\r\n"
							  "25 49 8d 00  |%%I..|\r\n"
							  "36 5a 9e 11  |6Z..|";

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			LOADSTRING(hInst, 257, Sample, sizeof(Sample));
			DisplayFont(&TextLogFont, TextFont, hDlg,
						IDC_FONTNAME_TEXT, IDC_FONTSAMPLE_TEXT);
			DisplayFont(&HexLogFont,  HexFont,  hDlg,
						IDC_FONTNAME_HEX, IDC_FONTSAMPLE_HEX);
			wsprintf(Buffer, TextSample, Sample);
			SetDlgItemText(hDlg, IDC_FONTSAMPLE_TEXT, Buffer);
			wsprintf(Buffer, HexSample, Sample);
			SetDlgItemText(hDlg, IDC_FONTSAMPLE_HEX,  Buffer);
			NewTextLogFont = TextLogFont;
			NewHexLogFont  = HexLogFont;
			Result = FALSE;
			break;

		case WM_COMMAND:
			switch (COMMAND) {
				case IDC_CHANGEFONT_TEXT:
					ChooseFontParam.FontTitle = 254+Language;
					if (SelectFont(hDlg, &NewTextLogFont)) {
						NewFont(&NewTextFont, &NewTextLogFont, &NewTextHeight);
						DisplayFont(&NewTextLogFont, NewTextFont, hDlg,
									IDC_FONTNAME_TEXT, IDC_FONTSAMPLE_TEXT);
						InvalidateRect(GetDlgItem(hDlg, IDC_FONTSAMPLE_TEXT),
									   NULL, TRUE);
					}
					break;

				case IDC_CHANGEFONT_HEX:
					ChooseFontParam.FontTitle = 255+Language;
					if (SelectFont(hDlg, &NewHexLogFont)) {
						NewFont(&NewHexFont, &NewHexLogFont, &NewHexHeight);
						DisplayFont(&NewHexLogFont, NewHexFont, hDlg,
									IDC_FONTNAME_HEX, IDC_FONTSAMPLE_HEX);
						InvalidateRect(GetDlgItem(hDlg, IDC_FONTSAMPLE_HEX),
									   NULL, TRUE);
					}
					break;

				case IDC_STORE:
					ShowFont();
					SaveFontSettings();
					DlgResult = TRUE;
					break;

				case ID_APPLY:
					/*FALLTHROUGH*/
				case IDOK:
					ShowFont();
					DlgResult = TRUE;
					if (COMMAND!=IDOK) break;
					/*FALLTHROUGH*/
				case IDCANCEL:
					EndDialog(hDlg, 0);
					if (NewTextFont) {
						DeleteObject(NewTextFont);
						NewTextFont = 0;
					}
					if (NewHexFont) {
						DeleteObject(NewHexFont);
						NewHexFont = 0;
					}
					break;
			}
			break;

		default:
			Result = FALSE;
	}
	return (Result);
}

BOOL FAR PASCAL SubSetCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	BOOL Result = FALSE, GotResult = FALSE;

	if (TabSheets[CurrTab].Callback != NULL) {
		Result = (*TabSheets[CurrTab].Callback)(hDlg, uMsg, wPar, lPar);
		GotResult = TRUE;
	}
	switch (uMsg) {
		case WM_INITDIALOG:
			if (!GotResult) Result = TRUE;
			break;

		case WM_COMMAND:
			if (!FocusOnFirstControl && NOTIFICATION==EN_SETFOCUS)
				switch (COMMAND) {
					case 100:
						SetFocus(GetDlgItem(hwndSettings, IDC_SETTINGSTAB));
						break;
					case 101:
						{	INT nRadio = IsDlgButtonChecked
											(hwndSettings, IDC_SAVEPERMANENTLY)
										 ? IDC_SAVEPERMANENTLY
										 : IDC_ONETIMESETTING;

							SetFocus(GetDlgItem(hwndSettings, nRadio));
							CheckDefButton();
						}
			}
			if (!GotResult) Result = TRUE;
	}
	return (Result);
}

static VOID SwitchTab(HWND hctlTab, INT NewTab)
{
	if (FocusOnFirstControl) SuppressFocusChange = FALSE;
	else SuppressFocusChange = GetFocus() == hctlTab;
	if (NewTab != CurrTab) {
		HWND NewFocus;

		ShowWindow(hwndSubSet, SW_HIDE);
		CurrTab = NewTab;
		hwndSubSet = TabSheets[CurrTab].hDlg;
		if (!hwndSubSet) {
			hwndSubSet = TabSheets[CurrTab].hDlg
				= CreateDialog(hInst, MAKEINTRES(TabSheets[CurrTab].Dialog),
							   /*hctlTab*/ hwndSettings, SubSetDlgProc);
		} else if (CurrTab==3 || CurrTab==5) {
			/*handle pushbutton in "Colors" or "Print" tabsheet...*/
			HWND Button;

			SendMessage(hwndSubSet, DM_SETDEFID, 102, 0);
			Button = GetDlgItem(hwndSubSet,
								CurrTab==3 ? IDC_BITMAPBROWSE : IDC_PRINTFONT);
			if (Button) SendMessage(Button,BM_SETSTYLE,(WPARAM)BS_PUSHBUTTON,0);
		}
		ShowWindow(hwndSubSet, SW_SHOW);
		BringWindowToTop(hwndSubSet);
		if (SuppressFocusChange) NewFocus = hctlTab;
		else {
			INT MaxLoop	= 6;
			INT Ctl		= TabSheets[CurrTab].FirstControl;

			do {
				NewFocus = GetDlgItem(hwndSubSet, Ctl++);
				if (GetWindowLong(NewFocus, GWL_STYLE) & WS_TABSTOP) break;
			} while (--MaxLoop);
		}
		SetFocus(NewFocus);
		SuppressFocusChange = FALSE;
		CheckDefButton();
	}
}

HWND GetCurrentTabSheet(VOID)
{
	return (TabSheets[CurrTab].hDlg);
}

RECT DlgRect = {-10000};

BOOL FAR PASCAL SettingsCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	INT			 i, Ctl, Inc;
	static BOOL	 Restart = FALSE;
	static HFONT hFont;

	switch (uMsg) {
		case WM_SETFONT:
			hFont = (HFONT)wPar;
			break;

		case WM_INITDIALOG:
			#if defined(OWN_TABCTRL)
				TabCtrlInit(hDlg, hFont);
			#endif
			if (DlgRect.left != -10000)
				SetWindowPos(hDlg, 0, DlgRect.left, DlgRect.top, 0, 0,
							 SWP_NOZORDER | SWP_NOSIZE | SWP_NOREDRAW);
			SendMessage(GetDlgItem(hDlg, IDC_PROFILENAME),
						CB_INSERTSTRING, 0, (LPARAM)(LPCSTR)"(default)");
			SendMessage(GetDlgItem(hDlg, IDC_PROFILENAME), CB_SETCURSEL, 0, 0);
			ShowWindow(GetDlgItem(hDlg, IDC_PROFILENAME),	 SW_HIDE);
			ShowWindow(GetDlgItem(hDlg, IDC_PROFILEMANAGER), SW_HIDE);
			ShowWindow(GetDlgItem(hDlg, IDC_PROFILETITLE),	 SW_HIDE);
			hDefParent = hDlg;
			nDefId = IDOK;
			{	TC_ITEM		TcItem = {TCIF_TEXT};
				HWND		hctlTab = GetDlgItem(hDlg, IDC_SETTINGSTAB);
				static UINT	LabelsLang;

				if (LabelsLang != Language) {
					PSTR		p, pNext;
					static CHAR	Labels[100];

					LOADSTRING(hInst, 912, Labels, sizeof(Labels));
					LabelsLang = Language;
					pNext = Labels;
					i = 0;
					while (*(p = SeparateString(&pNext)))
						TabSheets[i++].Label = p;
				}
				i = 0;
				while (TabSheets[i].Dialog) {
					TcItem.pszText = TabSheets[i].Label;
					TcItem.cchTextMax = lstrlen(TcItem.pszText);
					TabCtrl_InsertItem(hctlTab, i, &TcItem);
					++i;
				}
				TabCtrl_SetCurSel(hctlTab, CurrTab);
				hwndSubSet = TabSheets[CurrTab].hDlg =
					CreateDialog(hInst, MAKEINTRES(TabSheets[CurrTab].Dialog),
								 /*hctlTab*/ hDlg, SubSetDlgProc);
				CheckDlgButton(hDlg, IDC_ONETIMESETTING, TRUE);
				ShowWindow(hwndSubSet, SW_SHOW);
				BringWindowToTop(hwndSubSet);
			}
			{	HWND hCtl;
				INT  MaxLoop = 6;

				Ctl = TabSheets[CurrTab].FirstControl;
				do {
					hCtl = GetDlgItem(hwndSubSet, Ctl);
					if (GetWindowLong(hCtl, GWL_STYLE) & WS_TABSTOP) break;
					++Ctl;
				} while (--MaxLoop);
				PostMessage(hDlg, WM_COMMAND, 4567, (LPARAM)(UINT)hCtl);
			}
			return (FALSE);

		case WM_COMMAND:
			Ctl = 0;
			switch (COMMAND) {
				case 4567:
					ShowWindow(GetDlgItem(hDlg, IDC_PROFILENAME),	 SW_HIDE);
					ShowWindow(GetDlgItem(hDlg, IDC_PROFILEMANAGER), SW_HIDE);
					ShowWindow(GetDlgItem(hDlg, IDC_PROFILETITLE),	 SW_HIDE);
					/*...already hidden, but hide once more for Wine*/
					SetFocus((HWND)lPar);
					CheckDefButton();
					break;

				case 4568:
					CheckDefButton();
					return (TRUE);

				case 102:
					/*should not occur, only happens in NT 3.51*/
					QueryTime = GetTickCount();
					wsprintf(QueryString, "[102]");
					wPar = IDOK;
					/*FALLTHROUGH*/
				case ID_APPLY:
				case IDOK:
					{	INT  Curr=CurrTab;
						BOOL Save=IsDlgButtonChecked(hDlg, IDC_SAVEPERMANENTLY);
						HWND hctlTab = GetDlgItem(hDlg, IDC_SETTINGSTAB);
						UINT OldLang = Language;

						if (Save) for (i=0; TabSheets[i].Dialog; ++i)
							if (TabSheets[i].hDlg && TabSheets[i].Callback) {
								CurrTab = i;
								DlgResult = FALSE;
								SendMessage(TabSheets[i].hDlg,
											WM_COMMAND, IDC_STORE, 0);
								if (!DlgResult) {
									if (i != Curr) {
										ShowWindow(TabSheets[Curr].hDlg,
												   SW_HIDE);
										ShowWindow(TabSheets[i].hDlg, SW_SHOW);
										BringWindowToTop(TabSheets[i].hDlg);
										TabCtrl_SetCurSel(hctlTab, i);
										CheckDefButton();
									}
									return (FALSE);
								}
							}
						for (i=0; TabSheets[i].Dialog; ++i)
							if (TabSheets[i].hDlg && TabSheets[i].Callback) {
								CurrTab = i;
								DlgResult = FALSE;
								SendMessage(TabSheets[i].hDlg,
											WM_COMMAND, wPar, lPar);
								if (!DlgResult) {
									if (i != Curr) {
										ShowWindow(TabSheets[Curr].hDlg,
												   SW_HIDE);
										ShowWindow(TabSheets[i].hDlg, SW_SHOW);
										BringWindowToTop(TabSheets[i].hDlg);
										TabCtrl_SetCurSel(hctlTab, i);
										CheckDefButton();
									}
									return (FALSE);
								}
							}
						CurrTab = Curr;
						if (COMMAND == ID_APPLY) {
							if (Language == OldLang) return (TRUE);
							Restart = TRUE;
						}
					}
					/*FALLTHROUGH*/
				case IDCANCEL:
					hwndSubSet = 0;
					{	INT Curr = CurrTab;

						for (i=0; TabSheets[i].Dialog; ++i) {
							if (TabSheets[i].hDlg) {
								CurrTab = i;
								SendMessage(TabSheets[i].hDlg, uMsg, wPar,lPar);
								DestroyWindow(TabSheets[i].hDlg);
								TabSheets[i].hDlg = 0;
							}
						}
						CurrTab = Curr;
					}
					hFont = 0;
					PostMessage(hDlg, WM_CLOSE, 0, 0);
					/*FALLTHROUGH*/
				default:
					return (TRUE);

				case 100:
					Ctl = TabSheets[CurrTab].LastControl;
					Inc = -1;
					break;
				case 101:
					Ctl = TabSheets[CurrTab].FirstControl;
					Inc = 1;
			}
			if (hwndSubSet && Ctl) {
				HWND hCtl;
				INT  MaxLoop = 6;

				do {
					hCtl = GetDlgItem(hwndSubSet, Ctl);
					if (GetWindowLong(hCtl, GWL_STYLE) & WS_TABSTOP) break;
					Ctl += Inc;
				} while (--MaxLoop);
				if (Ctl == IDC_BITMAP)
					hCtl = GetDlgItem(hwndSubSet, Ctl=IDC_BITMAPBROWSE);
				/*...special hack: bitmap radio button is not the last control
				 *   on colors sub-dialog, set focus to following pushbutton.
				 */
				SetFocus(hCtl);
				CheckDefButton();
			}
			return (TRUE);

		case WM_NOTIFY:
			{	LPNMHDR	pNmHdr = (LPNMHDR)lPar;
				INT		NewTab;

				if (pNmHdr->code == TCN_SELCHANGE) {
					/*Tab sheet change...*/
					NewTab = (int)
						SendMessage(pNmHdr->hwndFrom, TCM_GETCURSEL, 0, 0);
					SwitchTab(pNmHdr->hwndFrom, NewTab);
				}
			}
			break;

		case WM_ACTIVATE:
			PostMessage(hDlg, WM_COMMAND, 4568, 0);
			break;

		case WM_CLOSE:
			if (hwndSubSet) SendMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
			if (hwndSettings) {
				EnableWindow(hwndMain, TRUE);
				SetFocus(hwndMain);
				GetWindowRect(hDlg, &DlgRect);
				DestroyWindow(hDlg);
				hwndSettings = 0;
				if (Restart) {
					PostMessage(hwndMain, WM_COMMAND, ID_SETTINGS, 0);
					Restart = FALSE;
				}
			}
			return (TRUE);
	}
	return (FALSE);
}

void Settings(void)
{
	if (!SubSetDlgProc) {
		 SubSetDlgProc =
			(DLGPROC)MakeProcInstance((FARPROC)SubSetCallback, hInst);
		 SettingsDlgProc =
			(DLGPROC)MakeProcInstance((FARPROC)SettingsCallback, hInst);
	}
	hwndSettings = CreateDialog(hInst, MAKEINTRES(IDD_SETTINGS),
								hwndMain, SettingsDlgProc);
	if (!hwndSettings) {
		return;
	}
	ShowWindow(hwndSettings, SW_SHOW);
	EnableWindow(hwndMain, FALSE);
}

BOOL IsSettingsDialogMessage(PMSG pMsg)
{	HWND hParent;

	if (!hwndSettings || !hwndSubSet || (hParent = pMsg->hwnd)==0)
		return (FALSE);
	if (hParent!=hwndSettings && hParent!=hwndSubSet) {
		hParent = GetParent(hParent);
		if (hParent!=hwndSettings && hParent!=hwndSubSet) {
			if (hParent != 0) hParent = GetParent(hParent);
			if (hParent!=hwndSettings && hParent!=hwndSubSet) return (FALSE);
		}
	}
	switch (pMsg->message) {
		case WM_KEYDOWN:
		case WM_CHAR:
			{	WPARAM		 wPar = pMsg->wParam;
				extern char  QueryString[150];
				extern DWORD QueryTime;

				if (wPar==VK_TAB) {
					if (GetKeyState(VK_CONTROL)<0) {
						HWND hctlTab = GetDlgItem(hwndSettings,IDC_SETTINGSTAB);

						if (hctlTab) {
							BOOL FocusOnTab = GetFocus()==hctlTab;
							INT  NewTab;

							if (!FocusOnTab) FocusOnFirstControl = TRUE;
							if (GetKeyState(VK_SHIFT) < 0) {
								if	  ((NewTab = CurrTab-1) == 0-1) NewTab = 6;
							} else if ((NewTab = CurrTab+1) == 6+1) NewTab = 0;
							TabCtrl_SetCurSel(hctlTab, NewTab);
							SwitchTab(hctlTab, NewTab);
							PostMessage(hwndSettings, WM_COMMAND,
										FocusOnTab ? 4567 : 4568,
										(LPARAM)(UINT)hctlTab);
							FocusOnFirstControl = FALSE;
							return (TRUE);
						}
					}
					PostMessage(hwndSettings, WM_COMMAND, 4568, 0);
				}
				if (wPar!=VK_RETURN && wPar!=VK_ESCAPE) break;
				QueryTime = GetTickCount();
				if (wPar==VK_RETURN) {
					hParent = hwndSettings;
					wPar = LOWORD(SendMessage(hParent, DM_GETDEFID, 0, 0));
					if (wPar == 102) {
						hParent = hwndSubSet;
						wPar = LOWORD(SendMessage(hParent, DM_GETDEFID, 0, 0));
						wsprintf(QueryString, "[4] message=0x%04x, wPar=0x%x",
								 pMsg->message, wPar);
					} else wsprintf(QueryString,
									"[5] message=0x%04x, wPar=%d",
									pMsg->message, wPar);
					switch (wPar) {
						case IDC_CHANGEFONT_TEXT:
						case IDC_CHANGEFONT_HEX:
						case IDC_BITMAPBROWSE:
						case IDC_PRINTFONT:
							hParent = hwndSubSet;
							break;
						default:
							hParent	= hDefParent;
							wPar	= nDefId;
					}
				} else {
					hParent = hwndSettings;
					wPar = IDCANCEL;
					wsprintf(QueryString,
							 "[3] message=0x%04x, wPar=0x%02x, WM_COMMAND, "
							 "wPar=%d", pMsg->message, pMsg->wParam, wPar);
				}
				PostMessage(hParent, WM_COMMAND, wPar, 0);
			}
			return (TRUE);

		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
			{	static CHAR Accelerators[8];
				static UINT AccelLang;

				if (AccelLang != Language) {
					LOADSTRING(hInst, 913, Accelerators, sizeof(Accelerators));
					AccelLang = Language;
				}
				if (strchr(Accelerators, pMsg->wParam) != NULL) {
					if (hParent==hwndSubSet) {
						hParent = hwndSettings;
						pMsg->hwnd = GetDlgItem(hwndSettings, IDOK);
					}
				} else if (hParent==hwndSettings) {
					hParent = hwndSubSet;
					pMsg->hwnd =
						GetDlgItem(hwndSubSet, TabSheets[CurrTab].FirstControl);
				}
			}
			break;

		case WM_KEYUP:
		case WM_LBUTTONDOWN:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
			PostMessage(hwndSettings, WM_COMMAND, 4568, 0);
	}
	return (IsDialogMessage(hParent, pMsg));
}
