/*************************************************************\
|*                                                           *|
|*   Program:              WinVi                             *|
|*   Author:               Raphael Molle                     *|
|*   Compiler:             Vc1.52c (this file 16-bit only)   *|
|*   --------------------- -------------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin          *|
|*    Copyleft by GPL     See file COPYING for more         *|
|*                         info about terms of use           *|
|*                                                           *|
\*************************************************************/

/* revision log:
 *		3-Dec-2000	first publication of this source file
 */

#include <windows.h>
#include <string.h>
#include <malloc.h>
#include "WinVi.h"
#include "TabCtrl.h"
#include "Status.h"
#include "resource.h"
#include "ctl3d.h"

#if defined(OWN_TABCTRL)

extern INT    CurrTab;
extern HPEN   HilitePen, ShadowPen, BlackPen;
extern HBRUSH DlgBkBrush;

static struct {
	PSTR Label;
	INT  Position, Width;
} TabLabels[12];
static INT	 CurrentLabel, Base = 10;
static HFONT hTabFont;
static RECT  TabCtrlFocusRect;
static BOOL  TabCtrlHasFocus;

BOOL TabInsertItem(HWND hTab, INT Which, TC_ITEM *TcItem)
{	INT   Len = TcItem->cchTextMax;
	LPSTR Str = TcItem->pszText;

	(VOID)hTab;		/*parameter never used*/
	if (Len>0 && *Str==' ') {
		++Str;
		--Len;
	}
	if (Len>0 && Str[Len-1]==' ') --Len;
	memmove(TabLabels+Which+1, TabLabels+Which, sizeof(TabLabels[0])
			* (sizeof(TabLabels)/sizeof(TabLabels[0])-Which-1));
	TabLabels[Which].Label = malloc(Len + 1);
	if (TabLabels[Which].Label != NULL) {
		HDC hDC;

		_fmemcpy(TabLabels[Which].Label, Str, Len);
		TabLabels[Which].Label[Len] = '\0';
		TabLabels[Which].Position = Which
			? TabLabels[Which-1].Position + TabLabels[Which-1].Width : 2;
		TabLabels[Which].Width = 24;
		if (hTabFont) {
			hDC = GetDC(NULL);
			if (hDC) {
				HFONT OldFont;
				SIZE  Size;

				OldFont = SelectObject(hDC, hTabFont);
				GetTextExtentPoint(hDC, Str, Len, &Size);
				SelectObject(hDC, OldFont);
				ReleaseDC(NULL, hDC);
				TabLabels[Which].Width = Size.cx + 15;
				Base = Size.cy + 7;
			}
		}
		while (TabLabels[++Which].Label != NULL)
			TabLabels[Which].Position
				= TabLabels[Which-1].Position + TabLabels[Which-1].Width;
	}
	return (TRUE);
}

BOOL TabSetCurSel(HWND hTab, INT Which)
{
	if (CurrentLabel != Which) {
		RECT  Rect;
		NMHDR NmHdr;

		Rect.left	 = TabLabels[CurrentLabel].Position - 2;
		Rect.top	 = 0;
		Rect.right	 = Rect.left + TabLabels[CurrentLabel].Width + 4;
		Rect.bottom	 = Base + 1;
		CurrentLabel = Which;
		InvalidateRect(hTab, &Rect, TRUE);
		Rect.left  = TabLabels[Which].Position - 2;
		Rect.right = Rect.left + TabLabels[Which].Width + 4;
		InvalidateRect(hTab, &Rect, TRUE);
		NmHdr.code	   = TCN_SELCHANGE;
		NmHdr.hwndFrom = hTab;
		SendMessage(GetParent(hTab), WM_NOTIFY, 0, (LPARAM)(LPNMHDR)&NmHdr);
	}
	return (TRUE);
}

BOOL TabCtrlPaint(HWND hWnd, HDC hDC, PPAINTSTRUCT Ps)
{	INT	  i;
	HPEN  OldPen;
	HFONT OldFont;
	RECT  Rect;

	(VOID)Ps;		/*parameter never used*/
	if (!hTabFont) hTabFont = (HFONT)SendMessage(hWnd, WM_GETFONT, 0, 0);
	GetClientRect(hWnd, &Rect);
	OldPen  = SelectObject(hDC, HilitePen);
	OldFont = SelectObject(hDC, hTabFont);
	SetBkMode(hDC, TRANSPARENT);
	#if defined(WIN32)
		SetTextColor(hDC, GetSysColor(COLOR_BTNTEXT));
	#else
		SetTextColor(hDC, GetSysColor(Ctl3dEnabled() ? COLOR_BTNTEXT
													 : COLOR_WINDOWTEXT));
	#endif
	MoveTo(hDC, 0, Rect.bottom);
	LineTo(hDC, 0, Base);
	if (CurrTab) LineTo(hDC, 2, Base);
	for (i=0; TabLabels[i].Label!=NULL; ++i) {
		if (i == CurrTab) {
			TextOut(hDC, TabLabels[i].Position+7, 4,
					TabLabels[i].Label, lstrlen(TabLabels[i].Label));
			TabCtrlFocusRect.left = TabLabels[i].Position+2;
			TabCtrlFocusRect.right = TabLabels[i].Position+TabLabels[i].Width-3;
			TabCtrlFocusRect.top = 3;
			TabCtrlFocusRect.bottom = Base-1;
			if (TabCtrlHasFocus) DrawFocusRect(hDC, &TabCtrlFocusRect);
			/*extend borders of tab by 2 pixels to the left, up and right...*/
			SelectObject(hDC, HilitePen);
			MoveTo(hDC, TabLabels[i].Position-2, Base);
			LineTo(hDC, TabLabels[i].Position-2, 2);
			LineTo(hDC, TabLabels[i].Position,   0);
			LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width,   0);
			SelectObject(hDC, BlackPen);
			MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width,   1);
			LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width+1, 2);
			LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width+1, Base);
			SelectObject(hDC, ShadowPen);
			MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width,   Base);
			LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width,   1);
			SelectObject(hDC, HilitePen);
		} else {
			TextOut(hDC, TabLabels[i].Position+7, 5,
					TabLabels[i].Label, lstrlen(TabLabels[i].Label));
			SelectObject(hDC, HilitePen);
			if (i == CurrTab+1) MoveTo(hDC, TabLabels[i].Position+2, 2);
			else {
				MoveTo(hDC, TabLabels[i].Position,   Base);
				LineTo(hDC, TabLabels[i].Position,   4);
				LineTo(hDC, TabLabels[i].Position+2, 2);
			}
			LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width-2, 2);
			if (i == CurrTab-1) {
				MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width-2, Base);
			} else {
				SelectObject(hDC, BlackPen);
				MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width-2, 3);
				LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width-1, 4);
				LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width-1, Base);
				SelectObject(hDC, ShadowPen);
				MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width-2, Base);
				LineTo(hDC, TabLabels[i].Position+TabLabels[i].Width-2, 3);
				SelectObject(hDC, HilitePen);
				MoveTo(hDC, TabLabels[i].Position+TabLabels[i].Width,   Base);
			}
			LineTo(hDC, TabLabels[i].Position, Base);
		}
	}
	MoveTo(hDC, TabLabels[i-1].Position+TabLabels[i-1].Width+1, Base);
	LineTo(hDC, Rect.right-1, Base);
	SelectObject(hDC, BlackPen);
	LineTo(hDC, Rect.right-1, Rect.bottom-1);
	LineTo(hDC, -1,			  Rect.bottom-1);
	SelectObject(hDC, ShadowPen);
	MoveTo(hDC, 1,			  Rect.bottom-2);
	LineTo(hDC, Rect.right-2, Rect.bottom-2);
	LineTo(hDC, Rect.right-2, Base);
	if (OldPen)  SelectObject(hDC, OldPen);
	if (OldFont) SelectObject(hDC, OldFont);
	return (TRUE);
}

LRESULT CALLBACK TabCtrlProc(HWND hWnd, UINT uMsg, WPARAM wPar, LPARAM lPar)
{
	switch(uMsg) {
		case WM_PAINT:
			{	static PAINTSTRUCT Ps;
				HDC				   hDC = BeginPaint(hWnd, &Ps);

				TabCtrlPaint(hWnd, hDC, &Ps);
				EndPaint(hWnd, &Ps);
			}
			break;

		case WM_ERASEBKGND:
			{	HDC  hDC = (HDC)wPar;
				RECT r;

				GetClientRect(hWnd, &r);
				r.bottom = Base + 1;
				FillRect(hDC, &r, DlgBkBrush);
			}
			return (TRUE);

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
			if (HIWORD(lPar) < (WORD)Base) {
				INT i;

				for (i=0; TabLabels[i].Label!=NULL; ++i) {
					if (LOWORD(lPar) >= (WORD)TabLabels[i].Position
							&& LOWORD(lPar) < (WORD)(TabLabels[i].Position
													 + TabLabels[i].Width)) {
						if (i == CurrentLabel) SetFocus(hWnd);
						else TabSetCurSel(hWnd, i);
						break;
					}
				}
			}
			break;

		case WM_KEYDOWN:
			if (wPar == VK_LEFT) {
				if (CurrentLabel) TabSetCurSel(hWnd, CurrentLabel-1);
			} else if (wPar==VK_RIGHT && TabLabels[CurrentLabel+1].Label!=NULL)
				TabSetCurSel(hWnd, CurrentLabel+1);
			break;

		case TCM_GETCURSEL:
			return (CurrentLabel);

		case WM_SETFOCUS:
			TabCtrlHasFocus = TRUE;
			InvalidateRect(hWnd, &TabCtrlFocusRect, FALSE);
			break;

		case WM_KILLFOCUS:
			TabCtrlHasFocus = FALSE;
			InvalidateRect(hWnd, &TabCtrlFocusRect, TRUE);
			break;

		case WM_GETDLGCODE:
			return (DLGC_WANTARROWS);

		default:
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));
	}
	return (0);
}

BOOL TabCtrlInit(HWND hParent, HFONT hFont)
{	INT i;

	(VOID)hParent;	/*parameter never used*/
	for (i=0; TabLabels[i].Label!=NULL; ++i) {
		free(TabLabels[i].Label);
		TabLabels[i].Label = NULL;
	}
	hTabFont = hFont;
	return (TRUE);
}

#endif
