/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <string.h>
#include <assert.h>
#include <malloc.h>
#include "resource.h"
#include "winvi.h"
#include "page.h"
#include "paint.h"
#include "status.h"
#include "toolbar.h"

#if defined(WIN32)
#	define ROW_WIDTH	   338
#	define ROW_HEIGHT	   22
#	define NEXT_ROW		   ROW_HEIGHT
#	define PUSHED_LOCKED   ROW_WIDTH
#	define PUSHED_DISABLED (ROW_WIDTH+24)
#else
#	define ROW_WIDTH	   339
#	define ROW_HEIGHT	   22
#	define NEXT_ROW		   (ROW_HEIGHT-1)
#	define PUSHED_LOCKED   (ROW_WIDTH-1)
#	define PUSHED_DISABLED (ROW_WIDTH-1+24)
#	if !defined(WS_EX_TOOLWINDOW)
#		define  WS_EX_TOOLWINDOW 0x00000080L
#	endif
#endif

struct {
	int  Id, xPos, Width, SwitchOff;
	BOOL Pushed, Locked, Disabled;
	PSTR Tip, StatusTip;
} ToolButtons[] = {
	{ID_NEW,	  0,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_OPEN,	 24,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_SAVE,	 48,24,	 -1,	 FALSE,FALSE,TRUE,  NULL,NULL},
	{ID_CUT,	 82,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_COPY,	106,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_PASTE,	130,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_UNDO,	164,24,	 -1,	 FALSE,FALSE,TRUE,  NULL,NULL},
	{ID_SEARCH,	198,24,	 -1,	 FALSE,FALSE,FALSE, NULL,NULL},
	{ID_SRCHAGN,222,24,	 -1,	 FALSE,FALSE,TRUE,  NULL,NULL},
	{ID_ANSI,	256,24, IDB_OEM, TRUE, TRUE, FALSE, NULL,NULL},
	{ID_OEM,	280,24, IDB_ANSI,FALSE,FALSE,FALSE, NULL,NULL},
	{ID_HEX,	314,24,	IDB_HEX, FALSE,FALSE,FALSE, NULL,NULL},
	{0,0,0,0,0,0,0,0}
};

void SetToolbarColor(LPRGBQUAD lpRgbQuad, int SysColor)
{	COLORREF   ColorRef = GetSysColor(SysColor);

	lpRgbQuad->rgbRed   = GetRValue(ColorRef);
	lpRgbQuad->rgbGreen = GetGValue(ColorRef);
	lpRgbQuad->rgbBlue  = GetBValue(ColorRef);
}

HBITMAP Buttons;

void PrepareBitmap(HDC hdcScreen)
{	HRSRC		 BitmapResource;
	HGLOBAL		 LoadedBitmap = NULL;
	LPBITMAPINFO LockedBitmap = NULL, NewBitmap = NULL;

	yButtonOffset = YBUTTON_OFFSET;
	nToolbarHeight = TOOLS_HEIGHT;
	if (WinVersion >= MAKEWORD(95,3)) {
		++yButtonOffset;
		++nToolbarHeight;
	}
	if (Buttons) {
		DeleteObject(Buttons);
		Buttons = NULL;
	}

	/*try to create a bitmap with 3D system colors...*/
	do {	/*for break only*/
		BitmapResource = FindResource(hInst, MAKEINTRESOURCE(100), RT_BITMAP);
		if (!BitmapResource) break;
		LoadedBitmap = LoadResource(hInst, BitmapResource);
		if (!LoadedBitmap) break;
		LockedBitmap = (LPBITMAPINFO)LockResource(LoadedBitmap);
		if (!LockedBitmap) break;
		{	INT Size = (INT)SizeofResource(hInst, BitmapResource);

			if (Size < 1000 || Size > 100000) break;
			NewBitmap = _fcalloc(1, Size);
			if (!NewBitmap) break;
			hmemcpy(NewBitmap, LockedBitmap, Size);
		}

#		if defined(WIN32)
			/*bmpColors[] before redefinition:
			 *  6=RGB(000,128,128)	dark cyan
			 *  7=RGB(192,192,192)	light gray,
			 *  8=RGB(128,128,128)	dark gray,
			 * 14=RGB(000,255,255)	light cyan
			 */
			NewBitmap->bmiColors[6] = LockedBitmap->bmiColors[7];
			SetToolbarColor(&NewBitmap->bmiColors[7],  COLOR_BTNFACE);
			SetToolbarColor(&NewBitmap->bmiColors[8],  COLOR_BTNSHADOW);
			SetToolbarColor(&NewBitmap->bmiColors[14], COLOR_BTNHIGHLIGHT);
#		else
			/*bmpColors[] before redefinition:
			 *  0=RGB(000,000,000)	black,
			 *  4=RGB(000,000,128)	dark blue,
			 *  7=RGB(192,192,192)	light gray,
			 *  8=RGB(128,128,128)	dark gray,
			 * 15=RGB(255,255,255)	white
			 */
			SetToolbarColor(&NewBitmap->bmiColors[0],  COLOR_WINDOWFRAME);
			SetToolbarColor(&NewBitmap->bmiColors[4],  COLOR_BTNTEXT);
			SetToolbarColor(&NewBitmap->bmiColors[7],  COLOR_BTNFACE);
			SetToolbarColor(&NewBitmap->bmiColors[8],  COLOR_BTNSHADOW);
			SetToolbarColor(&NewBitmap->bmiColors[15], COLOR_BTNHIGHLIGHT);
#		endif

		Buttons = CreateDIBitmap(hdcScreen, &NewBitmap->bmiHeader, CBM_INIT,
								 &NewBitmap->bmiColors[16], NewBitmap,
								 DIB_RGB_COLORS);

	} while (FALSE);
	if (NewBitmap)    _ffree(NewBitmap);
	if (LockedBitmap) (VOID)UnlockResource(LoadedBitmap);
	if (LoadedBitmap) FreeResource(LoadedBitmap);
	if (Buttons) return;

	/*system color bitmap creation failed, use normal bitmap resource*/
	Buttons = LoadBitmap(hInst, MAKEINTRESOURCE(100));
}

void RepaintButton(int Which)
{
	if (hwndMain) {
		RECT r;
		r.left   = XBUTTON_OFFSET + ToolButtons[Which].xPos;
		r.top    = yButtonOffset;
		r.right  = r.left + ToolButtons[Which].Width;
		r.bottom = r.top + ROW_HEIGHT;
		InvalidateRect(hwndMain, &r, FALSE);
		UpdateWindow(hwndMain);
	}
}

static char	TipTexts[200], StatusTips1[200], StatusTips2[160], StatusTips3[160];

void LoadTips(void)
{	PSTR p = TipTexts;
	int  i;

	LOADSTRING(hInst, 905, TipTexts,	sizeof(TipTexts));
	LOADSTRING(hInst, 906, StatusTips1,	sizeof(StatusTips1));
	LOADSTRING(hInst, 907, StatusTips2,	sizeof(StatusTips2));
	LOADSTRING(hInst, 908, StatusTips3,	sizeof(StatusTips3));
	for (i=0; ToolButtons[i].Id; ++i)
		ToolButtons[i].Tip		 = SeparateString(&p);
	p = StatusTips1;
	for (i=0; ToolButtons[i].Id != ID_UNDO; ++i)
		ToolButtons[i].StatusTip = SeparateString(&p);
	p = StatusTips2;
	for (; ToolButtons[i].Id != ID_ANSI; ++i)
		ToolButtons[i].StatusTip = SeparateString(&p);
	p = StatusTips3;
	for (; ToolButtons[i].Id; ++i)
		ToolButtons[i].StatusTip = SeparateString(&p);
}

void PaintTools(HDC hDC, PRECT pRect)
{	int			 i;
	HDC			 hCompDC;
	static POINT Line1[2], Line2[2], Line3[2];
	HPEN		 OldPen;
	RECT		 Background;

	if (!*TipTexts) LoadTips();
	Background.left   = pRect->left;
	Background.top	  = 1;
	Background.right  = pRect->right;
	Background.bottom = nToolbarHeight - 2;
	FillRect(hDC, &Background, FaceBrush);
	Line1[1].x = Line2[1].x = Line3[1].x = pRect->right;
	Line2[0].y = Line2[1].y = nToolbarHeight - 2;
	Line3[0].y = Line3[1].y = nToolbarHeight - 1;
	OldPen = SelectObject(hDC, ShadowPen);
	if (WinVersion >= MAKEWORD(95,3)) {
		Line1[0].y = Line1[1].y = 0;
		Polyline(hDC, Line1, 2);
		Line1[0].y = ++Line1[1].y;
	}
	Polyline(hDC, Line2, 2);
	SelectObject(hDC, HilitePen);
	Polyline(hDC, Line1, 2);
	SelectObject(hDC, BlackPen);
	Polyline(hDC, Line3, 2);
	SelectObject(hDC, OldPen);
	hCompDC = CreateCompatibleDC(hDC);
	if (hCompDC) {
		HBITMAP OldBitmap = SelectObject(hCompDC, Buttons);
		BitBlt(hDC, XBUTTON_OFFSET,  yButtonOffset, ROW_WIDTH, ROW_HEIGHT,
			   hCompDC, 0, Disabled>1 ? 2*NEXT_ROW : 0, SRCCOPY);
		SelectObject(hCompDC, OldBitmap);
		DeleteDC(hCompDC);
	}
	for (i=0; ToolButtons[i].Id; ++i)
		if (ToolButtons[i].Pushed || ToolButtons[i].Disabled && Disabled<=1) {
			HDC hDC = GetDC(hwndMain);

			if (hDC) {
				HDC hCompDC = CreateCompatibleDC(hDC);

				if (hCompDC) {
					int		SrcX = ToolButtons[i].xPos, SrcY;
					HBITMAP	OldBitmap = SelectObject(hCompDC, Buttons);

					if (Disabled>1 || ToolButtons[i].Disabled) {
						SrcY = 2 * NEXT_ROW;
						if (ToolButtons[i].Pushed) {
							if (i == IDB_ANSI) {
								SrcX = PUSHED_DISABLED;
								SrcY = 0;
							} else if (i == IDB_OEM) {
								SrcX = PUSHED_DISABLED;
								SrcY = NEXT_ROW;
							} else if (i == IDB_HEX) {
								SrcX = PUSHED_DISABLED;
							}
						}
					} else {
						SrcY = NEXT_ROW;
						if (ToolButtons[i].Locked) {
							if (i == IDB_ANSI) {
								SrcX = PUSHED_LOCKED;
								SrcY = 0;
							} else if (i == IDB_OEM) {
								SrcX = PUSHED_LOCKED;
							} else if (i == IDB_HEX) {
								SrcX = PUSHED_LOCKED;
								SrcY = NEXT_ROW * 2;
							}
						}
					}
					BitBlt(hDC, XBUTTON_OFFSET + ToolButtons[i].xPos,
								yButtonOffset, ToolButtons[i].Width, ROW_HEIGHT,
						   hCompDC, SrcX, SrcY, SRCCOPY);
					SelectObject(hCompDC, OldBitmap);
					DeleteDC(hCompDC);
				}
				ReleaseDC(hwndMain, hDC);
			}
		}
}

static int DisabledCount = 0;

void Disable(BOOL Repaint)
{	HDC hDC, hCompDC;

	HideEditCaret();
	++DisabledCount;
	if (Disabled > Repaint) return;
	if (Repaint && (hDC = GetDC(hwndMain)) != 0) {
		hCompDC = CreateCompatibleDC(hDC);
		if (hCompDC) {
			HBITMAP	OldBitmap = SelectObject(hCompDC, Buttons);
			int		i;

			BitBlt(hDC, XBUTTON_OFFSET, yButtonOffset, ROW_WIDTH-1, ROW_HEIGHT,
				   hCompDC, 0, 2*NEXT_ROW, SRCCOPY);
			ToolButtons[IDB_OPEN].Pushed = ToolButtons[IDB_SAVE].Pushed = FALSE;
			if (!ToolButtons[i=IDB_ANSI].Pushed) ++i;
			BitBlt(hDC, XBUTTON_OFFSET + ToolButtons[i].xPos, yButtonOffset,
						ToolButtons[i].Width, ROW_HEIGHT,
				   hCompDC, PUSHED_DISABLED, i==IDB_OEM ?NEXT_ROW :0, SRCCOPY);
			if (ToolButtons[IDB_HEX].Pushed)
				BitBlt(hDC,
					   XBUTTON_OFFSET+ToolButtons[IDB_HEX].xPos, yButtonOffset,
					   ToolButtons[IDB_HEX].Width, ROW_HEIGHT,
					   hCompDC, PUSHED_DISABLED, 2*NEXT_ROW, SRCCOPY);
			SelectObject(hCompDC, OldBitmap);
			DeleteDC(hCompDC);
		}
		ReleaseDC(hwndMain, hDC);
	}
	SwitchCursor(SWC_HOURGLASS_ON);
	Disabled = Repaint + 1;
	Interruptable = TRUE;
}

void Enable(void)
{	static RECT r;
	HDC hDC;

	if (DisabledCount && --DisabledCount) return;
	if (Disabled > 1) {
		Disabled = 0;  /*must be reset before painting tool bottons*/
		r.right  = ClientRect.right;
		r.bottom = nToolbarHeight;
		hDC = GetDC(hwndMain);
		if (hDC) {
			PaintTools(hDC, &r);
			ReleaseDC(hwndMain, hDC);
		}
	} else Disabled = 0;
	Interruptable = FALSE;
	SwitchCursor(SWC_HOURGLASS_OFF);
}

BOOL IsToolButtonEnabled(int Which)
{
	return(!ToolButtons[Which].Disabled);
}

void EnableToolButton(int Which, BOOL Enable)
{
	if (ToolButtons[Which].Disabled != !Enable) {
		ToolButtons[Which].Disabled = !Enable;
		if (hwndMain) {
			RECT r;
			r.left   = XBUTTON_OFFSET + ToolButtons[Which].xPos;
			r.top    = yButtonOffset;
			r.right  = r.left + ToolButtons[Which].Width;
			r.bottom = r.top + ROW_HEIGHT;
			InvalidateRect(hwndMain, &r, FALSE);
			UpdateWindow(hwndMain);
		}
	}
}

void ToggleHexMode(void)
{	extern BOOL	HexEditTextSide;
	extern int	HexEditFirstNibble;
	extern int	HexTextHeight, TextTextHeight;
	extern BOOL IsFixedFont;

	HexEditTextSide = FALSE;
	HexEditFirstNibble = -1;
	if (HexEditMode == ToolButtons[IDB_HEX].Locked) {
		/*command got via menu or ex command, button not yet changed*/
		ToolButtons[IDB_HEX].Locked = ToolButtons[IDB_HEX].Pushed ^= TRUE;
		RepaintButton(IDB_HEX);
	}
	HideEditCaret();
	HexEditMode = ToolButtons[IDB_HEX].Locked;
	if (HexEditMode) {
		extern LONG xOffset;
		int		    BytePos = (int)CountBytes(&CurrPos) & 15;

		ScreenStart = CurrPos;
		GoBack(&ScreenStart, 16*CurrLine + BytePos);
		xOffset = 0;
		TextHeight = HexTextHeight;
	} else {
		int c, LinesBack = CurrLine;
		if (CharFlags[c = CharAt(&CurrPos)] & 1) {
			if (c == '\n') {
				/*check if position is between CR and LF...*/
				GoBack(&CurrPos, 1);
				if ((c = CharAt(&CurrPos)) != C_CRLF && c != C_EOF)
					Advance(&CurrPos, 1);
			}
			if (Mode==CommandMode && CharFlags[c = GoBackAndChar(&CurrPos)] & 1)
				if (c != C_EOF) CharAndAdvance(&CurrPos);
		}
		ScreenStart = CurrPos;
		do do; while (!(CharFlags[c = GoBackAndChar(&ScreenStart)] & 1));
		while (--LinesBack >= 0);
		if (c != C_EOF) CharAndAdvance(&ScreenStart);
		VerifyPageStructure();
		TextHeight = TextTextHeight;
	}
	AveCharWidth = 0;		/*set again in first paint operation*/
	IsFixedFont  = FALSE;	/*...*/
	AdjustWindowParts(ClientRect.bottom, ClientRect.bottom);
	NewScrollPos();
	InvalidateText();
	NewPosition(&CurrPos);
	ShowEditCaret();
	GetXPos(&CurrCol);
}

void DisableMenuItems(int MenuIndex, HMENU hMenu)
{
	switch (MenuIndex) {
		case 0:	EnableMenuItem(hMenu, ID_SAVE,	 ToolButtons[IDB_SAVE].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_ALTERNATE, *AltFileName=='\0'
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				break;
		case 1:	EnableMenuItem(hMenu, ID_TAG, !(CharFlags[CharAt(&CurrPos)] & 8)
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_TAGBACK, IsTagStackEmpty()
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_UNDO,	 ToolButtons[IDB_UNDO].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_CUT,	 ToolButtons[IDB_CUT].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_COPY,	 ToolButtons[IDB_COPY].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_DELETE, ToolButtons[IDB_CUT].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_PASTE,  ToolButtons[IDB_PASTE].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				break;
		case 2: EnableMenuItem(hMenu, ID_SEARCH,
							   ToolButtons[IDB_SEARCH].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
				EnableMenuItem(hMenu, ID_SRCHAGN,
							   ToolButtons[IDB_SRCHAGN].Disabled
							   ? MF_BYCOMMAND | MF_GRAYED : MF_BYCOMMAND);
		case 3: CheckMenuItem(hMenu, ID_ANSI,
							  ToolButtons[IDB_ANSI].Pushed
							  ? MF_BYCOMMAND | MF_CHECKED : MF_BYCOMMAND);
				CheckMenuItem(hMenu, ID_OEM,
							  ToolButtons[IDB_OEM].Pushed
							  ? MF_BYCOMMAND | MF_CHECKED : MF_BYCOMMAND);
				CheckMenuItem(hMenu, ID_HEX,
							  ToolButtons[IDB_HEX].Pushed
							  ? MF_BYCOMMAND | MF_CHECKED : MF_BYCOMMAND);
	}
}

void CheckClipboard(void)
{	RECT r;

	if (!SelectCount != ToolButtons[IDB_COPY].Disabled) {
		extern HWND hwndSearch;
		ToolButtons[IDB_COPY].Disabled ^= TRUE^FALSE;
		r.left  = ToolButtons[IDB_COPY].xPos;
		if (ViewOnlyFlag) {
			if (!ToolButtons[IDB_CUT].Disabled) {
				ToolButtons[IDB_CUT].Disabled = TRUE;
				r.left  = ToolButtons[IDB_CUT].xPos;
			}
		} else {
			ToolButtons[IDB_CUT].Disabled = ToolButtons[IDB_COPY].Disabled;
			r.left  = ToolButtons[IDB_CUT].xPos;
		}
		r.right = ToolButtons[IDB_COPY].xPos + ToolButtons[IDB_COPY].Width;
		if (hwndSearch) {
			HWND Button = GetDlgItem(hwndSearch, IDC_REPLACE);

			if (Button && IsWindowVisible(Button))
				EnableWindow(Button, SelectCount!=0);
		}
	} else r.left = 0;
	if ((!IsClipboardFormatAvailable(CF_TEXT) || ViewOnlyFlag)
			!= ToolButtons[IDB_PASTE].Disabled) {
		ToolButtons[IDB_PASTE].Disabled ^= TRUE^FALSE;
		if (!r.left) r.left = ToolButtons[IDB_PASTE].xPos;
		r.right = ToolButtons[IDB_PASTE].xPos + ToolButtons[IDB_PASTE].Width;
	}
	if (r.left && hwndMain) {
		r.left	+= XBUTTON_OFFSET;
		r.right += XBUTTON_OFFSET;
		r.bottom = (r.top = yButtonOffset+2) + ROW_HEIGHT-5;
		InvalidateRect(hwndMain, &r, FALSE);
	}
}

void SetCharSet(UINT CSet, UINT Flags)
	/*Flags: 1=Invalidate Text and show cursor
	 *		 2=Update toolbar buttons
	 */
{
	if (CharSet != CSet) {
		CharSet = CSet; /*0=Ansi, 1=Oem*/
		SearchNewCset();
		if (ToolButtons[IDB_ANSI].Pushed != (CSet==0)) Flags |= 2;
		if (Flags & 2) {
			RECT r;
			int i = IDB_ANSI + CSet;
			ToolButtons[i].Locked = ToolButtons[i].Pushed = TRUE;
			i = ToolButtons[i].SwitchOff;
			ToolButtons[i].Locked = ToolButtons[i].Pushed = FALSE;
			if (hwndMain) {
				r.left  = ToolButtons[IDB_ANSI].xPos + XBUTTON_OFFSET;
				r.right = ToolButtons[IDB_OEM].xPos  +
						  ToolButtons[IDB_OEM].Width + XBUTTON_OFFSET;
				r.bottom = (r.top = yButtonOffset)   + NEXT_ROW;
				InvalidateRect(hwndMain, &r, FALSE);
			}
		}
		if (Flags & 1) {
			InvalidateText();
			ShowEditCaret();
		}
	}
}

int	 ActButton = -1;
PSTR SaveStatus0String;
int	 SaveStatus0Style;
BOOL HideCmd;

static BOOL CheckToolButtonCaptureRelease(UINT uMsg)
{	extern BOOL SuppressStatusBeep;

	if (uMsg != WM_LBUTTONUP) return (FALSE);
	ActButton = -1;
	ReleaseCapture();
	if (hwndCmd) {
		HideCmd = FALSE;
		ShowWindow(hwndCmd, SW_SHOW);
		SetFocus(hwndCmd);
	}
	SuppressStatusBeep = TRUE;
	NewStatus(0, SaveStatus0String, SaveStatus0Style);
	SuppressStatusBeep = FALSE;
	return (TRUE);
}

int CursorButton(LPARAM lPar)
{	int i;

	for (i=0; ToolButtons[i].Id; ++i) {
		int x = XBUTTON_OFFSET + ToolButtons[i].xPos;
			if ((short)LOWORD(lPar) >= x &&
				(short)LOWORD(lPar) < x+ToolButtons[i].Width)
					return (i);
	}
	return (-1);
}

int		 TooltipSuccess, TooltipButton;
LPARAM	 TooltipPosition;
COLORREF TooltipFgColor, TooltipBgColor;
HWND	 TooltipWindow;

LRESULT CALLBACK TooltipProc(HWND hWnd, UINT uMsg, WPARAM wPar, LPARAM lPar)
{
	switch(uMsg) {
		case WM_ERASEBKGND:
			{	RECT   r;
				HBRUSH BkBrush;

				GetClientRect(hWnd, &r);
				BkBrush = CreateSolidBrush
								((COLORREF)GetWindowLong(hWnd, sizeof(LONG)));
				if (BkBrush) {
					FillRect((HDC)wPar, &r, BkBrush);
					DeleteObject(BkBrush);
				}
			}
			return (TRUE);

		case WM_PAINT:
			{	PAINTSTRUCT	ps;
				HDC			hDC = BeginPaint(hWnd, &ps);

				if (hDC) {
					HFONT hfontOld = SelectObject(hDC,
									(HFONT)GetWindowLong(hWnd, 2*sizeof(LONG)));
					char  Buf[300];

					SetTextColor(hDC, (COLORREF)GetWindowLong(hWnd, 0));
					SetBkMode(hDC, TRANSPARENT);
					GetWindowText(hWnd, Buf, sizeof(Buf));
					TextOut(hDC, 2, (int)GetWindowLong(hWnd, 3*sizeof(LONG)),
							Buf, lstrlen(Buf));
					SelectObject(hDC, hfontOld);
					EndPaint(hWnd, &ps);
				}
			}
			break;

		case WM_SETFOCUS:
			SetFocus(hwndMain);
			break;

		default:
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));
	}
	return (0);
}

int TooltipPopupTime = 500;

void TooltipTimer(void)
{	POINT pt;

	if (TooltipSuccess && TooltipButton >= 0)
		if (++TooltipSuccess == 2 && GetFocus() == hwndMain) {
			/*show tooltip window...*/
			HDC   hDC;
			SIZE  Size;
			if (!TooltipFont) {
				#if defined(WIN32)
				{	static NONCLIENTMETRICS ncmInfo;

					TooltipFgColor = GetSysColor(COLOR_INFOTEXT);
					TooltipBgColor = GetSysColor(COLOR_INFOBK);
					ncmInfo.cbSize = sizeof(ncmInfo);
					if (!SystemParametersInfo(SPI_GETNONCLIENTMETRICS,
											  0, &ncmInfo, 0)) {
						HDC   hDC;

						lstrcpyn(ncmInfo.lfStatusFont.lfFaceName,
								 "MS Sans Serif",
								 sizeof(ncmInfo.lfStatusFont.lfFaceName));
						hDC = GetDC(NULL);
						if (hDC) {
							ncmInfo.lfStatusFont.lfHeight =
								-MulDiv(ncmInfo.lfStatusFont.lfHeight,
										GetDeviceCaps(hDC, LOGPIXELSY), 72);
							ReleaseDC(NULL, hDC);
						} else ncmInfo.lfStatusFont.lfHeight = 10;
					}
					TooltipFont = CreateFontIndirect(&ncmInfo.lfStatusFont);
				}
				#else
				{	static LOGFONT LogFont;
					HDC hDC;

					TooltipFgColor = TooltipBgColor = RGB(0,0,0);
					lstrcpyn(LogFont.lfFaceName, "MS Sans Serif",
							 sizeof(LogFont.lfFaceName));
					if (hDC = GetDC(NULL)) {
						LogFont.lfHeight = -MulDiv(8,
											GetDeviceCaps(hDC, LOGPIXELSY), 72);
						ReleaseDC(NULL, hDC);
					} else LogFont.lfHeight = 10;
					TooltipFont = CreateFontIndirect(&LogFont);
				}
				#endif
				if (TooltipFgColor == TooltipBgColor) {
					TooltipFgColor = RGB(000,000,000);
					TooltipBgColor = RGB(255,255,192);
				}
				if (!TooltipFont) return;
			}
			hDC = GetDC(NULL);
			if (hDC) {
				HFONT hfontOld = SelectObject(hDC, TooltipFont);
				GetTextExtentPoint(hDC,    ToolButtons[TooltipButton].Tip,
								   lstrlen(ToolButtons[TooltipButton].Tip),
								   &Size);
				SelectObject(hDC, hfontOld);
				ReleaseDC(NULL, hDC);
			} else Size.cx = 0;
			if (GetKeyState(VK_LBUTTON) < 0) return;
			GetCursorPos(&pt);
			ScreenToClient(hwndMain, &pt);
			if (!Size.cx) return;
			if (	pt.y <  yButtonOffset ||
					pt.y >= yButtonOffset  + ROW_HEIGHT ||
					pt.x <  XBUTTON_OFFSET + ToolButtons[TooltipButton].xPos ||
					pt.x >= XBUTTON_OFFSET + ToolButtons[TooltipButton].xPos
										   + ToolButtons[TooltipButton].Width)
				return;
			ClientToScreen(hwndMain, &pt);
			TooltipWindow = CreateWindow("WinVi - ToolTip",
										 ToolButtons[TooltipButton].Tip,
										 WS_BORDER | WS_POPUP,
										 pt.x, pt.y+19, Size.cx+7, Size.cy+4,
										 0 /*no parent*/, NULL, hInst, NULL);
			SetWindowLong(TooltipWindow, 0*sizeof(LONG), TooltipFgColor);
			SetWindowLong(TooltipWindow, 1*sizeof(LONG), TooltipBgColor);
			SetWindowLong(TooltipWindow, 2*sizeof(LONG), (UINT)TooltipFont);
			SetWindowLong(TooltipWindow, 3*sizeof(LONG), 1);
			SetWindowLong(TooltipWindow, GWL_EXSTYLE,
						  GetWindowLong(TooltipWindow, GWL_EXSTYLE)
						  | WS_EX_TOOLWINDOW);
			ShowWindow(TooltipWindow, SW_SHOWNA);
			SetCapture(hwndMain);
			SetTimer(hwndMain, 101, 5000, NULL);
		} else {
			if (TooltipSuccess <  3) TooltipSuccess = 3;
			if (TooltipSuccess == 3) {
				if (TooltipWindow) {
					DestroyWindow(TooltipWindow);
					TooltipWindow = NULL;
					if (GetCapture() == hwndMain) ReleaseCapture();
				}
				TooltipPopupTime = 500;
				KillTimer(hwndMain, 101);
			}
		}
}

void TooltipHide(int i)
{
	TooltipSuccess = 2;
	TooltipTimer();
	if (i != -1) TooltipButton  = i;
	UpdateWindow(hwndMain);
}

BOOL MouseForToolbar(UINT uMsg, WPARAM wPar, LPARAM lPar)
{	BOOL IsWithin = (short)HIWORD(lPar) <  yButtonOffset + ROW_HEIGHT
				 && (short)HIWORD(lPar) >= yButtonOffset;

	(VOID)wPar;		/*parameter never used*/
	if (ActButton >= 0) {
		BOOL Within = IsWithin
						&& (short)LOWORD(lPar) >= XBUTTON_OFFSET
												+ ToolButtons[ActButton].xPos
						&& (short)LOWORD(lPar) <  XBUTTON_OFFSET
												+ ToolButtons[ActButton].Width
												+ ToolButtons[ActButton].xPos;
		if (ToolButtons[ActButton].Pushed && !Disabled) {
			HDC hDC;
			int a = ActButton;

			if (uMsg == WM_LBUTTONUP) {
				if (ToolButtons[ActButton].SwitchOff != -1) {
					if (Within) {
						PostMessage(hwndMain, WM_COMMAND,
									ToolButtons[ActButton].Id, 0L);
						a = ToolButtons[ActButton].SwitchOff;
						if (a == ActButton) ToolButtons[a].Locked ^= TRUE;
						else {
							ToolButtons[ActButton].Locked = TRUE;
							ToolButtons[a].Locked		  = FALSE;
						}
						hDC = GetDC(hwndMain);
						if (hDC) {
							HDC hCompDC = CreateCompatibleDC(hDC);

							if (hCompDC) {
								HBITMAP OldBitmap =
										SelectObject(hCompDC, Buttons);
								int y = 0;

								switch (ActButton) {
									case IDB_HEX:
										y += NEXT_ROW;
										/*FALLTHROUGH*/
									case IDB_OEM:
										y += NEXT_ROW;
										/*FALLTHROUGH*/
									case IDB_ANSI:
										BitBlt(hDC,XBUTTON_OFFSET +
												   ToolButtons[ActButton].xPos,
												   yButtonOffset,
												   ToolButtons[ActButton].Width,
												   ROW_HEIGHT,
											   hCompDC, PUSHED_LOCKED, y,
												   SRCCOPY);
										/*FALLTHROUGH*/
									default:
										break;
								}
								SelectObject(hCompDC, OldBitmap);
								DeleteDC(hCompDC);
							}
							ReleaseDC(hwndMain, hDC);
						}
					} else if (ToolButtons[ActButton].Locked)
						a = ToolButtons[ActButton].SwitchOff;
				} else PostMessage
						(hwndMain, WM_COMMAND, ToolButtons[ActButton].Id, 0L);
			} else if (uMsg != WM_MOUSEMOVE || Within
					   || ToolButtons[ActButton].Locked)
				return (TRUE);
			if (!ToolButtons[a].Locked) {
				ToolButtons[a].Pushed = FALSE;
				hDC = GetDC(hwndMain);
				if (hDC) {
					HDC hCompDC = CreateCompatibleDC(hDC);

					if (hCompDC) {
						HBITMAP OldBitmap = SelectObject(hCompDC, Buttons);

						BitBlt(hDC,
							   XBUTTON_OFFSET + ToolButtons[a].xPos,
							   yButtonOffset, ToolButtons[a].Width,
							   ROW_HEIGHT, hCompDC,
							   ToolButtons[a].xPos, 0, SRCCOPY);
						SelectObject(hCompDC, OldBitmap);
						DeleteDC(hCompDC);
					}
					ReleaseDC(hwndMain, hDC);
				}
			}
			CheckToolButtonCaptureRelease(uMsg);
			return (TRUE);
		}
		if (CheckToolButtonCaptureRelease(uMsg)) return (TRUE);
		if (!Within) return (TRUE);
		uMsg = WM_LBUTTONDOWN;	/*fake for next if condition*/
	}
	if (uMsg == WM_LBUTTONDOWN || uMsg == WM_LBUTTONDBLCLK) {
		if (IsWithin) {
			int i = CursorButton(lPar);
			if (i >= 0) {
				HDC hDC;
				TooltipHide(i);
				if (Disabled) {
					MessageBeep(MB_OK);
					return (TRUE);
				}
				if (ToolButtons[i].Disabled) return (TRUE);
				hDC = GetDC(hwndMain);
				if (hDC) {
					HDC hCompDC = CreateCompatibleDC(hDC);

					if (hCompDC) {
						HBITMAP OldBitmap = SelectObject(hCompDC, Buttons);
						int		x, y;
						x = ToolButtons[i].xPos;
						y = NEXT_ROW;
						if (ToolButtons[i].Locked) {
							if (i == IDB_ANSI) {
								x = PUSHED_LOCKED;
								y = 0;
							} else if (i == IDB_OEM) {
								x = PUSHED_LOCKED;
							} else if (i == IDB_HEX) {
								x = PUSHED_LOCKED;
								y = NEXT_ROW*2;
							}
						}
						BitBlt(hDC, XBUTTON_OFFSET + ToolButtons[i].xPos,
									yButtonOffset,
									ToolButtons[i].Width, ROW_HEIGHT,
							   hCompDC, x, y, SRCCOPY);
						SelectObject(hCompDC, OldBitmap);
						DeleteDC(hCompDC);
					}
					ReleaseDC(hwndMain, hDC);
				}
				if (ActButton == -1) {
					SaveStatus0String = StatusFields[0].Text;
					SaveStatus0Style  = StatusFields[0].Style;
					if (hwndCmd) {
						HideCmd = TRUE;
						SetFocus(hwndMain);
						ShowWindow(hwndCmd, SW_HIDE);
					}
					NewStatus(0, ToolButtons[i].StatusTip, NS_TOOLTIP);
					ActButton = i;
					SetCapture(hwndMain);
				}
				ToolButtons[i].Pushed = TRUE;
			}
		}
	} else if (IsWithin) {
		int Curr = CursorButton(lPar);
		if (TooltipSuccess <= 1) {
			if (Curr >= 0) {
				TooltipSuccess	= 1;
				TooltipPosition	= lPar;
				if (TooltipButton == -1) {
					TooltipButton = Curr;
					TooltipTimer();
				} else TooltipButton = Curr;
				SetTimer(hwndMain, 101, TooltipPopupTime, NULL);
			}
		} else if (Curr != TooltipButton) {
			if (TooltipButton >= 0) {
				TooltipSuccess = 2;
				TooltipTimer();
				TooltipSuccess = 0;
			}
			if ((TooltipButton	 = Curr) >= 0) {
				TooltipSuccess	 = 1;
				TooltipPosition	 = lPar;
				TooltipPopupTime = 20;
				SetTimer(hwndMain, 101, TooltipPopupTime, NULL);
			} else {
				TooltipPopupTime = 500;
				KillTimer(hwndMain, 101);
			}
		} else if (TooltipSuccess == 2) SetTimer(hwndMain, 101, 5000, NULL);
	} else if (TooltipSuccess) {
		TooltipButton = 0;
		TooltipSuccess = 2;
		TooltipTimer();
		TooltipSuccess = 0;
	}
	return (FALSE);
}
