/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#include <windows.h>
#include <shellapi.h>
#include <commdlg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <assert.h>
#include "ctl3d.h"
#include "winvi.h"
#include "colors.h"
#include "paint.h"
#include "resource.h"
#include "page.h"
#include "intercom.h"
#include "printh.h"
#include "toolbar.h"
#include "status.h"
#include "exec.h"
#include "tabctrl.h"

#if !defined(WM_MOUSEWHEEL)
	#define  WM_MOUSEWHEEL  0x020a
#endif
#if !defined(WM_CONTEXTMENU)
	#define  WM_CONTEXTMENU 0x007B
#endif

PCSTR	 ApplName = "WinVi";
PCSTR	 ClassName = "Ramo's WinVi";
RECT	 ClientRect, WholeWindow;
INT 	 x = 40, y = 60, Width = 400, Height = 300;
BOOL	 Maximized;
LONG	 PageThreshold = 500;
BOOL	 UnsafeSettings = FALSE;
BOOL	 RefreshFlag = FALSE;

BOOL AllocString(LPSTR *Destination, LPCSTR Source)
{	LPSTR p;

	if ((p = _fcalloc(lstrlen(Source) + 1, 1)) == NULL) return (FALSE);
	if (*Destination) _ffree(*Destination);
	lstrcpy(p, Source);
	*Destination = p;
	return(TRUE);
}

VOID MakeBackgroundBrushes(VOID)
{	HDC hDC;

	if (SelBkgndBrush) DeleteObject(SelBkgndBrush);
	SelBkgndBrush = CreateSolidBrush(SelBkgndColor);
	if (BkgndBrush) DeleteObject(BkgndBrush);
	hDC = GetDC(NULL);
	if (hDC) {
		BkgndBrush = CreateSolidBrush
					 (GetNearestColor((HDC)hDC, BackgroundColor));
		ReleaseDC(NULL, hDC);
	} else BkgndBrush = CreateSolidBrush(BackgroundColor);
}

extern BOOL	IsFixedFont;

BOOL SelectBkBitmap(HWND hParent)
{	PSTR				p;
	static CHAR			BitmapDir[260], Filter[50], Title[40];
	static OPENFILENAME	Ofn = {
		sizeof(OPENFILENAME), 0, 0, NULL, NULL, 0, 1,
		BitmapFilename, sizeof(BitmapFilename), NULL, 0, NULL, NULL,
		OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_NOCHANGEDIR,
		0, 0, ".bmp", 0, NULL, NULL
	};

	if (*BitmapFilename) lstrcpy(BitmapDir, BitmapFilename);
	else GetModuleFileName(hInst, BitmapDir, sizeof(BitmapDir));
	for (p=BitmapDir+lstrlen(BitmapDir); p>BitmapDir && *p!='/' && *p!='\\';
										 --p);
	*p = '\0';
	if (!*Filter) {
		LOADSTRING(hInst, 281, Title,  sizeof(Title));
		LOADSTRING(hInst, 292, Filter, sizeof(Filter));
		SeparateAllStrings(Filter);
	}
	Ofn.lpstrInitialDir = *BitmapDir ? (LPSTR)BitmapDir : (LPSTR)NULL;
	Ofn.hwndOwner	= hParent;
	Ofn.hInstance	= hInst;
	Ofn.lpstrFilter	= Filter;
	Ofn.lpstrTitle	= Title;
	if (!GetOpenFileName(&Ofn)) return (FALSE);
	return (TRUE);
}

BOOL GetEditLong(HWND hDlg, INT EditId, INT Min, LONG Max, LONG *Value)
{	LONG i;
	PSTR p;
	static CHAR Buf[30];

	GetDlgItemText(hDlg, EditId, Buf, sizeof(Buf));
	i = strtol(Buf, &p, 10);
	while (*p == ' ') ++p;
	if (*p) {
		LOADSTRING(hInst, 309, Buf, sizeof(Buf));
		MessageBox(hDlg, Buf, NULL, MB_OK | MB_ICONSTOP);
		SetFocus(GetDlgItem(hDlg, EditId));
		return (FALSE);
	}
	if (i < Min || i > Max) {
		LOADSTRING(hInst, 310 + (i >= Min), Buf, sizeof(Buf));
		MessageBox(hDlg, Buf, NULL, MB_ICONSTOP | MB_OK);
		SetFocus(GetDlgItem(hDlg, EditId));
		return (FALSE);
	}
	*Value = i;
	return (TRUE);
}

#define	SETFLAG(f, id)								\
	if (f != (BOOL)IsDlgButtonChecked(hDlg, id)) {	\
		f ^= TRUE;									\
	}
#define	SETFLAG_AND_UPDATE(f, id, cond)				\
	if (f != (BOOL)IsDlgButtonChecked(hDlg, id)) {	\
		f ^= TRUE;									\
		if (cond) Update = TRUE;					\
	}

BOOL DisplaySettingsCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	CHAR Buf[260];
	LONG i;

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			if (TerseFlag)		CheckDlgButton(hDlg, IDC_TERSE,			TRUE);
			if (BreakLineFlag)	CheckDlgButton(hDlg, IDC_BREAKLINES,	TRUE);
			if (AutoRestoreFlag)CheckDlgButton(hDlg, IDC_AUTORESTORE,	TRUE);
			if (ShowMatchFlag)	CheckDlgButton(hDlg, IDC_SHOWMATCH,		TRUE);
			if (NumberFlag)		CheckDlgButton(hDlg, IDC_LINENUMBERS,	TRUE);
			if (HexNumberFlag)	CheckDlgButton(hDlg, IDC_HEXADDRESSES,	TRUE);
			if (RefreshFlag)	CheckDlgButton(hDlg, IDC_REFRESHBKG,	TRUE);
			wsprintf(Buf, "%u", TabStop);
			SetDlgItemText(hDlg, IDC_TABSTOP, Buf);
			wsprintf(Buf, "%u", MaxScroll);
			SetDlgItemText(hDlg, IDC_MAXSCROLL, Buf);
			return (TRUE);

		case WM_COMMAND:
			switch (COMMAND) {
				case IDOK: case IDC_STORE: case ID_APPLY:
					SETFLAG(TerseFlag,		 IDC_TERSE);
					SETFLAG(AutoRestoreFlag, IDC_AUTORESTORE);
					SETFLAG(ShowMatchFlag,	 IDC_SHOWMATCH);
					SETFLAG(RefreshFlag,	 IDC_REFRESHBKG);
					if (BreakLineFlag !=
							(BOOL)IsDlgButtonChecked(hDlg, IDC_BREAKLINES)) {
						BreakLineFlag ^= TRUE;
						InvalidateText();
					}
					{	INT Update = FALSE;

						SETFLAG_AND_UPDATE(NumberFlag,	 IDC_LINENUMBERS,
														 !HexEditMode);
						SETFLAG_AND_UPDATE(HexNumberFlag,IDC_HEXADDRESSES,
														 HexEditMode);
						if (!GetEditLong(hDlg, IDC_TABSTOP, 1, 80, &i))
							return (FALSE);
						if (i != TabStop) {
							TabStop = (INT)i;
							Update = TRUE;
						}
						if (!GetEditLong(hDlg, IDC_MAXSCROLL, 0, 80, &i))
							return (FALSE);
						if (i != MaxScroll) MaxScroll = (INT)i;
						if (Update) InvalidateText();
					}
					if (COMMAND == IDC_STORE) SaveDisplaySettings();
					DlgResult = TRUE;
					if (COMMAND != IDOK) return (TRUE);
					/*FALLTHROUGH*/
				case IDCANCEL:
					EndDialog(hDlg, 0);
			}
			return (TRUE);
	}
	return (FALSE);
}

static VOID DisplaySettings(VOID)
{	static DLGPROC DlgProc;

	if (!DlgProc)
		 DlgProc
		  = (DLGPROC)MakeProcInstance((FARPROC)DisplaySettingsCallback, hInst);
	DialogBox(hInst, MAKEINTRES(IDD_DISPLAYSETTINGS), hwndMain, DlgProc);
}

BOOL EditSettingsCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	CHAR Buf[260];
	LONG i;

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			if (AutoIndentFlag)	CheckDlgButton(hDlg, IDC_AUTOINDENT,	TRUE);
			if (AutoWriteFlag)	CheckDlgButton(hDlg, IDC_AUTOWRITE,		TRUE);
			if (WrapScanFlag)	CheckDlgButton(hDlg, IDC_WRAPSCAN,		TRUE);
			if (WrapPosFlag)	CheckDlgButton(hDlg, IDC_NOTEPADNL,		TRUE);
			if (IgnoreCaseFlag)	CheckDlgButton(hDlg, IDC_IGNORECASE,	TRUE);
			if (DefaultInsFlag)	CheckDlgButton(hDlg, IDC_DEFAULTINSERT,	TRUE);
			if (ReadOnlyFlag)	CheckDlgButton(hDlg, IDC_READONLY,		TRUE);
			if (MagicFlag)		CheckDlgButton(hDlg, IDC_MAGIC,			TRUE);
			if (FullRowFlag)	CheckDlgButton(hDlg, IDC_FULLROWEXTEND, TRUE);
			if (UndoLimit == (ULONG)-1)
				LOADSTRING(hInst, 336, Buf, sizeof(Buf));
			else wsprintf(Buf, "%lu", UndoLimit/1024);
			SetDlgItemText(hDlg, IDC_UNDOLIMIT, Buf);
			wsprintf(Buf, "%u", ShiftWidth);
			SetDlgItemText(hDlg, IDC_SHIFTWIDTH, Buf);
			return (TRUE);

		case WM_COMMAND:
			switch (COMMAND) {

				case IDOK: case IDC_STORE: case ID_APPLY:
					SETFLAG(AutoIndentFlag,	 IDC_AUTOINDENT);
					SETFLAG(AutoWriteFlag,	 IDC_AUTOWRITE);
					SETFLAG(WrapScanFlag,	 IDC_WRAPSCAN);
					SETFLAG(WrapPosFlag,	 IDC_NOTEPADNL);
					SETFLAG(IgnoreCaseFlag,	 IDC_IGNORECASE);
					SETFLAG(DefaultInsFlag,  IDC_DEFAULTINSERT);
					SETFLAG(MagicFlag,		 IDC_MAGIC);
					SETFLAG(FullRowFlag,	 IDC_FULLROWEXTEND);
					if (ReadOnlyFlag !=
							(BOOL)IsDlgButtonChecked(hDlg, IDC_READONLY)) {
						ReadOnlyFlag ^= TRUE;
						ReadOnlyFile = ReadOnlyFlag;
					}
					if (GetDlgItemText(hDlg, IDC_UNDOLIMIT, Buf, sizeof(Buf))) {
						PSTR p;

						for (p=Buf; *p==' '; ++p);
						if (*p>='0' && *p<='9')
							if ((UndoLimit = strtol(p, NULL, 10)) >= 4194304)
								UndoLimit = (ULONG)-1;
							else UndoLimit *= 1024;
						else UndoLimit = (ULONG)-1;
					}
					if (!GetEditLong(hDlg, IDC_SHIFTWIDTH, 1, 80, &i))
						return (FALSE);
					if (i != ShiftWidth) ShiftWidth = (INT)i;
					if (COMMAND == IDC_STORE) SaveEditSettings();
					DlgResult = TRUE;
					if (COMMAND != IDOK) return (TRUE);
					/*FALLTHROUGH*/
				case IDCANCEL:
					EndDialog(hDlg, 0);
			}
			return (TRUE);
	}
	return (FALSE);
}

static VOID EditSettings(VOID)
{	static DLGPROC DlgProc;

	if (!DlgProc)
		 DlgProc
		  = (DLGPROC)MakeProcInstance((FARPROC)EditSettingsCallback, hInst);
	DialogBox(hInst, MAKEINTRES(IDD_EDITSETTINGS), hwndMain, DlgProc);
}

BOOL FileSettingsCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	CHAR Buf[260];
	LONG i;

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			if (LcaseFlag)		CheckDlgButton(hDlg, IDC_LOWERCASE,		TRUE);
			if (QuitIfMoreFlag)	CheckDlgButton(hDlg, IDC_MOREFILESQUIT,	TRUE);
			if (DefaultTmpFlag) {
				CheckDlgButton(hDlg, IDC_DEFAULTTMP, TRUE);
				EnableWindow(GetDlgItem(hDlg, IDC_TMPTITLE),  FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_TMPFOLDER), FALSE);
				GetSystemTmpDir(Buf, sizeof(Buf));
				AllocString(&TmpDirectory, Buf);
			}
			if (TmpDirectory) SetDlgItemText(hDlg, IDC_TMPFOLDER, TmpDirectory);
			if (Tags)		  SetDlgItemText(hDlg, IDC_TAGS,	  Tags);
			wsprintf(Buf, "%ld", PageThreshold);
			SetDlgItemText(hDlg, IDC_PAGE_THRESHOLD, Buf);
#if 1
			ShowWindow(GetDlgItem(hDlg, IDC_BKUPTITLE1),		SW_HIDE);
			ShowWindow(GetDlgItem(hDlg, IDC_BKUPTITLE2),		SW_HIDE);
			ShowWindow(GetDlgItem(hDlg, IDC_NUMBER_OF_BACKUPS),	SW_HIDE);
#endif
			return (TRUE);

		case WM_COMMAND:
			switch (COMMAND) {
				case IDC_DEFAULTTMP:
					{	BOOL Checked = IsDlgButtonChecked(hDlg, IDC_DEFAULTTMP);
						EnableWindow(GetDlgItem(hDlg, IDC_TMPTITLE),  !Checked);
						EnableWindow(GetDlgItem(hDlg, IDC_TMPFOLDER), !Checked);
					}
					break;

				case IDOK: case IDC_STORE: case ID_APPLY:
					SETFLAG(LcaseFlag,		 IDC_LOWERCASE);
					SETFLAG(QuitIfMoreFlag,	 IDC_MOREFILESQUIT);
					SETFLAG(DefaultTmpFlag,  IDC_DEFAULTTMP);
					GetDlgItemText(hDlg, IDC_TMPFOLDER, Buf, sizeof(Buf));
					AllocString(&TmpDirectory, Buf);
					GetDlgItemText(hDlg, IDC_TAGS, Buf, sizeof(Buf));
					AllocString(&Tags, Buf);
					if (!GetEditLong(hDlg, IDC_PAGE_THRESHOLD, 8, 32767, &i))
						return (FALSE);
					if (i != PageThreshold) PageThreshold = i;
					if (COMMAND == IDC_STORE) SaveFileSettings();
					DlgResult = TRUE;
					if (COMMAND != IDOK) return (TRUE);
					/*FALLTHROUGH*/
				case IDCANCEL:
					EndDialog(hDlg, 0);
			}
			return (TRUE);
	}
	return (FALSE);
}

static VOID SetMenuRadioStyle(VOID)
{
	#if defined(WIN32)
		static CHAR Buf[40];
		static MENUITEMINFO miiRadioMenu = {sizeof(MENUITEMINFO), MIIM_TYPE, 0,
											0,0,NULL,NULL,NULL,0,Buf,0};
		HMENU hMenu = GetMenu(hwndMain);

		miiRadioMenu.cch = sizeof(Buf);
		GetMenuItemInfo(hMenu, ID_ANSI, FALSE, &miiRadioMenu);
		miiRadioMenu.fType |= MFT_RADIOCHECK;
		SetMenuItemInfo(hMenu, ID_ANSI, FALSE, &miiRadioMenu);

		miiRadioMenu.cch = sizeof(Buf);
		GetMenuItemInfo(hMenu, ID_OEM,  FALSE, &miiRadioMenu);
		miiRadioMenu.fType |= MFT_RADIOCHECK;
		SetMenuItemInfo(hMenu, ID_OEM,  FALSE, &miiRadioMenu);
	#endif
}

HMENU hPopupMenu;

BOOL LanguageCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	static UINT  NewLanguage;
	static LPSTR NewLangStr;
	static CHAR  CurrLang[6] = "-";
	extern CHAR  ViLang[];

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			if (*CurrLang == '-') lstrcpy(CurrLang, ViLang);
			if		(CurrLang[0]=='f' && CurrLang[1]=='r') wPar = IDC_FRENCH;
			else if (CurrLang[0]=='e' && CurrLang[1]=='s') wPar = IDC_SPANISH;
			else if (CurrLang[0]=='d' && CurrLang[1]=='e') wPar = IDC_GERMAN;
			else if (CurrLang[0]=='e' && CurrLang[1]=='n') wPar = IDC_ENGLISH;
			else										   wPar = IDC_AUTOLANG;
			CheckDlgButton(hDlg, wPar, TRUE);
			/*FALLTHROUGH*/
		case WM_COMMAND:
			switch (COMMAND) {
				case IDC_ENGLISH:
					NewLanguage = 9000;
					NewLangStr  = "en";
					break;
				case IDC_SPANISH:
					NewLanguage = 10000;
					NewLangStr  = "es";
					break;
				case IDC_FRENCH:
					NewLanguage = 12000;
					NewLangStr  = "fr";
					break;
				case IDC_GERMAN:
					NewLanguage = 7000;
					NewLangStr  = "de";
					break;
				case IDC_AUTOLANG:
					NewLanguage = GetLanguage(TRUE);
					NewLangStr  = NULL;
					break;
				case IDC_STORE:
					SaveLanguage(NewLangStr);
					DlgResult = TRUE;
					break;
				case ID_APPLY:
				case IDOK:
					if (Language != NewLanguage) {
						HMENU NewMenu;

						Language = NewLanguage;
						NewMenu = LoadMenu(hInst, MAKEINTRES(100));
						if (NewMenu) {
							HMENU OldMenu;
							extern VOID	InitFileList(VOID);
							extern CHAR ErrorTitle[];

							*ErrorTitle = '\0';
							OldMenu = GetMenu(hwndMain);
							DestroyMenu(SetMenu(hwndMain, NewMenu)
										? OldMenu : NewMenu);
							SetMenuRadioStyle();
							InitFileList();
						}
						if (hPopupMenu) {
							DestroyMenu(hPopupMenu);
							hPopupMenu = 0;
						}
						NewCapsString();
						GenerateTitle(TRUE);
						LoadTips();
					}
					lstrcpy(CurrLang, NewLangStr==NULL ? "" : NewLangStr);
					DlgResult = TRUE;
					if (COMMAND != IDOK) return (TRUE);
					/*FALLTHROUGH*/
				case IDCANCEL:
					EndDialog(hDlg, COMMAND!=IDCANCEL);
			}
			break;

		default:
			return (FALSE);
	}
	return (TRUE);
}

static VOID SelectLanguage(VOID)
{	static DLGPROC DlgProc;

	if (!DlgProc)
		 DlgProc = (DLGPROC)MakeProcInstance((FARPROC)LanguageCallback, hInst);
	DialogBox(hInst, MAKEINTRES(IDD_LANGUAGE), hwndMain, DlgProc);
}

INT WrapScrollStyle, WrapPos = IDC_WRAP_WORD1;
#define WSS_SHOW_SCROLLBAR		 -1
#define WSS_NONE				  0
#define WSS_WRAP_AT_RIGHT_BORDER  1

BOOL FAR PASCAL WrapScrollCallback
		(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	CHAR Buf[16];
	BOOL SetEnable = FALSE, Enabled;
	static BOOL SaveWrapPos;

	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			switch (WrapScrollStyle) {
				case WSS_NONE:
					CheckDlgButton(hDlg, IDC_NO_WRAP_OR_SCROLLBAR, TRUE);
					break;
				case WSS_WRAP_AT_RIGHT_BORDER:
					CheckDlgButton(hDlg, IDC_WRAP_AT_RIGHT_BORDER, TRUE);
					break;
				case WSS_SHOW_SCROLLBAR:
					CheckDlgButton(hDlg, IDC_SHOW_SCROLLBAR,	   TRUE);
					break;
				default:
					CheckDlgButton(hDlg, IDC_WRAP_AT_COLUMN,	   TRUE);
					wsprintf(Buf, "%d", WrapScrollStyle);
					SetDlgItemText(hDlg, IDC_WRAP_COLUMN, Buf);
			}
			if (WrapScrollStyle <= WSS_NONE) {
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_CHAR),  FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_WORD1), FALSE);
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_WORD2), FALSE);
			} else CheckDlgButton(hDlg, WrapPos, TRUE);
			SaveWrapPos = WrapPos;
			break;

		case WM_COMMAND:
			switch (COMMAND) {
				static INT NewStyle = 0;
				case IDC_WRAP_COLUMN:
					if (NOTIFICATION != EN_CHANGE) break;
					CheckRadioButton(hDlg, IDC_WRAP_AT_RIGHT_BORDER,
										   IDC_NO_WRAP_OR_SCROLLBAR,
										   IDC_WRAP_AT_COLUMN);
					/*FALLTHROUGH*/
				case IDC_WRAP_AT_RIGHT_BORDER:
				case IDC_WRAP_AT_COLUMN:
					SetEnable = Enabled = TRUE;
					break;

				case IDC_NO_WRAP_OR_SCROLLBAR:
				case IDC_SHOW_SCROLLBAR:
					SetEnable = TRUE;
					Enabled   = FALSE;
					break;

				case IDC_WRAP_CHAR:
				case IDC_WRAP_WORD1:
				case IDC_WRAP_WORD2:
					WrapPos = COMMAND;
					break;

				case IDOK:
					if (IsDlgButtonChecked(hDlg, IDC_NO_WRAP_OR_SCROLLBAR))
						NewStyle = WSS_NONE;
					else if (IsDlgButtonChecked(hDlg, IDC_WRAP_AT_RIGHT_BORDER))
						NewStyle = WSS_WRAP_AT_RIGHT_BORDER;
					else if (IsDlgButtonChecked(hDlg, IDC_SHOW_SCROLLBAR))
						NewStyle = WSS_SHOW_SCROLLBAR;
					else if (IsDlgButtonChecked(hDlg, IDC_WRAP_AT_COLUMN)) {
						GetDlgItemText(hDlg, IDC_WRAP_COLUMN, Buf, sizeof(Buf));
						NewStyle = (INT)strtol(Buf, NULL, 10);
					}
					if (WrapScrollStyle != NewStyle || WrapPos != SaveWrapPos) {
						WrapScrollStyle  = NewStyle;
						InvalidateText();
					}
					BreakLineFlag = WrapScrollStyle==WSS_WRAP_AT_RIGHT_BORDER;
					EndDialog(hDlg, TRUE);
					break;

				case IDCANCEL:
					WrapPos = SaveWrapPos;
					EndDialog(hDlg, FALSE);
			}
			if (SetEnable && IsWindowEnabled(GetDlgItem(hDlg, IDC_WRAP_CHAR))
							 != Enabled) {
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_CHAR),  Enabled);
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_WORD1), Enabled);
				EnableWindow(GetDlgItem(hDlg, IDC_WRAP_WORD2), Enabled);
				CheckDlgButton(hDlg, WrapPos, Enabled);
			}
			break;

		default:
			return (FALSE);
	}
	return (TRUE);
}

static VOID SelectWrapScroll(VOID)
{	static DLGPROC DlgProc;

	if (!DlgProc)
		 DlgProc = (DLGPROC)MakeProcInstance((FARPROC)WrapScrollCallback,hInst);
	DialogBox(hInst, MAKEINTRES(IDD_WRAP_SCROLL), hwndMain, DlgProc);
}

LOGFONT AboutLogFont = {
	0,0,0,0,0, 0,0,0,0,0,0,0,0, "Arial"
};

BOOL FAR PASCAL AboutCallback(HWND hDlg, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	static HFONT hFont;
	HDC			 hDC;

	(VOID)wPar;		/*parameter never used*/
	(VOID)lPar;		/*parameter never used*/
	switch (uMsg) {
		case WM_INITDIALOG:
			#if defined(WIN32)
				SetDlgItemText(hDlg, IDS_VERSIONBITS, "(32 bit)");
			#else
				SetDlgItemText(hDlg, IDS_VERSIONBITS, "(16 bit)");
			#endif
			SetDlgItemText(hDlg, IDS_VERSION, GetViVersion());
			hDC = GetDC(NULL);
			if (hDC) {
				AboutLogFont.lfHeight =
						-MulDiv(18/*pt*/, GetDeviceCaps(hDC, LOGPIXELSY), 72);
				hFont = CreateFontIndirect(&AboutLogFont);
				if (hFont)
					SendDlgItemMessage(hDlg, IDS_VERSIONTITLE,
									   WM_SETFONT, (WPARAM)hFont, 0);
				ReleaseDC(NULL, hDC);
			}
			return (TRUE);
		case WM_COMMAND:
			EndDialog(hDlg, 0);
			return (TRUE);
		case WM_DESTROY:
			if (hFont) {
				DeleteObject(hFont);
				hFont = NULL;
			}
			return (TRUE);
	}
	return (FALSE);
}

VOID About(HWND hwndParent)
{	static DLGPROC Callback;

	if (!Callback)
		 Callback = (DLGPROC)MakeProcInstance((FARPROC)AboutCallback, hInst);
	DialogBox(hInst, MAKEINTRES(IDD_ABOUT), hwndParent, Callback);
}

PSTR SeparateString(PSTR *p)
{	PSTR p2 = *p;

	if ((*p = strchr(p2, ',')) != NULL) *(*p)++ = '\0';
	else *p = "";
	return (p2);
}

VOID SeparateAllStrings(PSTR p)
{
	while (*p) SeparateString(&p);
}

VOID ContextMenu(HWND hWnd, INT xPos, INT yPos)
{
	if (!hPopupMenu && (hPopupMenu = CreatePopupMenu()) != 0) {
		static CHAR	MenuStrings[160];
		PSTR		p;

		LOADSTRING(hInst, 904, MenuStrings, sizeof(MenuStrings));
		p = MenuStrings;
		#if defined(WIN32)
		{	MENUITEMINFO miiRadioMenu =
				{sizeof(MENUITEMINFO), MIIM_ID | MIIM_TYPE,
				 MFT_RADIOCHECK | MFT_STRING};
			miiRadioMenu.wID = ID_INSMODE;
			miiRadioMenu.dwTypeData = SeparateString(&p);
			if (InsertMenuItem(hPopupMenu, 0, TRUE, &miiRadioMenu)) {
				miiRadioMenu.wID = ID_REPMODE;
				miiRadioMenu.dwTypeData = SeparateString(&p);
				InsertMenuItem(hPopupMenu, 1, TRUE, &miiRadioMenu);
				miiRadioMenu.wID = ID_CMDMODE;
				miiRadioMenu.dwTypeData = SeparateString(&p);
				InsertMenuItem(hPopupMenu, 2, TRUE, &miiRadioMenu);
			} else {
				AppendMenu(hPopupMenu,MF_STRING,ID_INSMODE,
						   miiRadioMenu.dwTypeData);
				AppendMenu(hPopupMenu,MF_STRING,ID_REPMODE, SeparateString(&p));
				AppendMenu(hPopupMenu,MF_STRING,ID_CMDMODE, SeparateString(&p));
			}
		}
		#else
		{	AppendMenu(hPopupMenu, MF_STRING, ID_INSMODE, SeparateString(&p));
			AppendMenu(hPopupMenu, MF_STRING, ID_REPMODE, SeparateString(&p));
			AppendMenu(hPopupMenu, MF_STRING, ID_CMDMODE, SeparateString(&p));
		}
		#endif
		AppendMenu(hPopupMenu, MF_SEPARATOR, 0, NULL);
		AppendMenu(hPopupMenu, MF_STRING,	 ID_TAG,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_STRING,	 ID_TAGBACK,  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_SEPARATOR, 0, NULL);
		AppendMenu(hPopupMenu, MF_STRING,	 ID_UNDO,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_SEPARATOR, 0, NULL);
		AppendMenu(hPopupMenu, MF_STRING,	 ID_CUT,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_STRING,	 ID_COPY,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_STRING,	 ID_PASTE,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_STRING,	 ID_DELETE,	  SeparateString(&p));
		AppendMenu(hPopupMenu, MF_SEPARATOR, 0, NULL);
		AppendMenu(hPopupMenu, MF_STRING,	 ID_SELECTALL,SeparateString(&p));
		AppendMenu(hPopupMenu, MF_STRING,	 ID_REFRESH,  SeparateString(&p));
	}
	if (hPopupMenu) {
		POINT p;
		MouseSelectText(WM_LBUTTONUP, 0, 0, FALSE);
		CheckMenuItem(hPopupMenu, ID_INSMODE,
					  Mode==InsertMode  ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hPopupMenu, ID_REPMODE,
					  Mode==ReplaceMode ? MF_CHECKED : MF_UNCHECKED);
		CheckMenuItem(hPopupMenu, ID_CMDMODE,
					  Mode==CommandMode ? MF_CHECKED : MF_UNCHECKED);
		DisableMenuItems(1, hPopupMenu);
		p.x = xPos;
		p.y = yPos;
		ClientToScreen(hWnd, &p);
		TrackPopupMenu(hPopupMenu, TPM_LEFTALIGN | TPM_RIGHTBUTTON,
					   p.x, p.y, 0, hWnd, NULL);
	}
}

LPARAM PopupLparam;

VOID SwitchMode(MODEENUM NewMode)
{
	if (Mode != NewMode) {
		if (Mode != CommandMode) {
			   GotChar('\033', 0);
			   GotChar(' ',	   0);
		} else GotChar('\033', 0);
		if (NewMode != CommandMode) GotChar(NewMode==InsertMode ? 'i' : 'R', 0);
		if (PopupLparam && !SelectCount) {
			MouseHandle(WM_LBUTTONDOWN,	0, PopupLparam);
			MouseHandle(WM_LBUTTONUP,	0, PopupLparam);
		}
	}
	PopupLparam = 0;
}

extern INT nCurrFileListEntry, nNumFileListEntries;
BOOL	   RestoreCaret, SizeGrip;
BOOL	   SessionEnding = FALSE, XOffState = FALSE;
UINT	   MsWheelRollMsg = 0, WheelAmount = 3, ScrollMsg = 0;
HWND	   WheelHandler;

LONG FAR PASCAL MainProc(HWND hWnd, UINT uMsg, WPARAM wPar, LPARAM lPar)
{	extern LONG xOffset;

	switch (uMsg) {
		case WM_QUERYENDSESSION:
			if (!AskForSave(hwndMain, 0)) break;
			SaveSettings();
			return (TRUE);

		case WM_SYSCOMMAND:
			RemoveStatusTip();
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));

		case WM_CLOSE:
			if (!AskForSave(hwndMain, 0)) break;
			QuitCmd(TRUE);
			break;

		case WM_ENDSESSION:
			if (!wPar) break;
			RemoveTmpFile();
			SessionEnding = TRUE;
			break;

		case WM_DESTROY:
			ReleasePaintResources();
			ReleaseBmp(0);
			if (hPopupMenu) DestroyMenu(hPopupMenu);
			PathExpCleanup();
			PostQuitMessage(0);
			RemoveTmpFile();
			hwndMain = 0;
			break;

		case WM_MOVE:
			if (!IsIconic(hWnd) && !IsZoomed(hWnd)) {
				GetWindowRect(hWnd, &WholeWindow);
				UnsafeSettings = TRUE;
			}
			break;

		case WM_GETMINMAXINFO:
			{	MINMAXINFO FAR *lpMmi = (MINMAXINFO FAR *)lPar;
				extern INT StatusHeight;
				RECT r;
				GetWindowRect(hWnd, &r);
				if (r.right-r.left <= 40) return (1);
				if (Disabled && r.bottom > r.top+StatusHeight+33) {
					lpMmi->ptMinTrackSize.x = lpMmi->ptMaxTrackSize.x =
						r.right - r.left;
					lpMmi->ptMinTrackSize.y = lpMmi->ptMaxTrackSize.y =
						r.bottom - r.top;
				} else {
					lpMmi->ptMinTrackSize.x = 271 +
						(r.right - r.left - ClientRect.right);
					lpMmi->ptMinTrackSize.y = StatusHeight + 33 + TextHeight +
						(r.bottom - r.top - ClientRect.bottom);
				}
			}
			break;

		case WM_SIZE:
			if (wPar != SIZE_MINIMIZED) {
				RECT OldRect;
				BOOL ShowCaretAgain = FALSE;

				OldRect = ClientRect;
				GetClientRect(hWnd, &ClientRect);
				SizeGrip = (wPar != SIZE_MAXIMIZED);
				if (ClientRect.right != OldRect.right) {
					HideEditCaret();
					RecalcStatusRow();
					AdjustWindowParts(OldRect.bottom, ClientRect.bottom);
					if (ClientRect.bottom != OldRect.bottom) NewScrollPos();
					ShowCaretAgain = TRUE;
				} else if (ClientRect.bottom != OldRect.bottom) {
					AdjustWindowParts(OldRect.bottom, ClientRect.bottom);
					NewScrollPos();
				}
				if (SizeGrip) {
					RECT New;
					GetWindowRect(hWnd, &New);
					if (	   New.left   != WholeWindow.left
							|| New.top	  != WholeWindow.top
							|| New.right  != WholeWindow.right
							|| New.bottom != WholeWindow.bottom) {
						WholeWindow = New;
						UnsafeSettings = TRUE;
					}
				}
				if (WrapScrollStyle) {
					extern LONG xOffset, ExpandRedraw, ShrinkRedraw;
					extern INT  RightOffset;
					LONG EndX = ClientRect.right + xOffset - RightOffset;
					extern BOOL Exiting;

					if (Exiting) {
						assert(!Exiting);
						break;
					}
					if (EndX >= ExpandRedraw || EndX < ShrinkRedraw) {
						InvalidateText();
						while (LineInfo && ComparePos(&LineInfo[CrsrLines].Pos,
													  &CurrPos) <= 0) {
							HideEditCaret();
							ScrollLine(1);
							ShowCaretAgain = TRUE;
						}
					}
				}
				if (ShowCaretAgain) ShowEditCaret();
			}
			break;

		case WM_SETFOCUS:
			if (hwndCmd) {
				extern BOOL HideCmd;
				if (!HideCmd) SetFocus(hwndCmd);
			} else {
				ShowEditCaret();
				CheckClipboard();
			}
			break;

		case WM_KILLFOCUS:
			MouseInTextArea(0, 0, 0, 0, 0, TRUE);	/*stop scrolling*/
			HideEditCaret();
			break;

		case WM_ACTIVATEAPP:
			if (wPar) {
				GotKeydown(VK_NUMLOCK, 0);	/*update status if keys changed*/
				GotKeydown(VK_CAPITAL, 0);
			} else TooltipHide(-1);
			break;

		case WM_KEYDOWN:
			TooltipHide(-1);
			if (Interruptable) {
				if (wPar == VK_CANCEL) {
					/*cancel button in search and replace dialogbox (???)*/
					#if defined(WIN32)
						ExecInterrupt(CTRL_BREAK_EVENT);
					#endif
					Interrupted	= TRUE;
					break;
				}
				if (Disabled && wPar != VK_NUMLOCK && wPar != VK_CAPITAL &&
								wPar != VK_SHIFT   && wPar != VK_CONTROL) {
					if (wPar!='C' || GetKeyState(VK_CONTROL)>=0)
						MessageBeep(MB_OK);
						/*...a ctrl+c message beep will be generated later*/
					break;
				} else if (wPar=='C' || wPar=='Q' || wPar=='S') break;
			}
			GotKeydown(wPar, lPar);
			break;

		case WM_CHAR:
			if (!Disabled) {
				if (Interruptable) {
					switch (wPar) {
						case CTRL('s'): XOffState = TRUE;
										break;
										#if defined(WIN32)
						case CTRL('c'):		Interrupted	= TRUE;
											ExecInterrupt(CTRL_C_EVENT);
										#endif
										/*FALLTHROUGH*/
						case CTRL('q'): XOffState = FALSE;
										break;
					}
				} else GotChar(wPar, 0);
			} else if (wPar == CTRL('c'))
				#if defined(WIN32)
					Interrupted = CTRL_C_EVENT+1;
				#else
					Interrupted	= TRUE;
				#endif
			break;

		case WM_SYSKEYDOWN:
			TooltipHide(-1);
			if (wPar == VK_F3) {
				PostMessage(hWnd, WM_COMMAND, ID_SEARCH, 0);
				break;
			} else if (wPar == VK_F10 && GetKeyState(VK_SHIFT) < 0) {
				if (GetKeyState(VK_CONTROL) < 0) {
					/*status context menu...*/
					POINT p;
					p.x = 20;
					p.y = ClientRect.bottom;
					ClientToScreen(hwndMain, &p);
					StatusContextMenu(p.x, p.y);
				} else ContextMenu(hwndMain, (INT)(CaretX-xOffset),
											 (INT)(CaretY+CaretHeight));
				break;
			}
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));

		case WM_SYSCHAR:
			if (wPar == '\b') {
				if (Disabled) MessageBeep(MB_OK);
				else Undo(1);
				break;
			}
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));

		case WM_VSCROLL:
			if (!Disabled) {
				#if defined(WIN32)
					Scroll(LOWORD(wPar), HIWORD(wPar));
				#else
					Scroll(wPar, LOWORD(lPar));
				#endif
			}
			break;

		case WM_MOUSEACTIVATE:
			MouseActivated = TRUE;
			return (DefWindowProc(hWnd, uMsg, wPar, lPar));

		case WM_TIMER:
			if (wPar == 100) MouseSelectText(uMsg, wPar, 0, FALSE);
			else TooltipTimer();
			break;

		case WM_LBUTTONDOWN:
		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONUP:
		case WM_MOUSEMOVE:
			MouseHandle(uMsg, wPar, lPar);
			break;

		case WM_RBUTTONDOWN:
			if (HIWORD(lPar) <  (WORD)nToolbarHeight) break;
			if (HIWORD(lPar) >= (WORD)ClientRect.bottom-StatusHeight) {
				MouseHandle(WM_RBUTTONDOWN, wPar, lPar);
				break;
			}
			if (!SelectCount) {
				PopupLparam = lPar;
				MouseHandle(WM_LBUTTONDOWN,	wPar, lPar);
				MouseHandle(WM_LBUTTONUP,	wPar, lPar);
			}
			ContextMenu(hWnd, (short)LOWORD(lPar), (short)HIWORD(lPar));
			break;

		case WM_CONTEXTMENU:
			if (lPar == -1) ContextMenu(hwndMain, (INT)(CaretX-xOffset),
												  (INT)(CaretY+CaretHeight));
			break;

		case WM_DROPFILES:
			{	HANDLE		hDrop = (HANDLE)wPar;
				INT			i, iDropped, nNumDropped;
				extern CHAR	FileNameBuf2[260];
				BOOL		ShiftPressed;

				ShiftPressed = GetKeyState(VK_SHIFT) < 0;
				nNumDropped = DragQueryFile(hDrop, (UINT)-1, NULL, 0);
				if (nNumDropped) {
					for (iDropped = 0; iDropped < nNumDropped; ++iDropped) {
						DragQueryFile(hDrop, iDropped,
									  FileNameBuf2, sizeof(FileNameBuf2));
						LowerCaseFname(FileNameBuf2);
						if (ShiftPressed) NewViEditCmd(FileNameBuf2);
						else {
							i = FindOrAppendToFileList(FileNameBuf2);
							if (!iDropped) nCurrFileListEntry = i;
						}
					}
					if (!ShiftPressed)
						Open(GetFileListEntry(nCurrFileListEntry), 0);
				}
				DragFinish(hDrop);
				if (AutoRestoreFlag && !ShiftPressed) {
					if (IsIconic(hWnd)) ShowWindow(hWnd, SW_NORMAL);
					else {
						#if defined(WIN32)
							SetForegroundWindow(hWnd);
						#else
							BringWindowToTop(hWnd);
						#endif
					}
				}
			}
			break;

		case WM_ERASEBKGND:
			if (hwndMain) {
				RECT Rect;
				if (CaretActive && !RestoreCaret) {
					HideEditCaret();
					RestoreCaret = TRUE;
				}
				{	RECT SaveClientRect;
					SaveClientRect = ClientRect;
					GetClientRect(hWnd, &Rect);
					ClientRect = Rect;
					TextAreaOnly(&Rect, FALSE);
					ClientRect = SaveClientRect;
				}
				if (!PaintBmp((HDC)wPar, 0, &Rect, 0,
							  FirstVisible * TextHeight)) {
					UnrealizeObject(BkgndBrush);
					FillRect((HDC)wPar, &Rect, BkgndBrush);
				}
			}
			return (1L);

		case WM_PAINT:
			{	PAINTSTRUCT Ps;
				HDC hDC = BeginPaint(hWnd, &Ps);
				if (hwndMain) Paint(hDC, &Ps.rcPaint);
				EndPaint(hWnd, &Ps);
			}
			break;

		case WM_PALETTECHANGED:
			{	RECT Rect;
				Rect = ClientRect;
				TextAreaOnly(&Rect, FALSE);
				InvalidateRect(hWnd, &Rect, TRUE);
			}
			break;

		case WM_COMMAND:
			if (Disabled && COMMAND!=ID_EXIT) MessageBeep(MB_OK);
			else switch (COMMAND) {
				case ID_NEW:		New(hWnd, TRUE);					break;
				case ID_NEWVI:		CommandExec(":vi");					break;
				case ID_OPEN:		Open("", 0);						break;
				case ID_SAVE:		Save("");							break;
				case ID_SAVEAS:		Save(NULL);							break;
				case ID_ALTERNATE:	CommandExec(":e#");					break;
				case ID_READFILE:	ReadCmd(-1, "");					break;
				case ID_FILELIST:	ArgsCmd();							break;
				case ID_PRINTSETTINGS:
									{	extern INT CurrTab;

										CurrTab = 5; Settings();
									}									break;
				case ID_PRINT:		PrintFromMenu(hWnd);				break;
				case ID_EXIT:		PostMessage(hWnd, WM_CLOSE, 0, 0L);	break;

				case ID_INSMODE:	SwitchMode(InsertMode);				break;
				case ID_REPMODE:	SwitchMode(ReplaceMode);			break;
				case ID_CMDMODE:	SwitchMode(CommandMode);			break;
				case ID_UNDO:		Undo(1);							break;
				case ID_CUT:		ClipboardCut();						break;
				case ID_COPY:		ClipboardCopy();					break;
				case ID_PASTE:		ClipboardPaste();					break;
				case ID_DELETE:		if (SelectCount) GotChar('\b', 0);	break;
				case ID_SELECTALL:	SelectAll();						break;
				case ID_REFRESH:	GotKeydown(VK_F5, 0);				break;
				case ID_TAG:		TagCurrentPosition();				break;
				case ID_TAGBACK:	PopTag(TRUE);						break;

				case ID_SEARCH:		SearchDialog();						break;
				case ID_SRCHAGN:	SearchAgain(0);						break;

				case ID_SETTINGS:	Settings();							break;
				case ID_WRAPSCROLL: SelectWrapScroll();					break;
				case ID_LANGUAGE:	SelectLanguage();					break;
				case ID_ANSI:		SetCharSet(0, 1); /*Ansi*/			break;
				case ID_OEM:		SetCharSet(1, 1); /*Pc*/			break;
				case ID_HEX:		ToggleHexMode();					break;

				case ID_ABOUT:		About(hWnd);						break;

				case ID_COMMAND:	if (hwndCmd &&
										  !GetWindowTextLength(hwndCmd) &&
										  IsWindowVisible(hwndCmd))
										PostMessage(hwndMain, WM_USER+1234, 0, 0);
																		break;
				case 910:
				case 911:			StatusContextMenuCommand(COMMAND);  break;

				default:			InterCommCommand(COMMAND);
									PathExpCommand  (COMMAND);
									FileListCommand (COMMAND);
			}
			break;

		case WM_USER+1234:
			CommandDone((BOOL)wPar);
			break;

		case WM_USER+1235:
			ExecInput(wPar, lPar);
			break;

		case WM_USER+1236:
			WatchMessage((LPSTR)lPar);
			break;

		case WM_INITMENUPOPUP:
			if (!HIWORD(lPar)) {
				if (LOWORD(lPar)==4) InterCommMenu(hWnd, (HMENU)wPar);
				else {
					extern HMENU PathExpMenu, hStatusPopupMenu;
					DisableMenuItems((INT)LOWORD(lPar), (HMENU)wPar);
					if (!LOWORD(lPar) && (HMENU)wPar!=PathExpMenu &&
										 (HMENU)wPar!=hPopupMenu  &&
										 (HMENU)wPar!=hStatusPopupMenu)
						InsertFileList((HMENU)wPar);
				}
			}
			break;

		case WM_SYSCOLORCHANGE:
			Ctl3dColorChange();
			SystemColors(TRUE);
			if (wUseSysColors) {
				if (wUseSysColors & 1)
					BackgroundColor	= GetSysColor(COLOR_WINDOW);
				if (wUseSysColors & 2)
					TextColor		= GetSysColor(COLOR_WINDOWTEXT);
				if (wUseSysColors & 4)
					SelBkgndColor	= GetSysColor(COLOR_HIGHLIGHT);
				if (wUseSysColors & 8)
					SelTextColor	= GetSysColor(COLOR_HIGHLIGHTTEXT);
				if (wUseSysColors & 3 || SelectCount)
					InvalidateRect(hWnd, NULL, TRUE);
			}
			break;

		default:
			if (uMsg!=MsWheelRollMsg || !uMsg) {
				LRESULT lr = InterCommWindowProc(hWnd, uMsg, wPar, lPar);
				if (lr != (LRESULT)-1) return (lr);
				return (DefWindowProc(hWnd, uMsg, wPar, lPar));
			}
			/*FALLTHROUGH*/
		case WM_MOUSEWHEEL:
			if (!Disabled) {
				INT			i;
				static UINT LastWheelMsg = 0;

				if (uMsg != LastWheelMsg) {
					if (LastWheelMsg /*only accept one type*/) break;
					LastWheelMsg = uMsg;
				}
				if (ScrollMsg)
					WheelAmount	= (INT)SendMessage(WheelHandler,ScrollMsg,0,0);
				if (wPar & 0x80008000)
					 for (i=WheelAmount; i>0; --i) Scroll(SB_LINEDOWN, 0);
				else for (i=WheelAmount; i>0; --i) Scroll(SB_LINEUP,   0);
			}
	}
	return (0);
}

LPCSTR ArgList, CurrArg;
CHAR   Title[270];

VOID GenerateTitle(BOOL ApplyToWindow)
{	PSTR   dst, LastPeriod = NULL;
	LPCSTR src = CurrFile;

	while (*src == ' ') ++src;
	if (WinVersion >= MAKEWORD(95,3)) dst = Title;
	else dst = Title + wsprintf(Title, "%s - ", (LPCSTR)ApplName);
	if (*src) {
		PSTR d2 = dst;
		for (;;) {
			switch (*src) {
				case '\0':	break;
				case '\\':
				case '/':	dst = d2;
							++src;
							LastPeriod = NULL;
							continue;
				case '.':	LastPeriod = dst;
							/*FALLTHROUGH*/
				default:	*dst++ = *src++;
							continue;
			}
			break;
		}
		#if defined(WIN32)
			if (LastPeriod != NULL) {
				DWORD HideFileExt;
				DWORD Size = sizeof(HideFileExt);
				HKEY  hKey;
				if (RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\Microsoft\\"
													"Windows\\CurrentVersion\\"
													"Explorer\\Advanced",
								 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS) {
					if (RegQueryValueEx(hKey, "HideFileExt", NULL, NULL,
										(LPBYTE)&HideFileExt, &Size)
									   == ERROR_SUCCESS && HideFileExt)
						dst = LastPeriod;
					RegCloseKey(hKey);
				}
			}
		#endif
		*dst = '\0';
	} else LOADSTRING(hInst, 901, dst, sizeof(Title)-(dst-Title)-1);
	{	extern BOOL SuppressAsterisk;
		if (Unsafe && !SuppressAsterisk) lstrcat(dst, "*");
	}
	if (WinVersion >= MAKEWORD(95,3)) {
		lstrcat(dst, " - ");
		lstrcat(dst, ApplName);
	}
	if (ApplyToWindow) SetWindowText(hwndMain, Title);
}

static DWORD GetVersion32(VOID)
{
	#if defined(WIN32)
		return (0);
	#elif LIBW_AVAILABLE
		extern DWORD FAR PASCAL LoadLibraryEx32W
								(LPCSTR lpszLibFile, DWORD hFile,DWORD dwFlags);
		extern DWORD FAR PASCAL FreeLibrary32W(DWORD hLibModule);
		extern DWORD FAR PASCAL GetProcAddress32W
								(DWORD hModule, LPCSTR lpszProc);
		extern DWORD FAR		CallProcEx32W(DWORD, DWORD, DWORD, ...);
		HMODULE					hKernel;
		DWORD					h32Dll = 0, p32GetVersion, Result = 0;

		do {
			if (!(hKernel = GetModuleHandle("KERNEL")))			   break;
			if (!GetProcAddress(hKernel, "GetProcAddress32W"))	   break;
			if (!(h32Dll = LoadLibraryEx32W("KERNEL32", NULL, 0))) break;
			if (!(p32GetVersion = GetProcAddress32W(h32Dll, "GetVersion")))
																   break;
			Result = CallProcEx32W(0, 0, p32GetVersion);
		} while (FALSE);
		if (h32Dll) FreeLibrary32W(h32Dll);
		return (Result);
	#else
		return (0);
	#endif
}

BOOL PrintOnly=FALSE, ReadOnlyParam=FALSE;

BOOL HandleFlag(LPSTR Arg, INT Len, INT *pnShow)
{
	switch (Arg[1]) {
		case 'R': /*readonly*/
			if (Len != 2) return (FALSE);
			ReadOnlyParam = TRUE;
			break;
		case 'p': case 'P': /*print file*/
			if (Len==2 || Arg[2]==':') {
				if (Len > 2 && !PrintSetPrinter(Arg+3, Len-3)) {
					ErrorBox(MB_ICONSTOP, 248);
					break;
				}
				*pnShow = SW_SHOWMINNOACTIVE;
				PrintOnly = TRUE;
				break;
			}
			/*FALLTHROUGH*/
		default:
			return (FALSE);
	}
	return (TRUE);
}

HGLOBAL SpareMemory;			/*released if memory problems occur*/

BOOL FreeSpareMemory(VOID)
{
	if (SpareMemory == NULL) return (FALSE);
	GlobalFree(SpareMemory);
	SpareMemory = NULL;
	ErrorBox(MB_ICONEXCLAMATION, 313);
	return (TRUE);
}

BOOL NotepadCompatible(VOID)
{	CHAR ModuleFileName[MAX_PATH];
	PSTR p, Tail;

	GetModuleFileName(NULL, ModuleFileName, sizeof(ModuleFileName));
	Tail = ModuleFileName;
	while ((p = strchr(Tail, '\\')) != NULL || (p = strchr(Tail, '/')) != NULL)
		Tail = p+1;
	return (*Tail=='N' || *Tail=='n');
}

BOOL InitInstance(INT nShow)
{
	if ((WinVersion = LOWORD(GetVersion())) == MAKEWORD(3,10)) {
		WORD Ver32  = LOWORD(GetVersion32());

		if (Ver32) WinVersion = Ver32;
	}
	WinVersion		= HIBYTE(WinVersion) + (LOBYTE(WinVersion) << 8);
	TextColor		= GetSysColor(COLOR_WINDOWTEXT);
	BackgroundColor	= GetSysColor(COLOR_WINDOW);
	MsWheelRollMsg	= RegisterWindowMessage("MSWHEEL_ROLLMSG");
	WheelHandler	= FindWindow("MouseZ", "Magellan MSWHEEL");
	if (WheelHandler)
		ScrollMsg	= RegisterWindowMessage("MSH_SCROLL_LINES_MSG");
	LoadSettings();
	#if !defined(SW_SHOWDEFAULT)
		#define  SW_SHOWDEFAULT 10
	#endif
	if (Maximized && (nShow==SW_SHOWNORMAL || nShow==SW_SHOWDEFAULT))
		nShow = SW_SHOWMAXIMIZED;
	if (ArgList) {
		if (NotepadCompatible()) {
			INT c;

			if (ArgList[0]=='/' && ((c=ArgList[1])=='P' || c=='p')
								&& ((c=ArgList[2])==' ' || c=='\0')) {
				nShow = SW_SHOWMINNOACTIVE;
				PrintOnly = TRUE;
				ArgList += 2;
				if (c) ++ArgList;
			}
			AppendToFileList(ArgList, lstrlen(ArgList), 1);
		} else if (ArgList) {
			LPSTR lp, lpStart;

			while (*ArgList == ' ') ++ArgList;
			lpStart = lp = (LPSTR)ArgList;
			for (;;) {
				switch (*lp++) {
					case '"':	do; while (*lp && *lp++ != '"');
								continue;
					case ' ':	if (*lpStart!='-' ||
									 !HandleFlag(lpStart, lp-lpStart-1, &nShow))
									AppendToFileList(lpStart, lp-lpStart-1, 1);
								while (*lp == ' ') ++lp;
								lpStart = lp;
								/*FALLTHROUGH*/
					default:	continue;
					case '\0':	/*break*/;
				}
				break;
			}
			if (*lpStart)
				if (*lpStart!='-' || !HandleFlag(lpStart, lp-lpStart-1, &nShow))
					AppendToFileList(lpStart, lp-lpStart-1, 1);
		}
	}
	nCurrFileListEntry = -1;
	do {
		if ((CurrFile = GetFileListEntry(++nCurrFileListEntry)) == NULL) {
			CurrFile = "";
			break;
		}
	} while (*CurrFile == '+');
	CurrArg = CurrFile;
	GenerateTitle(FALSE);
	hwndMain = CreateWindow(ClassName, Title, WS_OVERLAPPEDWINDOW,
							x, y, Width, Height, NULL, NULL, hInst, NULL);
	if (!hwndMain) {
		MessageBox(NULL, "CreateWindow failed!", NULL, MB_OK | MB_ICONSTOP);
		return (FALSE);
	}
	hwndErrorParent = hwndMain;
	DragAcceptFiles(hwndMain, TRUE);
	GetClientRect(hwndMain, &ClientRect);
	SizeGrip = TRUE;
	InitPaint();
	if (*CurrArg) {
		extern BOOL SuppressPaint;

		SuppressPaint = TRUE;
		SwitchCursor(SWC_HOURGLASS_ON);
		ShowWindow(hwndMain, nShow);
		if (!Open(CurrFile=CurrArg, 0)) {
			SuppressPaint = FALSE;
			InvalidateText();
		} /*else SuppressPaint has already been reset in Open()*/
	} else {
		New(hwndMain, FALSE);
		ShowWindow(hwndMain, nShow);
	}
	for (;;) {
		extern CHAR	CommandBuf[280];
		LPSTR		lp = GetFileListEntry(0);

		SwitchCursor(SWC_HOURGLASS_ON);
		if (!lp || *lp != '+') break;
		lstrcpy(CommandBuf, lp);
		RemoveFirstFileListEntry();
		CommandBuf[0] = ':';
		CommandExec(CommandBuf);
	}
	SetMenuRadioStyle();
	if (hwndMain) UpdateWindow(hwndMain);
	SwitchCursor(SWC_HOURGLASS_OFF);
	if (DefaultInsFlag) GotChar('i', 0);
	ShowEditCaret();
	InterCommInit(hwndMain);
	SpareMemory = GlobalAlloc(GMEM_MOVEABLE, 1024);
	return (TRUE);
}

BOOL InitApplication(VOID)
{	static WNDCLASS wc;

	wc.style		 = CS_DBLCLKS;
	wc.lpfnWndProc	 = MainProc;
	wc.cbClsExtra	 = 0;
	wc.cbWndExtra	 = 0;
	wc.hInstance	 = hInst;
	wc.hIcon		 = LoadIcon(hInst, MAKEINTRESOURCE(100));
	wc.hCursor		 = NULL;
	wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
	wc.lpszMenuName  = MAKEINTRES(100);
	wc.lpszClassName = ClassName;
	if (!RegisterClass(&wc)) {
		MessageBox(NULL, "RegisterClass failed!", NULL, MB_OK | MB_ICONSTOP);
		return (FALSE);
	}
	wc.hIcon		 = NULL;
	wc.lpszMenuName  = NULL;
	wc.hbrBackground = (HBRUSH)NULL;
	#if defined(OWN_TABCTRL)
		wc.lpfnWndProc   = TabCtrlProc;
		wc.lpszClassName = "WinVi - TabCtrl";
		RegisterClass(&wc);
	#endif
	wc.lpfnWndProc	 = ColorProc;
	wc.lpszClassName = "WinVi - ColorCtrl";
	RegisterClass(&wc);
	wc.lpfnWndProc	 = TooltipProc;
	wc.cbWndExtra	 = 4 * sizeof(LONG);
	wc.lpszClassName = "WinVi - ToolTip";
	RegisterClass(&wc);
	return (TRUE);
}

static BOOL SpecialCtrlVFlag = TRUE;

INT MessageLoop(BOOL Wait)
{	extern HWND	hwndSearch;
	MSG			Msg;
	static BOOL Escape;

	for (;;) {
		if (Wait && !RefreshBitmap) {
			if (!GetMessage (&Msg, NULL, 0, 0)) return (Msg.wParam);
		} else if (!PeekMessage(&Msg, NULL, 0, 0, PM_REMOVE)) {
			if (RefreshBitmap && hwndMain) {
				INT i;

				for (i=0; i<VisiLines; ++i)
					if (LineInfo[i].Flags & LIF_REFRESH_BKGND) {
						RECT r;

						r.left   = 0;
						r.right  = ClientRect.right;
						r.top    = nToolbarHeight + YTEXT_OFFSET +
								   i * TextHeight;
						r.bottom = r.top + TextHeight;
						LineInfo[i].Flags &= ~LIF_REFRESH_BKGND;
						if (RefreshFlag) InvalidateRect(hwndMain, &r, TRUE);
						else i = VisiLines;	/*skip the remaining part*/
						break;
					}
				if (++i >= VisiLines) {
					RefreshBitmap = FALSE;
					ShowEditCaret();
				}
				continue;
			}
			return (0);
		}
		if (Msg.message == WM_CHAR && Escape) {
			CHAR b[8];

			Escape = FALSE;
			wsprintf(b, Msg.wParam>=' ' && Msg.wParam<='~' || Msg.hwnd!=hwndCmd
						? "%c" : "\\%%%02x", Msg.wParam & 255);
			SendMessage(Msg.hwnd, EM_REPLACESEL, TRUE, (LPARAM)(LPSTR)b);
			continue;
		}
		if (Msg.message == WM_KEYDOWN && Msg.hwnd && !Escape) {
			if (Msg.hwnd == hwndMain) {
				#if 0
				//	if (SpecialCtrlVFlag && Mode==InsertMode && Msg.wParam=='V'
				//						 && GetKeyState(VK_CONTROL)<0) {
				//		Escape = TRUE;
				//		GotChar('^', 0);
				//		Position(1, 'h',  2);
				//		Position(1, ' ', 17);
				//	}
				#endif
			} else {
				if (Msg.hwnd == hwndCmd) {
					switch (Msg.wParam) {
						case '\t':
							if (GetKeyState(VK_SHIFT)   >= 0 &&
								GetKeyState(VK_CONTROL) >= 0) {
									ExpandPath(hwndCmd, 1);
									continue;
							}
							break;
						case '\r':
							PostMessage(hwndMain, WM_USER+1234, 0, 0);
							break;
						case '\033':
							PostMessage(hwndMain, WM_USER+1234, 1, 0);
							break;
						case 'V':
							if (SpecialCtrlVFlag && GetKeyState(VK_CONTROL)<0) {
								DWORD Sel;

								Escape = TRUE;
								SendMessage(hwndCmd, WM_CHAR, '^', 1);
								Sel = SendMessage(hwndCmd, EM_GETSEL, 0,0);
								#if defined(WIN32)
									SendMessage(hwndCmd, EM_SETSEL,
												LOWORD(Sel)-1,LOWORD(Sel));
								#else
									SendMessage(hwndCmd, EM_SETSEL, 1,
												MAKELPARAM(LOWORD(Sel),
														   LOWORD(Sel)-1));
								#endif
								continue;
							}
					}
				}
				if (Msg.wParam == VK_NUMLOCK || Msg.wParam == VK_CAPITAL)
					GotKeydown(Msg.wParam, Msg.lParam);
			}
			if (Msg.wParam=='\t' && hwndSearch && GetKeyState(VK_CONTROL)<0) {
				extern HWND hwndSettings;

				if (!hwndSettings || GetParent(Msg.hwnd)==hwndSearch) {
					SetActiveWindow(GetActiveWindow()!=hwndSearch
									? hwndSearch
									: hwndSettings ? hwndSettings
												   : hwndMain);
					continue;
				}
			}
		}
		if (Msg.message == WM_QUIT) {
			#if defined(WIN32)
				Interrupted = CTRL_CLOSE_EVENT+1;
			#else
				Interrupted = TRUE;
			#endif
			PostQuitMessage(Msg.wParam);
			return (0);
		}
		if (hwndSearch && IsDialogMessage(hwndSearch, &Msg)) continue;
		if (IsSettingsDialogMessage(&Msg)) continue;
		TranslateMessage(&Msg);
		DispatchMessage(&Msg);
		if (!Wait) {
			if (!hwndMain) return (0);
		} else if (RestoreCaret && !PeekMessage(&Msg, NULL, 0,0, PM_NOREMOVE))
			ShowEditCaret();
	}
}

INT PASCAL WinMain(HINST hCurr, HINST hPrev, LPSTR lpCmd, INT nShow)
{	INT Ret;

	hInst = hCurr;
	ArgList = lpCmd;
	Language = GetLanguage(FALSE);
	Ctl3dRegister(hInst);
	Ctl3dAutoSubclass(hInst);
	if (!hPrev && !InitApplication() || !InitInstance(nShow)) Ret = 1;
	else if (PrintOnly) {
		PrintImmediate(hwndMain);
		Ret = 0;
	} else Ret = MessageLoop(TRUE);
	Ctl3dUnregister(hInst);
	return (Ret);
}
