/********************************************************\
|*                                                      *|
|*   Program:              WinVi                        *|
|*   Author:               Raphael Molle                *|
|*   Compiler:             several (w/ WIN32 support)   *|
|*   --------------------- --------------------------   *|
|*    Copyright 1994-2000 by Raphael Molle, Berlin     *|
|*    Copyleft by GPL     See file COPYING for more    *|
|*                         info about terms of use      *|
|*                                                      *|
\********************************************************/

/* revision log:
 *		3-Dec-2000	first publication of source code
 */

#define HINST HINSTANCE
#if !defined(ULONG)
# define ULONG unsigned long
#endif
#if !defined(MAKEWORD)
# define MAKEWORD(lo,hi) (WORD)((WORD)((WORD)(hi) << 8) + (BYTE)(lo))
#endif

#if defined(WIN32)
# define COMMAND	  LOWORD(wPar)
# define NOTIFICATION HIWORD(wPar)
# define _fmemmove	  memmove
# define _fmemcpy	  memcpy
# define _fmemcmp	  memcmp
# define _fmemchr	  memchr
# define _fmemset	  memset
# define _fstrchr	  strchr
# define _fstrrchr	  strrchr
# define hmemcpy	  memcpy
# define huge
# define _fcalloc	  calloc
# define _frealloc	  realloc
# define _ffree		  free
#else
# define COMMAND	  wPar
# define NOTIFICATION HIWORD(lPar)
# define INT          int
# define CHAR         char
# if !defined(MAX_PATH)
#  define MAX_PATH	  129
# endif
#endif

#define LOADSTRING(a,b,c,d)	  LoadString(a,(b)+Language,c,d)
#define MAKEINTRES(a)		  MAKEINTRESOURCE((a)+Language)

#define CTRL(c)		((c) & 31)
#define LCASE(c)	((c) | ('a'-'A'))

/*toolbar index values...*/
#define IDB_NEW		0
#define IDB_OPEN	1
#define IDB_SAVE	2
#define IDB_CUT		3
#define IDB_COPY	4
#define IDB_PASTE	5
#define IDB_UNDO	6
#define IDB_SEARCH	7
#define IDB_SRCHAGN	8
#define IDB_ANSI	9
#define IDB_OEM	   10
#define IDB_HEX	   11

/*SwitchCursor flags...*/
#define SWC_STATUS_OR_TOOLBAR 1		/*mouse in non-text area*/
#define SWC_TEXTAREA		  2		/*mouse in text area*/
#define SWC_SIZEGRIP		  4		/*mouse in right lower corner*/
#define SWC_TEXTCURSOR		  8		/*text cursor switch between IBeam & Arrow*/
#define SWC_HOURGLASS_ON	 16		/*working, show hourglass*/
#define SWC_HOURGLASS_OFF	 32		/*switch back*/

/*NewStatus styles*/
#define NS_NORMAL		   0
#define NS_GRAY			   1
#define NS_ERROR		   2
#define NS_READING_WRITING 3
#define NS_TOOLTIP		   4

typedef char const *PCSTR;
typedef struct {
	LONG ScreenLine;	/*0 for first line, 1 and above for continuation lines*/
	LONG PixelOffset;	/*number of pixels from left border of leftmost char*/
} COLUMN, *PCOLUMN;

typedef enum {CommandMode, CmdGotFirstChar, GetDblQuoteChar, PositionTf,
			  PositionQuote, PositionBracket, InsertMode, ReplaceMode} MODEENUM;
extern MODEENUM Mode;

extern UINT		Language;
extern BOOL		HexEditMode, TfForward, TfIsFind, ReplaceSingle;
extern BOOL		Interruptable, XOffState, FullRowExtend;
extern UINT		Interrupted, CmdFirstChar, Number;
extern int		TextHeight;  /*screen line y difference in pixels*/
extern int		VisiLines;	 /*number of fully or partly displayed text lines*/
extern LONG		FirstVisible;/*line numbr (starting at 0) for bitmap alignment*/
extern int		AveCharWidth;
extern COLORREF TextColor;
extern HCURSOR	hcurCurrent, hcurArrow, hcurIBeam, hcurHourGlass, hcurSizeNwSe;
extern HFONT	TextFont, HexFont;
extern HINST	hInst;
extern HWND		hwndMain, hwndCmd, hwndErrorParent;
extern WORD		WinVersion;
extern RECT		ClientRect;
extern LPCSTR	CurrFile, CurrArg;
extern BOOL		IsAltFile, Unsafe, CommandWithExclamation, CaretActive;
extern BOOL		MouseActivated, LcaseFlag, BreakLineFlag, ShowMatchFlag;
extern BOOL		ReadOnlyFlag, AutoRestoreFlag, AutoIndentFlag, AutoWriteFlag;
extern BOOL		IgnoreCaseFlag, WrapScanFlag, TerseFlag, DefaultInsFlag;
extern BOOL		DefaultTmpFlag, QuitIfMoreFlag, MagicFlag, WrapPosFlag;
extern BOOL		ListFlag, NumberFlag, HexNumberFlag, FullRowFlag;
extern BOOL		MoreFilesWarned;
extern BOOL		ViewOnlyFlag;	/*view-only=restrictive read-only w/o editing*/
extern BOOL		ReadOnlyFile;	/*read-only for this file only*/
extern COLUMN	CurrCol;		/*aspired pixel position in logical line*/
extern ULONG	CurrColumn;	/*displayed character offset within line (for status)*/
extern ULONG	LinesInFile, SelectCount;
extern int		CrsrLines;		/*number of positionable displayed text lines*/
							/*CrsrLines==VisiLines || CrsrLines==VisiLines-1*/
extern int		CurrX, CurrLine;
extern int		CaretX, CaretY, CaretWidth, CaretHeight;
extern int		TabStop, ShiftWidth, MaxScroll, Disabled;
extern char		FileName[260];
extern LPSTR	TmpDirectory, Tags;
extern LONG		PageThreshold;
extern ULONG	UndoLimit;
extern PCSTR	ApplName;
extern UINT		CharSet;	/*0=Ansi, 1=Pc*/
extern BOOL		LButtonPushed;
extern long     RelativePosition;
extern BOOL		Indenting;
extern int		LinesVar, ColumnsVar;	/*used in ResizeWithLinesAndColumns()*/
extern BOOL		RefreshBitmap;
extern char		BitmapFilename[260];
extern char		AltFileName[260];
extern BOOL		DlgResult;

/*------ FUNCTION PROTOTYPES... ------*/
/*winvi.c*/
BOOL  AllocString(LPSTR *, LPCSTR);
PSTR  SeparateString(PSTR*);
void  SeparateAllStrings(PSTR);
void  ContextMenu(HWND, int, int);
void  GenerateTitle(BOOL);
BOOL  FreeSpareMemory(void);
int   MessageLoop(BOOL);
BOOL DisplaySettingsCallback(HWND, UINT, WPARAM, LPARAM);
BOOL EditSettingsCallback(HWND, UINT, WPARAM, LPARAM);
BOOL FileSettingsCallback(HWND, UINT, WPARAM, LPARAM);
BOOL LanguageCallback(HWND, UINT, WPARAM, LPARAM);

/*profile.c*/
void  SaveColors(void);
void  SaveSettings(void);
void  SaveFontSettings(void);
void  SaveEditSettings(void);
void  SaveDisplaySettings(void);
void  SaveFileSettings(void);
void  LoadSettings(void);
VOID  StorePrintSettings(VOID);
VOID  LoadPrintSettings(VOID);
void  InsertFileList(HMENU);
void  AddToFileList(LPCSTR);
void  FileListCommand(int);
int   GetLanguage(BOOL);
VOID  SaveLanguage(LPSTR);
BOOL  DebugMessages(VOID);

/*paint.c*/
void InitPaint(void);
void ReleasePaintResources(void);
void AdjustWindowParts(int, int);
void NewScrollPos(void);
void ExtendSelection(void);
void ScrollJump(int);
int  GetTextWidth(HDC, PSTR, int);
BOOL ScrollLine(int);
void InvalidateText(void);
void TextAreaOnly(RECT*, BOOL);
int  StartX(BOOL);
void TextLine(HDC, int, int);
void Paint(HDC, PRECT);
int  GetPixelLine(int, BOOL);
void GetXPos(PCOLUMN);
void CommandEdit(UINT);
void CommandDone(BOOL);
void Error(int, ...);
int  ErrorBox(int, int, ...);

/*status.c*/
void RecalcStatusRow(void);
void PaintStatus(HDC, PRECT, PRECT);
void NewStatus(int, PSTR, int);
void SetStatusCommand(PSTR, BOOL);

/*toolbar.c*/
void LoadTips(void);
void PaintTools(HDC, PRECT);
void Disable(BOOL);
void Enable(void);
BOOL IsToolButtonEnabled(int);
void EnableToolButton(int, BOOL);
void ToggleHexMode(void);
void DisableMenuItems(int, HMENU);
void CheckClipboard(void);
void SetCharSet(UINT, UINT);
BOOL MouseForToolbar(UINT, WPARAM, LPARAM);
void TooltipTimer(void);
void TooltipHide(int);
LRESULT CALLBACK TooltipProc(HWND, UINT, WPARAM, LPARAM);

/*font.c*/
BOOL NewFont(HFONT*, PLOGFONT, PINT);

/*file.c*/
void SetUnsafe(void);
void SetSafe(BOOL);
BOOL AskForSave(HWND, WORD);
void LowerCaseFname(LPSTR);
void New(HWND, BOOL);
void WatchMessage(LPSTR);
void ShowFileName(void);
void FileError(LPCSTR, int, int);
VOID JumpAbsolute(ULONG);
BOOL Open(LPCSTR, WORD);
void FilenameCmd(LPSTR);
void Save(LPCSTR);
BOOL EditCmd(LPCSTR);
BOOL ReadCmd(long, LPCSTR);
BOOL NextCmd(LPCSTR);
BOOL WriteCmd(long, long, LPCSTR, WORD);

/*insdel.c*/
void InsertChar(int);
BOOL DeleteChar(void);
void InsertEol(UINT, int);
void DeleteEol(BOOL);
void DeleteForward(UINT);
void Join(void);
void DeleteSelected(int);
BOOL Put(long, BOOL);
BOOL Shift(long, WORD);

/*undo.c*/
void StartUndoSequence(void);
void DisableUndo(void);
void SetSafeForUndo(BOOL);
void Undo(WORD);

/*input.c*/
BOOL IsViewOnly(void);
BOOL HideEditCaret(void);
void ShowEditCaret(void);
void SwitchCursor(WORD);
void GotKeydown(WPARAM, LPARAM);
void ShowSelect(void);
void CommandLineDone(BOOL, PSTR);
void Unselect(void);
VOID SelectAll(VOID);
void GotChar(int, WORD);
void Scroll(int, int);
VOID ShowMatch(WORD);

/*mouse.c*/
void MouseInTextArea(UINT, WPARAM, int, int, int, BOOL);
void MouseSelectText(UINT, WPARAM, LONG, BOOL);
BOOL MouseHandle(UINT, WPARAM, LPARAM);

/*command.c*/
BOOL NewViEditCmd(PSTR);
BOOL QuitCmd(BOOL);
void ResizeWithLinesAndColumns(void);
BOOL CommandExec(PSTR);

/*filelist.c*/
BOOL  RemoveFirstFileListEntry(void);
void  ClearFileList(void);
int   AppendToFileList(LPCSTR, int, int);
LPSTR GetFileListEntry(int);
void  ArgsCmd(void);
int   FindOrAppendToFileList(LPCSTR);

/*page.c*/
void RemoveTmpFile(void);

/*search.c*/
BOOL CheckInterrupt(void);
VOID StartInterruptCheck(void);
BOOL StartSearch(void);
PSTR BuildMatchList(PSTR, int, int *);
void SearchNewCset(void);
BOOL SubstituteCmd(LONG, LONG, PSTR);
BOOL SearchAgain(WORD);
void SearchDialog(void);
BOOL SaveMatchList(VOID);
BOOL RestoreMatchList(VOID);

/*clipbrd.c*/
BOOL ClipboardCopy(void);
BOOL ClipboardCut(void);
BOOL ClipboardPaste(void);

/*position.c*/
BOOL Position(long, int, int);

/*tags.c*/
VOID PopTag(BOOL);
BOOL IsTagStackEmpty(VOID);
BOOL Tag(PSTR, WORD);
VOID TagCurrentPosition(VOID);

/*bmp.c*/
void ReleaseBmp(int);
BOOL PaintBmp(HDC, int, PRECT, LONG, LONG);

/*pathexp.c*/
void PathExpCommand(UINT);
BOOL ExpandPath(HWND, WORD);
void PathExpCleanup(void);

/*settings.c*/
VOID CheckDefButton();
HWND GetCurrentTabSheet(VOID);
void Settings(void);
BOOL IsSettingsDialogMessage(PMSG);

/*map.c*/
BOOL MapCallback(HWND, UINT, WPARAM, LPARAM);

/*version.c*/
PSTR GetViVersion(void);
